%
% dticode/matlab/plotdti.m
%
% Written 2012 by Tuomo Valkonen <tuomov@iki.fi>, University of Graz.
%
% Plot a tensor field. Arguments:
%
%  sig          Tensor field u signature.
%  u            Tensor field to be plotted.
%  plotmode     Plot mode (string). One of:
%
%                colordir         Colour-coded principal eigenvector
%                fa-pev           Fractional anistropy in grayscale
%                                 superimposed by principal eigenvector
%                error-fa-peagle  Colour-coded error of fractional
%                                 anisotropy and principal eigenvector.
%                                 param=[max angle, max FA error]
%                component        Plot absolute value of individual 
%                                 tensor component in grayscale.
%                                 param=[index, divisor]
%  mask         Plotted area mask.
%  f0           Reference image for error plots
%  area         Area to plot: [x1 y1 x2 y2]. 
%               Empty matrix [] for full area.
%  param        plotmode-dependent parameters.
function [img, maskret]=plotdti(sig, u, plotmode, mask, f0, area, param)
    dim=sig(1);
    w=sig(3);
    h=sig(4);
    assert(sig(2)==2);
    assert(length(sig)==4 || (length(sig)==5 && sig(5)==1));

    if nargin<4 || isempty(mask)
        mask=true(1, size(u, 2));
    end
    
    if nargin<5
        f0=[];
    end
    
    if nargin>=6 & ~isempty(area)
        u=myzoom(u, h, w, area);
        f0=myzoom(f0, h, w, area);
        mask=myzoom(mask, h, w, area);
        w=area(3)-area(1)+1;
        h=area(4)-area(2)+1;
    end
    
    l=size(u, 2);
    
    [L, E]=sym2eigs_multi_sorted(u);
    
    fa=ev2fractanis(L);
    
    scale=max(max(abs(L)));
    e1=squeeze(E(1, :, :));
    %e1=zeros(size(E, 2), size(E, 3));
    %for i=1:size(E, 2)
    %    e1(i, :)=E(1, i, :);
    %end
    l1=L(1, :);
    major=vecmult(e1, l1/scale);

    
    img=[];
    
    cla;
    hold on;
    
    if strcmp(plotmode, 'ellipse')
        assert(sym2dim(u)==2);
        NPT=10;
        lx=zeros(size(u, 2), NPT);
        ly=zeros(size(u, 2), NPT);
        cx=cos(linspace(0, 2*pi, NPT));
        cy=sin(linspace(0, 2*pi, NPT));
        minor=squeeze(E(2, :, :)).*repmat(L(2, :), 2, 1)/scale;
        for i=1:size(u, 2)
            xi=mod(i-1, w)+1;
            yi=(i-xi)/w+1;
            x=(xi-1)+0.5;
            y=(yi-1)+0.5;
            
            P=major(:, i)*major(:, i)'/norm(major(:, i))+minor(:, i)*minor(:, i)'/norm(minor(:, i));
            tmp=(1/2)*P*[cx; cy]+repmat([x; y], 1, NPT);
            lx(i, :)=tmp(1, :);
            ly(i, :)=tmp(2, :);
        end
        ii=(L(1, :)>0 & L(2, :)>0);
        jj=(L(1, :)<0 & L(2, :)<0);
        h1=fill(lx(ii, :)', ly(ii, :)', 'b', 'edgecolor', 'b');
        h2=fill(lx(jj, :)', ly(jj, :)', 'r', 'edgecolor', 'r');
        h3=fill(lx(~ii & ~jj, :)', ly(~ii & ~jj, :)', 'g', 'edgecolor', 'g');
        h=[h1; h2; h3];
    elseif strcmp(plotmode, 'fa-pev')
        colormap(gray(256));
        image(256*reshape(fa.*mask, w, h)');
        
        [X, Y]=meshgrid(1+[0:w-1], 1+[0:h-1]);
        %contour(X, Y, reshape(fa.*mask, w, h), linspace(0, 1, 10));%, '-', 'Color', [0 0.3 0.3]);
        %set(gca, 'ColorOrder', [0.5 0 0.5]);
        X=flattenT(X'); Y=flattenT(Y');
        if 0
            e2=squeeze(E(2, :, :));
            minor=vecmult(e2, l1/scale);
            U=minor(1, :);
            V=minor(2, :);
            quiver(X(mask)-U(mask)/2, Y(mask)-V(mask)/2, U(mask), V(mask), 0, '.', 'LineWidth', 1.5, 'Color', [0, 0.4, 0]);
            e3=squeeze(E(3, :, :));
            minor=vecmult(e3, l1/scale);
            U=minor(1, :);
            V=minor(2, :);
            quiver(X(mask)-U(mask)/2, Y(mask)-V(mask)/2, U(mask), V(mask), 0, '.', 'LineWidth', 1.5, 'Color', [0, 0.4, 0]);
        end
        U=major(1, :);
        V=major(2, :);
        quiver(X(mask)-U(mask)/2, Y(mask)-V(mask)/2, U(mask), V(mask), 0, 'r.', 'LineWidth', 1.5);
        axis off;
        axis off;
    elseif strcmp(plotmode, 'ellipsoid')
        N=15;
        [X0,Y0,Z0]=sphere(N);
        OS=size(X0);
        S=prod(OS);
        X0=reshape(X0, 1, S);
        Y0=reshape(Y0, 1, S);
        Z0=reshape(Z0, 1, S);
        ee1=(squeeze(E(1, :, :)));
        ee2=(squeeze(E(2, :, :)));
        ee3=(squeeze(E(3, :, :)));
        for x=0:(w-1)
            for y=0:(h-1)
                i=1+x+y*w;
                sc=(2*scale);
                a=L(:, i)/sc;
                %ss=min(1, L(1, i)/(2*sc)+1/5);
                %a=(L(:, i)/sc)*ss/(L(1, i)/sc);
                R=a(1)*E(1, :, i)'*[1, 0, 0]+a(2)*E(2, :, i)'*[0, 1, 0]+a(3)*E(3, :, i)'*[0, 0, 1];
                tmp=R*[X0; Y0; Z0];
                X=tmp(1, :); Y=tmp(2, :); Z=tmp(3, :);
                X=reshape(X+(0.5+x), OS);
                Y=reshape(Y+(0.5+y), OS);
                Z=reshape(Z, OS);
                c=min(1, 1/3+fa(i))*abs(e1(:, i));
                C=repmat(reshape(c, 1, 1, 3), [OS(1) OS(2) 1]);
                surf(X, Y, Z, C);
            end
        end
        light('Position', [w/3 -h/3 30], 'Style', 'infinite');
        shading flat;
        lighting phong;
        axis on;
        whitebg('black')
        set(gca, 'CameraPosition', [-19.0077  -80.9439  100.1949], ...
            'XTick', [], 'YTick', [], 'ZTick', []);
    else
        if strcmp(plotmode, 'fa')
            rgb=repmat(fa.*mask, 3, 1);
        elseif strcmp(plotmode, 'eval')
            if length(param)<1
                idx=1;
            else
                idx=param(1);
            end
            if length(param)<2
                sc=scale;
            else
                sc=param(2);
            end
            %rgb=repmat(abs(L(param, :))/scale.*mask, 3, 1);
            s=min(abs(L(idx, :))/sc, 1);
            rgb=vecmult(cmapindex(s, jet(1024)), mask);
        elseif strcmp(plotmode, 'maxeig')
            ev=l1;
            Lf=sym2eigs_multi(f0);
            sc=max(max(abs(Lf)));
            s=min(1, abs(ev)/sc);
            rgb=vecmult(cmapindex(s, jet(1024)), mask);
        elseif strcmp(plotmode, 'colordir') || strcmp(plotmode, 'colordir2')
            red=abs(e1(1, :));
            green=abs(e1(2, :));
            if dim>=3
                blue=abs(e1(3, :));
            else
                blue=zeros(size(e1(1, :)));
            end
            %modul=fa.^(1/2)*(2/3)+1/3;
            if strcmp(plotmode, 'colordir')
                modul=min(1, 1/3+fa);%.^(1/2);
            else
                modul=abs(l1)/max(abs(l1));
            end
            rgb=vecmult([red; green; blue], mask.*modul);
        elseif strcmp(plotmode, 'error-fa-peangle')
            [L0, E0]=sym2eigs_multi_sorted(f0);
            fa0=ev2fractanis(L0);
            %fa0=sym2fractanis(f0);
            e10=squeeze(E0(1, :, :));
            devcos=min(1, abs(dot(e1, e10, 1)));
            if ~isempty(param)
                maxangle=param(1);
                fadiv=param(2);
            else
                maxangle=90;
                fadiv=0.50;
            end
            s=min(1, acos(devcos)/pi*180/maxangle);%.^(1/2);
            q=min(abs(fa-fa0)/fadiv, 1);
            rgb0=cmapindex(s, jet(1024));
            rgb=vecmult(max(rgb0, repmat(q, 3, 1)), mask);
        elseif strcmp(plotmode, 'error-fa')
            [L0, E0]=sym2eigs_multi_sorted(f0);
            fa0=ev2fractanis(L0);
            if isempty(param)
                param=0.1;
            end
            q=min(abs(fa-fa0)/param, 1);
            %rgb=vecmult(cmapindex(q, jet(1024)), mask);
            %rgb=vecmult(cmapindex(q, mymap(1024)), mask);
            %rgb(1, :)=q.*mask; rgb(2, :)=0; rgb(3, :)=0;
            %rgb=max(rgb, repmat(fa0.*mask, 3, 1));
            hsv(:, 1)=(0.5+q/2);
            hsv(:, 2)=q;
            hsv(:, 3)=(0.1+fa*0.9);
            rgb=vecmult(hsv2rgb(hsv)', mask);
        elseif strcmp(plotmode, 'error-peangle')
            [L0, E0]=sym2eigs_multi_sorted(f0);
            fa0=ev2fractanis(L0);
            e10=squeeze(E0(1, :, :));
            d=e1-e10;
            devcos=min(1, abs(dot(e1, e10, 1)));
            if isempty(param)
                param=90;
            end
            maxangle=param;
            q=min(1, acos(devcos)/pi*180/maxangle);%.^(1/2);
            hsv(:, 1)=(0.5+q/2);
            hsv(:, 2)=q;
            hsv(:, 3)=(0.1+fa*0.9);
            rgb=vecmult(hsv2rgb(hsv)', mask);
        elseif strcmp(plotmode, 'error-peval')
            L0=sym2eigs_multi_sorted(f0);
            %fa0=ev2fractanis(L0);
            l01=L0(1, :);
            d=l1-l01;
            if isempty(param)
                param=10;
            end
            q=min(1, param*abs(d)/max(l01));
            hsv(:, 1)=(0.5+q/2);
            hsv(:, 2)=q;
            hsv(:, 3)=(0.1+fa*0.9);
            rgb=vecmult(hsv2rgb(hsv)', mask);
        elseif strcmp(plotmode, 'error-frob')
            %L0=sym2eigs_multi_sorted(f0);
            %fa0=ev2fractanis(L0);
            d=u-f0;
            WM=ipweight([sym2dim(u), 2]);
            enorms=sum(d.*d.*repmat(WM, 1, size(u, 2)), 1);
            if isempty(param)
                param=10;
            end
            q=min(param*sqrt(enorms)/(dim*scale), 1);
            hsv(:, 1)=(0.5+q/2);
            hsv(:, 2)=q;
            hsv(:, 3)=(0.1+fa*0.9);
            rgb=vecmult(hsv2rgb(hsv)', mask);
        elseif strcmp(plotmode, 'posdef')
            negs=sum(L<-eps, 1);
            rgb(1, :)=double(negs+2*dim*(negs>0))/(3*dim);
            rgb(2, :)=double(mask)*0.4;
            rgb(3, :)=double(mask)*0.8;
        elseif strcmp(plotmode, 'component')
            if length(param)<1
                comp=u(1, :);
            else
                comp=u(param(1), :);
            end
            if length(param)<2
                scale=max(abs(comp));
            else
                scale=param(2);
            end
            comp=clamp(abs(comp)/scale, 0, 1);
            rgb(1, :)=comp;
            rgb(2, :)=comp;
            rgb(3, :)=comp;
        else
            error('Unknown plotmode')
        end
        % Results in transposed image:
        %img=reshape(rgb', w, w, 3);
        % => Need to do this:
        img=zeros(h, w, 3);
        img(:, :, 1)=reshape(rgb(1, :), w, h)';
        img(:, :, 2)=reshape(rgb(2, :), w, h)';
        img(:, :, 3)=reshape(rgb(3, :), w, h)';
        maskret=reshape(mask, w, h)';
        image(img);
    end

    %title(['scale = ', num2str(scale), ', median = ', num2str(md), ', mean = ', num2str(me)])
    set(gca, 'YDir', 'normal');
    hold off;
    axis equal;
    axis tight;
end

function r=iscale(y, mask)
    mi=min(y(mask));
    ma=max(y(mask));
    me=mean(y(mask));
    
    pt=0.15*me;
    
    r=sigmoid(-2+(y-pt)/(ma-pt)*8)/sigmoid(6);
    %r=exp(-2+(y-pt)/(ma-pt)*2)/exp(0);
    
    %c=10;
    %t=(y-me)*c;
    %scma=sigmoid((ma-me)*c);
    %scmi=sigmoid((mi-me)*c);
    %r=(sigmoid(t)-scmi)/(scma-scmi);
end

function r=iscale2(y, mask)
    % y \in [0, 1]!
    %r=sigmoid(-2+y*8)/sigmoid(6);
    r=sigmoid(-1+y*6)/sigmoid(5);
    %r=sigmoid(-2+y*8)/sigmoid(6);
end

function r=sigmoid(t)
    r=1./(1+exp(-t));
end

function e=posx(e)
    jj=e(1, :)<0;
    e(:, jj)=-e(:, jj);
end


function res=blowup(s, mask)
    mi=min(s(mask));
    ma=max(s(mask));
    res=((s-mi)/(ma-mi));
end


function imgz=myzoom(img, h, w, area)
    if ~isempty(area) & ~isempty(img);
        nh=area(3)-area(1)+1;
        nw=area(4)-area(2)+1;
        for i=1:size(img, 1)
            z=reshape(img(i, :), w, h);
            q=z(:, area(1):area(3), :);
            t=q(area(2):area(4), :, :);
            imgz(i, :)=reshape(t, 1, nw*nh);
        end
    else
        imgz=img;
    end
end

function rgb=mymap(count)
    rgb(:, 1)=linspace(0, 1, count);
    rgb(:, 2)=0;
    rgb(:, 3)=linspace(1, 0, count);
end
