%
% dticode/matlab/write_lut.m
%
% Written 2012 by Tuomo Valkonen <tuomov@iki.fi>, University of Graz & 
%                                                 University of Cambridge.
%
function write_lut()
    fs=[];
    fs(1)=fopen('../src/lut.c', 'w');
    fs(2)=fopen('../src/lut.h', 'w');
    
    max_dim=3;
    max_order=4;
    
    rules={struct('bdry', 0, 'alg', 'f'), ...
           struct('bdry', 0, 'alg', 'c'), ...
           struct('bdry', 0, 'alg', 'b'), ...
           struct('bdry', 1, 'alg', 'f')};
    
    for f=fs
        fprintf(f, '/* Automatically generated by write_lut.m */\n\n');
    end
    
    fprintf(fs(1), '#include "lut.h"\n\n');
    
    fprintf(fs(2), '#ifndef _LUT_H\n#define _LUT_H\n\n');
    fprintf(fs(2), '#include "sparse.h"\n\n');
    fprintf(fs(2), '#include "number.h"\n\n');
    fprintf(fs(2), '#define MAX_DIM %d\n', max_dim);
    fprintf(fs(2), '#define MAX_ORDER %d\n\n', max_order);

    write_diff(fs, max_dim, max_order, rules, 1, 'negsymdiv', ...
               @get_negsymdiv_sym, @get_negsymdiv_nonsym);
    write_diff(fs, max_dim, max_order, rules, 0, 'symgrad', ...
               @get_symgrad_sym, @get_symgrad_nonsym);
    write_diff(fs, max_dim, max_order, rules, 1, 'negdiv', ...
               @get_negdiv_sym, @get_negdiv_nonsym);
    write_diff(fs, max_dim, max_order, rules, 0, 'grad', ...
               @get_grad_sym, @get_grad_nonsym);
               
    write_diff(fs, max_dim, max_order, rules, 1, 'negsymdivcomplex', ...
               getcomplex(@get_negsymdiv_sym), getcomplex(@get_negsymdiv_nonsym));
    write_diff(fs, max_dim, max_order, rules, 0, 'symgradcomplex', ...
               getcomplex(@get_symgrad_sym), getcomplex(@get_symgrad_nonsym));
    write_diff(fs, max_dim, max_order, rules, 1, 'negdivcomplex', ...
               getcomplex(@get_negdiv_sym), getcomplex(@get_negdiv_nonsym));
    write_diff(fs, max_dim, max_order, rules, 0, 'gradcomplex', ...
               getcomplex(@get_grad_sym), getcomplex(@get_grad_nonsym));
    
    write_index(fs, 'tsize_sym', max_dim, max_order, 'int', @(n, a, b) sprintf('%d', tsize_sym([a b])), []);
    write_index(fs, 'tsize_nonsym', max_dim, max_order, 'int', @(n, a, b) sprintf('%d', tsize_nonsym([a b])), []);
    
    for dim=1:max_dim
        for order=0:max_order
            W=ipweight([dim order]);
            name=tenname('ipweight', dim, order);
            write_vector(fs(1), name, W);
            name=tenname('ipweightcomplex', dim, order);
            write_vector(fs(1), name, reshape([W, W]', 2*length(W), 1));
        end
    end
    write_index(fs, 'ipweight', max_dim, max_order, 'FNumber*', @tenname, []);
    write_index(fs, 'ipweightcomplex', max_dim, max_order, 'FNumber*', @tenname, []);
    
    fprintf(fs(2), '#endif /* _LUT_H */\n\n');
    
    fclose(fs(1));
    fclose(fs(2));
    
end


function name=tenname(n, a, b)
    name=sprintf('%s_%d_%d', n, a, b);
end

function name=refer_tenname(n, a, b)
    name=['&', tenname(n, a, b)];
end

function write_index(fs, name, a, b, type, fn, nullmap)
    fprintf(fs(2), 'extern %s %s[MAX_DIM][MAX_ORDER+1];\n\n', type, name);
    
    do_write_index(fs(1), name, a, b, type, fn, nullmap);
end

function do_write_index(f, name, max_dim, max_order, type, fn, nullmap)
    fprintf(f, '%s %s[MAX_DIM][MAX_ORDER+1] = {\n', type, name);
    
    comma1='';
    for dim=1:max_dim
        fprintf(f, [comma1, '    {']);
        comma='';
        for order=0:max_order
            if ~isempty(nullmap) && ~nullmap(dim, order+1)
                fwrite(f, [comma, 'NULL']);
            else
                fwrite(f, [comma, fn(name, dim, order)]);
            end
            comma=', ';
        end
        fprintf(f, '}');
        comma1=',\n';
    end
    fprintf(f, '\n};\n\n');
end

function write_sparse_matrix(f, name, M, staticstr)
    s=size(M);
    [jc, ir, pr]=get_sparse(M);
    
    fprintf(f, 'static size_t %s_jc[]=', name);
    do_write_vec(f, '%d', jc);
    fprintf(f, ';\n');
    
    fprintf(f, 'static size_t %s_ir[]=', name);
    do_write_vec(f, '%d', ir);
    fprintf(f, ';\n');
    
    fprintf(f, 'static FNumber %s_pr[]=', name);
    do_write_vec(f, '%.99g', pr);
    fprintf(f, ';\n');
    
    fprintf(f, '%sSparseMatrix %s = {%s_jc, %s_ir, %s_pr, %d, %d, %d};\n\n', ...
            staticstr, name, name, name, name, s(1), s(2), length(ir));
end


function write_vector(f, name, V)
    fprintf(f, 'static FNumber %s[]=', name);
    do_write_vec(f, '%d', V);
    fprintf(f, ';\n\n');
end

function do_write_vec(f, t, v)
    fprintf(f, '{');
    comma='';
    for i=1:length(v)
        fprintf(f, [comma, t], v(i));
        comma=', ';
    end
    fprintf(f, '}');
end

function res=decide(f, t, v)
    if v
        res=t;
    else
        res=f;
    end
end

function rule=make_diffrule(rule, reverse)
    rule=sprintf('diffrule_%s%c_%c', ...
                 decide('', 'reverse_', reverse), ...
                 rule.alg, ...
                 decide('0', '1', rule.bdry));
end

function res=commasep(prefix, c, fn)
    if nargin<3
        fn=@(x) x;
    end
    comma='';
    res='';
    for i=1:length(c)
        res=[res, comma, prefix, fn(c{i})];
        comma=', ';
    end
end

function do_write_calcdiff(f, name, M, dim, coords)
    needed=[];
    
    for j=1:size(M, 2)
        needed(j)=(sum(abs(M(:, j)))~=0);
    end
    
    assert(length(coords)==dim);
    
    srcdim=size(M,2)/dim;
    for i=1:length(coords)
        c=coords{i};
        fprintf(f, '#define %s_FILL%s(vec%s, where)\\\n', name, c, c);
        fprintf(f, '    {\\\n');
        for j=1:srcdim
            k=(i-1)*srcdim+j;
            if needed(k)
                fprintf(f, '        _val%d = vec%s(%d, where);\\\n', k-1, c, j-1);
            end
        end
        fprintf(f, '    }\n');
    end
    
    fprintf(f, '#define %s(thisres, %s, %s, %s)\\\n', name, ...
            commasep('', coords), commasep('max', coords), ...
            commasep('vec', coords));
    fprintf(f, '    {\\\n');
    vals=commasep('_val', num2cell(find(needed)-1), @num2str);
    fprintf(f, '        FNumber %s;\\\n', vals);
    for i=1:length(coords)
        c=coords{i};
        fprintf(f, '        SPLIT_OPTIM(%s, max%s, %s_FILL%s, vec%s);\\\n', c, c, name, c, c);
    end
    
    for i=1:size(M, 1)
        fprintf(f, '        DIFF_ASSIGN(thisres(%d), ', i-1);
        plus='';
        for j=1:size(M, 2)
            if M(i, j)~=0
                if M(i,j)==1
                    fprintf(f, '%s_val%d', plus, j-1);
                elseif M(i,j)==-1
                    fprintf(f, '%s(-_val%d)', plus, j-1);
                else
                    fprintf(f, '%sFNUMBER(%.99g)*_val%d', plus, M(i, j), j-1);
                end
                plus=' + ';
            end
        end
        if isempty(plus)
            fprintf(f, '0');
        end
        fprintf(f, ');\\\n');
    end
    fprintf(f, '    }\n');
end

function write_diff_calcdiff(f, name, dim, rule, M)
    calcdiff=['CALCDIFF_', name];
    
    % I don't know why this is needed for some versions of matlab.
    M=full(M);
    
    if dim==1
        do_write_calcdiff(f, calcdiff, M, dim, {'x'});
    elseif dim==2
        do_write_calcdiff(f, calcdiff, M, dim, {'x', 'y'});
    elseif dim==3
        do_write_calcdiff(f, calcdiff, M, dim, {'x', 'y', 'z'});
    end
end

function write_diff_worker(f, name, dim, rule, M)
    fprintf(f, '#define CALCDIFF%d CALCDIFF_%s\n', dim, name);
    fprintf(f, '#define DIFFRULE %s\n', rule);
    fprintf(f, '#define VOLUME_DIM %d\n', dim);
    fprintf(f, '#define DIM %d\n', size(M, 2)/dim);
    fprintf(f, '#define RESDIM %d\n', size(M, 1));
    fprintf(f, '#define WORKER CAT(worker_%s_%s, VOLUME_SIZE_NAME)\n', rule, name);
    fprintf(f, '#include "diff_worker_template.h"\n\n');
    fprintf(f, '#undef CALCDIFF%d\n', dim);
    fprintf(f, '#undef DIFFRULE\n', rule);
    fprintf(f, '#undef VOLUME_DIM\n');
    fprintf(f, '#undef DIM\n');
    fprintf(f, '#undef RESDIM\n');
    fprintf(f, '#undef WORKER\n');
    fprintf(f, '\n');
end


function res=wrapvolume(n)
    res=sprintf('CAT(%s, VOLUME_SIZE_NAME)', n);
end

function do_write_diff_selector(f, max_dim, max_order, rules, workers, sp)
    fprintf(f, '        if(0 < dim && dim<=MAX_DIM && 0 <= order && order<=MAX_ORDER){\n');
    els='';
    for i=1:length(workers)
        fprintf(f, '            %sif(alg==''%c'' && bdry==%d)\n', els, rules{i}.alg, rules{i}.bdry);
        fprintf(f, '            {\n');
        fprintf(f, '                *pm=%s[dim-1][order];\n', sp);
        fprintf(f, '                return %s[dim-1][order];\n', wrapvolume(workers{i}));
        fprintf(f, '            }\n');
        els='else ';
    end
    fprintf(f, '        }\n');
end

function workers=write_selector_index(f, max_dim, max_order, rulenames, name, nullmap)
    workers={};
    for i=1:length(rulenames)
        workers{i}=sprintf('worker_%s_%s', rulenames{i}, name);
        do_write_index(f, wrapvolume(workers{i}), ...
                       max_dim, max_order, 'static DiffWorker*', ...
                       @(n, a, b) wrapvolume(tenname(workers{i}, a, b)), ...
                       nullmap);
    end
end


function write_diffop(fsp, fd, dim, order, rulenames, name, getop)
    D=getop([dim, order]);
    tn=tenname(name, dim, order);
    write_sparse_matrix(fsp, ['diff_matrix_', tn], D, 'static ');
    for i=1:length(rulenames)
        write_diff_calcdiff(fd, tn, dim, rulenames{i}, D);
        write_diff_worker(fd, tn, dim, rulenames{i}, D);
    end
end

function write_diff(fs, max_dim, max_order, rules, orderstart, name, getopsym, getopnonsym)
    rulenames={};
    for i=1:length(rules)
        rulenames{i}=make_diffrule(rules{i}, orderstart==0);
    end

    f=fopen(['../src/diff_workers_', name, '.c'], 'w');
    fprintf(f, '/* Automatically generated by write_lut.m */\n\n');
    fprintf(f, '#include "diff_rules.h"\n');
    fprintf(f, '#include "diff_workers.h"\n\n');
    orders=orderstart:(max_order-1+orderstart);
    nullmap=zeros(max_dim, max_order+1);
    for dim=1:max_dim
        for order=orders
            nullmap(dim, order+1)=1;
            write_diffop(fs(1), f, dim, order, rulenames, ...
                         [name, '_sym'], getopsym);
            write_diffop(fs(1), f, dim, order, rulenames, ...
                         [name, '_nonsym'], getopnonsym);
        end
    end
    spsym=['diff_matrix_', name, '_sym'];
    spnonsym=['diff_matrix_', name, '_nonsym'];
    
    write_index(fs, spsym, max_dim, max_order, 'SparseMatrix*', ...
                @refer_tenname, nullmap);
    write_index(fs, spnonsym, max_dim, max_order, 'SparseMatrix*', ...
                @refer_tenname, nullmap);

    wsym=write_selector_index(f, max_dim, max_order, rulenames, ...
                              [name, '_sym'], nullmap);
    wnonsym=write_selector_index(f, max_dim, max_order, rulenames, ...
                                 [name, '_nonsym'], nullmap);
    
    fprintf(f, 'DiffWorker *%s(const Sig *sig, FNumber alpha, FNumber beta, int bdry, char alg, const SparseMatrix **pm)\n', ...
            wrapvolume(['diff_select_', name]));
    fprintf(f, '{\n');
    fprintf(f, '    int dim=sig->dim, order=sig->order;\n');
    fprintf(f, '    *pm=NULL;\n');
    fprintf(f, '    if(sig->symmetric){\n');
    do_write_diff_selector(f, max_dim, max_order, rules, wsym, spsym);
    fprintf(f, '    }else{\n');
    do_write_diff_selector(f, max_dim, max_order, rules, wnonsym, spnonsym);
    fprintf(f, '    }\n');
    fprintf(f, '    return NULL;\n');
    fprintf(f, '}\n\n');

    fclose(f);
end

function res=premult_transposition(D, dim)
    n=size(D, 1)/dim;
    v=1:n;
    res=[];
    for i=0:(dim-1)
        res=[res, D(i*n+v, :)];
    end
end

function A=blkdiagN(M, n)
    [r,c]=size(M);
    A=zeros(r*n, c*n);
    for i=0:(n-1)
        A(i*r+(1:r), i*c+(1:c))=M;
    end
end

% -symdiv(symmetric tensor)
function D=get_negsymdiv_sym(sig)
    D=calc_SYMPRES(orderdown(sig))*get_negdiv_sym(sig);
    %[div, divc]=calc_DIVsym(sig);
    %D=-div;
end

% -div(symmetric tensor)
function D=get_negdiv_sym(sig)
    D=get_negdiv_nonsym(sig)*blkdiagN(calc_EXPAND(sig), sig(1));
end

% symgrad(symmetric tensor)
function D=get_symgrad_sym(sig)
    D=calc_SYMPRES(orderup(sig))*get_grad_sym(sig);
    %[div, divc]=calc_DIVsym(orderup(sig));
    %D=premult_transposition(-divc, sig(1));
end

% grad(symmetric tensor)
function D=get_grad_sym(sig)
    D=get_grad_nonsym(sig)*blkdiagN(calc_EXPAND(sig), sig(1));
end

% -symdiv(general tensor)
function D=get_negsymdiv_nonsym(sig)
    D=calc_SYMPRES(orderdown(sig))*get_negdiv_nonsym(sig);
end

% -div(general tensor)
function D=get_negdiv_nonsym(sig)
    D=-calc_DIVnonsym(sig);
end

% symgrad(general tensor)
function D=get_symgrad_nonsym(sig)
    D=calc_SYMPRES(orderup(sig))*get_grad_nonsym(sig);
end

% grad(general tensor)
function D=get_grad_nonsym(sig)
    mdiv=get_negdiv_nonsym(orderup(sig));
    D=premult_transposition(mdiv', sig(1));
end

function D=do_getcomplex(fun, sig)
    D1=fun(sig);
    %dim=sig(1);
    %n=size(D1, 2)/dim;
    %v=1:n;
    %D=[];
    %Z=zeros(size(D1, 1), n);
    %for i=0:(dim-1)
    %    Di=D1(:, i*n+v);
    %    Rp=[Di; Z];
    %    Ip=[Z; Di];
    %    D=[D, Rp, Ip];
    %end

    %  D1 = a b  =>  D = a 0 b 0
    %       c d          0 a 0 b
    %                    c 0 d 0
    %                    0 c 0 d
    n=size(D1, 1);
    m=size(D1, 2);
    Z=zeros([n m]);
    %  D1 = a b  =>  TMP1 = a 0 b 0
    %       c d             c 0 d 0
    TMP1=reshape([D1; Z], [n 2*m]); % real part
    %  D1 = a b  =>  TMP2 = 0 a 0 b
    %       c d             0 c 0 d
    TMP2=[TMP1(:, end), TMP1(:, 1:end-1)];  % complex part
    D=reshape([TMP1, TMP2]', [2*m 2*n])';
end

function fun=getcomplex(fun)
    fun=@(sig) do_getcomplex(fun, sig);
end
