/*
 * dticode/src/combine.c
 *
 * Written 2012 by Tuomo Valkonen <tuomov@iki.fi>, University of Graz.
 *
 */

#include <libtu/misc.h>
#include <libtu/output.h>
#include "combine.h"
#include "splitvar.h"
#include "cpock_accel.h"
#include "fid_l2alt.h"

static void combined_k(CombinedDual *a, 
                       const CombinedPrimal *b,
                       CombinedParams *cp)
{
    cp->reg->k(COMBINED_REG(a), b, FALSE, cp->regp);
    cp->fid->k(COMBINED_FID(a), COMBINED_U(b), FALSE, cp->fidp);
}

static void combined_kc(CombinedPrimal *a, 
                        const CombinedDual *b,
                        CombinedParams *cp)
{
    cp->reg->kc(a, COMBINED_REG(b), FALSE, FALSE, cp->regp);
    cp->fid->kc(COMBINED_U(a), COMBINED_FID(b), TRUE, cp->fidp);
}

static void combined_fcres(CombinedDual *a, 
                           FNumber sigma, 
                           CombinedParams *cp)
{
    cp->reg->fcres(COMBINED_REG(a), sigma, cp->regp);
    cp->fid->fcres(COMBINED_FID(a), sigma, cp->fidp);
}

static void combined_gres(CombinedPrimal *a, 
                          FNumber sigma, 
                          CombinedParams *cp)
{
    cp->reg->gres(COMBINED_W(a), sigma, cp->regp);
    cp->fid->gres(COMBINED_U(a), sigma, cp->fidp);
}


void combined_pdgap(FNumber *pdgap, 
                    const CombinedPrimal *x, 
                    const CombinedDual *y,
                    const CombinedPrimal *kcy, 
                    const CombinedDual *kx, 
                    CombinedPrimal *xwork, 
                    unsigned int iter,
                    CombinedParams *cp)
{
    FNumber reg_pdgap[2];
    FNumber fid_pdgap[2];
    
    cp->reg->pdgap(reg_pdgap, x, COMBINED_REG(y), kcy, COMBINED_REG(kx), xwork, 
                   iter, cp->regp);
    cp->fid->pdgap(fid_pdgap, COMBINED_U(x), COMBINED_FID(y), COMBINED_U(kcy),
                   COMBINED_FID(kx), COMBINED_U(xwork), iter, cp->fidp);
    
    pdgap[0]=reg_pdgap[0]+fid_pdgap[0];
    pdgap[1]=reg_pdgap[1]+fid_pdgap[1];
}

CombinedDual *combined_alloc_dual(CombinedParams *cp)
{
    CombinedDual *d=alloc_splitvar(2);
    
    COMBINED_FID(d)=cp->fid->alloc_dual(cp->fidp);
    COMBINED_REG(d)=cp->reg->alloc_dual(cp->regp);
    
    return d;
}

CombinedPrimal *combined_alloc_primal(CombinedParams *cp)
{
    CombinedPrimal *d=alloc_splitvar(2);
        
    COMBINED_U(d)=cp->fid->alloc_primal(cp->fidp);
    COMBINED_W(d)=cp->reg->alloc_primal(cp->regp);
    
    return d;
}


CPockFn combined_fn={
    (CPockK*)combined_k,
    (CPockKC*)combined_kc,
    (CPockFCres*)combined_fcres,
    (CPockGres*)combined_gres,
    (CPockPDgap*)combined_pdgap,
};

void combined_init(CombinedPrimal *p0, CombinedDual *d0, 
                   FidInitPrimalStyle initstyle, const void *initparam, 
                   const Fid *fid, FidParams *fidp,
                   const Reg *reg, RegParams *regp)
{
    if(initstyle==FidInitPrimalFile){
        vector_load(COMBINED_U(p0), (const char*)initparam);
    }else if(fid->init_primal==NULL){
        if(initstyle!=FidInitPrimalZero)
            die("Initialisation needs fidelity function support.");
        vector_zero(COMBINED_U(p0));
    }else{
        fid->init_primal(COMBINED_U(p0), initstyle, fidp);
    }
    
    fid->init_dual(COMBINED_FID(d0), COMBINED_U(p0), fidp);
    reg->init_dual_extra(COMBINED_W(p0), COMBINED_REG(d0), COMBINED_U(p0), regp);
    //vector_zero(COMBINED_FID(d0));
    //vector_zero(COMBINED_W(p0));
    //vector_zero(COMBINED_REG(d0));
}

FNumber combined_estimate(Estimate estim,
                          const CombinedPrimal *p,
                          const Fid *fid, FidParams *fidp,
                          const Reg *reg, RegParams *regp)
{
    FNumber fid_estim=fid->estimate(estim, (p ? COMBINED_U(p) : NULL), fidp);
    FNumber reg_estim=reg->estimate(estim, regp);
    
    if(estim==EstimateKNorm){
        //return fid_estim+reg_estim;
        return fn_sqrt(fid_estim*fid_estim+reg_estim*reg_estim);
    }else if(estim==EstimateGUniformConvexity || estim==EstimateFCUniformConvexity){
        return fn_min(fid_estim, reg_estim);
    }else{
        die("Invalid estimate type");
        return 0;
    }
}                          

void combine_do_cpock(CPockParam *cpockp,
                      const Fid *fid, FidParams *fidp,
                      const Reg *reg, RegParams *regp,
                      FidInitPrimalStyle initstyle, 
                      const void *initparam,
                      CombinedPrimal **pp,
                      CombinedDual **dp)
{
    CombinedParams cp;
    CombinedPrimal *p, *p0, *ptmp;
    CombinedDual *d, *d0, *dtmp;
    FidPrimal *res;
    
    cp.reg=reg;
    cp.regp=regp;
    cp.fid=fid;
    cp.fidp=fidp;
    
    p=combined_alloc_primal(&cp);
    p0=combined_alloc_primal(&cp);
    ptmp=combined_alloc_primal(&cp);
    d=combined_alloc_dual(&cp);
    d0=combined_alloc_dual(&cp);
    dtmp=combined_alloc_dual(&cp);

    // Init
    combined_init(p0, d0, initstyle, initparam, fid, fidp, reg, regp);
    
    cpockp->knorm=combined_estimate(EstimateKNorm, p0, fid, fidp, reg, regp);
    cpockp->gamma=fn_max(combined_estimate(EstimateGUniformConvexity, p0, fid, fidp, reg, regp),
                         combined_estimate(EstimateFCUniformConvexity, p0, fid, fidp, reg, regp));
    
    // run the algorithm
    cpock_accel((Vector*)p, (Vector*)d, (Vector*)p0,
                (Vector*)d0, (Vector*)ptmp, (Vector*)dtmp, 
                cpockp, &combined_fn, &cp);
    
    // Pick result, and prevent it from being unallocated

    if(pp!=NULL)
        *pp=p;
    else    
        destroy_obj((Obj*)p);

    if(dp!=NULL)
        *dp=d;
    else    
        destroy_obj((Obj*)dp);
        
    destroy_obj((Obj*)p0);
    destroy_obj((Obj*)d0);
    
    destroy_obj((Obj*)ptmp);
    destroy_obj((Obj*)dtmp);
}

void noop()
{
}

void init_primal_zero(FidPrimal *p, const FidParams *fidp)
{
    vector_zero(p);
}
