/*
 * dticode/src/complexx.c
 *
 * Written 2012-2013 by Tuomo Valkonen <tuomov@iki.fi>, University of Cambridge.
 *
 */

#include <complex.h>
#include <assert.h>

#include "number.h"
#include "tensor.h"
#include "complexx.h"

void construct_complex(TensorField *y, 
                       const TensorField *rho, 
                       const TensorField *phi,
                       bool addto_y)
{                              
    assert(y->nelem==rho->nelem && y->nelem==phi->nelem);
    assert(y->sig.is_complex && y->elemsize==2);
    assert(!rho->sig.is_complex && rho->elemsize==1);
    assert(!phi->sig.is_complex && phi->elemsize==1);
    
    const FNumber *rho_p=rho->p;
    const FNumber *phi_p=phi->p;
    FNumber *y_p=y->p;
    const size_t nelem=y->nelem;

    #pragma omp parallel for
    #pragma acc kernels loop deviceptr(rho_p, phi_p, y_p), independent
    for(size_t j=0; j<nelem; j++){
        FNumber_complex a=rho_p[j]*fn_cfromarg(phi_p[j]);
        if(addto_y)
            set_complex(y_p, j, a+get_complex(y_p, j));
        else
            set_complex(y_p, j, a);
    }
}

void construct_complex_2phase(TensorField *y, const TensorField *rho, 
                              const FNumber phi1mult,
                              const TensorField *phi1, 
                              const FNumber phi2mult,
                              const TensorField *phi2,
                              bool addto_y)
{                              
    assert(y->nelem==rho->nelem && y->nelem==phi1->nelem && y->nelem==phi2->nelem);
    assert(y->sig.is_complex && y->elemsize==2);
    assert(!rho->sig.is_complex && rho->elemsize==1);
    assert(!phi1->sig.is_complex && phi1->elemsize==1);
    assert(!phi2->sig.is_complex && phi2->elemsize==1);

    const FNumber *rho_p=rho->p;
    const FNumber *phi1_p=phi1->p;
    const FNumber *phi2_p=phi2->p;
    FNumber *y_p=y->p;
    const size_t nelem=y->nelem;
    
    #pragma omp parallel for
    #pragma acc kernels loop deviceptr(rho_p, phi1_p, phi2_p, y_p), independent
    for(size_t j=0; j<nelem; j++){
        FNumber phi=phi1mult*phi1_p[j]+phi2mult*phi2_p[j];
        FNumber_complex a=rho_p[j]*fn_cfromarg(phi);
        if(addto_y)
            set_complex(y_p, j, a+get_complex(y_p, j));
        else
            set_complex(y_p, j, a);
    }
}


void decompose_complex(TensorField *rho, TensorField *phi, 
                       const TensorField *y)
{                              
    FNumber *rho_p=rho->p;
    FNumber *phi_p=phi->p;
    const FNumber *y_p=y->p;
    const size_t nelem=y->nelem;
    
    #pragma omp parallel for
    #pragma acc kernels loop deviceptr(rho_p, phi_p, y_p), independent
    for(size_t j=0; j<nelem; j++){
        FNumber_complex a=get_complex(y_p, j);
        rho_p[j]=fn_cabs(a);
        phi_p[j]=fn_carg(a);
    }
}

void decompose_complex_2phase(TensorField *rho1, 
                              TensorField *rho2, 
                              const FNumber phi1mult1,
                              const FNumber phi1mult2,
                              TensorField *phi1,
                              const FNumber phi2mult1,
                              const FNumber phi2mult2,
                              TensorField *phi2,
                              const TensorField *y1, 
                              const TensorField *y2)
{
    const FNumber div=phi1mult1*phi2mult2-phi2mult1*phi1mult2;
    
    assert(y1->nelem==y2->nelem);

    FNumber *rho1_p=rho1->p;
    FNumber *rho2_p=rho2->p;
    FNumber *phi1_p=phi1->p;
    FNumber *phi2_p=phi2->p;
    const FNumber *y1_p=y1->p;
    const FNumber *y2_p=y2->p;
    const size_t nelem=y1->nelem;
    
    #pragma omp parallel for
    #pragma acc kernels loop deviceptr(rho1_p, rho2_p, phi1_p, phi2_p, y1_p, y2_p), independent
    for(size_t j=0; j<nelem; j++){
        FNumber_complex a=get_complex(y1_p, j);
        FNumber_complex b=get_complex(y2_p, j);
        rho1_p[j]=fn_cabs(a);
        rho2_p[j]=fn_cabs(b);
        // A=[a b; c d] => A^{-1}=[d -b; -c a]/[ad-bc];
        phi1_p[j]=( phi2mult2*fn_carg(a)-phi2mult1*fn_carg(b))/div;
        phi2_p[j]=(-phi1mult2*fn_carg(a)+phi1mult1*fn_carg(b))/div;
    }
}

void apply_grad_construct_complex(TensorField *y,
                                  const TensorField *rho, 
                                  const TensorField *phi,
                                  const TensorField *drho,
                                  const TensorField *dphi,
                                  bool add)
{
    const FNumber *rho_p=rho->p;
    const FNumber *phi_p=phi->p;
    const FNumber *drho_p=drho->p;
    const FNumber *dphi_p=dphi->p;
    FNumber *y_p=y->p;
    const size_t nelem=y->nelem;
    
    #pragma omp parallel for
    #pragma acc kernels loop deviceptr(rho_p, phi_p, drho_p, dphi_p, y_p), independent
    for(size_t j=0; j<nelem; j++){
        FNumber_complex a=(drho_p[j] + I*dphi_p[j]*rho_p[j])*fn_cfromarg(phi_p[j]);
        if(add)
            a+=get_complex(y_p, j);
        set_complex(y_p, j, a);
    }
}                                  

void apply_grad_construct_complex_2phase(TensorField *y,
                                         const TensorField *rho, 
                                         const FNumber phi1mult,
                                         const TensorField *phi1,
                                         const FNumber phi2mult,
                                         const TensorField *phi2,
                                         const TensorField *drho,
                                         const TensorField *dphi1,
                                         const TensorField *dphi2,
                                         bool add)
{
    const FNumber *rho_p=rho->p;
    const FNumber *phi1_p=phi1->p;
    const FNumber *phi2_p=phi2->p;
    const FNumber *drho_p=drho->p;
    const FNumber *dphi1_p=dphi1->p;
    const FNumber *dphi2_p=dphi2->p;
    FNumber *y_p=y->p;
    const size_t nelem=y->nelem;
    
    #pragma omp parallel for
    #pragma acc kernels loop deviceptr(rho_p, phi1_p, phi2_p, drho_p, dphi1_p, dphi2_p, y_p), independent
    for(size_t j=0; j<nelem; j++){
        FNumber phi=phi1mult*phi1_p[j]+phi2mult*phi2_p[j];
        FNumber dphi=phi1mult*dphi1_p[j]+phi2mult*dphi2_p[j];
        FNumber_complex a=(drho_p[j] + I*dphi*rho_p[j])*fn_cfromarg(phi);
        if(add)
            a+=get_complex(y_p, j);
        set_complex(y_p, j, a);
    }
}                                  

void apply_gradc_construct_complex(TensorField *drho, 
                                   TensorField *dphi, 
                                   const TensorField *rho, 
                                   const TensorField *phi,
                                   const TensorField *y,
                                   bool add)
{
    const FNumber *rho_p=rho->p;
    const FNumber *phi_p=phi->p;
    FNumber *drho_p=drho->p;
    FNumber *dphi_p=dphi->p;
    const FNumber *y_p=y->p;
    const size_t nelem=y->nelem;
    
    #pragma omp parallel for
    #pragma acc kernels loop deviceptr(rho_p, phi_p, drho_p, dphi_p, y_p), independent
    for(size_t j=0; j<nelem; j++){
        // (dre, dim)=[grad G(rho, phi)] (drho, dphi)
        // dre = drho cos(phi) - dphi rho sin(phi)
        // dim = drho sin(phi) + dphi rho cos(phi)
        // ==> 
        // <re, drho cos(phi) - dphi rho sin(phi)> + <im, drho sin(phi) + dphi rho cos(phi)>
        // = <re cos(phi) + im sin(phi), drho> + <im rho cos(phi) - re rho sin(phi), dphi>
        FNumber_complex yi=get_complex(y_p, j);
        FNumber cosphi=fn_cos(phi_p[j]);
        FNumber sinphi=fn_sin(phi_p[j]);
        FNumber drhoj=fn_real(yi)*cosphi + fn_imag(yi)*sinphi;
        FNumber dphij=rho_p[j]*(fn_imag(yi)*cosphi - fn_real(yi)*sinphi);
        if(add){
            drho_p[j]+=drhoj;
            dphi_p[j]+=dphij;
        }else{
            drho_p[j]=drhoj;
            dphi_p[j]=dphij;
        }
    }
}

void apply_gradc_construct_complex_2phase(TensorField *drho, 
                                          TensorField *dphi1,
                                          TensorField *dphi2,
                                          const TensorField *rho, 
                                          const FNumber phi1mult,
                                          const TensorField *phi1,
                                          const FNumber phi2mult,
                                          const TensorField *phi2,
                                          const TensorField *y,
                                          bool add_drho, bool add_dphi)
{
    const FNumber *rho_p=rho->p;
    const FNumber *phi1_p=phi1->p;
    const FNumber *phi2_p=phi2->p;
    FNumber *drho_p=drho->p;
    FNumber *dphi1_p=dphi1->p;
    FNumber *dphi2_p=dphi2->p;
    const FNumber *y_p=y->p;
    const size_t nelem=y->nelem;
    
    #pragma omp parallel for
    #pragma acc kernels loop deviceptr(rho_p, phi1_p, phi2_p, drho_p, dphi1_p, dphi2_p, y_p), independent
    for(size_t j=0; j<nelem; j++){
        FNumber_complex yi=get_complex(y_p, j);
        FNumber phi=phi1mult*phi1_p[j]+phi2mult*phi2_p[j];
        FNumber cosphi=fn_cos(phi);
        FNumber sinphi=fn_sin(phi);
        FNumber drhoj=fn_real(yi)*cosphi + fn_imag(yi)*sinphi;
        FNumber dphij=rho_p[j]*(fn_imag(yi)*cosphi - fn_real(yi)*sinphi);
        if(add_drho){
            drho_p[j]+=drhoj;
        }else{
            drho_p[j]=drhoj;
        }
        if(add_dphi){
            dphi1_p[j]+=phi1mult*dphij;
            dphi2_p[j]+=phi2mult*dphij;
        }else{
            dphi1_p[j]=phi1mult*dphij;
            dphi2_p[j]=phi2mult*dphij;
        }
    }
}

void tensorfield_conjugate(TensorField *t, const TensorField *s)
{
    size_t n=t->nelem*t->elemsize/2;
    FNumber *p=t->p;
    FNumber *q=s->p;
    
    assert(t->sig.is_complex && t->elemsize%2==0);
    assert(t->nelem == s->nelem && t->elemsize==s->elemsize);
    
    #pragma omp parallel
    #pragma acc kernels loop deviceptr(p, q)
    for(size_t i=0; i<n; i++){
        set_complex(p, i, fn_conj(get_complex(q, i)));
    }
}
