/*
 * dticode/src/complexx.h
 *
 * Written 2012-2013 by Tuomo Valkonen <tuomov@iki.fi>, University of Cambridge.
 *
 */

#ifndef _COMPLEXX_H
#define _COMPLEXX_H

#include <complex.h>

#include "number.h"
#include "tensor.h"

static __inline__ FNumber_complex get_complex(const FNumber *p, size_t j)
{
    double x_re=p[2*j];
    double x_im=p[2*j+1];
    return x_re+I*x_im;
}

#if _OPENACC
// PGCC OpenACC implementation hates functions
#define set_complex(p, j, a) {p[2*j]=fn_real(a); p[2*j+1]=fn_imag(a);}
#define add_complex(p, j, a) {p[2*j]+=fn_real(a); p[2*j+1]+=fn_imag(a);}
#else
static __inline__ void set_complex(FNumber *p, size_t j, FNumber_complex a)
{
    p[2*j]=fn_real(a);
    p[2*j+1]=fn_imag(a);
}
static __inline__ void add_complex(FNumber *p, size_t j, FNumber_complex a)
{
    p[2*j]+=fn_real(a);
    p[2*j+1]+=fn_imag(a);
}
#endif

static __inline__ FNumber get_real(const FNumber *p, size_t idx)
{
    return p[idx];
}

static __inline__ void set_real(FNumber *p, size_t idx, FNumber v)
{
    p[idx]=v;
}

static __inline__ void add_real(FNumber *p, size_t idx, FNumber v)
{
    p[idx]+=v;
}

static __inline__ FNumber_complex get_complex_conj(const FNumber *p, size_t idx)
{
    return fn_conj(get_complex(p, idx));
}

static __inline__ void set_real_part(FNumber *p, size_t idx, FNumber_complex v)
{
    set_real(p, idx, fn_real(v));
}

static __inline__ void add_real_part(FNumber *p, size_t idx, FNumber_complex v)
{
    add_real(p, idx, fn_real(v));
}

static __inline__ FNumber ciprod(FNumber_complex a, FNumber_complex b)
{
    FNumber a_re=fn_real(a);
    FNumber a_im=fn_imag(a);
    FNumber b_re=fn_real(b);
    FNumber b_im=fn_imag(b);
    return a_re*b_re+a_im*b_im;
}

static __inline__ FNumber cabs2(FNumber_complex a)
{
    return ciprod(a, a);
}

void construct_complex(TensorField *y, const TensorField *rho, 
                       const TensorField *phi, bool addto_y);

void construct_complex_2phase(TensorField *y, const TensorField *rho, 
                              const FNumber phi1mult,
                              const TensorField *phi1, 
                              const FNumber phi2mult,
                              const TensorField *phi2, 
                              bool addto_y);

void decompose_complex(TensorField *rho, TensorField *phi, 
                       const TensorField *y);

void decompose_complex_2phase(TensorField *rho1, 
                              TensorField *rho2, 
                              const FNumber phi1mult1,
                              const FNumber phi1mult2,
                              TensorField *phi1,
                              const FNumber phi2mult1,
                              const FNumber phi2mult2,
                              TensorField *phi2,
                              const TensorField *y1, 
                              const TensorField *y2);

void apply_grad_construct_complex(TensorField *y,
                                  const TensorField *rho, 
                                  const TensorField *phi,
                                  const TensorField *drho,
                                  const TensorField *dphi,
                                  bool add);

void apply_grad_construct_complex_2phase(TensorField *y,
                                         const TensorField *rho, 
                                         const FNumber phi1mult,
                                         const TensorField *phi1,
                                         const FNumber phi2mult,
                                         const TensorField *phi2,
                                         const TensorField *drho,
                                         const TensorField *dphi1,
                                         const TensorField *dphi2,
                                         bool add);

void apply_gradc_construct_complex(TensorField *drho, 
                                   TensorField *dphi, 
                                   const TensorField *rho, 
                                   const TensorField *phi,
                                   const TensorField *y,
                                   bool add);

void apply_gradc_construct_complex_2phase(TensorField *drho, 
                                          TensorField *dphi1,
                                          TensorField *dphi2,
                                          const TensorField *rho, 
                                          const FNumber phi1mult,
                                          const TensorField *phi1,
                                          const FNumber phi2mult,
                                          const TensorField *phi2,
                                          const TensorField *y,
                                          bool add_drho, bool add_dphi);

extern void tensorfield_conjugate(TensorField *t, const TensorField *s);

#endif /* _COMPLEXX_H */
