/*
 * dticode/src/cpock_accel.c
 *
 * Written 2012 by Tuomo Valkonen <tuomov@iki.fi>, University of Graz.
 *
 */

#include <math.h>
#include <stdio.h>
#include <libtu/output.h>
#include "cpock_accel.h"
#include "timing.h"
#include "number.h"
#include "vector.h"
#if _MPI
#include <mpi.h>
#endif

#define SWAP(TYPE, A, B) {TYPE tmp=A; A=B; B=tmp; }

void cpock_accel(Primal *xres, Dual *yres, Primal *x, Dual *y, 
                 Primal *xtmp, Dual *ytmp,
                 const CPockParam *p, 
                 const CPockFn *fn,
                 ModelParams *mp)
{

Primal *xnext=xres;
    Dual *ynext=yres;
    FNumber tau, sigma;
    FNumber theta=1;
    FNumber pdgap[2]={1,0};
    int stop=FALSE;
    unsigned int iter=1;
    FNumber x0norm=vector_norm(x);
    FNumber y0norm=vector_norm(y);

    tau=(p->tau<=0 ? 0.95 : p->tau)/p->knorm;
    sigma=(p->sigma<=0 ? 0.95 : p->sigma)/p->knorm;

#if _MPI
    int rank;
    MPI_Comm_rank(MPI_COMM_WORLD, &rank);
#else
    const int rank=0;
#endif
   
    /* We're trying to minimise needed dual and primal variables to three,
     * without allocating anything during the iterations.
     * Hence the messy of the variables below.
     *
     * For help, we include comments:
     *
     * | OUR_VARIABLE := real contents 
     */

    /* | x := x{0} = xbar{0} */
    /* | y := y{0} */
    
    fn->k(ytmp, x, mp);
    /* | ytmp := K(x{0}) */
    vector_copy(ynext, ytmp);
    /* | ynext := K(x{-1}).
     * Initialising theta=0 would suffice instead,
     * if we ensured ynext!=NaN.
     */

     
    /* Calculate initial pdgap */
    if(p->rho>0 && p->calculate_pdgap0){
        fn->kc(xtmp, y, mp); /* only calculated for initial pdgap */       
fn->pdgap(pdgap, x, y, xtmp, ytmp, xres, 0, mp);
    }
    
    do{
        vector_scaleplus(ynext, ytmp, ynext, (1+theta), -theta, 1);
        /* | ynext := K(xbar{i}) = K((1+theta)x{i} - theta x{i-1}) */
        vector_scaleplus(ynext, y, ynext, 1, sigma, 1);
        /* | ynext := y{i} + sigma K(xbar{i}) */
        fn->fcres(ynext, sigma, mp); /* in-place on ynext */
        /* | ynext := y{i+1} */
        
        fn->kc(xtmp, ynext, mp);
        /* | xtmp := K^c(y{i+1}) */
        vector_scaleplus(xnext, x, xtmp, 1, -tau, 1);
        fn->gres(xnext, tau, mp); /* in-place */
        /* | xnext := x{i+1} */
      
        if(p->gamma>0 && p->gamma!=INFINITY){
            theta=1/fn_sqrt(1+p->gamma*2*tau);
            tau=theta*tau;
            sigma=sigma/theta;
        }else{
            theta=p->theta;
        }
        
        SWAP(Dual*, ytmp, y);
        /* | y := K(x{i}); */
        
        fn->k(ytmp, xnext, mp);
        /* | ytmp := K(x{i+1}) */

        /* Current state:
         * | xnext := x{i+1}     [needed pdgap, next iter.]
         * | ynext := y{i+1}     [needed pdgap, next iter.]
         * | xtmp := K^c(y{i+1}) [needed pdgap]
         * | ytmp := K(x{i+1})   [needed pdgap, next iter. for K(xbar_next).]
         * | x is free
         * | y := K(x{i})        [needed next iter. for K(xbar_next)]
         *
         * We need to swap variables (without copying!) in order to force
         *
         * | ynext := K(x{i})
         * | y := y{i+1}
         * | x := x{i+1}
         */
        SWAP(Dual*, y, ynext);
        SWAP(Primal*, x, xnext);
        
        stop=(iter>=p->maxiter);
        
        if(p->rho>0){
            fn->pdgap(pdgap, x, y, xtmp, ytmp, xnext, iter, mp);
           //stop|=(pdgap[0]/pdgap[1]<=0.0001);
        }
        
        if(rank==0 && p->verbose && (stop || iter%p->verbose_granularity==0)){
            size_t l=printf("[%d/%d] ", iter, p->maxiter);
            printf("[gap: %f, gap0: %f | %f]    \n", pdgap[0], pdgap[1], pdgap[0]/pdgap[1]);
            printf("%*c[solution norm %f, init %f]\n", (int)l, ' ', vector_norm(x), x0norm);
            printf("%*c[dual norm %f, init %f]\n", (int)l, ' ', vector_norm(y), y0norm);
        }

        iter++;
#if _MPI
        /* Rank 0 decides in the end, to be sure that everyone agrees,
         * although everyone calculates iter and pdgap above
         */
        MPI_Bcast(&stop, 1, MPI_INT, 0, MPI_COMM_WORLD);
#endif
    }while(!stop);
    
    /* Due to the swapping above, we may need to copy some data around now */
    if(x!=xres)
        vector_copy(xres, x);
        
    if(y!=yres)
        vector_copy(yres, y);
        
    printf("\n");
}
