/*
 * dticode/src/cpock_nonlin.c
 *
 * Written 2012-2013 by Tuomo Valkonen <tuomov@iki.fi>, University of Graz and 
 *                                                      University of Cambridge.
 *
 */

#include <math.h>
#include <stdio.h>
#include <libtu/output.h>
#include "cpock_nonlin.h"
#include "timing.h"
#include "number.h"
#include "splitvar.h"
#if _MPI
#include <mpi.h>
#endif

#define SWAP(TYPE, A, B) {TYPE tmp=A; A=B; B=tmp; }

void printsplit(Vector *x, int lvl, int minlvl, int maxlvl)
{
    if(lvl>maxlvl){
        printf("*%f", vector_norm(x));
        return;
    }
    
    if(OBJ_IS(x, SplitVar)){
        SplitVar *xs=(SplitVar*)x;
        if(lvl+1>=minlvl)
            printf("[");
        for(size_t i=0; i<xs->n; i++){
            printsplit(xs->p[i], lvl+1, minlvl, maxlvl);
            if(i+1<xs->n && lvl+1>=minlvl)
                printf(", ");
        }
        if(lvl+1>=minlvl)
            printf("]");
    }else if(lvl>=minlvl){
        printf("%f", vector_norm(x));
    }
}

void cpock_nonlin(Primal *xres, Dual *yres, 
                  Primal *x, Dual *y, 
                  Primal *xtmp, Dual *ytmp, 
                  const CPockParam *p, 
                  const CPockNLFn *fn,
                  ModelParams *mp)
{   
    Primal *xnext=xres;
    Dual *ynext=yres;
    FNumber tau, sigma;
    FNumber theta=1;
    FNumber pdgap[2]={1,0};
    int stop=FALSE;
    unsigned int iter=1;
    FNumber accum_step_length=0;
    FNumber x0norm=vector_norm(x);
    FNumber y0norm=vector_norm(y);
    FNumber x0norm0=x0norm;
    NLCPockMode mode=p->mode;
    
    tau=(p->tau<=0 ? 0.95 : p->tau)/p->knorm;
    sigma=(p->sigma<=0 ? 0.95 : p->sigma)/p->knorm;

#if _MPI
    int rank;
    MPI_Comm_rank(MPI_COMM_WORLD, &rank);
#else
    const int rank=0;
#endif

#if DEBUG
    printf("[x0 norm parts: ");
    printsplit(x, 1, 3, 3);
    printf("]\n");
#endif
    
    /* We're trying to minimise needed dual and primal variables to three,
     * without allocating anything during the iterations.
     * Hence the messy of the variables below.
     *
     * For help, we include comments:
     *
     * | OUR_VARIABLE := real contents 
     */

    /* | x := x{0} = xbar{0} */
    /* | y := y{0} */
    
    /* Calculate initial pdgap */
    //if(p->rho>0 && p->calculate_pdgap0){
    //    fn->kc(xtmp, y, mp); /* only calculated for initial pdgap */
    //    fn->pdgap(pdgap, x, y, xtmp, ytmp, xres, 0, mp);
    //}
    
    do{
        FNumber knorm=fn->nl_estimate(EstimateKNorm, x, mp);
        tau=(p->tau<=0 ? 0.95 : p->tau)/knorm;
        sigma=(p->sigma<=0 ? 0.95 : p->sigma)/knorm;

        if(mode==NLCPockInterp){
            fn->nl_k(ytmp, x, mp);
            /* | ytmp := K(x{i}) */
            if(iter>1)
                vector_scaleplus(ynext, ytmp, ynext, (1+theta), -theta, 1);
            else
                vector_copy(ynext, ytmp);
            /* | ynext := K(xbar{i}) = K((1+theta)x{i}) - theta K(x{i-1}) */
        }else if(mode==NLCPockExact){
            vector_scaleplus(xtmp, x/*x{i}*/, xnext/*x{i-1}*/, (1+theta), -theta, 1);
            /* | xtmp := xbar{i}=(1+theta)x{i}-theta x{i-1} */
            fn->nl_k(ynext, xtmp, mp);
            /* | ynext := K(xbar{i}) */
        }else/*if(mode==NLCPockLin)*/{
            fn->nl_k(ynext, xnext, mp);
            /* | ynext := K(x{i-1}) */
            if(iter>1){
                vector_scaleplus(xtmp, x/*x{i}*/, xnext/*x{i-1}*/, (1+theta), -(1+theta), 1);
                /* | xtmp := (1+theta)(x{i}-theta x{i-1}) */
                fn->nl_kag(ytmp, xnext, xtmp, mp);
                /* | ytmp := [grad K](x{i-1})(xbar{i}-x{i-1}) */
                vector_scaleplus(ynext, ytmp, ynext, 1, 1, 1);
                /* | ynext := K(x{i-1})+grad K(x{i-1})(xbar{i}-x{i-1}) */
            }
        }
        
        vector_scaleplus(ynext, y, ynext, 1, sigma, 1);
        /* | ynext := y{i} + sigma Kapprox */
        fn->fcres(ynext, sigma, mp); /* in-place on ynext */
        /* | ynext := y{i+1} */
        
        fn->nl_kc(xtmp, x, ynext, mp);
        /* | xtmp := K^c(y{i+1}) */
        vector_scaleplus(xnext, x, xtmp, 1, -tau, 1);
        fn->gres(xnext, tau, mp); /* in-place */
        /* | xnext := x{i+1} */
        
        if(p->gamma>0 && p->gamma!=INFINITY){
            theta=1/fn_sqrt(1+p->gamma*2*tau);
            tau=theta*tau;
            sigma=sigma/theta;
        }else{
            theta=p->theta;
        }
        
        SWAP(Dual*, ytmp, y);
        /* | y := K(x{i}) IF mode==NLCPockInterp OTHERWISE no effect */
        
        /* Current state:
         * | xnext := x{i+1}     [needed next iter.]
         * | ynext := y{i+1}     [needed next iter.]
         * | xtmp := K^c(y{i+1}) 
         * | ytmp := unused
         * | x := x{i}
         * | y := K(x{i})        [needed next iter. for K(xbar_next) IF mode==NLCPockInterp ]
         *
         * We need to swap variables (without copying!) in order to force
         *
         * | ynext := K(x{i}) IF mode==NLCPockInterp OTHERWISE no effect
         * | y := y{i+1}
         * | x := x{i+1}
         * | xnext := x{i}
         */
        SWAP(Dual*, y, ynext);
        SWAP(Primal*, x, xnext);
        
        stop=(iter>=p->maxiter);
        
        //if(p->rho>0){
        //    fn->pdgap(pdgap, x, y, xtmp, ytmp, xnext, iter, mp);
        //    stop|=(pdgap[0]/pdgap[1]<=p->rho);
        //}
        FNumber step_length=vector_dist(x, xnext);
        accum_step_length+=step_length;
        
        if(x0norm0==0)
            x0norm=fn_max(x0norm, vector_norm(x));
            
        if(rank==0 && p->verbose && (stop || iter%p->verbose_granularity==0)){
            if(x0norm>0 && iter>1)
                stop|=(accum_step_length/(FNumber)p->verbose_granularity < p->rho*x0norm*0.1);
            //printf("[%d/%d] [gap: %f, gap0: %f | %f]    \n", iter, p->maxiter,
            //       pdgap[0], pdgap[1], pdgap[0]/pdgap[1]);
            size_t l=printf("[%d/%d] ", iter, p->maxiter);
            printf("[primal step: this iter %f, accumulated %f, fractional %f]\n", 
                   step_length, accum_step_length, step_length/x0norm);
            printf("%*c[solution norm %f, init %f]\n", (int)l, ' ', vector_norm(x), x0norm);
            accum_step_length=0;
#if DEBUG
            printf("%*c[dual step %f, norm %f, init %f]\n", (int)l, ' ', vector_dist(y, ynext), vector_norm(y), y0norm);
            printf("%*c[Norm, parts: ", (int)l, ' ');
            printsplit(x, 1, 3, 3);
            printf("]\n");
#endif
        }

        iter++;
#if _MPI
        /* Rank 0 decides in the end, to be sure that everyone agrees,
         * although everyone calculates iter and pdgap above
         */
        MPI_Bcast(&stop, 1, MPI_INT, 0, MPI_COMM_WORLD);
#endif
    }while(!stop);
    
    /* Due to the swapping above, we may need to copy some data around now */
    if(x!=xres)
        vector_copy(xres, x);
    if(y!=yres)
        vector_copy(yres, y);
    
    printf("\n");
}
