/*
 * dticode/src/ksreg-nl.c
 *
 * Written 2012-2013 by Tuomo Valkonen <tuomov@iki.fi>, University of Cambridge.
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <libtu/util.h>
#include <libtu/output.h>
#include <libtu/optparser.h>
#include <libtu/misc.h>

#include "parallel.h"
#include "cpock_nonlin.h"
#include "tensor.h"
#include "reg.h"
#include "reg_tdiv.h"
#include "fid.h"
#include "fid_l2alt.h"
#include "fid_l1.h"
#include "fid_oper.h"
#include "fid_split.h"
#include "timing.h"
#include "matrix.h"
#include "diffoper.h"
#include "diffoper_complex.h"
#include "diffoper_phasediff.h"
#include "diffoper_norm.h"
#include "diffoper_splitsame.h"
#include "linoper_id.h"
#include "linoper_sparsemat.h"
#include "method.h"

static OptParserOpt opts[]={
    QUICK_OPT('R', "alpha",   "Weight of regulariser of u"),
    QUICK_OPT('Q', "beta",    "Weight of regulariser of div u"),
    QUICK_OPT('6', "tvdiv",   "TV term weight divisor"),
    QUICK_OPT('7', "huber",   "Huberisation factor"),
    QUICK_OPT('8', "huberdiv", "Huberisation factor of divergence"),
    QUICK_OPT('E', "eps",     "Smoothing of pointwise norm forward operator"),
    QUICK_OPT('1', "infile",  "Input file (raw doubles)"),
    QUICK_OPT('9', "blurmat", "Blur operator file (sparse matrix)"),
    QUICK_OPT('5', "divmat",  "Divergence operator file (sparse matrix)"),
    QUICK_OPT('^', "divmatcount", "Count of divergence operators"),
    QUICK_OPT('O', "order",   "Tensor order"),
    QUICK_OPT('D', "dim",     "Tensor dimension"),
    METHOD_OPTS,
    END_OPTPARSEROPTS
};

#define MAX_OPDIV 4

int main(int argc, char *argv[])
{
    MethodParam mp=DEFAULT_METHOD_PARAM;
    FNumber alpha=1;
    FNumber beta=2;
    FNumber tvdiv=0;
    FNumber huber=0;
    FNumber huberdiv=0;
    FNumber eps=10*FNUMBER_EPS;
    const char *infile=NULL;
    const char *blurfile=NULL;
    const char *divfile=NULL;
    TensorField *f=NULL;
    int opt;
    bool dualdiv=FALSE;
    
    FidNL *fidnl=NULL;
    Fid *fid=NULL;
    FidOperParams fidp;
    Fid *fid_base=NULL;
    FidP fidp_base;
    Fid *fid_base_split[2];
    FidP fidp_base_split[2];
    void *fidp_base_split_p[2]={&(fidp_base_split[0]), &(fidp_base_split[1])};
    Reg *reg;
    RegP regp;
    Reg *reg_split[1+MAX_OPDIV];
    RegP regp_split[1+MAX_OPDIV];
    void *regp_split_p[1+MAX_OPDIV];
    DOper *doper;
    DOper *dops[2];
    LinOper *opblur=NULL, *opdiv[MAX_OPDIV]={NULL,}, *opid2=NULL;
    size_t opdiv_count=0;
    
    for(size_t i=0; i<1+MAX_OPDIV; i++)
        regp_split_p[i]=&regp_split[i];
    
    parallel_init1(&argc, &argv);
    
    libtu_init(argv[0]);
    optparser_init(argc, argv, OPTP_MIDLONG, opts);
    
    while((opt=optparser_get_opt())){
        switch(opt){
            case OPT_ID('R'):
                alpha=atof(optparser_get_arg());
                break;
            case OPT_ID('Q'):
                beta=atof(optparser_get_arg());
                break;
            case OPT_ID('6'):
                tvdiv=atof(optparser_get_arg());
                break;
            case OPT_ID('7'):
                huber=atof(optparser_get_arg());
                break;
            case OPT_ID('8'):
                huberdiv=atof(optparser_get_arg());
                break;
            case OPT_ID('E'):
                eps=atof(optparser_get_arg());
                break;
            case OPT_ID('1'):
                infile=optparser_get_arg();
                break;
            case OPT_ID('9'):
                blurfile=optparser_get_arg();
                break;
            case OPT_ID('5'):
                divfile=optparser_get_arg();
                break;
            case OPT_ID('^'):
                opdiv_count=atoi(optparser_get_arg());
                if(opdiv_count>MAX_OPDIV)
                    die("Too many divergence operators.");
                break;
            case OPT_ID('O'):
                mp.sig.order=atoi(optparser_get_arg());
                break;
            case OPT_ID('D'):
                mp.sig.dim=atoi(optparser_get_arg());
                break;
            default:
                if(!method_do_opt(&mp, opt, opts)){
                    optparser_print_error();
                    return 1;
                }
                break;
        }
    }
    
    
    if(infile==NULL)
        die("infile missing\n");

    if(huberdiv==0)
        huberdiv=huber;
        
    method_params_sanity_check(&mp);
    
    parallel_init2();

    {
        Sig fsig=mp.sig;
        fsig.order=0;
        f=load_tensorfield(&fsig, infile);
    }

    if(blurfile==NULL)
        opblur=(LinOper*)alloc_linoper_id(&f->sig);
    else
        opblur=(LinOper*)load_linoper_sparsemat(blurfile, &f->sig, &f->sig);

    if(divfile!=NULL){
        if(opdiv_count==0){
            opdiv[0]=(LinOper*)load_linoper_sparsemat(divfile, &f->sig, &mp.sig);
            opdiv_count=1;
        }else{
            for(size_t i=0; i<opdiv_count; i++){
                char *tmp;
                libtu_asprintf(&tmp, "%s.%d", divfile, i);
                opdiv[i]=(LinOper*)load_linoper_sparsemat(tmp, &f->sig, &mp.sig);
                free(tmp);
            }
        }
    }
    
    opid2=(LinOper*)alloc_linoper_id(&mp.sig);

    fid_l2alt_init((FidL2AltPrimal*)f, opblur, 
                   &(fid_base_split[0]), &(fidp_base_split[0].l2alt));
    fid_l1_init(NULL, opid2, alpha, huber,
                &(fid_base_split[1]), &(fidp_base_split[1].l1));
                
    fidp_base.split.n=2;
    fidp_base.split.fid=fid_base_split;
    fidp_base.split.params=fidp_base_split_p;
    fid_split_init(&fid_base, &fidp_base.split);
        
    // Init nonlinear operator
    dops[0]=(DOper*)create_diffoper_norm(&mp.sig, eps);
    dops[1]=(DOper*)alloc_linoper_id(&mp.sig);

    doper=(DOper*)create_diffoper_splitsame(dops, 2);
    
    // Init nonlinear fidelity 
    if(mp.gaussnewton){
        fid_oper_init(fid_base, (FidParams*)&fidp_base, 
                      doper, &fid, &fidp);
    }else{
        fid_oper_init_nonlin(fid_base, (FidParams*)&fidp_base, 
                             doper, &fidnl, &fidp);
    }

    // Init regulariser(s)

    if(divfile==NULL){
        reg_tdiv_init(&mp.sig, FALSE, beta, huberdiv, mp.diffalg, &(reg_split[0]), &(regp_split[0].tdiv));
    }else{
        for(size_t i=0; i<opdiv_count; i++)
            reg_operl1_init(opdiv[i], beta, huberdiv, &(reg_split[i]), &(regp_split[i].operl1));
    }
    
    if(tvdiv==0 && opdiv_count<=1){
        reg=reg_split[0];
        regp=regp_split[0];
    }else{
        regp.splitsame.n=opdiv_count;
        regp.splitsame.reg=reg_split;
        regp.splitsame.params=regp_split_p;
        if(tvdiv!=0){
            reg_tv_init(&mp.sig, FALSE, beta/tvdiv, huberdiv, mp.diffalg, &(reg_split[opdiv_count]), &(regp_split[opdiv_count].tv));
            regp.splitsame.n++;
        }
        reg_splitsame_init(&reg, &regp.splitsame);
    }
    
    if(mp.gaussnewton)
        method_do(&mp, fid, &fidp, reg, &regp);
    else
        method_do_nl(&mp, fidnl, &fidp, reg, &regp);

    parallel_finalise();
    
    return 0;
}
