/*
 * dticode/src/diff.c
 *
 * Written 2012 by Tuomo Valkonen <tuomov@iki.fi>, University of Graz.
 *
 */

#include <string.h>
#include <stddef.h>
#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <libtu/output.h>
#if _MPI
#include <limits.h>
#include <mpi.h>
#endif

#include "sym2multi3.h"
#include "tensor.h"
#include "misc.h"
#include "lut.h"
#include "acc_support.h"
#include "number.h"
#include "diff_workers.h"
#include "tensornorm.h"

extern DiffSelector diff_select_symgrad;
extern DiffSelector diff_select_negsymdiv;
extern DiffSelector diff_select_grad;
extern DiffSelector diff_select_negdiv;
extern DiffSelector diff_select_symgradcomplex;
extern DiffSelector diff_select_negsymdivcomplex;
extern DiffSelector diff_select_gradcomplex;
extern DiffSelector diff_select_negdivcomplex;

static void tensorfield_do_diff(TensorField *res, const TensorField *u, 
                                FNumber alpha, FNumber beta,
                                int bdry, char alg,
                                DiffSelector *selector,
                                DiffSelector *selectorcomplex,
                                const char *opername)
{
    const Sig *sig=&(u->sig);
    const SparseMatrix *pm=NULL;
    DiffWorker *worker=(u->sig.is_complex
                        ? selectorcomplex(sig, alpha, beta, bdry, alg, &pm)
                        : selector(sig, alpha, beta, bdry, alg, &pm));
    
    assert(res->nelem==u->nelem);
    assert(res->sig.dim==u->sig.dim);
    
    if(worker==NULL){
        die("No %s worker found for requested tensor type (%d, %d) and domain (%dx%dx%d)",
            opername, sig->dim, sig->order, sig->w, sig->h, sig->d);
    }

    if(res->elemsize!=pm->m || u->elemsize*sig->dim!=pm->n)
        die("Differentiation matrix dimension mismatch");

#if _MPI
    int rank, numtasks;
    MPI_Comm_rank(MPI_COMM_WORLD, &rank);
    MPI_Comm_size(MPI_COMM_WORLD, &numtasks);

    if(numtasks>1){
        size_t slicesize=sig_slice_size(&u->sig);
        size_t myslices=sig_slice_count(&u->sig);
        FNumber * const DEVPTR first=u->p;
        FNumber * const DEVPTR last=u->p+(myslices-1)*slicesize;
        FNumber * const DEVPTR firstm1=u->fullp;
        FNumber * const DEVPTR lastp1=last+slicesize;

        int tag1=1, tag2=2;
        MPI_Request rqsend[2], rqrecv[2];
        int sendcount=0, recvcount=0;
        
#if _OPENACC
        FNumber * const t_first=u->transferbuf;
        FNumber * const t_last=u->transferbuf+slicesize;
        FNumber * const t_firstm1=u->transferbuf+2*slicesize;
        FNumber * const t_lastp1=u->transferbuf+3*slicesize;
        
        veccopy_from_dev(t_first, first, slicesize);
        veccopy_from_dev(t_last, last, slicesize);
#else
        FNumber * const t_first=first;
        FNumber * const t_last=last;
        FNumber * const t_firstm1=firstm1;
        FNumber * const t_lastp1=lastp1;
#endif

        /* receive last+1 slice */
        if(rank<numtasks-1){
            MPI_Irecv(t_lastp1, 1, u->slicetype, rank+1, tag1,
                      MPI_COMM_WORLD, &rqrecv[recvcount]);
            recvcount++;
        }
        /* receive first-1 slice */
        if(rank>0){
            MPI_Irecv(t_firstm1, 1, u->slicetype, rank-1, tag2,
                      MPI_COMM_WORLD, &rqrecv[recvcount]);
            recvcount++;
        }
        /* send first slice */
        if(rank>0){
            MPI_Isend(t_first, 1, u->slicetype, rank-1, tag1,
                      MPI_COMM_WORLD, &rqsend[sendcount]);
            sendcount++;
        }
        /* send last slice */
        if(rank<numtasks-1){
            MPI_Isend(t_last, 1, u->slicetype, rank+1, tag2,
                      MPI_COMM_WORLD, &rqsend[sendcount]);
            sendcount++;
        }

        MPI_Waitall(recvcount, rqrecv, MPI_STATUSES_IGNORE);
        
#if _OPENACC
        veccopy_to_dev(firstm1, t_firstm1, slicesize);
        veccopy_to_dev(lastp1, t_lastp1, slicesize);
#endif

        if(rank==0){
            worker(res->elemsize, u->elemsize, sig->w, sig->h, sig->d+1, 
                   sig->xres, sig->yres, sig->zres,
                   res->p, u->p, alpha, beta, pm);
        }else if(rank==numtasks-1){
            worker(res->elemsize, u->elemsize, sig->w, sig->h, sig->d+1, 
                   sig->xres, sig->yres, sig->zres,
                   res->fullp, u->fullp, alpha, beta, pm);
        }else{
            worker(res->elemsize, u->elemsize, sig->w, sig->h, sig->d+2, 
                   sig->xres, sig->yres, sig->zres,
                   res->fullp, u->fullp, alpha, beta, pm);
        }

        MPI_Waitall(sendcount, rqsend, MPI_STATUSES_IGNORE);
        
    }else
#endif /* _MPI */
    {
        worker(res->elemsize, u->elemsize, sig->w, sig->h, sig->d, 
               sig->xres, sig->yres, sig->zres,
               res->p, u->p, alpha, beta, pm);
    }
}

void tensorfield_symgrad(TensorField *res, const TensorField *u, 
                         FNumber alpha, FNumber beta,
                         int bdry, char alg)
{
    assert(res->sig.order==u->sig.order+1);
    
    tensorfield_do_diff(res, u, alpha, beta, bdry, alg, 
                        diff_select_symgrad, 
                        diff_select_symgradcomplex,
                        "symgrad");
}

void tensorfield_negsymdiv(TensorField *res, const TensorField *u, 
                           FNumber alpha, FNumber beta,
                           int bdry, char alg)
{
    assert(res->sig.order==u->sig.order-1);
    
    tensorfield_do_diff(res, u, alpha, beta, bdry, alg, 
                        diff_select_negsymdiv, 
                        diff_select_negsymdivcomplex,
                        "negsymdiv");
}

void tensorfield_grad(TensorField *res, const TensorField *u, 
                      FNumber alpha, FNumber beta,
                      int bdry, char alg)
{
    assert(res->sig.order==u->sig.order+1);
    
    tensorfield_do_diff(res, u, alpha, beta, bdry, alg, 
                        diff_select_grad,
                        diff_select_gradcomplex,
                        "grad");
}

void tensorfield_negdiv(TensorField *res, const TensorField *u, 
                        FNumber alpha, FNumber beta,
                        int bdry, char alg)
{
    assert(res->sig.order==u->sig.order-1);
    
    tensorfield_do_diff(res, u, alpha, beta, bdry, alg, 
                        diff_select_negdiv,
                        diff_select_negdivcomplex,
                        "negdiv");
}

FNumber calc_l2fact(const Sig *sig, char alg)
{
    FNumber res=fn_min(fn_min(sig->xres, sig->yres), sig->zres);
    FNumber res2=res*res;
    int essdim;
    
    if(sig->dim==3 && sig->d==1){
        essdim=sig->dim-1;
    }else if(sig->dim==2 && sig->h==1 && sig->d==1){
        essdim=sig->dim-1;
    }else{
        essdim=sig->dim;
    }

    if(alg=='f' || alg=='b'){
        return 4*essdim/res2;
    }else if(alg=='c'){
        return essdim/res2;
    }else{
        die("Unsupported differentiation algorithm");
        return 0;
    }
}
