/*
 * dticode/src/diffoper_complex.c
 *
 * Written 2013 by Tuomo Valkonen <tuomov@iki.fi>, University of Cambridge.
 *
 */

#include <assert.h>
#include <libtu/objp.h>
#include <libtu/types.h>
#include <libtu/misc.h>
#include <libtu/output.h>
#include "diffoper_complex.h"
#include "complexx.h"
#include "splitvar.h"
#include "tensornorm.h"
#include "normalise.h"

void get_rho_phi(TensorField **rho, TensorField **phi, const DOperComplexDom *x)
{
    assert(x->n==2);
    
    *rho=(TensorField*)x->p[0];
    *phi=(TensorField*)x->p[1];
}

// calculate y = G(rho, phi) 
// where G(rho, phi) = rho exp(i phi)
// pointwise
void diffoper_complex_calculate(DOperComplex *op, 
                                DOperComplexCod *y, 
                                const DOperComplexDom *x,
                                bool addto_y)
{
    TensorField *rho, *phi;
    
    get_rho_phi(&rho, &phi, x);
    
    construct_complex(y, rho, phi, addto_y);
}

// calculate (rho,phi) s.t. y=G(rho, phi)
void diffoper_complex_backproj(DOperComplex *op,
                               DOperComplexDom *x, 
                               const DOperComplexCod *y)
{
    TensorField *rho, *phi;
    
    get_rho_phi(&rho, &phi, x);
    decompose_complex(rho, phi, y);
}

// calculate dy = [grad G(rho, phi)] (drho, dphi)
//               = (drho + i dphi rho) exp(i phi)
void diffoper_complex_apply_grad(DOperComplex *op,
                                 DOperComplexCod *y, 
                                 const DOperComplexDom *x0, 
                                 const DOperComplexDom *x, 
                                 bool addto_y)
{
    TensorField *rho, *phi, *drho, *dphi;
    
    get_rho_phi(&rho, &phi, x0);
    get_rho_phi(&drho, &dphi, x);

    apply_grad_construct_complex(y, rho, phi, drho, dphi, addto_y);
}

// calculate (drho, dphi) = [grad G(rho, phi)]^* dy
//                        = ???
void diffoper_complex_apply_gradc(DOperComplex *op,
                                  DOperComplexDom *x, 
                                  const DOperComplexDom *x0, 
                                  const DOperComplexCod *y, 
                                  bool addto_x)
{
    TensorField *rho, *phi, *drho, *dphi;
    
    get_rho_phi(&rho, &phi, x0);
    get_rho_phi(&drho, &dphi, x);
    
    apply_gradc_construct_complex(drho, dphi, rho, phi, y, addto_x);
}

DOperComplexDom *diffoper_complex_alloc_dom(DOperComplex *op)
{
    Sig sig=op->sig;
    sig.is_complex=FALSE;
    
    return alloc_splitvar_tensorfields(2, &sig);
}

DOperComplexCod *diffoper_complex_alloc_cod(DOperComplex *op)
{
    Sig sig=op->sig;
    sig.is_complex=TRUE;
    
    return (DOperComplexCod*)alloc_tensorfield(&sig);
}

FNumber diffoper_complex_constraint_scale(DOperComplex *op,
                                          const DOperComplexDom *x)
{
    TensorField *rho, *phi;
    
    get_rho_phi(&rho, &phi, x);
    
    return fn_max(tensorfield_frobinf(rho), tensorfield_frobinf(phi));
}

FNumber diffoper_complex_constraint_conjugate(DOperComplex *op,
                                              const DOperComplexDom *x)
{
    TensorField *rho, *phi;
    
    get_rho_phi(&rho, &phi, x);
    
    return (tensorfield_frob1(rho)+tensorfield_frob1(phi));
}

// Estimate |grad G(rho, phi)|.
// We have pointwise
//    [grad G(rho, phi)] (drho, dphi) = (drho + i dphi rho) exp(i phi)
// so
//    |[grad G(rho, phi)] (drho, dphi)| <= |drho + i dphi rho|
//                                      = sqrt(|drho|^2 + |dphi rho|^2).
//                                      <= max{1, |rho|} sqrt(|drho|^2 + |dphi|^2).
//
// Mmaximising <beta, alpha> with |alpha|_2<=1 and
// beta_x=max{1,|rho_x|}, we get the result |beta|_2, i.e.
//
//   |grad G(rho, phi)| <= |max{1, rho}|_2
//
FNumber diffoper_complex_estimate_grad_norm(DOperComplex *op,
                                            const DOperComplexDom *x)
{
    TensorField *rho, *phi;
    FNumber normsq=0;
    
    get_rho_phi(&rho, &phi, x);
    
    assert(sig_tensor_size(&(rho->sig))==1);
    
    #pragma omp parallel for
    for(size_t j=0; j<rho->nelem; j++){
        FNumber b=fn_max(rho->p[j], 1);
        normsq+=b*b;
    }
    
    return fn_sqrt(normsq);
}

bool diffoper_complex_init(DOperComplex *op, const Sig *sig)
{
    op->sig=*sig;
    return diffoper_init(&op->doper);
}

void diffoper_complex_deinit(DOperComplex *op)
{
    diffoper_deinit(&op->doper);
}

DOperComplex *create_diffoper_complex(const Sig *sig)
{
    CREATEOBJ_IMPL(DOperComplex, diffoper_complex, (p, sig));
}

static DynFunTab diffoper_complex_dynfuntab[]={
    {diffoper_calculate, diffoper_complex_calculate},
    {diffoper_backproj, diffoper_complex_backproj},
    {diffoper_apply_grad, diffoper_complex_apply_grad},
    {diffoper_apply_gradc, diffoper_complex_apply_gradc},
    {(DynFun*)diffoper_alloc_dom, (DynFun*)diffoper_complex_alloc_dom},
    {(DynFun*)diffoper_alloc_cod, (DynFun*)diffoper_complex_alloc_cod},
    {(DynFun*)diffoper_constraint_scale, (DynFun*)diffoper_complex_constraint_scale},
    {(DynFun*)diffoper_constraint_conjugate, (DynFun*)diffoper_complex_constraint_conjugate},
    {(DynFun*)diffoper_estimate_grad_norm, (DynFun*)diffoper_complex_estimate_grad_norm},
    END_DYNFUNTAB,
};

IMPLCLASS(DOperComplex, DOper, diffoper_complex_deinit, diffoper_complex_dynfuntab);
