/*
 * dticode/src/diffoper_compose.c
 *
 * Written 2013 by Tuomo Valkonen <tuomov@iki.fi>, University of Cambridge.
 *
 */

#include <assert.h>
#include <libtu/objp.h>
#include <libtu/types.h>
#include <libtu/misc.h>
#include <libtu/output.h>
#include "linoper.h"
#include "diffoper.h"
#include "diffoper_compose.h"

void diffoper_compose_calculate(DOperCompose *op,
                                DOperCod *y, 
                                const DOperDom *x, 
                                bool addto_y)
{
    diffoper_calculate(op->dop, op->tmp, x, FALSE);
    diffoper_calculate(op->lop, y, op->tmp, addto_y);
}

void diffoper_compose_backproj(DOperCompose *op,
                               DOperDom *x, 
                               const DOperCod *y)
{
    diffoper_backproj(op->lop, op->tmp, y);
    diffoper_backproj(op->dop, x, op->tmp);
}

void diffoper_compose_apply_grad(DOperCompose *op,
                                 DOperCod *y, 
                                 const DOperDom *x0, 
                                 const DOperDom *x, 
                                 bool addto_y)
{
    diffoper_apply_grad(op->dop, op->tmp, x0, x, FALSE);
    if(OBJ_IS(op->lop, LinOper)){
        linoper_calculate((LinOper*)op->lop, y, op->tmp, addto_y);
    }else{
        diffoper_calculate(op->dop, op->tmp2, x0, FALSE);
        diffoper_apply_grad(op->lop, y, op->tmp2, op->tmp, addto_y);
    }
}

void diffoper_compose_apply_gradc(DOperCompose *op,
                                  DOperDom *x, 
                                  const DOperDom *x0, 
                                  const DOperCod *y, 
                                  bool addto_x)
{
    if(OBJ_IS(op->lop, LinOper)){
        linoper_calculatec((LinOper*)op->lop, op->tmp, y, FALSE);
    }else{
        diffoper_calculate(op->dop, op->tmp2, x0, FALSE);
        diffoper_apply_gradc(op->lop, op->tmp, op->tmp2, y, FALSE);
    }
    diffoper_apply_gradc(op->dop, x, x0, op->tmp, addto_x);
}


DOperDom *diffoper_compose_alloc_dom(DOperCompose *op)
{
    return diffoper_alloc_dom(op->dop);
}

DOperCod *diffoper_compose_alloc_cod(DOperCompose *op)
{
    return diffoper_alloc_cod(op->lop);
}

void diffoper_compose_projdom(DOperCompose *op, DOperDom *x)
{
    diffoper_projdom(op->dop, x);
}

FNumber diffoper_compose_constraint_scale(DOperCompose *op, const DOperDom *x)
{
    return diffoper_constraint_scale(op->dop, x);
}

FNumber diffoper_compose_constraint_conjugate(DOperCompose *op, DOperDom *x)
{
    return diffoper_constraint_conjugate(op->dop, x);
}

FNumber diffoper_compose_estimate_grad_norm(DOperCompose *op, const DOperDom *x)
{
    FNumber lopn, dopn;
    
    dopn=diffoper_estimate_grad_norm(op->dop, x);
    if(OBJ_IS(op->lop, LinOper)){
        lopn=linoper_estimate_norm((LinOper*)op->lop);
    }else{
        diffoper_calculate(op->dop, op->tmp2, x, FALSE);
        lopn=diffoper_estimate_grad_norm(op->lop, op->tmp2);
    }
        
    return lopn*dopn;
}

bool diffoper_compose_init(DOperCompose *op, 
                           DOper *lop, DOper *dop)
{
    op->lop=lop;
    op->dop=dop;
    op->tmp=diffoper_alloc_cod(dop);
    if(OBJ_IS(lop, LinOper))
        op->tmp2=NULL;
    else
        op->tmp2=diffoper_alloc_dom(lop);
    
    return diffoper_init(&op->doper);
}

void diffoper_compose_deinit(DOperCompose *op)
{
    destroy_obj((Obj*)op->tmp);
    if(op->tmp2!=NULL)
        destroy_obj((Obj*)op->tmp2);
    diffoper_deinit(&op->doper);
}

DOperCompose *create_diffoper_compose(DOper *lop, DOper *dop)
{
    CREATEOBJ_IMPL(DOperCompose, diffoper_compose, (p, lop, dop));
}

static DynFunTab diffoper_compose_dynfuntab[]={
    {diffoper_calculate, diffoper_compose_calculate},
    {diffoper_backproj, diffoper_compose_backproj},
    {diffoper_apply_grad, diffoper_compose_apply_grad},
    {diffoper_apply_gradc, diffoper_compose_apply_gradc},
    {diffoper_projdom, diffoper_compose_projdom},
    {(DynFun*)diffoper_alloc_dom, (DynFun*)diffoper_compose_alloc_dom},
    {(DynFun*)diffoper_alloc_cod, (DynFun*)diffoper_compose_alloc_cod},
    {(DynFun*)diffoper_constraint_scale, (DynFun*)diffoper_compose_constraint_scale},
    {(DynFun*)diffoper_constraint_conjugate, (DynFun*)diffoper_compose_constraint_conjugate},
    {(DynFun*)diffoper_estimate_grad_norm, (DynFun*)diffoper_compose_estimate_grad_norm},
    END_DYNFUNTAB,
};

IMPLCLASS(DOperCompose, DOper, diffoper_compose_deinit, diffoper_compose_dynfuntab);
