/*
 * dticode/src/diffoper_dti.c
 *
 * Written 2013 by Tuomo Valkonen <tuomov@iki.fi>, University of Cambridge.
 *
 */

#include <assert.h>
#include <libtu/objp.h>
#include <libtu/types.h>
#include <libtu/misc.h>
#include <libtu/output.h>
#include "diffoper_dti.h"
#include "complexx.h"
#include "splitvar.h"
#include "tensornorm.h"
#include "normalise.h"
#include "project_posdef.h"

#include "stejskal-tanner.h"

// calculate s = G(u) = s_0 exp(-<b,u>) pointwise for multiple signals
void diffoper_dti_calculate(DOperDTI *op,
                            DOperDTICod *y, 
                            const DOperDTIDom *x, 
                            bool addto_y)
{
    // #define DTI_CALC_A(y, s0, x, z, bmat, 
    //                    addto_y, yset_type, gets0, getx, getz, C) 
    if(!op->s0){
        DTI_CALC_A(y, NULLTF, x, NULLTF, op->bmat, 
                   addto_y, real, GETONE, get_real, GETONE, 1);
    }else if(!op->s0->sig.is_complex){
        DTI_CALC_A(y, op->s0, x, NULLTF, op->bmat, 
                   addto_y, real, get_real, get_real, GETONE, 1);
    }else{
        DTI_CALC_A(y, op->s0, x, NULLTF, op->bmat, 
                   addto_y, complex, get_complex, get_real, GETONE, 2);
    }
}

// calculate (u) s.t. y=G(u)
void diffoper_dti_backproj(DOperDTI *op,
                           DOperDTIDom *x, 
                           const DOperDTICod *y)
{
    die("diffoper_dti_backproj unimplemented");
}

// calculate grad G(u) du =  s_0 exp(<b, u>) <b, du> pointwise, multiple signals
void diffoper_dti_apply_grad(DOperDTI *op,
                             DOperDTICod *y, 
                             const DOperDTIDom *x0, 
                             const DOperDTIDom *x, 
                             bool addto_y)
{
    // #define DTI_APPLY_GRAD_A(y, s0, x, z, x0, z0, bmat, 
    //                          addto_y, yset_type, gets0, getx, getz, C)
    if(!op->s0){
        DTI_APPLY_GRAD_A(y, NULLTF, x, NULLTF, x0, NULLTF, op->bmat, 
                         addto_y, real, GETONE, get_real, GETONE, 1);
    }else if(!op->s0->sig.is_complex){
        DTI_APPLY_GRAD_A(y, op->s0, x, NULLTF, x0, NULLTF, op->bmat, 
                         addto_y, real, get_real, get_real, GETONE, 1);
    }else{
        DTI_APPLY_GRAD_A(y, op->s0, x, NULLTF, x0, NULLTF, op->bmat, 
                         addto_y, complex, get_complex, get_real, GETONE, 2);
    }
}


// calculate du = [grad G(u)]^* dy =  s_0 exp(<b, u>) (b^* dy) 
//                                 =  b^* (s_0 exp(<b, u>) dy)
// pointwise, multiple signals
void diffoper_dti_apply_gradc(DOperDTI *op,
                              DOperDTIDom *x, 
                              const DOperDTIDom *x0, 
                              const DOperDTICod *y, 
                              bool addto_x)
{
    // #define DTI_APPLY_GRADC_A(x, z, s0, y, x0, z0, bmat, 
    //                           addto_x, xset_type, zset_type, gets0, getx, getz, gety, C) 
    if(!op->s0){
        DTI_APPLY_GRADC_A(x, NULLTF, NULLTF, y, x0, NULLTF, op->bmat, 
                          addto_x, real, nothing, GETONE, get_real, GETONE, get_real, 1);
    }else if(!op->s0->sig.is_complex){
        DTI_APPLY_GRADC_A(x, NULLTF, op->s0, y, x0, NULLTF, op->bmat, 
                          addto_x, real, nothing, get_real, get_real, GETONE, get_real, 1);
    }else{
        DTI_APPLY_GRADC_A(x, NULLTF, op->s0, y, x0, NULLTF, op->bmat, 
                          addto_x, real_part, nothing, get_complex_conj, get_real, GETONE, get_complex, 2);
    }
}

DOperDTIDom *diffoper_dti_alloc_dom(DOperDTI *op)
{
    Sig sig=op->sig;
    sig.order=op->bmat->sigm.order;
    sig.dim=op->bmat->sigm.dim;
    sig.symmetric=op->bmat->sigm.symmetric;
    sig.is_complex=FALSE;

    return alloc_tensorfield(&sig);
}

DOperDTICod *diffoper_dti_alloc_cod(DOperDTI *op)
{
    Sig sig=op->sig;
    sig.order=op->bmat->sign.order;
    sig.dim=op->bmat->sign.dim;
    sig.symmetric=op->bmat->sigm.symmetric;
    sig.is_complex=(op->s0 && op->s0->sig.is_complex);
    
    return alloc_tensorfield(&sig);
}

FNumber diffoper_dti_constraint_scale(DOperDTI *op, const DOperDTIDom *x)
{
    return tensorfield_frobinf(x);
}

FNumber diffoper_dti_constraint_conjugate(DOperDTI *op, const DOperDTIDom *x)
{
    return tensorfield_frob1(x);
}


FNumber diffoper_dti_estimate_grad_norm(DOperDTI *op, const DOperDTIDom *x0)
{
    FNumber s0expmax=0;
    
    // #define DTI_EST_GRAD_NORM(s0expmax, s0expmaxnz, s0, x0, z0, bmat, 
    //                           gets0, getx, getz, abs2, C, DONZ)
    if(!op->s0){
        DTI_EST_GRAD_NORM(s0expmax, unused, NULLTF, x0, NULLTF, op->bmat, 
                          GETONE, get_real, GETONE, fn_pow2, 1, DO_NO);
    }else if(!op->s0->sig.is_complex){
        DTI_EST_GRAD_NORM(s0expmax, unused, op->s0, x0, NULLTF, op->bmat, 
                          get_real, get_real, GETONE, fn_pow2, 1, DO_NO);
    }else{
        DTI_EST_GRAD_NORM(s0expmax, unused, op->s0, x0, NULLTF, op->bmat, 
                          get_complex, get_real, GETONE, fn_cabs2, 2, DO_NO);
    }
    
    return s0expmax*op->bmat_opnorm;
}

bool diffoper_dti_init(DOperDTI *op, 
                       const Sig *sig,
                       const Matrix *bmat,
                       const TensorField *s0)
{
    if(bmat->sign.is_complex || bmat->sigm.is_complex)
        die("Complex DTI B matrix not supported.");
        
    op->sig=*sig;
    op->bmat=bmat;
    op->bmat_opnorm=matrix_opnorm(bmat);
    op->s0=s0;
    
    return diffoper_init(&op->doper);
}

void diffoper_dti_deinit(DOperDTI *op)
{
    diffoper_deinit(&op->doper);
}

DOperDTI *create_diffoper_dti(const Sig *sig,
                              const Matrix *bmat,
                              const TensorField *s0)
{
    CREATEOBJ_IMPL(DOperDTI, diffoper_dti, (p, sig, bmat, s0));
}

static DynFunTab diffoper_dti_dynfuntab[]={
    {diffoper_calculate, diffoper_dti_calculate},
    {diffoper_backproj, diffoper_dti_backproj},
    {diffoper_apply_grad, diffoper_dti_apply_grad},
    {diffoper_apply_gradc, diffoper_dti_apply_gradc},
    {(DynFun*)diffoper_alloc_dom, (DynFun*)diffoper_dti_alloc_dom},
    {(DynFun*)diffoper_alloc_cod, (DynFun*)diffoper_dti_alloc_cod},
    {(DynFun*)diffoper_constraint_scale, (DynFun*)diffoper_dti_constraint_scale},
    {(DynFun*)diffoper_constraint_conjugate, (DynFun*)diffoper_dti_constraint_conjugate},
    {(DynFun*)diffoper_estimate_grad_norm, (DynFun*)diffoper_dti_estimate_grad_norm},
    END_DYNFUNTAB,
};

IMPLCLASS(DOperDTI, DOper, diffoper_dti_deinit, diffoper_dti_dynfuntab);

// Non-negativity constraint.

void diffoper_dti_nonneg_projdom(DOperDTINonNeg *op, DOperDTIDom *x)
{
    tensorfield_project_posdef(x);
}

FNumber diffoper_dti_nonneg_constraint_scale(DOperDTINonNeg *op, const DOperDTIDom *x)
{
    return tensorfield_frobinf(x);
}

FNumber diffoper_dti_nonneg_constraint_conjugate(DOperDTINonNeg *op, DOperDTIDom *x)
{
    diffoper_dti_nonneg_projdom(op, x);
    return tensorfield_frob1(x);
}


bool diffoper_dti_nonneg_init(DOperDTINonNeg *op, 
                              const Sig *sig,
                              const Matrix *bmat,
                              const TensorField *s0)
{
    return diffoper_dti_init(&op->dti, sig, bmat, s0);
}

void diffoper_dti_nonneg_deinit(DOperDTINonNeg *op)
{
    diffoper_dti_deinit(&op->dti);
}

DOperDTINonNeg *create_diffoper_dti_nonneg(const Sig *sig,
                                           const Matrix *bmat,
                                           const TensorField *s0)
{
    CREATEOBJ_IMPL(DOperDTINonNeg, diffoper_dti_nonneg, (p, sig, bmat, s0));
}

static DynFunTab diffoper_dti_nonneg_dynfuntab[]={
    {diffoper_projdom, diffoper_dti_nonneg_projdom},
    {(DynFun*)diffoper_constraint_scale, (DynFun*)diffoper_dti_nonneg_constraint_scale},
    {(DynFun*)diffoper_constraint_conjugate, (DynFun*)diffoper_dti_nonneg_constraint_conjugate},
    END_DYNFUNTAB,
};

IMPLCLASS(DOperDTINonNeg, DOperDTI, diffoper_dti_nonneg_deinit, diffoper_dti_nonneg_dynfuntab);
