/*
 * dticode/src/diffoper_dtis.c
 *
 * Written 2013 by Tuomo Valkonen <tuomov@iki.fi>, University of Cambridge.
 *
 */

#include <assert.h>
#include <libtu/objp.h>
#include <libtu/types.h>
#include <libtu/misc.h>
#include <libtu/output.h>
#include "diffoper_dtis.h"
#include "complexx.h"
#include "splitvar.h"
#include "tensornorm.h"
#include "normalise.h"
#include "project_posdef.h"

#include "stejskal-tanner.h"

static __inline__ void get_dom1_dom2(TensorField **dom1, TensorField **dom2, const DOperDTISDom *x)
{
    assert(x->n==2);
    
    *dom1=(TensorField*)x->p[0];
    *dom2=(TensorField*)x->p[1];
}

#define get_real_z(s0p, ii) (get_real(s0p, ii)*zp[ii])
#define get_complex_z(s0p, ii) (get_complex(s0p, ii)*zp[ii])
#define get_real_z0(s0p, ii) (get_real(s0p, ii)*z0p[ii])
#define get_complex_z0(s0p, ii) (get_complex(s0p, ii)*z0p[ii])
#define get_complex_conj_z0(s0p, ii) fn_conj(get_complex(s0p, ii)*z0p[ii])
#define get_real_y(s0p, ii) (get_real(s0p, ii)*yp__[ii])
#define get_complex_conj_y(s0p, ii) (fn_conj(get_complex(s0p, ii))*get_complex(yp__, ii))

void diffoper_dtis_calculate(DOperDTIS *op,
                             DOperDTISCod *y, 
                             const DOperDTISDom *xz, 
                             bool addto_y)
{
    TensorField *x, *z;
    FNumber *zp;
    
    get_dom1_dom2(&x, &z, xz);
    
    assert(!z->sig.is_complex);
    assert(z->nelem==x->nelem && z->elemsize==1);
    
    zp=z->p;
    
    // #define DTI_CALC_A(y, s0, x, z, bmat, 
    //                    addto_y, yset_type, gets0, getx, getz, C) 
    if(!op->s0){
        DTI_CALC_A(y, NULLTF, x, z, op->bmat, 
                   addto_y, real, GETONE, get_real, get_real, 1);
    }else if(!op->s0->sig.is_complex){
        DTI_CALC_A(y, op->s0, x, z, op->bmat, 
                   addto_y, real, get_real, get_real, get_real, 1);
    }else{
        DTI_CALC_A(y, op->s0, x, z, op->bmat, 
                   addto_y, complex, get_complex, get_real, get_real, 2);
    }
}

void diffoper_dtis_backproj(DOperDTIS *op,
                            DOperDTISDom *x, 
                            const DOperDTISCod *y)
{
    die("diffoper_dtis_backproj unimplemented");
}

void diffoper_dtis_apply_grad(DOperDTIS *op,
                              DOperDTISCod *y, 
                              const DOperDTISDom *xz0, 
                              const DOperDTISDom *xz, 
                              bool addto_y)
{
    TensorField *x, *z, *x0, *z0;
    FNumber *zp, *z0p;
    
    get_dom1_dom2(&x, &z, xz);
    get_dom1_dom2(&x0, &z0, xz0);
    
    assert(!z->sig.is_complex && !z0->sig.is_complex);
    assert(z->nelem==x->nelem && z->elemsize==1);
    assert(z0->nelem==x0->nelem && z0->elemsize==1);

    zp=z->p;
    z0p=z0->p;

    // #define DTI_APPLY_GRAD_A(y, s0, x, z, x0, z0, bmat, 
    //                          addto_y, yset_type, gets0, getx, getz, C)
    if(!op->s0){
        DTI_APPLY_GRAD_A(y, NULLTF, x, z, x0, z0, op->bmat, 
                         addto_y, real, GETONE, get_real, get_real, 1);
    }else if(!op->s0->sig.is_complex){
        DTI_APPLY_GRAD_A(y, op->s0, x, z, x0, z0, op->bmat, 
                         addto_y, real, get_real, get_real, get_real, 1);
    }else{
        DTI_APPLY_GRAD_A(y, op->s0, x, z, x0, z0, op->bmat, 
                         addto_y, complex, get_complex, get_real, get_real, 2);
    }
}


void diffoper_dtis_apply_gradc(DOperDTIS *op,
                               DOperDTISDom *xz, 
                               const DOperDTISDom *xz0, 
                               const DOperDTISCod *y, 
                               bool addto_x)
{
    TensorField *x, *z, *x0, *z0;
    FNumber *zp, *z0p, *yp__;
    
    get_dom1_dom2(&x, &z, xz);
    get_dom1_dom2(&x0, &z0, xz0);
    
    assert(!z->sig.is_complex && !z0->sig.is_complex);
    assert(z->nelem==x->nelem && z->elemsize==1);
    assert(z0->nelem==x0->nelem && z0->elemsize==1);

    zp=z->p;
    z0p=z0->p;
    yp__=y->p;
    
    // #define DTI_APPLY_GRADC_A(x, z, s0, y, x0, z0, bmat, 
    //                           addto_x, xset_type, zset_type, gets0, getx, getz, gety, C) 
    if(!op->s0){
        DTI_APPLY_GRADC_A(x, z, NULLTF, y, x0, z0, op->bmat,
                          addto_x, real, real, GETONE, get_real, get_real, get_real, 1);
    }else if(!op->s0->sig.is_complex){
        DTI_APPLY_GRADC_A(x, z, op->s0, y, x0, z0, op->bmat,
                          addto_x, real, real, get_real, get_real, get_real, get_real, 1);
    }else{
        DTI_APPLY_GRADC_A(x, z, op->s0, y, x0, z0, op->bmat,
                          addto_x, real_part, real_part, get_complex_conj, get_real, get_real, get_complex, 2);
    }
}

DOperDTISDom *diffoper_dtis_alloc_dom(DOperDTIS *op)
{
    SplitVar *sv=alloc_splitvar(2);
    
    Sig sig=op->sig;
    sig.order=op->bmat->sigm.order;
    sig.dim=op->bmat->sigm.dim;
    sig.symmetric=op->bmat->sigm.symmetric;
    sig.is_complex=FALSE;

    sv->p[0]=(Vector*)alloc_tensorfield(&sig);
    if(sv->p[0]==NULL)
        die_err();
    
    sig.order=0;
    sv->p[1]=(Vector*)alloc_tensorfield(&sig);
    if(sv->p[1]==NULL)
        die_err();

    return sv;
}

DOperDTISCod *diffoper_dtis_alloc_cod(DOperDTIS *op)
{
    Sig sig=op->sig;
    sig.order=op->bmat->sign.order;
    sig.dim=op->bmat->sign.dim;
    sig.symmetric=op->bmat->sigm.symmetric;
    sig.is_complex=(op->s0 && op->s0->sig.is_complex);
    
    return alloc_tensorfield(&sig);
}

FNumber diffoper_dtis_constraint_scale(DOperDTIS *op, const DOperDTISDom *xz)
{
    TensorField *x, *z;
    get_dom1_dom2(&x, &z, xz);
    
    return fn_max(tensorfield_frobinf(x), tensorfield_frobinf(z));
}

FNumber diffoper_dtis_constraint_conjugate(DOperDTIS *op, const DOperDTISDom *xz)
{
    TensorField *x, *z;
    get_dom1_dom2(&x, &z, xz);

    return tensorfield_frob1(x)+tensorfield_frob1(z);
}


FNumber diffoper_dtis_estimate_grad_norm(DOperDTIS *op, const DOperDTISDom *xz0)
{
    FNumber s0expmax=0;
    FNumber s0expmaxnz=0;
    TensorField *x0, *z0;
    FNumber *z0p;
    
    get_dom1_dom2(&x0, &z0, xz0);

    assert(!z0->sig.is_complex);
    assert(z0->nelem==x0->nelem && z0->elemsize==1);
    
    z0p=z0->p;
    
    // #define DTI_EST_GRAD_NORM(s0expmax, s0expmaxnz, s0, x0, z0, bmat, 
    //                           gets0, getx, getz, abs2, C, DONZ)
    if(!op->s0){
        DTI_EST_GRAD_NORM(s0expmax, s0expmaxnz, NULLTF, x0, z0, op->bmat, 
                          GETONE, get_real, get_real, fn_pow2, 1, DO_YES);
    }else if(!op->s0->sig.is_complex){
        DTI_EST_GRAD_NORM(s0expmax, s0expmaxnz, op->s0, x0, z0, op->bmat, 
                          get_real, get_real, get_real, fn_pow2, 1, DO_YES);
    }else{
        DTI_EST_GRAD_NORM(s0expmax, s0expmaxnz, op->s0, x0, z0, op->bmat, 
                          get_complex, get_real, get_real, fn_cabs2, 2, DO_YES);
    }
    
    return fn_max(s0expmax*op->bmat_opnorm, s0expmaxnz);
}

bool diffoper_dtis_init(DOperDTIS *op, 
                        const Sig *sig,
                        const Matrix *bmat,
                        const TensorField *s0)
{
    if(bmat->sign.is_complex || bmat->sigm.is_complex)
        die("Complex DTIS B matrix not supported.");
        
    op->sig=*sig;
    op->bmat=bmat;
    op->bmat_opnorm=matrix_opnorm(bmat);
    op->s0=s0;
    
    return diffoper_init(&op->doper);
}

void diffoper_dtis_deinit(DOperDTIS *op)
{
    diffoper_deinit(&op->doper);
}

DOperDTIS *create_diffoper_dtis(const Sig *sig,
                                const Matrix *bmat,
                                const TensorField *s0)
{
    CREATEOBJ_IMPL(DOperDTIS, diffoper_dtis, (p, sig, bmat, s0));
}

static DynFunTab diffoper_dtis_dynfuntab[]={
    {diffoper_calculate, diffoper_dtis_calculate},
    {diffoper_backproj, diffoper_dtis_backproj},
    {diffoper_apply_grad, diffoper_dtis_apply_grad},
    {diffoper_apply_gradc, diffoper_dtis_apply_gradc},
    {(DynFun*)diffoper_alloc_dom, (DynFun*)diffoper_dtis_alloc_dom},
    {(DynFun*)diffoper_alloc_cod, (DynFun*)diffoper_dtis_alloc_cod},
    {(DynFun*)diffoper_constraint_scale, (DynFun*)diffoper_dtis_constraint_scale},
    {(DynFun*)diffoper_constraint_conjugate, (DynFun*)diffoper_dtis_constraint_conjugate},
    {(DynFun*)diffoper_estimate_grad_norm, (DynFun*)diffoper_dtis_estimate_grad_norm},
    END_DYNFUNTAB,
};

IMPLCLASS(DOperDTIS, DOper, diffoper_dtis_deinit, diffoper_dtis_dynfuntab);

// Non-negativity constraint.

void diffoper_dtis_nonneg_projdom(DOperDTISNonNeg *op, DOperDTISDom *xz)
{
    TensorField *x, *z;
    get_dom1_dom2(&x, &z, xz);
    
    tensorfield_project_posdef(x);
}

FNumber diffoper_dtis_nonneg_constraint_scale(DOperDTISNonNeg *op, const DOperDTISDom *xz)
{
    return diffoper_dtis_constraint_scale(&op->dtis, xz);
}

FNumber diffoper_dtis_nonneg_constraint_conjugate(DOperDTISNonNeg *op, DOperDTISDom *xz)
{
    diffoper_dtis_nonneg_projdom(op, xz);
    return diffoper_dtis_constraint_conjugate(&op->dtis, xz);
}


bool diffoper_dtis_nonneg_init(DOperDTISNonNeg *op, 
                               const Sig *sig,
                               const Matrix *bmat,
                               const TensorField *s0)
{
    return diffoper_dtis_init(&op->dtis, sig, bmat, s0);
}

void diffoper_dtis_nonneg_deinit(DOperDTISNonNeg *op)
{
    diffoper_dtis_deinit(&op->dtis);
}

DOperDTISNonNeg *create_diffoper_dtis_nonneg(const Sig *sig,
                                             const Matrix *bmat,
                                             const TensorField *s0)
{
    CREATEOBJ_IMPL(DOperDTISNonNeg, diffoper_dtis_nonneg, (p, sig, bmat, s0));
}

static DynFunTab diffoper_dtis_nonneg_dynfuntab[]={
    {diffoper_projdom, diffoper_dtis_nonneg_projdom},
    {(DynFun*)diffoper_constraint_scale, (DynFun*)diffoper_dtis_nonneg_constraint_scale},
    {(DynFun*)diffoper_constraint_conjugate, (DynFun*)diffoper_dtis_nonneg_constraint_conjugate},
    END_DYNFUNTAB,
};

IMPLCLASS(DOperDTISNonNeg, DOperDTIS, diffoper_dtis_nonneg_deinit, diffoper_dtis_nonneg_dynfuntab);
