/*
 * dticode/src/diffoper_multiply.c
 *
 * Written 2013 by Tuomo Valkonen <tuomov@iki.fi>, University of Cambridge.
 *
 */

#include <assert.h>
#include <libtu/objp.h>
#include <libtu/types.h>
#include <libtu/misc.h>
#include <libtu/output.h>
#include "diffoper_multiply.h"
#include "multiply.h"
#include "complexx.h"
#include "splitvar.h"
#include "tensornorm.h"
#include "normalise.h"


static __inline__ void get_dom1_dom2(TensorField **dom1, TensorField **dom2, const DOperMultiplyDom *x)
{
    assert(x->n==2);
    
    *dom1=(TensorField*)x->p[0];
    *dom2=(TensorField*)x->p[1];
}

// calculate y = G(dom1, dom2) 
// where G(dom1, dom2) = dom1 exp(i dom2)
// pointwise
void diffoper_multiply_calculate(DOperMultiply *op, 
                                 DOperMultiplyCod *y, 
                                 const DOperMultiplyDom *x,
                                 bool addto_y)
{
    TensorField *dom1, *dom2;
    
    get_dom1_dom2(&dom1, &dom2, x);
    
    tensorfield_multiply_pointwise(y, dom1, dom2, addto_y);
    //construct_complex(y, dom1, dom2, addto_y);
}

// calculate (dom1,dom2) s.t. y=G(dom1, dom2)
void diffoper_multiply_backproj(DOperMultiply *op,
                               DOperMultiplyDom *x, 
                               const DOperMultiplyCod *y)
{
    die("diffoper_multiply_backproj unimplemented.");
}

// calculate dy = [grad G(dom1, dom2)] (ddom1, ddom2)
void diffoper_multiply_apply_grad(DOperMultiply *op,
                                  DOperMultiplyCod *y, 
                                  const DOperMultiplyDom *x0, 
                                  const DOperMultiplyDom *x, 
                                  bool addto_y)
{
    TensorField *dom1, *dom2, *ddom1, *ddom2;
    
    get_dom1_dom2(&dom1, &dom2, x0);
    get_dom1_dom2(&ddom1, &ddom2, x);

    tensorfield_multiply_pointwise(y, dom1, ddom2, addto_y);
    tensorfield_multiply_pointwise(y, dom2, ddom1, TRUE);

    //apply_grad_construct_complex(y, dom1, dom2, ddom1, ddom2, addto_y);
}

// calculate (ddom1, ddom2) = [grad G(dom1, dom2)]^* dy
void diffoper_multiply_apply_gradc(DOperMultiply *op,
                                   DOperMultiplyDom *x, 
                                   const DOperMultiplyDom *x0, 
                                   const DOperMultiplyCod *y, 
                                   bool addto_x)
{
    TensorField *dom1, *dom2, *ddom1, *ddom2;
    
    get_dom1_dom2(&dom1, &dom2, x0);
    get_dom1_dom2(&ddom1, &ddom2, x);

    tensorfield_multiply_pointwise_conj(ddom1, dom1, y, addto_x);
    tensorfield_multiply_pointwise_conj(ddom2, dom2, y, addto_x);
}

DOperMultiplyDom *diffoper_multiply_alloc_dom(DOperMultiply *op)
{
    SplitVar *sv=alloc_splitvar(2);
    
    Sig sig=op->sig;
    
    sig.is_complex=op->dom1_is_complex;
    sv->p[0]=(Vector*)alloc_tensorfield(&sig);
    if(sv->p[0]==NULL)
        die_err();
    
    sig.is_complex=op->dom2_is_complex;
    sv->p[1]=(Vector*)alloc_tensorfield(&sig);
    if(sv->p[1]==NULL)
        die_err();
    
    return sv;
}

DOperMultiplyCod *diffoper_multiply_alloc_cod(DOperMultiply *op)
{
    Sig sig=op->sig;
    sig.is_complex=(op->dom1_is_complex || op->dom2_is_complex);
    
    return (DOperMultiplyCod*)alloc_tensorfield(&sig);
}

FNumber diffoper_multiply_constraint_scale(DOperMultiply *op,
                                           const DOperMultiplyDom *x)
{
    TensorField *dom1, *dom2;
    
    get_dom1_dom2(&dom1, &dom2, x);
    
    return fn_max(tensorfield_frobinf(dom1), tensorfield_frobinf(dom2));
}

FNumber diffoper_multiply_constraint_conjugate(DOperMultiply *op,
                                               const DOperMultiplyDom *x)
{
    TensorField *dom1, *dom2;
    
    get_dom1_dom2(&dom1, &dom2, x);
    
    return (tensorfield_frob1(dom1)+tensorfield_frob1(dom2));
}

// Estimate |grad G(dom1, dom2)|.
FNumber diffoper_multiply_estimate_grad_norm(DOperMultiply *op,
                                             const DOperMultiplyDom *x)
{
    TensorField *dom1, *dom2;
    
    get_dom1_dom2(&dom1, &dom2, x);
    
    return fn_max(tensorfield_frobinf(dom1), tensorfield_frobinf(dom2));
}

bool diffoper_multiply_init(DOperMultiply *op, 
                            const Sig *sig,
                            bool dom1_is_complex,
                            bool dom2_is_complex)
{
    op->sig=*sig;
    op->dom1_is_complex=dom1_is_complex;
    op->dom2_is_complex=dom2_is_complex;
    
    return diffoper_init(&op->doper);
}

void diffoper_multiply_deinit(DOperMultiply *op)
{
    diffoper_deinit(&op->doper);
}

DOperMultiply *create_diffoper_multiply(const Sig *sig,
                                        bool dom1_is_complex,
                                        bool dom2_is_complex)
{
    CREATEOBJ_IMPL(DOperMultiply, diffoper_multiply, 
                   (p, sig, dom1_is_complex, dom2_is_complex));
}

static DynFunTab diffoper_multiply_dynfuntab[]={
    {diffoper_calculate, diffoper_multiply_calculate},
    {diffoper_backproj, diffoper_multiply_backproj},
    {diffoper_apply_grad, diffoper_multiply_apply_grad},
    {diffoper_apply_gradc, diffoper_multiply_apply_gradc},
    {(DynFun*)diffoper_alloc_dom, (DynFun*)diffoper_multiply_alloc_dom},
    {(DynFun*)diffoper_alloc_cod, (DynFun*)diffoper_multiply_alloc_cod},
    {(DynFun*)diffoper_constraint_scale, (DynFun*)diffoper_multiply_constraint_scale},
    {(DynFun*)diffoper_constraint_conjugate, (DynFun*)diffoper_multiply_constraint_conjugate},
    {(DynFun*)diffoper_estimate_grad_norm, (DynFun*)diffoper_multiply_estimate_grad_norm},
    END_DYNFUNTAB,
};

IMPLCLASS(DOperMultiply, DOper, diffoper_multiply_deinit, diffoper_multiply_dynfuntab);
