/*
 * dticode/src/diffoper_norm.c
 *
 * Written 2013 by Tuomo Valkonen <tuomov@iki.fi>, University of Cambridge.
 *
 */

#include <assert.h>
#include <libtu/objp.h>
#include <libtu/types.h>
#include <libtu/misc.h>
#include <libtu/output.h>
#include "diffoper_norm.h"
#include "tensornorm.h"
#include "normalise.h"

#define WEIGHT_APPLY(W, X) ((W)*(X))
#define WEIGHT_ONE(W, X) (X)
#define ADDTO(Y, X) Y+=(X)
#define SETTO(Y, X) Y=(X)

#define DO_CALC(WEIGHT, UPD) \
    _Pragma("pragma omp parallel for") \
    _Pragma("pragma acc kernels loop deviceptr(yp,xp,wp), independent") \
    for(size_t i=0; i<nelem; i++){ \
        FNumber tmp=0; \
        for(size_t j=0; j<elemsize; j++){ \
            tmp=tmp+WEIGHT(wp[j], fn_pow2(xp[i*elemsize+j])); \
        } \
        UPD(yp[i], fn_sqrt(eps+tmp)); \
    }

void diffoper_norm_calculate(DOperNorm *op, 
                             DOperNormCod *y, 
                             const DOperNormDom *x,
                             bool addto_y)
{
    size_t nelem=x->nelem;
    size_t elemsize=x->elemsize;
    FNumber * DEVPTR yp=y->p;
    const FNumber * DEVPTR xp=x->p;
    const FNumber * DEVPTR wp=sig_get_weight(&x->sig);
    const FNumber eps=op->eps;

    assert(y->nelem==nelem);
    assert(y->elemsize==1);
    
    if(addto_y){
        if(wp!=NULL){
            DO_CALC(WEIGHT_APPLY, ADDTO);
        }else{
            DO_CALC(WEIGHT_ONE, ADDTO);
        }
    }else{
        if(wp!=NULL){
            DO_CALC(WEIGHT_APPLY, SETTO);
        }else{
            DO_CALC(WEIGHT_ONE, SETTO);
        }
    }
}

void diffoper_norm_backproj(DOperNorm *op,
                            DOperNormDom *x, 
                            const DOperNormCod *y)
{
    die_obj("diffoper_norm_backproj", "Unimplemented");
}

#define DO_APPLY_GRAD(WEIGHT, UPD) \
    _Pragma("pragma omp parallel for") \
    _Pragma("pragma acc kernels loop deviceptr(yp,xp,wp), independent") \
    for(size_t i=0; i<nelem; i++){ \
        FNumber tm0=0, tmp=0; \
        for(size_t j=0; j<elemsize; j++){ \
            tm0=tm0+WEIGHT(wp[j], fn_pow2(x0p[i*elemsize+j])); \
            tmp=tmp+WEIGHT(wp[j], xp[i*elemsize+j]*x0p[i*elemsize+j]); \
        } \
        UPD(yp[i], tmp/fn_sqrt(eps+tm0)); \
    }

void diffoper_norm_apply_grad(DOperNorm *op,
                              DOperNormCod *y, 
                              const DOperNormDom *x0, 
                              const DOperNormDom *x, 
                              bool addto_y)
{
    size_t nelem=x->nelem;
    size_t elemsize=x->elemsize;
    FNumber * DEVPTR yp=y->p;
    const FNumber * DEVPTR xp=x->p;
    const FNumber * DEVPTR x0p=x0->p;
    const FNumber * DEVPTR wp=sig_get_weight(&x->sig);
    const FNumber eps=op->eps;

    assert(y->nelem==nelem && x0->nelem==nelem);
    assert(x0->elemsize==elemsize);
    assert(y->elemsize==1);
    
    if(addto_y){
        if(wp!=NULL){
            DO_APPLY_GRAD(WEIGHT_APPLY, ADDTO);
        }else{
            DO_APPLY_GRAD(WEIGHT_ONE, ADDTO);
        }
    }else{
        if(wp!=NULL){
            DO_APPLY_GRAD(WEIGHT_APPLY, SETTO);
        }else{
            DO_APPLY_GRAD(WEIGHT_ONE, SETTO);
        }
    }
}

#define DO_APPLY_GRADC(WEIGHT, UPD) \
    _Pragma("pragma omp parallel for") \
    _Pragma("pragma acc kernels loop deviceptr(yp,xp,wp), independent") \
    for(size_t i=0; i<nelem; i++){ \
        FNumber tm0=0, tmp=0; \
        for(size_t j=0; j<elemsize; j++){ \
            tm0=tm0+WEIGHT(wp[j], fn_pow2(x0p[i*elemsize+j])); \
        } \
        tm0=fn_sqrt(eps+tm0); \
        for(size_t j=0; j<elemsize; j++){ \
            FNumber tmp=yp[i]*x0p[i*elemsize+j]; \
            UPD(xp[i*elemsize+j], tmp/tm0); \
        } \
    }
    
void diffoper_norm_apply_gradc(DOperNorm *op,
                               DOperNormDom *x, 
                               const DOperNormDom *x0, 
                               const DOperNormCod *y, 
                               bool addto_x)
{
    size_t nelem=x->nelem;
    size_t elemsize=x->elemsize;
    const FNumber * DEVPTR yp=y->p;
    FNumber * DEVPTR xp=x->p;
    const FNumber * DEVPTR x0p=x0->p;
    const FNumber * DEVPTR wp=sig_get_weight(&x->sig);
    const FNumber eps=op->eps;

    assert(y->nelem==nelem && x0->nelem==nelem);
    assert(x0->elemsize==elemsize);
    assert(y->elemsize==1);
    
    if(addto_x){
        if(wp!=NULL){
            DO_APPLY_GRADC(WEIGHT_APPLY, ADDTO);
        }else{
            DO_APPLY_GRADC(WEIGHT_ONE, ADDTO);
        }
    }else{
        if(wp!=NULL){
            DO_APPLY_GRADC(WEIGHT_APPLY, SETTO);
        }else{
            DO_APPLY_GRADC(WEIGHT_ONE, SETTO);
        }
    }
}

DOperNormDom *diffoper_norm_alloc_dom(DOperNorm *op)
{
    Sig sig=op->sig;
    
    return alloc_tensorfield(&sig);
}

DOperNormCod *diffoper_norm_alloc_cod(DOperNorm *op)
{
    Sig sig=op->sig;
    sig.is_complex=FALSE;
    sig.order=0;
    
    return alloc_tensorfield(&sig);
}

FNumber diffoper_norm_constraint_scale(DOperNorm *op,
                                       const DOperNormDom *x)
{
    return tensorfield_frobinf(x);
}

FNumber diffoper_norm_constraint_conjugate(DOperNorm *op,
                                           const DOperNormDom *x)
{
    return tensorfield_frob1(x);
}

// Estimate |grad G(rho, phi)|.
// We have pointwise
//    [grad G(rho, phi)] (drho, dphi) = (drho + i dphi rho) exp(i phi)
// so
//    |[grad G(rho, phi)] (drho, dphi)| <= |drho + i dphi rho|
//                                      = sqrt(|drho|^2 + |dphi rho|^2).
//                                      <= max{1, |rho|} sqrt(|drho|^2 + |dphi|^2).
//
// Mmaximising <beta, alpha> with |alpha|_2<=1 and
// beta_x=max{1,|rho_x|}, we get the result |beta|_2, i.e.
//
//   |grad G(rho, phi)| <= |max{1, rho}|_2
//
FNumber diffoper_norm_estimate_grad_norm(DOperNorm *op,
                                            const DOperNormDom *x)
{
    return 1;
}

bool diffoper_norm_init(DOperNorm *op, const Sig *sig, FNumber eps)
{
    op->sig=*sig;
    op->eps=eps;
    return diffoper_init(&op->doper);
}

void diffoper_norm_deinit(DOperNorm *op)
{
    diffoper_deinit(&op->doper);
}

DOperNorm *create_diffoper_norm(const Sig *sig, FNumber eps)
{
    CREATEOBJ_IMPL(DOperNorm, diffoper_norm, (p, sig, eps));
}

static DynFunTab diffoper_norm_dynfuntab[]={
    {diffoper_calculate, diffoper_norm_calculate},
    {diffoper_backproj, diffoper_norm_backproj},
    {diffoper_apply_grad, diffoper_norm_apply_grad},
    {diffoper_apply_gradc, diffoper_norm_apply_gradc},
    {(DynFun*)diffoper_alloc_dom, (DynFun*)diffoper_norm_alloc_dom},
    {(DynFun*)diffoper_alloc_cod, (DynFun*)diffoper_norm_alloc_cod},
    {(DynFun*)diffoper_constraint_scale, (DynFun*)diffoper_norm_constraint_scale},
    {(DynFun*)diffoper_constraint_conjugate, (DynFun*)diffoper_norm_constraint_conjugate},
    {(DynFun*)diffoper_estimate_grad_norm, (DynFun*)diffoper_norm_estimate_grad_norm},
    END_DYNFUNTAB,
};

IMPLCLASS(DOperNorm, DOper, diffoper_norm_deinit, diffoper_norm_dynfuntab);
