/*
 * dticode/src/diffoper_phasediff.c
 *
 * Written 2013 by Tuomo Valkonen <tuomov@iki.fi>, University of Cambridge.
 *
 */

#include <assert.h>
#include <libtu/objp.h>
#include <libtu/types.h>
#include <libtu/misc.h>
#include <libtu/output.h>
#include "diffoper_phasediff.h"
#include "complexx.h"
#include "splitvar.h"
#include "tensornorm.h"
#include "normalise.h"

static void get_rho1_phi1(TensorField **rho1, TensorField **phi1, 
                          const DOperPhaseDiffDom *x)
{
    assert(x->n==4);
    
    *rho1=(TensorField*)x->p[0];
    *phi1=(TensorField*)x->p[1];
}

static void get_rho2_phi2(TensorField **rho2, TensorField **phi2,
                          const DOperPhaseDiffDom *x)
{
    assert(x->n==4);
    
    *rho2=(TensorField*)x->p[2];
    *phi2=(TensorField*)x->p[3];
}

static void get_rho1_rho2(TensorField **rho1, TensorField **rho2, 
                          const DOperPhaseDiffDom *x)
{
    assert(x->n==4);
    
    *rho1=(TensorField*)x->p[0];
    *rho2=(TensorField*)x->p[2];
}

void get_res_i(TensorField **res, const DOperPhaseDiffDom *y, size_t i)
{
    assert(y->n==2);
    
    *res=(TensorField*)y->p[i];
}


// calculate y = G(rho, phi) 
// where G(rho, phi) = rho exp(i phi)
// pointwise
void diffoper_phasediff_calculate(DOperPhaseDiff *op,
                                  DOperPhaseDiffCod *y, 
                                  const DOperPhaseDiffDom *x, 
                                  bool addto_y)
{
    /*const*/ TensorField *rho1, *phi1, *rho2, *phi2;
    TensorField *res1, *res2;

    get_rho1_phi1(&rho1, &phi1, x);
    get_rho2_phi2(&rho2, &phi2, x);
    
    // First component
    get_res_i(&res1, y, 0);
    construct_complex_2phase(res1, rho1, 
                             op->phi1mult1, phi1, 
                             op->phi2mult1, phi2,
                             addto_y);

    // Second component
    get_res_i(&res2, y, 1);
    construct_complex_2phase(res2, rho2, 
                             op->phi1mult2, phi1, 
                             op->phi2mult2, phi2,
                             addto_y);
}

// calculate (rho,phi) s.t. y=G(rho, phi)
void diffoper_phasediff_backproj(DOperPhaseDiff *op,
                                 DOperPhaseDiffDom *x, 
                                 const DOperPhaseDiffCod *y)
{
    TensorField *rho1, *phi1, *rho2, *phi2;
    /*const*/ TensorField *a1, *a2;

    get_rho1_phi1(&rho1, &phi1, x);
    get_rho2_phi2(&rho2, &phi2, x);
    get_res_i(&a1, y, 0);
    get_res_i(&a2, y, 1);
    
    decompose_complex_2phase(rho1, rho2,
                             op->phi1mult1, op->phi1mult2, phi1,
                             op->phi2mult1, op->phi2mult2, phi2,
                             a1, a2);
}

// calculate dy = [grad G(rho, phi)] (drho, dphi)
//               = (drho + i dphi rho) exp(i phi)
void diffoper_phasediff_apply_grad(DOperPhaseDiff *op,
                                   DOperPhaseDiffCod *y, 
                                   const DOperPhaseDiffDom *x0, 
                                   const DOperPhaseDiffDom *x, 
                                   bool addto_y)
{
    TensorField *res1, *res2;
    /*const*/ TensorField *rho1, *phi1, *rho2, *phi2;
    /*const*/ TensorField *drho1, *dphi1, *drho2, *dphi2;

    get_rho1_phi1(&rho1, &phi1, x0);
    get_rho2_phi2(&rho2, &phi2, x0);
    get_rho1_phi1(&drho1, &dphi1, x);
    get_rho2_phi2(&drho2, &dphi2, x);

    // First component
    get_res_i(&res1, y, 0);
    apply_grad_construct_complex_2phase(res1, rho1, 
                                        op->phi1mult1, phi1, 
                                        op->phi2mult1, phi2, 
                                        drho1, dphi1, dphi2,
                                        addto_y);
    
    // Second component
    get_res_i(&res2, y, 1);
    apply_grad_construct_complex_2phase(res2, rho2, 
                                        op->phi1mult2, phi1, 
                                        op->phi2mult2, phi2, 
                                        drho2, dphi1, dphi2,
                                        addto_y);
}


// calculate (drho, dphi) = [grad G(rho, phi)]^* dy
//                        = ???
void diffoper_phasediff_apply_gradc(DOperPhaseDiff *op,
                                    DOperPhaseDiffDom *x, 
                                    const DOperPhaseDiffDom *x0, 
                                    const DOperPhaseDiffCod *y, 
                                    bool addto_x)
{
    TensorField *a1, *a2;
    /*const*/ TensorField *rho1, *phi1, *rho2, *phi2;
    /*const*/ TensorField *drho1, *dphi1, *drho2, *dphi2;

    get_rho1_phi1(&rho1, &phi1, x0);
    get_rho2_phi2(&rho2, &phi2, x0);
    get_rho1_phi1(&drho1, &dphi1, x);
    get_rho2_phi2(&drho2, &dphi2, x);

    // First component
    get_res_i(&a1, y, 0);
    apply_gradc_construct_complex_2phase(drho1, dphi1, dphi2, rho1,
                                         op->phi1mult1, phi1,
                                         op->phi2mult1, phi2,
                                         a1, addto_x, addto_x);
    
    // Second component
    get_res_i(&a2, y, 1);
    apply_gradc_construct_complex_2phase(drho2, dphi1, dphi2, rho2,
                                         op->phi1mult2, phi1,
                                         op->phi2mult2, phi2,
                                         a2, addto_x, TRUE);
}


DOperPhaseDiffDom *diffoper_phasediff_alloc_dom(DOperPhaseDiff *op)
{
    Sig sig=op->sig;
    sig.is_complex=FALSE;
    
    return alloc_splitvar_tensorfields(4, &sig);
}

DOperPhaseDiffCod *diffoper_phasediff_alloc_cod(DOperPhaseDiff *op)
{
    Sig sig=op->sig;
    sig.is_complex=TRUE;
    
    return alloc_splitvar_tensorfields(2, &sig);
}

FNumber diffoper_phasediff_constraint_scale(DOperPhaseDiff *op,
                                            const DOperPhaseDiffDom *x)
{
    /*const*/ TensorField *rho1, *phi1, *rho2, *phi2;
    
    get_rho1_phi1(&rho1, &phi1, x);
    get_rho2_phi2(&rho2, &phi2, x);
    
    FNumber a=fn_max(tensorfield_frobinf(rho1), tensorfield_frobinf(rho2));
    FNumber b=fn_max(tensorfield_frobinf(phi1), tensorfield_frobinf(phi2));
    return fn_max(a, b);
}

FNumber diffoper_phasediff_constraint_conjugate(DOperPhaseDiff *op,
                                                const DOperPhaseDiffDom *x)
{
    /*const*/ TensorField *rho1, *phi1, *rho2, *phi2;
    
    get_rho1_phi1(&rho1, &phi1, x);
    get_rho2_phi2(&rho2, &phi2, x);
    
    return ( tensorfield_frob1(rho1)+tensorfield_frob1(rho2)
            +tensorfield_frob1(phi1)+tensorfield_frob1(phi2));
}

FNumber diffoper_phasediff_estimate_grad_norm(DOperPhaseDiff *op,
                                              const DOperPhaseDiffDom *x)
{
    TensorField *rho1, *rho2;
    FNumber normsq=0;
    
    get_rho1_rho2(&rho1, &rho2, x);
    
    assert(sig_tensor_size(&(rho1->sig))==1 && sig_tensor_size(&(rho2->sig))==1);
    assert(rho1->nelem==rho2->nelem);
    
    #pragma omp parallel for
    for(size_t j=0; j<rho1->nelem; j++){
        FNumber b1=fn_max(fn_abs(rho1->p[j]), 1);
        FNumber b2=fn_max(fn_abs(rho2->p[j]), 1);
        //normsq+=b1*b1+b2*b2;
        normsq=fn_max(fn_max(normsq, b1*b1), b2*b2);
    }
    
    return fn_sqrt(normsq);
}

bool diffoper_phasediff_init(DOperPhaseDiff *op, const Sig *sig)
{
    op->sig=*sig;
    op->phi1mult1=.5;
    op->phi2mult1=.5;
    op->phi1mult2=-.5;
    op->phi2mult2=.5;
    
    return diffoper_init(&op->doper);
}

void diffoper_phasediff_deinit(DOperPhaseDiff *op)
{
    diffoper_deinit(&op->doper);
}

DOperPhaseDiff *create_diffoper_phasediff(const Sig *sig)
{
    CREATEOBJ_IMPL(DOperPhaseDiff, diffoper_phasediff, (p, sig));
}

static DynFunTab diffoper_phasediff_dynfuntab[]={
    {diffoper_calculate, diffoper_phasediff_calculate},
    {diffoper_backproj, diffoper_phasediff_backproj},
    {diffoper_apply_grad, diffoper_phasediff_apply_grad},
    {diffoper_apply_gradc, diffoper_phasediff_apply_gradc},
    {(DynFun*)diffoper_alloc_dom, (DynFun*)diffoper_phasediff_alloc_dom},
    {(DynFun*)diffoper_alloc_cod, (DynFun*)diffoper_phasediff_alloc_cod},
    {(DynFun*)diffoper_constraint_scale, (DynFun*)diffoper_phasediff_constraint_scale},
    {(DynFun*)diffoper_constraint_conjugate, (DynFun*)diffoper_phasediff_constraint_conjugate},
    {(DynFun*)diffoper_estimate_grad_norm, (DynFun*)diffoper_phasediff_estimate_grad_norm},
    END_DYNFUNTAB,
};

IMPLCLASS(DOperPhaseDiff, DOper, diffoper_phasediff_deinit, diffoper_phasediff_dynfuntab);
