/*
 * dticode/src/diffoper_split.c
 *
 * Written 2013 by Tuomo Valkonen <tuomov@iki.fi>, University of Cambridge.
 *
 */

#include <assert.h>
#include <libtu/objp.h>
#include <libtu/types.h>
#include <libtu/misc.h>
#include <libtu/output.h>
#include "linoper.h"
#include "diffoper.h"
#include "diffoper_split.h"

void diffoper_split_calculate(DOperSplit *op,
                              DOperSplitCod *y, 
                              const DOperSplitDom *x, 
                              bool addto_y)
{
    assert(x->n==op->n_dops);
    assert(y->n==op->n_dops);
    
    #pragma omp parallel for
    for(size_t i=0; i<op->n_dops; i++)
        diffoper_calculate(op->dops[i], y->p[i], x->p[i], addto_y);
}

void diffoper_split_backproj(DOperSplit *op,
                             DOperSplitDom *x, 
                             const DOperSplitCod *y)
{
    assert(x->n==op->n_dops);
    assert(y->n==op->n_dops);
    
    #pragma omp parallel for
    for(size_t i=0; i<op->n_dops; i++)
        diffoper_backproj(op->dops[i], x->p[i], y->p[i]);
}

// calculate grad G(u) du = - s_0 exp(-<b, u>) <b, du> pointwise, multiple signals
void diffoper_split_apply_grad(DOperSplit *op,
                               DOperSplitCod *y, 
                               const DOperSplitDom *x0, 
                               const DOperSplitDom *x, 
                               bool addto_y)
{
    assert(x->n==op->n_dops);
    assert(x0->n==op->n_dops);
    assert(y->n==op->n_dops);
    
    #pragma omp parallel for
    for(size_t i=0; i<op->n_dops; i++)
        diffoper_apply_grad(op->dops[i], y->p[i], x0->p[i], x->p[i], addto_y);
}


// calculate du = [grad G(u)]^* dy = - s_0 exp(-<b, u>) (b^* dy) 
//                                 = - b^* (s_0 exp(-<b, u>) dy)
// pointwise, multiple signals
void diffoper_split_apply_gradc(DOperSplit *op,
                                DOperSplitDom *x, 
                                const DOperSplitDom *x0, 
                                const DOperSplitCod *y, 
                                bool addto_x)
{
    assert(x->n==op->n_dops);
    assert(x0->n==op->n_dops);
    assert(y->n==op->n_dops);
    
    #pragma omp parallel for
    for(size_t i=0; i<op->n_dops; i++)
        diffoper_apply_gradc(op->dops[i], x->p[i], x0->p[i], y->p[i], addto_x);
}


DOperSplitDom *diffoper_split_alloc_dom(DOperSplit *op)
{
    SplitVar *sv=alloc_splitvar(op->n_dops);
    
    #pragma omp parallel for
    for(size_t i=0; i<op->n_dops; i++)
        sv->p[i]=diffoper_alloc_dom(op->dops[i]);
        
    return sv;
}

DOperSplitCod *diffoper_split_alloc_cod(DOperSplit *op)
{
    SplitVar *sv=alloc_splitvar(op->n_dops);
    
    #pragma omp parallel for
    for(size_t i=0; i<op->n_dops; i++)
        sv->p[i]=diffoper_alloc_cod(op->dops[i]);
        
    return sv;
}

void diffoper_split_projdom(DOperSplit *op, DOperSplitDom *x)
{
    #pragma omp parallel for
    for(size_t i=0; i<op->n_dops; i++)
        diffoper_projdom(op->dops[i], x->p[i]);
}

FNumber diffoper_split_constraint_scale(DOperSplit *op, const DOperSplitDom *x)
{
    FNumber sc=0;

    assert(x->n==op->n_dops);
    
    #pragma omp parallel for
    for(size_t i=0; i<op->n_dops; i++)
        sc=fn_max(sc, diffoper_constraint_scale(op->dops[i], x->p[i]));
        
    return sc;
}

FNumber diffoper_split_constraint_conjugate(DOperSplit *op, DOperSplitDom *x)
{
    FNumber res=0;
    
    assert(x->n==op->n_dops);

    #pragma omp parallel for
    for(size_t i=0; i<op->n_dops; i++)
        res+=diffoper_constraint_conjugate(op->dops[i], x->p[i]);
    
    return res;
}

FNumber diffoper_split_estimate_grad_norm(DOperSplit *op, const DOperSplitDom *x)
{
    FNumber tmp=0;
    
    assert(x->n==op->n_dops);

    #pragma omp parallel for
    for(size_t i=0; i<op->n_dops; i++){
        FNumber g=diffoper_estimate_grad_norm(op->dops[i], x->p[i]);
        tmp=fn_max(tmp, g);
    }
    
    return tmp;
}

bool diffoper_split_init(DOperSplit *op, 
                         DOper **dops,
                         size_t n_dops)
{
    op->dops=ALLOC_N(DOper*, n_dops);
    
    for(size_t i=0; i<n_dops; i++)
        op->dops[i]=dops[i];
        
    op->n_dops=n_dops;
    
    return diffoper_init(&op->doper);
}

void diffoper_split_deinit(DOperSplit *op)
{
    for(size_t i=0; i<op->n_dops; i++)
        destroy_obj((Obj*)op->dops[i]);
    diffoper_deinit(&op->doper);
}

DOperSplit *create_diffoper_split(DOper **dops, size_t n_dops)
{
    CREATEOBJ_IMPL(DOperSplit, diffoper_split, (p, dops, n_dops));
}

static DynFunTab diffoper_split_dynfuntab[]={
    {diffoper_calculate, diffoper_split_calculate},
    {diffoper_backproj, diffoper_split_backproj},
    {diffoper_apply_grad, diffoper_split_apply_grad},
    {diffoper_apply_gradc, diffoper_split_apply_gradc},
    {diffoper_projdom, diffoper_split_projdom},
    {(DynFun*)diffoper_alloc_dom, (DynFun*)diffoper_split_alloc_dom},
    {(DynFun*)diffoper_alloc_cod, (DynFun*)diffoper_split_alloc_cod},
    {(DynFun*)diffoper_constraint_scale, (DynFun*)diffoper_split_constraint_scale},
    {(DynFun*)diffoper_constraint_conjugate, (DynFun*)diffoper_split_constraint_conjugate},
    {(DynFun*)diffoper_estimate_grad_norm, (DynFun*)diffoper_split_estimate_grad_norm},
    END_DYNFUNTAB,
};

IMPLCLASS(DOperSplit, DOper, diffoper_split_deinit, diffoper_split_dynfuntab);
