/*
 * dticode/src/diffoper_splitsame.c
 *
 * Written 2013 by Tuomo Valkonen <tuomov@iki.fi>, University of Cambridge.
 *
 */

#include <assert.h>
#include <libtu/objp.h>
#include <libtu/types.h>
#include <libtu/misc.h>
#include <libtu/output.h>
#include "linoper.h"
#include "diffoper.h"
#include "diffoper_splitsame.h"

void diffoper_splitsame_calculate(DOperSplitSame *op,
                                  DOperSplitSameCod *y, 
                                  const DOperSplitSameDom *x, 
                                  bool addto_y)
{
    assert(y->n==op->n_dops);
    
    #pragma omp parallel for
    for(size_t i=0; i<op->n_dops; i++)
        diffoper_calculate(op->dops[i], y->p[i], x, addto_y);
}

void diffoper_splitsame_backproj(DOperSplitSame *op,
                                 DOperSplitSameDom *x, 
                                 const DOperSplitSameCod *y)
{
    die("diffoper_splitsame_backproj unsupported.");
}

// calculate grad G(u) du = - s_0 exp(-<b, u>) <b, du> pointwise, multiple signals
void diffoper_splitsame_apply_grad(DOperSplitSame *op,
                                   DOperSplitSameCod *y, 
                                   const DOperSplitSameDom *x0, 
                                   const DOperSplitSameDom *x, 
                                   bool addto_y)
{
    assert(y->n==op->n_dops);
    
    #pragma omp parallel for
    for(size_t i=0; i<op->n_dops; i++)
        diffoper_apply_grad(op->dops[i], y->p[i], x0, x, addto_y);
}


// calculate du = [grad G(u)]^* dy = - s_0 exp(-<b, u>) (b^* dy) 
//                                 = - b^* (s_0 exp(-<b, u>) dy)
// pointwise, multiple signals
void diffoper_splitsame_apply_gradc(DOperSplitSame *op,
                                    DOperSplitSameDom *x, 
                                    const DOperSplitSameDom *x0, 
                                    const DOperSplitSameCod *y, 
                                    bool addto_x)
{
    assert(y->n==op->n_dops);
    
    for(size_t i=0; i<op->n_dops; i++){
        diffoper_apply_gradc(op->dops[i], x, x0, y->p[i], addto_x);
        addto_x=TRUE;
    }
}


DOperSplitSameDom *diffoper_splitsame_alloc_dom(DOperSplitSame *op)
{
    return diffoper_alloc_dom(op->dops[0]);
}

DOperSplitSameCod *diffoper_splitsame_alloc_cod(DOperSplitSame *op)
{
    SplitVar *sv=alloc_splitvar(op->n_dops);
    
    for(size_t i=0; i<op->n_dops; i++)
        sv->p[i]=diffoper_alloc_cod(op->dops[i]);
        
    return sv;
}

FNumber diffoper_splitsame_constraint_scale(DOperSplitSame *op, const DOperSplitSameDom *x)
{
#if 0
    FNumber sc=0;

    #pragma omp parallel for
    for(size_t i=0; i<op->n_dops; i++)
        sc=fn_max(sc, diffoper_constraint_scale(op->dops[i], x));

    return sc;
#else
    #warning "Assuming that all the operators withing DOperSplitSame have the same domain."
    return diffoper_constraint_scale(op->dops[0], x);
#endif
}

FNumber diffoper_splitsame_constraint_conjugate(DOperSplitSame *op, DOperSplitSameDom *x)
{
#if 0
    FNumber res=INFINITY;
    
    #warning "An estimate"
    #pragma omp parallel for
    for(size_t i=0; i<op->n_dops; i++)
        res=fn_min(res, diffoper_constraint_conjugate(op->dops[i], x));
    
    return res;
#else
    return diffoper_constraint_conjugate(op->dops[0], x);
#endif
}

void diffoper_splitsame_projdom(DOperSplitSame *op, DOperSplitSameDom *x)
{
    diffoper_projdom(op->dops[0], x);
}

FNumber diffoper_splitsame_estimate_grad_norm(DOperSplitSame *op, const DOperSplitSameDom *x)
{
    FNumber tmp=0;
    
    #pragma omp parallel for
    for(size_t i=0; i<op->n_dops; i++){
        FNumber g=diffoper_estimate_grad_norm(op->dops[i], x);
        tmp+=g*g;
    }
    
    return fn_sqrt(tmp);
}

bool diffoper_splitsame_init(DOperSplitSame *op, 
                             DOper **dops,
                             size_t n_dops)
{
    op->dops=ALLOC_N(DOper*, n_dops);
    
    for(size_t i=0; i<n_dops; i++)
        op->dops[i]=dops[i];
        
    op->n_dops=n_dops;
    
    return diffoper_init(&op->doper);
}

void diffoper_splitsame_deinit(DOperSplitSame *op)
{
    for(size_t i=0; i<op->n_dops; i++)
        destroy_obj((Obj*)op->dops[i]);
    diffoper_deinit(&op->doper);
}

DOperSplitSame *create_diffoper_splitsame(DOper **dops, size_t n_dops)
{
    CREATEOBJ_IMPL(DOperSplitSame, diffoper_splitsame, (p, dops, n_dops));
}

static DynFunTab diffoper_splitsame_dynfuntab[]={
    {diffoper_calculate, diffoper_splitsame_calculate},
    {diffoper_backproj, diffoper_splitsame_backproj},
    {diffoper_apply_grad, diffoper_splitsame_apply_grad},
    {diffoper_apply_gradc, diffoper_splitsame_apply_gradc},
    {diffoper_projdom, diffoper_splitsame_projdom},
    {(DynFun*)diffoper_alloc_dom, (DynFun*)diffoper_splitsame_alloc_dom},
    {(DynFun*)diffoper_alloc_cod, (DynFun*)diffoper_splitsame_alloc_cod},
    {(DynFun*)diffoper_constraint_scale, (DynFun*)diffoper_splitsame_constraint_scale},
    {(DynFun*)diffoper_constraint_conjugate, (DynFun*)diffoper_splitsame_constraint_conjugate},
    {(DynFun*)diffoper_estimate_grad_norm, (DynFun*)diffoper_splitsame_estimate_grad_norm},
    END_DYNFUNTAB,
};

IMPLCLASS(DOperSplitSame, DOper, diffoper_splitsame_deinit, diffoper_splitsame_dynfuntab);
