/*
 * dticode/src/doubleio.c
 *
 * Written 2012 by Tuomo Valkonen <tuomov@iki.fi>, University of Graz.
 *
 */

// This has to be first due to features.h and _BSD_SOURCE etc. interaction
#if defined(__APPLE__)
#undef _POSIX_C_SOURCE
#define _USE_BSD
#include <sys/types.h>
#include <libkern/OSByteOrder.h>
#define bswap_32(x) OSSwapInt32(x)
#define bswap_64(x) OSSwapInt64(x)
#else
#undef _POSIX_C_SOURCE
#define _BSD_SOURCE
#include <endian.h>
#include <byteswap.h>
#endif

#include <stdlib.h>
#include <stddef.h>
#include <stdio.h>
#include <assert.h>
#include <string.h>
#include <stdint.h>


#if _OPENACC
#include <openacc.h>
#endif
#include <libtu/output.h>

#include "acc_support.h"

#include "number.h"
#include "doubleio.h"

#ifndef BYTE_ORDER
#error "BYTE_ORDER undefined"
#endif

static __inline__ double double_to_machine_endian(double a)
{
#if BYTE_ORDER == LITTLE_ENDIAN
    return a;
#else    
    double res;
    assert(sizeof(double)==sizeof(uint64_t));
    *(uint64_t*)&res=bswap_64(*(uint64_t*)&a);
    return res;
#endif
}

static __inline__ double double_to_little_endian(double a)
{
#if BYTE_ORDER == LITTLE_ENDIAN
    return a;
#else
    double res;
    assert(sizeof(double)==sizeof(uint64_t));
    *(uint64_t*)&res=bswap_64(*(uint64_t*)&a);
    return res;
#endif
}

bool readfn_simple(const double *buf, size_t idx, size_t bs, void *res_)
{
    FNumber *res=&((FNumber*)res_)[idx*bs];
    
    for(size_t i=0; i<bs; i++)
        res[i]=buf[i];

    return TRUE;
}

bool writefn_simple(double *buf, size_t idx, size_t bs, void *data_)
{
    FNumber *data=&((FNumber*)data_)[idx*bs];
    
    for(size_t i=0; i<bs; i++)
        buf[i]=data[i];

    return TRUE;
}

bool readfn_simple_acc(const double *buf, size_t idx, size_t bs, void *res_)
{
    FNumber *res=&((FNumber*)res_)[idx*bs];
    
    #pragma acc kernels loop copyin(buf[0:bs]), deviceptr(res), independent
    for(size_t i=0; i<bs; i++)
        res[i]=buf[i];

    return TRUE;
}

bool writefn_simple_acc(double *buf, size_t idx, size_t bs, void *data_)
{
    FNumber *data=&((FNumber*)data_)[idx*bs];
    
    #pragma acc kernels loop copyout(buf[0:bs]), deviceptr(data), independent
    for(size_t i=0; i<bs; i++)
        buf[i]=data[i];

    return TRUE;
}

size_t read_doubles(size_t nblocks, size_t blocksize, FILE *f, 
                    ReadFn *fn, void *fp)
{
    double *buf;
    size_t k=0;
    
    buf=malloc(sizeof(double)*blocksize);
    if(buf==NULL)
        return 0;
    
    for(k=0; k<nblocks; k++){
        size_t n=fread(buf, sizeof(double), blocksize, f);
        if(n<blocksize)
            break;
        for(size_t i=0; i<blocksize; i++)
            buf[i]=double_to_machine_endian(buf[i]);
        if(!fn(buf, k, blocksize, fp))
            break;
    }
    
    free(buf);
    
    return k;
}



size_t write_doubles(size_t nblocks, size_t blocksize, FILE *f,
                     WriteFn *fn, void *fp)
{
    double *buf;
    size_t k;

    buf=malloc(sizeof(double)*blocksize);
    if(buf==NULL)
        return 0;
    
    for(k=0; k<nblocks; k++){
        if(!fn(buf, k, blocksize, fp))
            break;
        for(size_t i=0; i<blocksize; i++)
            buf[i]=double_to_little_endian(buf[i]);
        size_t n=fwrite(buf, sizeof(double), blocksize, f);
        if(n<blocksize)
            break;
    }
    
    free(buf);
    
    return k;
}


void read_doublefile(size_t nblocks, size_t blocksize, const char *file,
                     ReadFn *fn, void *fp)
{
    FILE *f=fopen(file, "r");
    
    if(f==NULL)
        die_err_obj(file);

    size_t n=read_doubles(nblocks, blocksize, f, fn, fp);

    size_t at=ftell(f);

    if(n==nblocks){
        fseek(f, 0, SEEK_END);
        size_t end=ftell(f);
        if(at==end){
            fclose(f);
            return;
        }
        at=end;
    }
    
    die_obj(file, "Invalid file size (expected: %d, size: %d)", nblocks*blocksize*sizeof(double), at);
}

void write_doublefile(size_t nblocks, size_t blocksize, const char *file,
                      WriteFn *fn, void *fp)
{
    FILE *f=fopen(file, "w");
    
    if(f==NULL)
        die_err_obj(file);

    size_t n=write_doubles(nblocks, blocksize, f, fn, fp);

    if(n!=nblocks)
        die_err_obj(file);
        
    fclose(f);
}

