/*
 * dticode/src/feedback.c
 *
 * Written 2013 by Tuomo Valkonen <tuomov@iki.fi>, University of Cambridge.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <libtu/output.h>

#include "cpock_accel.h"
#include "combine.h"
#include "tensor.h"
#include "number.h"
#include "feedback.h"
#include "parallel.h"
#include "timing.h"


//
// Used to implement Bregman iterations and Gauss-Newton
//

FidPrimal *combined_do_feedback_cpock(MethodParam *mp, 
                                      const Fid *fid,
                                      FidParams *fidp,
                                      const Reg *reg,
                                      RegParams *regp,
                                      FidBaseshiftMode mode,
                                      const char *feedname)
{   
    TensorField *u=NULL;
    CombinedParams cp;
    CombinedPrimal *p, *p0, *ptmp, *pprev=NULL;
    CombinedDual *d, *d0, *dtmp;
    FidPrimal *res;

    uint fbiter=0;
    FNumber pinitnorm=0;
    
    if(mode!=FidBaseshiftNone && fid->baseshift==NULL)
        die("Feedback needs fidelity function to implement baseshift");
        
    cp.reg=reg;
    cp.regp=regp;
    cp.fid=fid;
    cp.fidp=fidp;
    
    combined_alloc_primal(&cp);
    
    p=combined_alloc_primal(&cp);
    p0=combined_alloc_primal(&cp);


    ptmp=combined_alloc_primal(&cp);
    d=combined_alloc_dual(&cp);
    d0=combined_alloc_dual(&cp);
    dtmp=combined_alloc_dual(&cp);
    
    printf("primal value is %f\n",p[1]);
    // Init
    combined_init(p0, d0, mp->initstyle, mp->initfile, fid, fidp, reg, regp);

    if(mp->fbeps>0){
        pprev=combined_alloc_primal(&cp);
        vector_copy((Vector*)pprev, (Vector*)p0);
        pinitnorm=vector_norm((Vector*)p0);
    }

    if(mode==FidBaseshiftSet)
        fid->baseshift(COMBINED_U(p0), FidBaseshiftSet, fidp);
   
    // Run Bregman or Gauss-Newton feedback iterations
    while(1){
        printf("Primal value is %f\n",p[1]); 
        if(mode!=FidBaseshiftNone && mp->cpockp.verbose)
            printf("%s iteration %d/max. %d:\n", feedname, fbiter+1, mp->maxfbiter);

        mp->cpockp.knorm=combined_estimate(EstimateKNorm, p0, fid, fidp, reg, regp);
        mp->cpockp.gamma=fn_max(combined_estimate(EstimateGUniformConvexity, p0, fid, fidp, reg, regp),
                                combined_estimate(EstimateFCUniformConvexity, p0, fid, fidp, reg, regp));
        
        // Run Chambolle-Pock
     
   
        cpock_accel((Vector*)p, (Vector*)d, (Vector*)p0, 
                    (Vector*)d0, (Vector*)ptmp, (Vector*)dtmp, 
                    &mp->cpockp, &combined_fn, &cp);

        if(mp->timing){
            timing_snapshot();
            parallel_timing_print();
        }
        
        if(mode==FidBaseshiftNone){
            break;
        }else{
            FNumber dist=fid->baseshift(COMBINED_U(p), mode, fidp);
            FNumber fstep=0;
            
            if(mp->fbeps>0){
                fstep=vector_dist((Vector*)pprev, (Vector*)p0)/fn_max(1, pinitnorm);
                vector_copy((Vector*)pprev, (Vector*)p0);
            }
            
            if(mp->cpockp.verbose){
                printf("Discrepancy %f (target %f)", dist, mp->disttarget);
                if(mp->fbeps>0){
                    printf("; Fractional step %f (target %f)", fstep, mp->fbeps);
                }
                printf("\n\n");
            }

            fbiter++;
            
            if(mp->outfile!=NULL){
                char *tmp=NULL;
                libtu_asprintf(&tmp, "%s.fbiter-%d", mp->outfile, fbiter);
                vector_save(COMBINED_U(p), tmp);
                free(tmp);
            }


            if(dist<mp->disttarget || fstep<=mp->fbeps || fbiter>=mp->maxfbiter)
                break;

            // Calculate duality gap based on first initialisation to be consistent in quality!
            mp->cpockp.calculate_pdgap0=FALSE;
            
            // Initialise at previous iterate

            
            vector_copy((Vector*)p0, (Vector*)p);
            vector_copy((Vector*)d0, (Vector*)d);
        }
    }

    res=COMBINED_U(p);
    COMBINED_U(p)=NULL;
    
    destroy_obj((Obj*)p);
    destroy_obj((Obj*)p0);
    destroy_obj((Obj*)ptmp);
    destroy_obj((Obj*)d);
    destroy_obj((Obj*)d0);
    destroy_obj((Obj*)dtmp);
    if(pprev!=NULL)
        destroy_obj((Obj*)pprev);
    
    return res;
}
