/*
 * dticode/src/fft.c
 *
 * Written 2012-2013 by Tuomo Valkonen <tuomov@iki.fi>, University of Cambridge.
 *
 */

#include <fftw3.h>
#include <unistd.h>
#include <sys/types.h>
#include <pwd.h>
#include <stdlib.h>
#include <libtu/output.h>
#include <libtu/types.h>

#if _OPENMP
#include <omp.h>
#endif

#if _OPENACC
#ifdef __PGI
#define __align__(X)
#define __location__(X)
#define CUDARTAPI
#endif
#include <cufft.h>
#endif

#include "number.h"
#include "tensor.h"
#include "fft.h"

// inverse order for FFTW
#define DIMS(X) dims[rank-1-(X)]

#if USE_FLOAT
    #error "Not implemented for single-precision"
#endif

#ifndef CF_FFTW_MODE
#define CF_FFTW_MODE FFTW_MEASURE
#endif

void fftshift(int rank, const int *dims, 
              FNumber *res, const FNumber *src, 
              bool inverse)
{
    int prod=1;
    int strides[rank];
    
    for(int j=0; j<rank; j++){
        prod*=DIMS(j);
        if(j>0)
            strides[j]=strides[j-1]*DIMS(j-1);
        else
            strides[j]=2;
    }
    
    #pragma acc data copyin(strides[0:rank-1],dims[0:rank-1]), deviceptr(res, src)
    #pragma acc loop seq
    for(int j=0; j<rank; j++){
        int n=DIMS(j);
        int rows=prod/n;
        int stride=strides[j];
        
        #pragma omp parallel for
        #pragma acc kernels loop, independent
        for(int r=0; r<rows; r++){
            int rowoff=0;
            int z=r;
            
            #pragma acc loop seq
            for(int l=0; l<rank; l++){
                if(l!=j){
                    int a=z%DIMS(l);
                    z=z/DIMS(l);
                    rowoff+=a*strides[l];
                }
            }
            
            if(n%2==0){
                //int shift=stride*((n+1)/2);
                int shift=stride*(n/2);

                #pragma acc loop seq
                for(int i=0; i<n/2; i++){
                    int k=stride*i+rowoff;
                    FNumber tmp1, tmp2;
                    // allow src==res
                    tmp1=src[k+shift];       // real
                    tmp2=src[k+shift+1];     // imaginary
                    res[k+shift]=src[k];     // real
                    res[k+shift+1]=src[k+1]; // imaginary
                    res[k]=tmp1;                 // real
                    res[k+1]=tmp2;               // imaginary
                }
            }else if(!inverse){
                int shift=stride*(n/2);
                int mid=rowoff+shift;
                int end=rowoff+stride*(n-1);
                
                FNumber save1=src[end];   // real
                FNumber save2=src[end+1]; // imaginary

                res[end]=src[mid];        // real
                res[end+1]=src[mid+1];    // imaginary

                #pragma acc loop seq
                for(int i=n/2-1; i>=0; i--){
                    int k=i*stride+rowoff;
                    FNumber tmp1, tmp2;
                    // allow src==res
                    tmp1=src[k];              // real
                    tmp2=src[k+1];            // imaginary
                    res[k]=save1;             // real
                    res[k+1]=save2;           // imaginary
                    save1=src[k+shift];   // real
                    save2=src[k+shift+1]; // imaginary
                    res[k+shift]=tmp1;    // real
                    res[k+shift+1]=tmp2;  // imaginary
                }
            }else{
                int shift=stride*(n/2+1);
                int mid=rowoff+stride*(n/2);
                int end=rowoff+stride*(n-1);
                
                FNumber save1=src[mid];   // real
                FNumber save2=src[mid+1]; // imaginary

                res[mid]=src[end];        // real
                res[mid+1]=src[end+1];    // imaginary

                #pragma acc loop seq
                for(int i=0; i<n/2; i++){
                    int k=i*stride+rowoff;
                    FNumber tmp1, tmp2;
                    // allow src==res
                    tmp1=src[k];              // real
                    tmp2=src[k+1];            // imaginary
                    res[k]=save1;             // real
                    res[k+1]=save2;           // imaginary
                    save1=src[k+shift];   // real
                    save2=src[k+shift+1]; // imaginary
                    res[k+shift]=tmp1;    // real
                    res[k+shift+1]=tmp2;  // imaginary
                }
            }
        }
        // next direction, src is now res
        src=res;
    }
}

void do_fft(int rank, const int *dims, 
            FNumber *res, const FNumber *src)
{
    fftshift(rank, dims, res, src, 0);
    
#if _OPENACC
    cufftHandle plan;
    
    if(cufftPlanMany(&plan, rank, (int*)dims,
                     NULL, 0, 1, 
                     NULL, 0, 1,
                     CUFFT_C2C, 1)!=CUFFT_SUCCESS){
        die("Error creating CUFFT plan.");
    }
    
    // res -> res due to fftshift above
    if(cufftExecC2C(plan,
                    (cufftComplex*)res, 
                    (cufftComplex*)res, 
                    CUFFT_FORWARD)!=CUFFT_SUCCESS){
        die("Error executing CUFFT plan.");
    }
    
    if(cudaThreadSynchronize()!=cudaSuccess)
        die("Error synchronising cuda.");
        
    cufftDestroy(plan);
#else
    fftw_plan plan;
    
    #pragma omp critical(dticode_fftw_entry)
    {
#if _OPENMP
        fftw_plan_with_nthreads(omp_get_max_threads());
#endif

        // res -> res due to fftshift above
        plan=fftw_plan_dft(rank, dims,
                           (fftw_complex*)res, 
                           (fftw_complex*)res, 
                           FFTW_FORWARD, CF_FFTW_MODE);
    }
    
    fftw_execute(plan);
    
    #pragma omp critical(dticode_fftw_entry)
    {
        fftw_destroy_plan(plan);
    }
#endif
}

void do_ifft(int rank, const int *dims, 
             FNumber *res, const FNumber *src)
{
#if _OPENACC
    cufftHandle plan;
    
    if(cufftPlanMany(&plan, rank, (int*)dims,
                     NULL, 0, 1, 
                     NULL, 0, 1,
                     CUFFT_C2C, 1)!=CUFFT_SUCCESS){
        die("Error creating CUFFT plan.");
    }
    
    if(cufftExecC2C(plan, 
                    (cufftComplex*)src,
                    (cufftComplex*)res,
                    CUFFT_INVERSE)!=CUFFT_SUCCESS){
        die("Error executing CUFFT plan.");
    }
    
    if(cudaThreadSynchronize()!=cudaSuccess)
        die("Error synchronising cuda.");
        
    cufftDestroy(plan);
#else
    fftw_plan plan;
    
    #pragma omp critical(dticode_fftw_entry)
    {
#if _OPENMP
        fftw_plan_with_nthreads(omp_get_max_threads());
#endif

        plan=fftw_plan_dft(rank, dims,
                           (fftw_complex*)src, 
                           (fftw_complex*)res, 
                           FFTW_BACKWARD, CF_FFTW_MODE);
    }
    
    fftw_execute(plan);
    
    #pragma omp critical(dticode_fftw_entry)
    {
        fftw_destroy_plan(plan);
    }
#endif
    
    fftshift(rank, dims, res, res, 1);
}

void fft_init_params(size_t *rank, int dims[3], FNumber *normfact, 
                     const Sig *sig)
{
    if(sig->h<=1 && sig->d<=1){
        *rank=1;
        dims[0]=sig->w;
        *normfact=sig->w;
    }else if(sig->d<=1){
        *rank=2;
        // FFTW stores the indices in inverse order: height first, then width.
        dims[0]=sig->h;
        dims[1]=sig->w;
        *normfact=sig->w*sig->h;
    }else{
        *rank=3;
        dims[0]=sig->d;
        dims[1]=sig->h;
        dims[2]=sig->w;
        *normfact=sig->w*sig->h*sig->d;
    }
}

static char *get_wisdom_filename()
{
    struct passwd *pw = getpwuid(getuid());
    char *wisdom;
    
    libtu_asprintf(&wisdom, "%s/.dticode_fftw_wisdom", pw->pw_dir);
    
    return wisdom;
}

void fft_init()
{
#if !_OPENACC
    char *wisdom=get_wisdom_filename();

    fftw_init_threads();

#if 0
    fftw_import_wisdom_from_filename(wisdom);
#else
    // DAMTP uses an obsolete linux distro...
    FILE *f=fopen(wisdom, "r");
    if(f!=NULL){
        fftw_import_wisdom_from_file(f);
        fclose(f);
    }
#endif
    
    free(wisdom);
#endif // !_OPENACC
}

void fft_finalise()
{
#if !_OPENACC
    char *wisdom=get_wisdom_filename();

#if 0    
    fftw_export_wisdom_to_filename(wisdom);
#else
    // DAMTP uses an obsolete linux distro...
    FILE *f=fopen(wisdom, "w");
    if(f!=NULL){
        fftw_export_wisdom_to_file(f);
        fclose(f);
    }else{
        warn_err_obj(wisdom);
    }
#endif
    
    free(wisdom);
#endif // !_OPENACC
}
