/*
 * dticode/src/fid_l1.c
 *
 * Written 2013 by Tuomo Valkonen <tuomov@iki.fi>, University of Cambridge.
 *
 */

#include <math.h>
#include <stdio.h>
#include <string.h>
#include <libtu/types.h>
#include <libtu/output.h>
#include <libtu/misc.h>

#include "tensor.h"
#include "combine.h"
#include "combine_nonlin.h"
#include "fid_l1.h"
#include "scaleplus.h"
#include "project_posdef.h"
#include "tensornorm.h"
#include "posdef.h"
#include "sumdot1.h"
#include "normalise.h"
#include "linoper.h"

#define TENSOR_F(fp) ((TensorField*)fp->f)

//
// The approach:
//
// |A u-f|_F + delta_{u>=0}
//  = max_lambda  delta_{u>=0} + <lambda, A u-f> - delta_{|lambda|<=1}
//  = max_lambda  delta_{u>=0} + <lambda,A u>    - <lambda,f> - delta_{|lambda|<=1}
//  = max_lambda  G(u)         + <lambda,K u>  - F^*(lambda)
//
// F(y) = sup_x <y,x> - <x,f>-delta_{|x|<=1}
//      = |y-f|_1
//
// G(u) = delta_{u>=0}(u) 
//
// G^*(y) = sup_x <y,x>-delta_{x>=0}(x) = 
//      { INFINITY, y has pos eigenvalues
//      { 0,        y neg.semi.def
//     
// Because G^* can be infinite, we need an extra bound |u(x)|<M, 
// where M > sup_x |f(x)| to make the duality gap practical. 
//
// Therefore, we take actually:
//
// G(u) = delta_{u>=0 & |u|<=M}(u)
//
// G^*(y) = sup_x <y,x>-delta_{x>=0 & |x|<=M}(x)
//        = M<normalise_pointwise(P_{>=0}(y)), y>
//        = M|P_{>=0}(y)|_{L1-frob}
// G^*(-K^*y) = M*|P_{>=0}(-K^*y)|
//

static void fid_l1_k(FidL1Dual *a, const FidL1Primal *b, 
                     bool addto_a, FidL1Params *fp)
{
    linoper_calculate(fp->op, (LinOperCod*)a, b, addto_a);
}


static void fid_l1_kc(FidPrimal *a, const FidL1Dual *b, 
                      bool addto_a, FidL1Params *fp)
{
    linoper_calculatec(fp->op, a, (LinOperCod*)b, addto_a);
}


static void fid_l1_gres(FidPrimal *u, FNumber tau, 
                        FidL1Params *fp)
{
    // Project if necessary.
    linoper_projdom(fp->op, u);
}

static void fid_l1_fcres(FidL1Dual *lambda, FNumber tau,
                         FidL1Params *fp)
{
    // min_x{ |lambda-x|^2/2tau+<f,x>+delta_{\alpha B}(x) } 
    // => (x-lambda)/tau + f + N_{\alpha B x} = 0
    //  => x +  N_{\alpha B x} = lambda - tau f;
    if(TENSOR_F(fp)!=NULL)
        tensorfield_scaleplus(lambda, lambda, TENSOR_F(fp), 1, -tau, 1);
    tensorfield_project_frob2(lambda, 1);
}

static void fid_l1_pdgap(FNumber *pdgap,
                         const FidL1Primal *u, 
                         const FidL1Dual *y,
                         const FidL1Primal *kcy,
                         const FidL1Dual *ku,
                         FidPrimal *uwork,
                         unsigned int iter,
                         FidL1Params *fp)
{
    FNumber ab, c, mx;
    
    /* F(Ku) + G(u) +G^*(-K*y)+F^*(y) */

    if(u!=NULL){
        // Calculate M such that G_M(u)<infty, where G_M has the constraint set scaled by M.
        mx=diffoper_constraint_scale(&fp->op->doper, u);
        mx=fn_max(mx, fp->primal_bound);
        fp->primal_bound=mx;
        
        // Calculate G^*. 
        vector_scaleplus(uwork, kcy, kcy, -1.0, 0, 1.0);
        c=diffoper_constraint_conjugate(&fp->op->doper, uwork);
    }else{
        // Called through, e.g., fid_oper that implements the same as
        // above on the operator parameter domain.
        c=0;
        mx=0;
    }
    
    if(TENSOR_F(fp)==NULL){
        ab=tensorfield_sumdot1(ku, ku, FALSE, 2);                        // F(Ku)
    }else{
        ab=tensorfield_sumdot1(ku, TENSOR_F(fp), TRUE, 2)                // F(Ku)
           +0.0                                                          // G(u)
           +(tensorfield_iprod(y, TENSOR_F(fp))+0.0);                    // F^*(y)
    }
  	
    
    
    if(iter==0){
        fp->pdgap0_ab=ab;
        fp->pdgap0_c=c;
    }
    
    pdgap[0]=ab+mx*c;
    pdgap[1]=fp->pdgap0_ab+mx*fp->pdgap0_c;
}

static FidPrimal *fid_l1_alloc_primal(FidL1Params *fp)
{
    return (FidPrimal*)linoper_alloc_dom(fp->op);
}

static FidDual *fid_l1_alloc_dual(FidL1Params *fp)
{
     LinOperCod *p=linoper_alloc_cod(fp->op);
     
     assert(OBJ_IS(p, TensorField));
     
     return (FidDual*)p;
}

void fid_l1_init_primal(FidL1Primal *u, FidInitPrimalStyle initstyle, 
                        FidL1Params *fp)
{
    if(initstyle==FidInitPrimalZero){
        vector_zero(u);
    }else if(initstyle==FidInitPrimalBackproj){
        linoper_backproj(fp->op, u, fp->f);
    }else{
        die("Unknown FidInitPrimalStyle");
    }
}

void fid_l1_init_dual(FidL1Dual *b, const FidL1Primal *a, 
                      FidL1Params *fp)
{
    fid_l1_k(b, a, FALSE, fp);
    if(TENSOR_F(fp)!=NULL)
        tensorfield_scaleplus(b, b, TENSOR_F(fp), 1, -1, 1);
    tensorfield_normalise_frob2(b, 1);
}


/*
static FNumber fid_l1_baseshift(FidL1Primal *u, FidBaseshiftMode mode, 
                                   FidL1Params *fp)
{
    return fid_l2_baseshift(u, mode, fp);
}
*/

static FNumber fid_l1_estimate(Estimate estim, const FidPrimal *u, FidL1Params *fp)
{
    if(estim==EstimateKNorm)
        return linoper_estimate_norm(fp->op);
    else if(estim==EstimateGUniformConvexity)
        return 0;
    else if(estim==EstimateFCUniformConvexity)
        return INFINITY;
    else
        return 0;
}

static bool fid_l1_grad(TensorField *grad, const TensorField *u, 
                        FidL1Params *fp)
{
    if(TENSOR_F(fp)!=NULL)
        tensorfield_scaleplus(grad, u, TENSOR_F(fp), 1, -1, 1);
    tensorfield_normalise_frob2(grad, 1);
    return FALSE; // Only a subgradient
}

static Fid fid_l1={
    (FidK*)fid_l1_k,
    (FidKC*)fid_l1_kc,
    (FidFCres*)fid_l1_fcres,
    (FidGres*)fid_l1_gres,
    (FidPDgap*)fid_l1_pdgap,
    (FidAllocPrimal*)fid_l1_alloc_primal,
    (FidAllocDual*)fid_l1_alloc_dual,
    (FidInitPrimal*)fid_l1_init_primal,
    (FidInitDual*)fid_l1_init_dual,
    (FidBaseshift*)NULL, //fid_l1_baseshift,
    (FidEstimate*)fid_l1_estimate,
    (FidGrad*)fid_l1_grad,
};



void fid_l1_init(const LinOperCod *f,
                 LinOper *op,
                 Fid **fid, 
                 FidL1Params *fp)
{
    assert(f==NULL || OBJ_IS(f, TensorField));
    
    fid_l2alt_init(f, op, fid, fp);
    //param->ubound=1+10*tensorfield_frobinf(f)/sqrt(matrix_normsquared_lowerbound(amat));
    
    *fid=&fid_l1;
}
