/*
 * dticode/src/fid_l2.c
 *
 * Written 2012 by Tuomo Valkonen <tuomov@iki.fi>, University of Cambridge.
 *
 */

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <libtu/types.h>
#include <libtu/output.h>
#include <libtu/misc.h>
#include <string.h>

#include "tensor.h"
#include "combine.h"
#include "fid_l2.h"
#include "scaleplus.h"
#include "project_posdef.h"
#include "tensornorm.h"
#include "posdef.h"
#include "sumdot1.h"
#include "normalise.h"
#include "doubleio.h"

#define OPL2(FP) ((LinOperL2*)((FP)->op))

//
// |f-Fu|^2/2 + |Mu|^2/2
// = |f-Fu|^2/2 + <lambda, Mu> - |lambda|^2/2
// = G(u) + <lambda, Ku> - F(lambda)
//
// NOTE: We use symmetric scaling of the DFT.
// This is in particular for better convergence properties
// of the Chambolle-Pock algorithm.
//
//  Fx=(1/sqrt(N)) sum_{j=0}^{N-1} x_j exp(-i 2 pi j/N)
//  F^* x=(1/sqrt(N)) sum_{j=0}^{N-1} x_j exp(i 2 pi j/N)
//


static void fid_l2_k(FidL2Dual *a, const FidL2Primal *b, 
                     bool addto_a, FidL2Params *fp)
{
#if 0
    if(fp->mask_term_weight!=0){
        FNumber *ap=a->p;
        const FNumber *bp=b->p;
        const FNumber *mp=fp->mask->p;
        const FNumber w=fp->mask_term_weight;
        const size_t n=a->nelem;
        const size_t m=a->elemsize;
        
        if(addto_a){
            #pragma omp parallel for
            #pragma acc kernels loop deviceptr(ap, bp, mp), independent
            for(size_t i=0; i<n; i++){
                #pragma acc loop independent, vector
                for(size_t j=0; j<m; j++)
                    ap[i*m+j]+=bp[i*m+j]*mp[i]*w;
            }
        }else{
            #pragma omp parallel for
            #pragma acc kernels loop deviceptr(ap, bp, mp), independent
            for(size_t i=0; i<n; i++){
                #pragma acc loop independent, vector
                for(size_t j=0; j<m; j++)
                    ap[i*m+j]=bp[i*m+j]*mp[i]*w;
            }
        }
    }
#endif
}

static void fid_l2_kc(FidL2Primal *a, const FidL2Dual *b, 
                      bool addto_a, FidL2Params *fp)
{
#if 0
    if(fp->mask_term_weight!=0){
        FNumber *ap=a->p;
        const FNumber *bp=b->p;
        const FNumber *mp=fp->mask->p;
        const FNumber w=fp->mask_term_weight;
        const size_t n=a->nelem;
        const size_t m=a->elemsize;
        
        if(addto_a){
            #pragma omp parallel for
            #pragma acc kernels loop deviceptr(ap, bp, mp), independent
            for(size_t i=0; i<n; i++){
                #pragma acc loop independent, vector
                for(size_t j=0; j<m; j++)
                    ap[i*m+j]+=bp[i*m+j]*mp[i]*w;
            }
        }else{
            #pragma omp parallel for
            #pragma acc kernels loop deviceptr(ap, bp, mp), independent
            for(size_t i=0; i<n; i++){
                #pragma acc loop independent, vector
                for(size_t j=0; j<m; j++)
                    ap[i*m+j]=bp[i*m+j]*mp[i]*w;
            }
        }
    }
#endif
}

static void fid_l2_gres(FidL2Primal *u, FNumber tau, FidL2Params *fp)
{
    linoper_l2_gsolve(OPL2(fp), u, tau, 1.0, 1.0, fp->f);
}

static void fid_l2_fcres(FidL2Dual *lambda, FNumber tau, FidL2Params *fp)
{
#if 0
    // min_x{ |lambda-x|^2/2tau+|x|^2/2 } 
    // => (x-lambda)/tau + x = 0
    //  => x =  lambda/(1+tau)
    if(fp->mask_term_weight!=0)
        tensorfield_scale(lambda, 1/(1+tau));
#endif
}

static void fid_l2_pdgap(FNumber *pdgap,
                         const FidL2Primal *u, 
                         const FidL2Dual *y,
                         const FidL2Primal *kcy,
                         const FidL2Dual *ku,
                         FidL2Primal *uwork,
                         unsigned int iter,
                         FidL2Params *fp)
{
    FNumber g, gcm, fpart=0, ab, c, mx;

#if 0
    if(fp->mask_term_weight!=0){
        // F(Ku)+F^*(y)
        fpart=tensorfield_frob222(ku)+tensorfield_frob222(y);
    }
#endif

    // calculate G(u)
    mx=fp->primal_bound;
    g=linoper_l2_g(OPL2(fp), u, fp->f, &mx);
    printf("discrepancy is %f\n",g);
    mx=fn_max(mx, fp->primal_bound);
    fp->primal_bound=mx;

    // calculate G^*(-kcy)
    gcm=linoper_l2_gc(OPL2(fp), kcy, fp->f, uwork, &c);
    
    ab=g+gcm+fpart;
    
    if(iter==0){
        fp->pdgap0_ab=ab;
        fp->pdgap0_c=c;
    }

    pdgap[0]=ab+mx*c;
    pdgap[1]=fp->pdgap0_ab+mx*fp->pdgap0_c;
}


static FidL2Primal *fid_l2_alloc_primal(FidL2Params *fp)
{
  
    return linoper_alloc_dom(fp->op);
}


static FidL2Dual *fid_l2_alloc_dual(FidL2Params *fp)
{
#if 0
    if(fp->mask_term_weight!=0)
        return (Vector*)alloc_tensorfield(&fp->sig);
    else
#endif
        return alloc_dummyvector();

}


void fid_l2_init_dual(FidL2Dual *b, const FidL2Primal *a, 
                      FidL2Params *fp)
{
    // Nothing without the mask
#if 0
#endif
}

static FNumber fid_l2_estimate(Estimate estim, const FidPrimal *u, FidL2Params *fp)
{
    if(estim==EstimateKNorm){
        return (fp->mask==NULL ? 0 : fp->mask_term_weight); 
    }else if(estim==EstimateGUniformConvexity){
        // |Kx|>g|x|
        FNumber g=linoper_estimate_lower_bound(fp->op);
        return g*g;
    }else if(estim==EstimateFCUniformConvexity){
        return INFINITY;
    }else{
        return 0;
    }
}


static Fid fid_l2={
    (FidK*)fid_l2_k,
    (FidKC*)fid_l2_kc,
    (FidFCres*)fid_l2_fcres,
    (FidGres*)fid_l2_gres,
    (FidPDgap*)fid_l2_pdgap,
    (FidAllocPrimal*)fid_l2_alloc_primal,
    (FidAllocDual*)fid_l2_alloc_dual,
    (FidInitPrimal*)fid_l2alt_init_primal,
    (FidInitDual*)fid_l2_init_dual,
    (FidBaseshift*)fid_l2alt_baseshift,
    (FidEstimate*)fid_l2_estimate,
    (FidGrad*)fid_l2alt_grad,
};


void fid_l2_init(const FidL2Primal *f,
                 LinOperL2 *op,
                 //const TensorField *mask,
                 //FNumber mask_term_weight,
                 Fid **fid, 
                 FidL2Params *fp)
{
    fp->mask=NULL;//mask;
    fp->pdgap0_ab=0;
    fp->pdgap0_c=0;
    fp->mask_term_weight=0;//mask_term_weight;
    fp->f=f;
    fp->f_orig=f;
    fp->f_tmp=NULL;
    fp->op=(LinOper*)op;
    fp->primal_bound=0;
    
    *fid=&fid_l2;
}

