/*
 * dticode/src/fid_l2alt.c
 *
 * Written 2013 by Tuomo Valkonen <tuomov@iki.fi>, University of Cambridge.
 *
 */

#include <math.h>
#include <stdio.h>
#include <string.h>
#include <libtu/types.h>
#include <libtu/output.h>
#include <libtu/misc.h>

#include "tensor.h"
#include "combine.h"
#include "combine_nonlin.h"
#include "fid_l2alt.h"
#include "scaleplus.h"
#include "project_posdef.h"
#include "tensornorm.h"
#include "posdef.h"
#include "sumdot1.h"
#include "normalise.h"
#include "linoper.h"

//
// The approach:
//
// |Au-f|^2_F + delta_{u>=0}
//  = max_lambda  delta_{u>=0} + <lambda, Au-f> - |lambda|^2/2
//  = max_lambda  delta_{u>=0} + <lambda,A u>    - <lambda,f> - |lambda|^2/2
//  = max_lambda  G(u)         + <lambda,K u>  - F^*(lambda)
//
// F(y) = sup_x <y,x> - <x,f>-|lambda|^2/2
//      = |y-f|^2/2
//
// G(u) = delta_{u>=0}(u) 
//
// G^*(y) = sup_x <y,x>-delta_{x>=0}(x) = 
//      { INFINITY, y has pos eigenvalues
//      { 0,        y neg.semi.def
//     
// Because G^* can be infinite, we need an extra bound |u(x)|<M, 
// where M > sup_x |f(x)| to make the duality gap practical. 
//
// Therefore, we take actually:
//
// G(u) = delta_{u>=0 & |u|<=M}(u)
//
// G^*(y) = sup_x <y,x>-delta_{x>=0 & |x|<=M}(x)
//        = M<normalise_pointwise(P_{>=0}(y)), y>
//        = M|P_{>=0}(y)|_{L2Alt-frob}
// G^*(-K^*y) = M*|P_{>=0}(-K^*y)|
//

static void fid_l2alt_k(FidL2AltDual *a, const FidL2AltPrimal *b, 
                        bool addto_a, FidL2AltParams *fp)
{
    //FNumber in=vector_norm22(a);
    linoper_calculate(fp->op, a, b, addto_a);
    //printf("K %f  -> %f, %d [%f]\n", vector_norm22(b), vector_norm22(a), addto_a, in);
}


static void fid_l2alt_kc(FidPrimal *a, const FidL2AltDual *b, 
                         bool addto_a, FidL2AltParams *fp)
{
    //FNumber in=vector_norm22(a);
    linoper_calculatec(fp->op, a, b, addto_a);
    //printf("KC %f  -> %f, %d [%f]\n", vector_norm22(b), vector_norm22(a), addto_a, in);
}


static void fid_l2alt_gres(FidPrimal *u, FNumber tau, 
                           FidL2AltParams *fp)
{
    // Project if necessary
    linoper_projdom(fp->op, u);
}

static void fid_l2alt_fcres(FidL2AltDual *lambda, FNumber tau,
                            FidL2AltParams *fp)
{
    // min_x{ |lambda-x|^2/2tau+<f,x>+|x|^2/2 } 
    // => (x-lambda)/tau + f + x = 0
    //  => x =  (lambda-tau*f)/(1+tau)
    vector_scaleplus(lambda, lambda, fp->f, 1, -tau, 1+tau);
}

static void fid_l2alt_pdgap(FNumber *pdgap,
                            const FidL2AltPrimal *u, 
                            const FidL2AltDual *y,
                            const FidL2AltPrimal *kcy,
                            const FidL2AltDual *ku,
                            FidPrimal *uwork,
                            unsigned int iter,
                            FidL2AltParams *fp)
{
    FNumber ab, c, mx;
    
    /* F(Ku) + G(u) +G^*(-K*y)+F^*(y) */

    if(u!=NULL){
        // Calculate M such that G_M(u)<infty, where G_M has the constraint set scaled by M.
        mx=diffoper_constraint_scale(&fp->op->doper, u);
        mx=fn_max(mx, fp->primal_bound);
        fp->primal_bound=mx;
        
        // Calculate G^*(-kcy). 
        vector_scaleplus(uwork, kcy, kcy, -1.0, 0, 1.0);
        c=diffoper_constraint_conjugate(&fp->op->doper, uwork);
    }else{
        // Called through, e.g., fid_oper that implements the same as
        // above on the operator parameter domain.
        c=0;
        mx=0;
    }
    
    ab=vector_dist22(ku, fp->f)                    // F(Ku)
       +0.0                                        // G(u)
       +(vector_iprod(y, fp->f)+vector_norm22(y)); // F^*(y)
    
    //printf("%d: %f [%f] // %f, %f, %f[kcy %f] || %f, %f\n", iter, ab, fp->pdgap0_ab, vector_dist22(ku, fp->f), vector_iprod(y, fp->f), vector_norm22(y), vector_norm22(kcy), mx, c);
    
    if(iter==0){
        fp->pdgap0_ab=ab;
        fp->pdgap0_c=c;
    }
    
    pdgap[0]=ab+mx*c;
    pdgap[1]=fp->pdgap0_ab+mx*fp->pdgap0_c;
}

static FidPrimal *fid_l2alt_alloc_primal(FidL2AltParams *fp)
{
    return linoper_alloc_dom(fp->op);
}

static FidDual *fid_l2alt_alloc_dual(FidL2AltParams *fp)
{
    return (FidDual*)linoper_alloc_cod(fp->op);
}

void fid_l2alt_init_dual(FidL2AltDual *b, const FidL2AltPrimal *a, 
                         FidL2AltParams *fp)
{
    fid_l2alt_k(b, a, FALSE, fp);
    vector_scaleplus(b, b, fp->f, 1, -1, 1);
}


static FNumber fid_l2alt_estimate(Estimate estim, const FidPrimal *u, FidL2AltParams *fp)
{
    if(estim==EstimateKNorm)
        return linoper_estimate_norm(fp->op);
    else if(estim==EstimateGUniformConvexity)
        return 0;
    else if(estim==EstimateFCUniformConvexity)
        return INFINITY;
    else
        return 0;
}

FNumber fid_l2alt_baseshift(FidL2AltPrimal *u, FidBaseshiftMode mode, 
                            FidL2AltParams *fp)
{
    FNumber dist=0;
    LinOperDom *f_tmp=fp->f_tmp;
    const LinOperDom *f_orig=fp->f_orig;

    if(f_tmp==NULL){
        f_tmp=linoper_alloc_cod(fp->op);
        vector_copy(f_tmp, f_orig);
        fp->f=f_tmp;
        fp->f_tmp=f_tmp;
    }

    dist=linoper_calcdist(fp->op, u, f_orig, NULL);
    dist=dist*dist/2;
    
    if(mode==FidBaseshiftAdd){
        // f_tmp := f_tmp - Fu
        //vector_scaleplus(f_tmp, f_tmp, fu, 1, -1, 1);
        #warning "TODO: instead of addto_y, have weight parameter in linoper_calculate"
        vector_scale(f_tmp, -1);
        linoper_calculate(fp->op, f_tmp, u, TRUE);
        vector_scale(f_tmp, -1);
    }else if(mode==FidBaseshiftSet){
        // f_tmp := f_orig - Fu
        //vector_scaleplus(f_tmp, f_orig, fu, 1, -1, 1);
        linoper_calculate(fp->op, f_tmp, u, FALSE);
        vector_scaleplus(f_tmp, f_orig, f_tmp, 1, -1, 1);
    }else if(mode==FidBaseshiftBregman){
        // f_tmp := f_orig + f_tmp - Fu
        vector_scale(f_tmp, -1);
        linoper_calculate(fp->op, f_tmp, u, TRUE);
        vector_scaleplus(f_tmp, f_orig, f_tmp, 1, -1, 1);
    }else if(mode==FidBaseshiftNone){
        vector_copy(f_tmp, f_orig);
    }else{
        die("Unknown FidBaseshiftMode");
    }

    return dist;
}

void fid_l2alt_init_primal(FidL2AltPrimal *u, FidInitPrimalStyle initstyle, 
                           FidL2AltParams *fp)
{
    if(initstyle==FidInitPrimalZero){
        vector_zero(u);
    }else if(initstyle==FidInitPrimalBackproj){
        linoper_backproj(fp->op, u, fp->f);
    }else{
        die("Unknown FidInitPrimalStyle");
    }
}

bool fid_l2alt_grad(FidL2AltPrimal *grad, const FidL2AltPrimal *u, 
                    FidL2AltParams *fp)
{
    LinOperCod *tmp=linoper_alloc_cod(fp->op);
    
    linoper_calculate(fp->op, tmp, u, FALSE);
    vector_scaleplus(tmp, tmp, fp->f, 1, -1, 1);
    linoper_calculatec(fp->op, grad, tmp, FALSE);
    
    destroy_obj((Obj*)tmp);
    
    return TRUE;
}

static Fid fid_l2alt={
    (FidK*)fid_l2alt_k,
    (FidKC*)fid_l2alt_kc,
    (FidFCres*)fid_l2alt_fcres,
    (FidGres*)fid_l2alt_gres,
    (FidPDgap*)fid_l2alt_pdgap,
    (FidAllocPrimal*)fid_l2alt_alloc_primal,
    (FidAllocDual*)fid_l2alt_alloc_dual,
    (FidInitPrimal*)fid_l2alt_init_primal,
    (FidInitDual*)fid_l2alt_init_dual,
    (FidBaseshift*)fid_l2alt_baseshift,
    (FidEstimate*)fid_l2alt_estimate,
    (FidGrad*)fid_l2alt_grad,
};




void fid_l2alt_init(const LinOperCod *f,
                    LinOper *op,
                    Fid **fid, 
                    FidL2AltParams *fp)
{
    fp->mask=NULL;//mask;
    fp->pdgap0_ab=0;
    fp->pdgap0_c=0;
    fp->mask_term_weight=0;//mask_term_weight;
    fp->f=f;
    fp->f_orig=f;
    fp->f_tmp=NULL;
    printf("l2alt");
    printf("%p\n",op);   
fp->op=op;
    printf("%p\n",fp->op);  
    fp->primal_bound=0;
    
    *fid=&fid_l2alt;
}
