/*
 * dticode/src/fid_oper.c
 *
 * Written 2013 by Tuomo Valkonen <tuomov@iki.fi>, University of Cambridge.
 *
 */

#include <stdio.h>
#include <math.h>
#include <assert.h>
#include <libtu/misc.h>
#include <libtu/output.h>
#include "combine.h"
#include "diffoper.h"
#include "fid_oper.h"
#include "tensornorm.h"

////////

static void fid_oper_k(FidDual *a, const DOperDom *b, bool addto_a, FidOperParams *fp)
{
    diffoper_apply_grad(fp->oper, fp->tmp, fp->oper_u0, b, FALSE);
    fp->fid->k(a, (FidPrimal*)fp->tmp, addto_a, fp->fidp);
}

static void fid_oper_kc(DOperDom *a, const FidDual *b, bool addto_a, FidOperParams *fp)
{
    fp->fid->kc((FidPrimal*)fp->tmp, b, FALSE, fp->fidp);
    diffoper_apply_gradc(fp->oper, a, fp->oper_u0, fp->tmp, addto_a);
}

static void fid_oper_fcres(FidDual *a, FNumber tau, FidOperParams *fp)
{
    fp->fid->fcres(a, tau, fp->fidp);
}

static void fid_oper_gres(DOperDom *a, FNumber tau, FidOperParams *fp)
{
    if(fp->xi!=NULL){
        // G(x)=<xi,x> ==>
        // min_x{|a-x|^2/(2tau)+G(x)} 
        // optimality conditions (x-a)/tau+xi+N_dom(x)=0 ==> x+N_dom(x)=a-tau*xi
        vector_scaleplus(a, a, fp->xi, 1, -tau, 1);
        diffoper_projdom(fp->oper, a);
    }
}

static void fid_oper_pdgap(FNumber *pdgap, 
                           const DOperDom *x, 
                           const FidDual *y,
                           const DOperDom *kcy, 
                           const FidDual *kx,
                           DOperDom *xwork,
                           unsigned int iter,
                           FidOperParams *fp)
{
    FNumber mx=0, ab=0, c=0;
    
    // ASSUMPTION: no G, G implemented by fid_oper! ==> DOperDom variables not used

    mx=diffoper_constraint_scale(fp->oper, x);
    mx=fn_max(mx, fp->primal_bound);
    fp->primal_bound=mx;
    
    if(fp->xi!=NULL){
        ab=vector_iprod(fp->xi, x);     // G(x)=<xi,x>
        // G^*(-K*y) becomes sup_v{<-K^*y-xi, v> | |v| < mx},
        // so put xwork := -K^*y-xi
        vector_scaleplus(xwork, kcy, fp->xi, -1, -1, 1);
    }else{
        // Just put xwork := - K^*y 
        vector_scaleplus(xwork, kcy, kcy, -1, 0, 1);
    }
    
    c=diffoper_constraint_conjugate(fp->oper, xwork); // G^*(-K*y)
    
    fp->fid->pdgap(pdgap, NULL, y, NULL, kx, NULL, iter, fp->fidp);
    
    if(iter==0){
        fp->pdgap0_ab=ab;
        fp->pdgap0_c=c;
    }
    
    pdgap[0]+=ab+mx*c;
    pdgap[1]+=fp->pdgap0_ab+mx*fp->pdgap0_c;
}

////////


static FidPrimal *fid_oper_alloc_primal(FidOperParams *fp)
{
    return (FidPrimal*)diffoper_alloc_dom(fp->oper);
}

static FidDual *fid_oper_alloc_dual(FidOperParams *fp)
{
    return fp->fid->alloc_dual(fp->fidp);
}


static void fid_oper_init_dual(FidDual *psi, const FidPrimal *u, 
                               FidOperParams *fp)
{
    diffoper_apply_grad(fp->oper, fp->tmp, fp->oper_u0, u, FALSE);
    fp->fid->init_dual(psi, (FidPrimal*)fp->tmp, fp->fidp);
}


static FNumber fid_oper_baseshift(FidPrimal *u, FidBaseshiftMode mode, 
                                  FidOperParams *fp)
{
    DOperCod *tmp=fp->tmp;
    FNumber dist=INFINITY;

    assert(fp->fid->baseshift!=NULL);
    
    assert(fp->oper_u0!=NULL);

    //printf("Baseshift %d : %f\n", mode, vector_norm(u));
    
    if(mode==FidBaseshiftBregman || mode==FidBaseshiftSet){
        if(mode==FidBaseshiftBregman){
            bool add_to_xi=TRUE;
            // Shift linearisation point: Bregmanise through xi
            // xi := xi + grad H(u0)^* [\grad F](H(u0)+\grad H(u0) (u-u0)),
            //     = xi + grad H(u0)^* [\grad F](H(u0)+\grad H(u0) (u-u0)),
            // where H is the non-linearisation, and F the fidelity function.
            // The part H(u0)-\grad H(u0) u0 has been incorporated into F
            // with fid_baseshift (below, in next step, where we update it).
            DOperCod *tmp2=diffoper_alloc_cod(fp->oper);
            //diffoper_calculate(fp->oper, tmp2, u);
            diffoper_apply_grad(fp->oper, tmp2, fp->oper_u0, u, FALSE);
            fp->fid->grad((FidPrimal*)tmp, (FidPrimal*)tmp2, fp->fidp);
            destroy_obj((Obj*)tmp2);
            if(fp->xi==NULL){
                fp->xi=diffoper_alloc_dom(fp->oper);
                add_to_xi=FALSE;
            }
            diffoper_apply_gradc(fp->oper, fp->xi, fp->oper_u0, tmp, add_to_xi);
        }
        
        // New linearisation point
        vector_copy(fp->oper_u0, u);
#if 0
        // Shift fp->fid by H(u0) - grad H(u0) u0, where H=fp->oper.
        //
        // First set shift to H(u0), in order to calculate a reasonable distance function value.
        diffoper_calculate(fp->oper, tmp, u, FALSE);
        dist=fp->fid->baseshift((FidPrimal*)tmp, FidBaseshiftSet, fp->fidp);
        // Now add -grad G(x0) x0 to shift
        diffoper_apply_grad(fp->oper, tmp, u, u, FALSE);
        vector_scavle(tmp, -1);
        fp->fid->baseshift(tmp, FidBaseshiftAdd, fp->fidp);
#else
        diffoper_calculate(fp->oper, tmp, u, FALSE);
        dist=fp->fid->baseshift((FidPrimal*)tmp, FidBaseshiftNone, fp->fidp);
        vector_scale(tmp, -1);
        diffoper_apply_grad(fp->oper, tmp, u, u, TRUE);
        vector_scale(tmp, -1);
        fp->fid->baseshift(tmp, FidBaseshiftSet, fp->fidp);
#endif
    }else if(mode==FidBaseshiftNone){
        diffoper_calculate(fp->oper, tmp, u, FALSE);
        dist=fp->fid->baseshift(tmp, mode, fp->fidp);
        if(fp->xi){
            destroy_obj((Obj*)fp->xi);
            fp->xi=NULL;
        }
    }else{
        die("Unsupported base-shift for fid_oper.");
    }
    
    return dist;
}

static void fid_oper_init_primal(DOperDom *u, FidInitPrimalStyle initstyle, FidOperParams *fp)
{
    if(initstyle==FidInitPrimalZero){
        vector_zero(u);
    }else if(initstyle==FidInitPrimalBackproj){
        fp->fid->init_primal((FidPrimal*)fp->tmp, initstyle, fp->fidp);
        diffoper_backproj(fp->oper, u, fp->tmp);
    }else{
        die("Unknown FidInitPrimalStyle");
    }
    fid_oper_baseshift(u, FidBaseshiftSet, fp);
}

static FNumber fid_oper_estimate(Estimate estim, const CombinedPrimal *u, FidOperParams *fp)
{
    if(estim==EstimateKNorm){
        FNumber kn=fp->fid->estimate(estim, NULL, fp->fidp);
        FNumber gn=diffoper_estimate_grad_norm(fp->oper, fp->oper_u0);
        return kn*gn;
    }else if(estim==EstimateGUniformConvexity){
        return 0; // G=indicator function of virtual constraint
                  // set for pseudo-duality gap}
    }else if(estim==EstimateFCUniformConvexity){
        return fp->fid->estimate(EstimateFCUniformConvexity, NULL, fp->fidp);
    }else{
        return 0;
    }
}

static bool fid_oper_grad(FidPrimal *grad, const FidPrimal *u, 
                          FidOperParams *fp)
{
    DOperCod *tmp=fp->tmp;
    DOperCod *tmp2=diffoper_alloc_cod(fp->oper);
    bool not_subgrad;
    
    // Calculate gradient of the linearised functional!
    diffoper_apply_grad(fp->oper, tmp, fp->oper_u0, u, FALSE);
    not_subgrad=fp->fid->grad((FidPrimal*)tmp2, (FidPrimal*)tmp, fp->fidp);
    diffoper_apply_gradc(fp->oper, grad, fp->oper_u0, tmp, FALSE);
    if(fp->xi)
        vector_scaleplus(grad, grad, fp->xi, 1, 1, 1);
    
    destroy_obj((Obj*)tmp2);
    
    return not_subgrad;
}

////////

static Fid fid_oper={
    (FidK*)fid_oper_k,
    (FidKC*)fid_oper_kc,
    (FidFCres*)fid_oper_fcres,
    (FidGres*)fid_oper_gres,
    (FidPDgap*)fid_oper_pdgap,
    (FidAllocPrimal*)fid_oper_alloc_primal,
    (FidAllocDual*)fid_oper_alloc_dual,
    (FidInitPrimal*)fid_oper_init_primal,
    (FidInitDual*)fid_oper_init_dual,
    (FidBaseshift*)fid_oper_baseshift,
    (FidEstimate*)fid_oper_estimate,
    (FidGrad*)fid_oper_grad
};

void fid_oper_init(Fid *fid, FidParams *fidp, DOper *oper, 
                   Fid **fid_ret, FidOperParams *fp)
{
    if(oper==NULL)
        die("Operator must be given for fid_oper");

    fp->fid=fid;
    fp->fidp=fidp;
    fp->oper=oper;
    fp->oper_u0=diffoper_alloc_dom(oper);
    vector_zero(fp->oper_u0);
    
    fp->pdgap0_ab=0;
    fp->pdgap0_c=0;
    fp->primal_bound=0;
    fp->xi=NULL;
        
    fp->tmp=diffoper_alloc_cod(fp->oper);
    
    *fid_ret=&fid_oper;
}

////////

static void fid_oper_k_nl(FidDual *a, const DOperDom *b, bool addto_a, FidOperParams *fp)
{
    diffoper_calculate(fp->oper, fp->tmp, b, FALSE);
    fp->fid->k(a, (FidPrimal*)fp->tmp, addto_a, fp->fidp);
}

static void fid_oper_kag_nl(FidDual *a, const DOperDom *b0, const DOperDom *b, bool addto_a, FidOperParams *fp)
{
    diffoper_apply_grad(fp->oper, fp->tmp, b0, b, FALSE);
    fp->fid->k(a, (FidPrimal*)fp->tmp, addto_a, fp->fidp);
}

static void fid_oper_kc_nl(DOperDom *a, DOperDom *a0, const FidDual *b, bool addto_a, FidOperParams *fp)
{
    fp->fid->kc((FidPrimal*)fp->tmp, b, FALSE, fp->fidp);
    diffoper_apply_gradc(fp->oper, a, a0, fp->tmp, addto_a);
}

static void fid_oper_init_dual_nl(FidDual *psi, const FidPrimal *u, 
                               FidOperParams *fp)
{
    diffoper_calculate(fp->oper, fp->tmp, u, FALSE);
    fp->fid->init_dual(psi, (FidPrimal*)fp->tmp, fp->fidp);
}

static void fid_oper_init_primal_nl(DOperDom *u, FidInitPrimalStyle initstyle, FidOperParams *fp)
{
    if(initstyle==FidInitPrimalZero){
        vector_zero(u);
    }else if(initstyle==FidInitPrimalBackproj){
        fp->fid->init_primal((FidPrimal*)fp->tmp, initstyle, fp->fidp);
        diffoper_backproj(fp->oper, u, fp->tmp);
    }else{
        die("Unknown FidInitPrimalStyle");
    }
}

static FNumber fid_oper_baseshift_nl(FidPrimal *u, FidBaseshiftMode mode, 
                                    FidOperParams *fp)
{
    DOperCod *tmp=fp->tmp;
    FNumber dist=INFINITY;

    assert(fp->fid->baseshift!=NULL);
    
    if(mode==FidBaseshiftBregman){
        bool add_to_xi=TRUE;
        // Shift linearisation point: Bregmanise through xi
        DOperCod *tmp2=diffoper_alloc_cod(fp->oper);
        diffoper_calculate(fp->oper, tmp2, u, FALSE);
        fp->fid->grad((FidPrimal*)tmp, (FidPrimal*)tmp2, fp->fidp);
        if(fp->xi==NULL){
            fp->xi=diffoper_alloc_dom(fp->oper);
            add_to_xi=FALSE;
        }
        // xi{i+1} := xi{i} + grad[fid](u{i+1}) in grad[reg](u{i+1})
        // (Actually xi{i+1} := xi{i} - alpha grad(fidelity) in usual formulations,
        // of Bregman iteration, but then xi gets multiplied again by -alpha, through
        // alpha D_xi{i+1}(u, u{i+1}), and we do not have this here.)
        diffoper_apply_gradc(fp->oper, fp->xi, u, tmp, add_to_xi);
        
        dist=fp->fid->baseshift((FidPrimal*)tmp2, FidBaseshiftNone, fp->fidp);
        
        destroy_obj((Obj*)tmp2);
    }else{
        die("Unsupported base-shift");
    }
    
    return dist;
}

static FNumber fid_oper_estimate_nl(Estimate estim, const FidPrimal *u, FidOperParams *fp)
{
    if(estim==EstimateKNorm){
        assert(u!=NULL);
        FNumber kn=fp->fid->estimate(estim, NULL, fp->fidp);
        FNumber gn=diffoper_estimate_grad_norm(fp->oper, u);
        return kn*gn;
    }else if(estim==EstimateGUniformConvexity){
        return 0; // G=indicator function of virtual constraint
                  // set for pseudo-duality gap}
    }else if(estim==EstimateFCUniformConvexity){
        return fp->fid->estimate(EstimateFCUniformConvexity, NULL, fp->fidp);
    }else{
        return 0;
    }
}

static bool fid_oper_grad_nl(FidPrimal *grad, const FidPrimal *u, 
                             FidOperParams *fp)
{
    DOperCod *tmp=fp->tmp;
    DOperCod *tmp2=diffoper_alloc_cod(fp->oper);
    bool not_subgrad;
    
    // Calculate gradient of the non-linear functional!
    diffoper_calculate(fp->oper, tmp, u, FALSE);
    not_subgrad=fp->fid->grad((FidPrimal*)tmp2, (FidPrimal*)tmp, fp->fidp);
    diffoper_apply_gradc(fp->oper, grad, u, tmp, FALSE);
    if(fp->xi)
        vector_scaleplus(grad, grad, fp->xi, 1, 1, 1);
    
    destroy_obj((Obj*)tmp2);
    
    return not_subgrad;
}

////////

static FidNL fid_oper_nl={
    {
        (FidK*)NULL, 
        (FidKC*)NULL,
        (FidFCres*)fid_oper_fcres,
        (FidGres*)fid_oper_gres,
        (FidPDgap*)fid_oper_pdgap,
        (FidAllocPrimal*)fid_oper_alloc_primal,
        (FidAllocDual*)fid_oper_alloc_dual,
        (FidInitPrimal*)fid_oper_init_primal_nl,
        (FidInitDual*)fid_oper_init_dual_nl,
        (FidBaseshift*)fid_oper_baseshift_nl,
        (FidEstimate*)fid_oper_estimate_nl,
        (FidGrad*)fid_oper_grad_nl,
    },
    (FidNLK*)fid_oper_k_nl,
    (FidNLKAG*)fid_oper_kag_nl,
    (FidNLKC*)fid_oper_kc_nl,
};

void fid_oper_init_nonlin(Fid *fid, FidParams *fidp, DOper *oper, 
                          FidNL **fid_ret, FidOperParams *fp)
{
    if(oper==NULL)
        die("Operator must be given for fid_oper");

    fp->fid=fid;
    fp->fidp=fidp;
    fp->oper=oper;
    fp->oper_u0=NULL;
    
    fp->pdgap0_ab=0;
    fp->pdgap0_c=0;
    fp->primal_bound=0;
    fp->xi=NULL;
    
    fp->tmp=diffoper_alloc_cod(fp->oper);
    
    *fid_ret=&fid_oper_nl;
}


