/*
 * dticode/src/fid_po.c
 *
 */

#include <math.h>
#include <stdio.h>
#include <string.h>
#include <libtu/types.h>
#include <libtu/output.h>
#include <libtu/misc.h>

#include "tensor.h"
#include "combine.h"
#include "combine_nonlin.h"
#include "fid_po.h"
#include "scaleplus.h"
#include "project_posdef.h"
#include "tensornorm.h"
#include "posdef.h"
#include "sumdot1.h"
#include "normalise.h"
#include "linoper.h"

#define TENSOR_F(fp) ((TensorField*)fp->f)

//
// The approach:
//
// |A u-f|_F + delta_{u>=0}
//  = max_lambda  delta_{u>=0} + <lambda, A u-f> - delta_{|lambda|<=1}
//  = max_lambda  delta_{u>=0} + <lambda,A u>    - <lambda,f> - delta_{|lambda|<=1}
//  = max_lambda  G(u)         + <lambda,K u>  - F^*(lambda)
//
// F(y) = sup_x <y,x> - <x,f>-delta_{|x|<=1}
//      = |y-f|_1
//
// G(u) = delta_{u>=0}(u) 
//
// G^*(y) = sup_x <y,x>-delta_{x>=0}(x) = 
//      { INFINITY, y has pos eigenvalues
//      { 0,        y neg.semi.def
//     
// Because G^* can be infinite, we need an extra bound |u(x)|<M, 
// where M > sup_x |f(x)| to make the duality gap practical. 
//
// Therefore, we take actually:
//
// G(u) = delta_{u>=0 & |u|<=M}(u)
//
// G^*(y) = sup_x <y,x>-delta_{x>=0 & |x|<=M}(x)
//        = M<normalise_pointwise(P_{>=0}(y)), y>
//        = M|P_{>=0}(y)|_{L1-frob}
// G^*(-K^*y) = M*|P_{>=0}(-K^*y)|
//

static void fid_po_k(FidPODual *a, const FidPOPrimal *b, 
                     bool addto_a, FidPOParams *fp)
{
    linoper_calculate(fp->op, (LinOperCod*)a, b, addto_a);
}


static void fid_po_kc(FidPrimal *a, const FidPODual *b, 
                      bool addto_a, FidPOParams *fp)
{
    linoper_calculatec(fp->op, a, (LinOperCod*)b, addto_a);
}


static void fid_po_gres(FidPrimal *u, FNumber tau, 
                        FidPOParams *fp)
{
    // Project if necessary.
    linoper_projdom(fp->op, u);
}

static void fid_po_fcres(FidPODual *lambda, FNumber tau,
                         FidPOParams *fp)
{
    // min_x{ |lambda-x|^2/2tau+<f,x>+delta_{\alpha B}(x) } 
    // => (x-lambda)/tau + f + N_{\alpha B x} = 0
    //  => x +  N_{\alpha B x} = lambda - tau f;
    
    tensorfield_threshold_frob2(lambda,fp->lb,fp->rb,tau);
//vector_scaleplus(lambda, lambda, fp->f, 1, -tau, 1+tau);
}

static void fid_po_pdgap(FNumber *pdgap,
                         const FidPOPrimal *u, 
                         const FidPODual *y,
                         const FidPOPrimal *kcy,
                         const FidPODual *ku,
                         FidPOPrimal *uwork,
                         unsigned int iter,
                         FidPOParams *fp)
{
    FNumber ab, c, mx,alpha,f;
    /* F(Ku) + G(u) +G^*(-K*y)+F^*(y) */
    ab=0;
    if(u!=NULL){
        // Calculate M such that G_M(u)<infty, where G_M has the constraint set scaled by M.
        mx=diffoper_constraint_scale(&fp->op->doper, u);
        mx=fn_max(mx, fp->primal_bound);
        fp->primal_bound=mx;
        
        // Calculate G^*. 
        vector_scaleplus(uwork, kcy, kcy, -1.0, 0, 1.0);
	FNumber norm_uwork,norm_u;
	norm_uwork=vector_norm(uwork);
	norm_u=vector_norm(u);
	//printf("uwork is %f\n",norm_uwork);
	//printf("u is %f\n",norm_u);     
        c=diffoper_constraint_conjugate(&fp->op->doper, uwork);
	//c=0;
    }else{
        // Called through, e.g., fid_oper that implements the same as
        // above on the operator parameter domain.
        c=0;
        mx=0;
    }
        // Calculate alpha such that F*1(u)<infty 
	 if(TENSOR_F(fp)==NULL){
        ab=tensorfield_sumdot1(ku, ku, FALSE, 2);
	printf("doing the null path");                        // F(Ku)
    }else{
	alpha=vector_norm(y);
	printf("alpha is%f\n",alpha);        
	//printf("t-norm\n"); 
	tensorfield_threshold_norm(ku,fp->lb,fp->rb, &f);  // Calculate F1**
	//printf("t-prod\n");   	  
	tensorfield_threshold_prod(y,fp->lb,fp->rb, &ab);  // F^*(y)
	ab+=f*alpha;
        
	//printf("ab is %f\n",ab);   
     }
	//printf("c is %f\n",c);           
    if(iter==0){
        fp->pdgap0_ab=ab;
        fp->pdgap0_c=c;
    }
    
    pdgap[0]=ab+mx*c;
    pdgap[1]=fp->pdgap0_ab+mx*fp->pdgap0_c;
}

static FidPrimal *fid_po_alloc_primal(FidPOParams *fp)
{


return (FidPrimal*)linoper_alloc_dom(fp->op);
}

static FidDual *fid_po_alloc_dual(FidPOParams *fp)
{
     LinOperCod *p=linoper_alloc_cod(fp->op);
     
     assert(OBJ_IS(p, TensorField));
     
     return (FidDual*)p;
}

void fid_po_init_primal(FidPOPrimal *u, FidInitPrimalStyle initstyle, 
                        FidPOParams *fp)
{
    if(initstyle==FidInitPrimalZero){
        vector_zero(u);
    }else if(initstyle==FidInitPrimalBackproj){
        linoper_backproj(fp->op, u, fp->f);
    }else{
        die("Unknown FidInitPrimalStyle");
    }
}

void fid_po_init_dual(FidPODual *b, const FidPOPrimal *a, 
                      FidPOParams *fp)
{
    fid_po_k(b, a, FALSE, fp);
    if(TENSOR_F(fp)!=NULL)
        tensorfield_scaleplus(b, b, TENSOR_F(fp), 1, -1, 1);
    tensorfield_normalise_frob2(b, 1);
}


/*
static FNumber fid_po_baseshift(FidL1Primal *u, FidBaseshiftMode mode, 
                                   FidL1Params *fp)
{
    return fid_l2_baseshift(u, mode, fp);
}
*/

static FNumber fid_po_estimate(Estimate estim, const FidPrimal *u, FidPOParams *fp)
{
    if(estim==EstimateKNorm)
        return linoper_estimate_norm(fp->op);
    else if(estim==EstimateGUniformConvexity)
        return 0;
    else if(estim==EstimateFCUniformConvexity)
        return INFINITY;
    else
        return 0;
}

static bool fid_po_grad(TensorField *grad, const TensorField *u, 
                        FidPOParams *fp)
{
    if(TENSOR_F(fp)!=NULL)
        tensorfield_scaleplus(grad, u, TENSOR_F(fp), 1, -1, 1);
    tensorfield_normalise_frob2(grad, 1);
    return FALSE; // Only a subgradient
}

static Fid fid_po={
    (FidK*)fid_po_k,
    (FidKC*)fid_po_kc,
    (FidFCres*)fid_po_fcres,
    (FidGres*)fid_po_gres,
    (FidPDgap*)fid_po_pdgap,
    (FidAllocPrimal*)fid_po_alloc_primal,
    (FidAllocDual*)fid_po_alloc_dual,
    (FidInitPrimal*)fid_po_init_primal,
    (FidInitDual*)fid_po_init_dual,
    (FidBaseshift*)NULL, //fid_po_baseshift,
    (FidEstimate*)fid_po_estimate,
    (FidGrad*)fid_po_grad,
};



void fid_po_init(const LinOperCod *lb,const LinOperCod *rb, const LinOperCod *f,
                 LinOper *op,
                 Fid **fid, 
                 FidPOParams *fp)
{
    assert(f==NULL || OBJ_IS(f, TensorField));
    printf("init fid\n");
    fp->mask=NULL;//mask;
    fp->pdgap0_ab=0;
    fp->pdgap0_c=0;
    fp->mask_term_weight=0;//mask_term_weight;
    fp->f=f;
    fp->f_orig=f;
    fp->f_tmp=NULL;
    fp->op=op;
    fp->primal_bound=0;
    fp->lb=lb;
    fp->rb=rb;
    //param->ubound=1+10*tensorfield_frobinf(f)/sqrt(matrix_normsquared_lowerbound(amat));
    
    *fid=&fid_po;
}
