/*
 * dticode/src/fid_split.c
 *
 * Written 2013 by Tuomo Valkonen <tuomov@iki.fi>, University of Cambridge.
 *
 */

#include <stdio.h>
#include <math.h>
#include <assert.h>
#include <libtu/misc.h>
#include <libtu/output.h>
#include "combine.h"
#include "fid_split.h"
#include "splitvar.h"

////////

#include "splitp.h"

////////

static void fid_split_k(SplitDual *a, const SplitPrimal *b, bool addto_a, FidSplitParams *spp)
{
    #pragma omp parallel for
    for(size_t i=0; i<spp->n; i++){
        FidDual *a_i;
        const FidPrimal *b_i;
        
        get_dual_i(&a_i, a, i);
        get_primal_i_const(&b_i, b, i);
        
        spp->fid[i]->k(a_i, b_i, addto_a, spp->params[i]);
    }
}

static void fid_split_kc(SplitPrimal *a, const SplitDual *b, bool addto_a, FidSplitParams *spp)
{
    #pragma omp parallel for
    for(size_t i=0; i<spp->n; i++){
        FidPrimal *a_i;
        const FidDual *b_i;
        
        get_primal_i(&a_i, a, i);
        get_dual_i_const(&b_i, b, i);
        
        spp->fid[i]->kc(a_i, b_i, addto_a, spp->params[i]);
    }
}

////////

static void fid_split_fcres(SplitDual *a, FNumber tau, FidSplitParams *spp)
{
    #pragma omp parallel for
    for(size_t i=0; i<spp->n; i++){
        FidDual *a_i;
        get_dual_i(&a_i, a, i);
        
        spp->fid[i]->fcres(a_i, tau, spp->params[i]);
    }
}

static void fid_split_gres(SplitPrimal *a, FNumber tau, FidSplitParams *spp)
{
    #pragma omp parallel for
    for(size_t i=0; i<spp->n; i++){
        FidPrimal *a_i;
        get_primal_i(&a_i, a, i);
        
        spp->fid[i]->gres(a_i, tau, spp->params[i]);
    }
}

////////

static void fid_split_pdgap(FNumber *pdgap, 
                            const SplitPrimal *x, 
                            const SplitDual *y,
                            const SplitPrimal *kcy, 
                            const SplitDual *kx,
                            SplitPrimal *xwork,
                            unsigned int iter,
                            FidSplitParams *spp)
{
    pdgap[0]=0;
    pdgap[1]=0;
    
    #pragma omp parallel for
    for(size_t i=0; i<spp->n; i++){
        const FidPrimal *x_i=NULL;
        const FidDual *y_i;
        const FidPrimal *kcy_i=NULL;
        const FidDual *kx_i;
        FidPrimal *xwork_i=NULL;
        FNumber pdgap_i[2];
        
        if(x!=NULL){
            // fid_oper calls with NULL
            get_primal_i_const(&x_i, x, i);
            get_primal_i_const(&kcy_i, kcy, i);
            get_primal_i(&xwork_i, xwork, i);
        }
        get_dual_i_const(&y_i, y, i);
        get_dual_i_const(&kx_i, kx, i);
        
        spp->fid[i]->pdgap(pdgap_i, x_i, y_i, kcy_i, kx_i, xwork_i, iter, spp->params[i]);
        
        pdgap[0]+=pdgap_i[0];
        pdgap[1]+=pdgap_i[1];
    }
}

////////


static SplitPrimal *fid_split_alloc_primal(FidSplitParams *spp)
{
    SplitDual *d=alloc_splitvar(spp->n);
    
    for(size_t i=0; i<spp->n; i++)
        d->p[i]=spp->fid[i]->alloc_primal(spp->params[i]);
        
    return d;
}

static SplitDual *fid_split_alloc_dual(FidSplitParams *spp)
{
    SplitDual *d=alloc_splitvar(spp->n);
    
    for(size_t i=0; i<spp->n; i++)
        d->p[i]=spp->fid[i]->alloc_dual(spp->params[i]);
        
    return d;
}

static void fid_split_init_dual(SplitDual *psi, const SplitPrimal *u, 
                                FidSplitParams *spp)
{
    for(size_t i=0; i<spp->n; i++){
        const FidPrimal *u_i;
        FidDual *psi_i;
        
        get_primal_i_const(&u_i, u, i);
        get_dual_i(&psi_i, psi, i);
        
        spp->fid[i]->init_dual(psi_i, u_i, spp->params[i]);
    }
}

static void fid_split_init_primal(SplitPrimal *u, 
                                  FidInitPrimalStyle initstyle,
                                  FidSplitParams *spp)
{
    for(size_t i=0; i<spp->n; i++){
        FidPrimal *u_i;
        
        get_primal_i(&u_i, u, i);
        
        spp->fid[i]->init_primal(u_i, initstyle, spp->params[i]);
    }
}

static FNumber fid_split_baseshift(SplitPrimal *u, FidBaseshiftMode mode, 
                                   FidSplitParams *spp)
{
    FNumber dist=0;
    
    #pragma omp parallel for
    for(size_t i=0; i<spp->n; i++){
        FidPrimal *u_i;
        
        get_primal_i(&u_i, u, i);
        
        dist+=spp->fid[i]->baseshift(u_i, mode, spp->params[i]);
    }

    return dist;
}

static FNumber fid_split_estimate(Estimate estim, SplitPrimal *u, FidSplitParams *spp)
{
    if(estim==EstimateKNorm){
        FNumber knorm=0;
        #pragma omp parallel for
        for(size_t i=0; i<spp->n; i++){
            const FidPrimal *u_i=NULL;
            if(u!=NULL)
                get_primal_i_const(&u_i, u, i);
            knorm+=spp->fid[i]->estimate(EstimateKNorm, u_i, spp->params[i]);
        }
        return knorm;
    }else if(estim==EstimateGUniformConvexity || estim==EstimateFCUniformConvexity){
        FNumber uc=0;
        #pragma omp parallel for
        for(size_t i=0; i<spp->n; i++){
            const FidPrimal *u_i=NULL;
            if(u!=NULL)
                get_primal_i_const(&u_i, u, i);
            uc=fn_min(uc, spp->fid[i]->estimate(estim, u_i, spp->params[i]));
        }
        return uc;
    }else{
        return 0;
    }
}

static bool fid_split_grad(SplitPrimal *grad, const SplitPrimal *u,
                           FidSplitParams *spp)
{
    bool not_subgrad=TRUE;
    
    #pragma omp parallel for
    for(size_t i=0; i<spp->n; i++){
        const FidPrimal *u_i;
        FidPrimal *grad_i;
        
        get_primal_i_const(&u_i, u, i);
        get_primal_i(&grad_i, grad, i);
        
        bool not_subgrad_i=spp->fid[i]->grad(grad_i, u_i, spp->params[i]);
        
        not_subgrad=not_subgrad && not_subgrad_i;
    }
    
    return not_subgrad;
}

////////

static Fid fid_split={
    (FidK*)fid_split_k,
    (FidKC*)fid_split_kc,
    (FidFCres*)fid_split_fcres,
    (FidGres*)fid_split_gres,
    (FidPDgap*)fid_split_pdgap,
    (FidAllocPrimal*)fid_split_alloc_primal,
    (FidAllocDual*)fid_split_alloc_dual,
    (FidInitPrimal*)fid_split_init_primal,
    (FidInitDual*)fid_split_init_dual,
    (FidBaseshift*)fid_split_baseshift,
    (FidEstimate*)fid_split_estimate,
    (FidGrad*)fid_split_grad,
};

////////

void fid_split_init(Fid **fid, FidSplitParams *spp)
{
    *fid=&fid_split;
}
