/*
 * dticode/src/ksreg-nl.c
 *
 * Written 2012-2013 by Tuomo Valkonen <tuomov@iki.fi>, University of Cambridge.
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <libtu/util.h>
#include <libtu/output.h>
#include <libtu/optparser.h>
#include <libtu/misc.h>

#include "parallel.h"
#include "cpock_nonlin.h"
#include "tensor.h"
#include "reg.h"
#include "reg_split.h"
#include "fid_l2alt.h"
#include "fid_oper.h"
#include "fid_split.h"
#include "fftop.h"
#include "timing.h"
#include "matrix.h"
#include "diffoper.h"
#include "diffoper_complex.h"
#include "diffoper_phasediff.h"
#include "fft.h"
#include "method.h"

static OptParserOpt opts[]={
    QUICK_OPT('R', "regrho",  "Regulariser of phi: tv|tgv2|none,alpha,beta"),
    QUICK_OPT('Q', "regphi",  "Regulariser of rho: tv|tgv2|none,alpha,beta"),
    QUICK_OPT('v', "regv",    "Regulariser of v: tv|tgv2|none,alpha,beta"),
    //QUICK_OPT('F', "fid",     "Fidelity: fourier-l2alt"),
    QUICK_OPT('1', "infile",  "Input file (raw doubles)"),
    QUICK_OPT('2', "infile2", "Input file (raw doubles)"),
    QUICK_OPT('k', "fdomain", "Data domain: kspace|kspace-sparse|kspace-nonuniform. Sparse format consists of a sequence of triples (idx, real, imag), each a double with the linear index idx assumed to have integral value. Nonuniform format consists of (freq1, freq2, real, imag) quadruples."),
    {OPT_ID('S'), "sparse", 0, NULL, "Same as -fdomain kspace-sparse"},
    //QUICK_OPT('M', "maskfile", "Mask file (raw doubles)"),
    //QUICK_OPT('W', "lambda",  "Mask weight"),
    METHOD_OPTS,
    END_OPTPARSEROPTS
};

                            
typedef union{
    FidL2AltParams l2alt;
    FidSplitParams split;
} FidBaseP;

int main(int argc, char *argv[])
{
    MethodParam mp=DEFAULT_METHOD_PARAM;
    
    mp.sig.is_complex=TRUE;
    
    const char *rho_regspec="tv,1";
    const char *phi_regspec="tv,1";
    const char *v_regspec=NULL;
    const char *infile=NULL;
    const char *infile2=NULL;
    TensorField *f=NULL, *f2=NULL;
    Vector *u=NULL;
    //TensorField *mask=NULL;
    //FNumber mask_term_weight=0;
    int opt;
    const char *datadomain="kspace";
    
    FidNL *fidnl=NULL;
    Fid *fid=NULL;
    FidOperParams fidp;
    Fid *fid_base=NULL;
    FidBaseP fidp_base;
    Fid *fid_base_split[2];
    FidBaseP fidp_base_split[2];
    void *fidp_base_split_p[2]={&fidp_base_split[0], &fidp_base_split[1]};
    Reg *reg;
    RegSplitParams regp;
    Reg *reg_split[4];
    RegP regp_split[4];
    void *regp_split_p[4]={&regp_split[0], &regp_split[1], &regp_split[2], &regp_split[3]};
    DOper *doper;
    LinOper *fftop, *fftop2;
    
    parallel_init1(&argc, &argv);
    
    libtu_init(argv[0]);
    optparser_init(argc, argv, OPTP_MIDLONG, opts);
    
    while((opt=optparser_get_opt())){
        switch(opt){
            case OPT_ID('R'):
                rho_regspec=optparser_get_arg();
                break;
            case OPT_ID('Q'):
                phi_regspec=optparser_get_arg();
                break;
            case OPT_ID('v'):
                v_regspec=optparser_get_arg();
                break;
            case OPT_ID('1'):
                infile=optparser_get_arg();
                break;
            case OPT_ID('2'):
                infile2=optparser_get_arg();
                break;
            //case OPT_ID('M'):
            //    maskfile=optparser_get_arg();
            //    break;
            case OPT_ID('k'):
                datadomain=optparser_get_arg();
                break;
            case OPT_ID('S'):
                datadomain="kspace-sparse";
                break;
            default:
                if(!method_do_opt(&mp, opt, opts)){
                    optparser_print_error();
                    return 1;
                }
                break;
        }
    }
    
    
    if(infile==NULL)
        die("infile missing\n");

    method_params_sanity_check(&mp);
        
    parallel_init2();
    fft_init();

    // Create FFT operators
    fftop=load_fftop(infile, &mp.sig, datadomain, &f);

    if(infile2!=NULL)
        fftop2=load_fftop(infile2, &mp.sig, datadomain, &f2);

    if(infile2==NULL){
        // Init base fidelity
        fid_l2alt_init((FidL2AltPrimal*)f, fftop, 
                        &fid_base, &fidp_base.l2alt);
        // Init nonlinear operator
        doper=(DOper*)create_diffoper_complex(&mp.sig);
    }else{
        // Init base fidelity
        fid_l2alt_init((FidL2AltPrimal*)f, fftop, 
                        &fid_base_split[0], &fidp_base_split[0].l2alt);
        fid_l2alt_init((FidL2AltPrimal*)f2, fftop2, 
                        &fid_base_split[1], &fidp_base_split[1].l2alt);
        fidp_base.split.n=2;
        fidp_base.split.fid=fid_base_split;
        fidp_base.split.params=fidp_base_split_p;
        fid_split_init(&fid_base, &fidp_base.split);
        
        // Init nonlinear operator
        doper=(DOper*)create_diffoper_phasediff(&mp.sig);
    }

    // Init fidelity nonlinearisation 
    if(mp.gaussnewton){
        fid_oper_init(fid_base, (FidParams*)&fidp_base, 
                      doper, &fid, &fidp);
    }else{
        fid_oper_init_nonlin(fid_base, (FidParams*)&fidp_base, 
                             doper, &fidnl, &fidp);
    }

    // Init regulariser
    Sig signc=mp.sig;
    signc.is_complex=FALSE;
        
    if(v_regspec==NULL)
        v_regspec=phi_regspec;
        
    if(infile2){
        init_regulariser(&reg_split[0], &regp_split[0], rho_regspec, &signc, mp.diffalg);
        init_regulariser(&reg_split[1], &regp_split[1], v_regspec, &signc, mp.diffalg);
        init_regulariser(&reg_split[2], &regp_split[2], rho_regspec, &signc, mp.diffalg);
        init_regulariser(&reg_split[3], &regp_split[3], phi_regspec, &signc, mp.diffalg);
        regp.n=4;
    }else{
        init_regulariser(&reg_split[0], &regp_split[0], rho_regspec, &signc, mp.diffalg);
        init_regulariser(&reg_split[1], &regp_split[1], phi_regspec, &signc, mp.diffalg);
        regp.n=2;
    }
    
    regp.reg=reg_split;
    regp.params=regp_split_p;
    reg_split_init(&reg, &regp);

    if(mp.gaussnewton)
        method_do(&mp, fid, &fidp, reg, &regp);
    else
        method_do_nl(&mp, fidnl, &fidp, reg, &regp);

    fft_finalise();
    
    parallel_finalise();
    
    return 0;
}
