/*
 * dticode/src/ksreg.c
 *
 * Written 2012 by Tuomo Valkonen <tuomov@iki.fi>, University of Cambridge.
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <libtu/util.h>
#include <libtu/output.h>
#include <libtu/optparser.h>
#include <libtu/misc.h>

#include "parallel.h"
#include "cpock_accel.h"
#include "tensor.h"
#include "fid.h"
#include "timing.h"
#include "combine.h"
#include "matrix.h"
#include "feedback.h"
#include "fft.h"
#include "fftop.h"
#include "method.h"
#include "linoper_multiply_by.h"
#include "linoper_compose.h"
#include "multiply.h"
#include "reg.h"

static OptParserOpt opts[]={
    QUICK_OPT('R', "reg",     "Regulariser: tv|tgv2|none,alpha,beta"),
    QUICK_OPT('F', "fid",     "Fidelity: l1|l2|l2alt"),
    QUICK_OPT('f', "infile",  "Input file (raw doubles)"),
    QUICK_OPT('E', "wvec",    "Weight vector w (raw doubles)"),
    //QUICK_OPT('M', "maskfile", "Mask file (raw doubles)"),
    //QUICK_OPT('W', "lambda",  "Mask weight"),
    QUICK_OPT('k', "fdomain", "Data domain: kspace|kspace-sparse|kspace-nonuniform. Sparse format consists of a sequence of triples (idx, real, imag), each a double with the linear index idx assumed to have integral value. Nonuniform format consists of (freq1, freq2, real, imag) quadruples."),
    {OPT_ID('S'), "sparse", 0, NULL, "Same as -fdomain kspace-sparse"},
    METHOD_OPTS,
    END_OPTPARSEROPTS
};


int main(int argc, char *argv[])
{
    MethodParam mp=DEFAULT_METHOD_PARAM;
    
    mp.sig.is_complex=TRUE;
    
    const char *regspec="tv,1,1";
    const char *fidname="l2";
    FNumber alpha=1;
    FNumber beta=0.5;
    const char *infile=NULL;
    //const char *maskfile=NULL;
    const char *datadomain="kspace";
    const char *wvecfile=NULL;
    TensorField *f=NULL;
    TensorField *u=NULL;
    //TensorField *mask=NULL;
    //FNumber mask_term_weight=0;
    int opt;
    LinOper *fftop=NULL;
    FidP fidp;
    RegP regp;
    Fid *fid;
    Reg *reg;
    
    parallel_init1(&argc, &argv);
    
    libtu_init(argv[0]);
    optparser_init(argc, argv, OPTP_MIDLONG, opts);
    
    while((opt=optparser_get_opt())){
        switch(opt){
            //case OPT_ID('W'):
            //    mask_term_weight=atof(optparser_get_arg());
            //    break;
            case OPT_ID('R'):
                regspec=optparser_get_arg();
                break;
            case OPT_ID('F'):
                fidname=optparser_get_arg();
                break;
            case OPT_ID('f'):
                infile=optparser_get_arg();
                break;
            //case OPT_ID('M'):
            //    maskfile=optparser_get_arg();
            //    break;
            case OPT_ID('k'):
                datadomain=optparser_get_arg();
                break;
            case OPT_ID('S'):
                datadomain="kspace-sparse";
                break;
            case OPT_ID('E'):
                wvecfile=optparser_get_arg();
                break;
            default:
                if(!method_do_opt(&mp, opt, opts)){
                    optparser_print_error();
                    return 1;
                }
                break;
        }
    }
    
    if(infile==NULL)
        die("infile missing\n");
        
    method_params_sanity_check(&mp);
        
    parallel_init2();
    
    fft_init();
    
    /*if(maskfile!=NULL){
        Sig masksig=mp.sig;
        masksig.is_complex=FALSE;
        mask=load_tensorfield(&masksig, maskfile);
    }*/

    fftop=load_fftop(infile, &mp.sig, datadomain, &f);
    
    if(fftop==NULL)
        die_err();
        
    if(wvecfile!=NULL){
        Sig wsig=f->sig; wsig.is_complex=FALSE;
        TensorField *wvec=load_tensorfield(&wsig, wvecfile);
        tensorfield_multiply_pointwise(f, wvec, f, FALSE);
        LinOper *wop=(LinOper*)create_linoper_multiply_by(&f->sig, wvec);
        fftop=(LinOper*)create_linoper_compose(wop, fftop);
    }

    /*{
        Vector *u=linoper_alloc_dom(fftop);
        Vector *d=linoper_alloc_cod(fftop);
        linoper_calculatec(fftop, u, f, FALSE);
        vector_save(u, "INVERSION");
        linoper_calculate(fftop, d, u, FALSE);
        vector_save(d, "FORWARD");
    }*/
    
    init_fidelity(&fid, &fidp, fidname, fftop, f, f ,f);

    init_regulariser(&reg, &regp, regspec, &mp.sig, mp.diffalg);
    
    method_do(&mp, fid, &fidp, reg, &regp);
    
    fft_finalise();
    
    parallel_finalise();

    return 0;
}
