/*
 * dticode/src/linoper.c
 *
 * Written 2013 by Tuomo Valkonen <tuomov@iki.fi>, University of Cambridge.
 *
 */

#include <libtu/types.h>
#include <libtu/objp.h>
#include "tensor.h"
#include "vector.h"
#include "number.h"
#include "linoper.h"

void linoper_calculate(LinOper *op,
                       LinOperCod *y, 
                       const LinOperDom *x,
                       bool addto_y)
{
    CALL_DYN(linoper_calculate, op, (op, y, x, addto_y));
}

void linoper_calculatec(LinOper *op,
                        LinOperDom *y, 
                        const LinOperCod *x,
                        bool addto_y)
{
    CALL_DYN(linoper_calculatec, op, (op, y, x, addto_y));
}

void linoper_backproj(LinOper *op,
                      LinOperDom *x, 
                      const LinOperCod *y)
{
    CALL_DYN(linoper_backproj, op, (op, x, y));
}

static void linoper_projdom_default(LinOper *op, LinOperDom *x)
{
}

void linoper_projdom(LinOper *op,
                     LinOperDom *x)
{
    CALL_DYN(linoper_projdom, op, (op, x));
}

LinOperDom *linoper_alloc_dom(LinOper *op)
{
    LinOperDom *ret=NULL;
    
    CALL_DYN_RET(ret, LinOperDom*, linoper_alloc_dom, op, (op));

    return ret;
}

LinOperCod *linoper_alloc_cod(LinOper *op)
{
    LinOperCod *ret=NULL;
    
    CALL_DYN_RET(ret, LinOperCod*, linoper_alloc_cod, op, (op));
    
    return ret;
}

FNumber linoper_estimate_norm_default(const LinOper *op)
{
    return INFINITY;
}

FNumber linoper_estimate_norm(const LinOper *op)
{
    FNumber ret=0;
    
    CALL_DYN_RET(ret, FNumber, linoper_estimate_norm, op, (op));
    
    return ret;
}

FNumber linoper_estimate_lower_bound_default(const LinOper *op)
{
    return 0;
}

FNumber linoper_estimate_lower_bound(const LinOper *op)
{
    FNumber ret=0;
    
    CALL_DYN_RET(ret, FNumber, linoper_estimate_lower_bound, op, (op));
    
    return ret;
}

bool linoper_init(LinOper *op)
{
    return diffoper_init(&op->doper);
}

void linoper_deinit(LinOper *op)
{
    diffoper_deinit(&op->doper);
}

//
// DiffOper defaults
//

void diffoper_linoper_calculate(LinOper *op,
                                LinOperCod *y, 
                                const LinOperDom *x, 
                                bool addto_y)
{
    linoper_calculate(op, y, x, addto_y);
}

void diffoper_linoper_backproj(LinOper *op,
                               LinOperDom *x, 
                               const LinOperCod *y)
{
    linoper_backproj(op, x,  y);
}

void diffoper_linoper_apply_grad(LinOper *op,
                                 LinOperCod *y, 
                                 const LinOperDom *x0, 
                                 const LinOperDom *x, 
                                 bool addto_y)
{
    linoper_calculate(op, y, x, addto_y);
}


// calculate du = [grad G(u)]^* dy = - s_0 exp(-<b, u>) (b^* dy) 
//                                 = - b^* (s_0 exp(-<b, u>) dy)
// pointwise, multiple signals
void diffoper_linoper_apply_gradc(LinOper *op,
                                  LinOperDom *x, 
                                  const LinOperDom *x0, 
                                  const LinOperCod *y, 
                                  bool addto_x)
{
    linoper_calculatec(op, x, y, addto_x);
}

FNumber linoper_calcdist_default(LinOper *op,
                                 const LinOperDom *x,
                                 const LinOperCod *f,
                                 FNumber *m)
{
    FNumber g, d;
    LinOperCod *tmp;
    
    tmp=linoper_alloc_cod(op);
    
    linoper_calculate(op, tmp, x, FALSE);
    d=vector_dist(f, tmp);
    g=d*d/2;
    
    destroy_obj((Obj*)tmp);
    
    if(m!=NULL)
        *m=diffoper_constraint_scale(&op->doper, x);
    
    return g;
}

FNumber linoper_calcdist(LinOper *op,
                         const LinOperDom *x,
                         const LinOperCod *f,
                         FNumber *m)
{
    FNumber ret=0;
    
    CALL_DYN_RET(ret, FNumber, linoper_calcdist, op, (op, x, f, m));
    
    return ret;
}                      

LinOperDom *diffoper_linoper_alloc_dom(LinOper *op)
{
    return linoper_alloc_dom(op);
}

LinOperCod *diffoper_linoper_alloc_cod(LinOper *op)
{
    return linoper_alloc_cod(op);
}

FNumber diffoper_linoper_constraint_scale(LinOper *op, const LinOperDom *x)
{
    // Dummy
    FNumber tmp=vector_norm(x);
    return tmp*tmp/2;
}

FNumber diffoper_linoper_constraint_conjugate(LinOper *op, const LinOperDom *x)
{
    // Dummy
    FNumber tmp=vector_norm(x);
    return tmp*tmp/2;
}

FNumber diffoper_linoper_estimate_grad_norm(LinOper *op, const LinOperDom *x)
{
    return linoper_estimate_norm(op);
}

static DynFunTab linoper_dynfuntab[]={
    {diffoper_calculate, diffoper_linoper_calculate},
    {diffoper_backproj, diffoper_linoper_backproj},
    {diffoper_apply_grad, diffoper_linoper_apply_grad},
    {diffoper_apply_gradc, diffoper_linoper_apply_gradc},
    {diffoper_projdom, diffoper_linoper_apply_gradc},
    {diffoper_projdom, linoper_projdom},
    {linoper_projdom, linoper_projdom_default},
    {(DynFun*)diffoper_alloc_dom, (DynFun*)diffoper_linoper_alloc_dom},
    {(DynFun*)diffoper_alloc_cod, (DynFun*)diffoper_linoper_alloc_cod},
    {(DynFun*)diffoper_constraint_scale, (DynFun*)diffoper_linoper_constraint_scale},
    {(DynFun*)diffoper_constraint_conjugate, (DynFun*)diffoper_linoper_constraint_conjugate},
    {(DynFun*)diffoper_estimate_grad_norm, (DynFun*)diffoper_linoper_estimate_grad_norm},
    {(DynFun*)linoper_estimate_norm, (DynFun*)linoper_estimate_norm_default},
    {(DynFun*)linoper_estimate_lower_bound, (DynFun*)linoper_estimate_lower_bound_default},
    {(DynFun*)linoper_calcdist, (DynFun*)linoper_calcdist_default},
    END_DYNFUNTAB,
};

IMPLCLASS(LinOper, DOper, linoper_deinit, linoper_dynfuntab);

// L2

FNumber linoper_l2_g_default(LinOperL2 *op,
                             const LinOperDom *x,
                             const LinOperCod *f,
                             FNumber *m)
{
    FNumber d=linoper_calcdist_default(&op->linoper, x, f, m);
    return d*d/2;
}

FNumber linoper_l2_calcdist_default(LinOperL2 *op,
                                    const LinOperDom *x,
                                    const LinOperCod *f,
                                    FNumber *m)
{
    FNumber g=linoper_l2_g(op, x, f, m);
    return fn_sqrt(2*g);
}

FNumber linoper_l2_g(LinOperL2 *op,
                     const LinOperDom *x,
                     const LinOperCod *f,
                     FNumber *m)
{
    FNumber ret=0;
    
    CALL_DYN_RET(ret, FNumber, linoper_l2_g, op, (op, x, f, m));
    
    return ret;
}                      

FNumber linoper_l2_gc_default(LinOperL2 *op,
                              const LinOperDom *minus_x,
                              const LinOperCod *f,
                              LinOperDom *xwork,
                              FNumber *c)
{
    // Does not in general work for problems with fake constraints!
    *c=0;
    
    vector_scaleplus(xwork, minus_x, minus_x, -1, 0, 1);
    linoper_l2_gsolve(op, xwork, 1.0, 0.0, 1.0, f);
    return (vector_iprod(xwork, minus_x) - linoper_l2_g(op, xwork, f, NULL));
}

FNumber linoper_l2_gc(LinOperL2 *op,
                      const LinOperDom *minus_x,
                      const LinOperCod *f,
                      LinOperDom *xwork,
                      FNumber *c)
{
    FNumber ret=0;
    
    CALL_DYN_RET(ret, FNumber, linoper_l2_gc, op, (op, minus_x, f, xwork, c));
    
    return ret;
}                      

// Solve for y: zeta y - xweight x + tau K^*(Ky-f) \in -N_{constraint set}
void linoper_l2_gsolve(LinOperL2 *op,
                       LinOperDom *x,
                       FNumber tau, FNumber zeta, FNumber xweight,
                       const LinOperCod *f)
{
    CALL_DYN(linoper_l2_gsolve, op, (op, x, tau, zeta, xweight, f));
}

#if 0
void linoper_l2_projdom(LinOperL2 *op, LinOperDom *x)
{
    // Solve for y: 1.0 y - 1.0 x \in -N_{constraint set}
    LinOperCod *tmp=linoper_alloc_cod(&op->linoper);
    linoper_l2_gsolve(op, x, 0.0, 1.0, 1.0, tmp);
    destroy_obj((Obj*)tmp);
}
#endif

bool linoper_l2_init(LinOperL2 *a)
{
    return linoper_init(&a->linoper);
}

void linoper_l2_deinit(LinOperL2 *a)
{
    linoper_deinit(&a->linoper);
}

static DynFunTab linoper_l2_dynfuntab[]={
    {(DynFun*)linoper_l2_gc, (DynFun*)linoper_l2_gc_default}, 
    {(DynFun*)linoper_l2_g, (DynFun*)linoper_l2_g_default},
    {(DynFun*)linoper_calcdist, (DynFun*)linoper_l2_calcdist_default},
#if 0
    {linoper_projdom, linoper_l2_projdom},
#endif
    END_DYNFUNTAB,
};

IMPLCLASS(LinOperL2, LinOper, linoper_l2_deinit, linoper_l2_dynfuntab);

