/*
 * dticode/src/linoper_nufft.c
 *
 * Written 2013 by Tuomo Valkonen <tuomov@iki.fi>, University of Cambridge.
 *
 */


#include <libtu/types.h>
#include <libtu/objp.h>
#include <libtu/misc.h>
#include <libtu/output.h>
#include <nfft3.h>
#include <fftw3.h>
#include "tensor.h"
#include "vector.h"
#include "number.h"
#include "linoper.h"
#include "linoper_nufft.h"
#include "scaleplus.h"
#include "complexx.h"
#include "sumdot1.h"
#include "tensornorm.h"
#include "doubleio.h"
#include "fft.h"

#define USE_NFFT_INIT_HACK 1

#if _OPENACC
    #error "linoper_nufft does not support OpenACC"
#endif

#if _MPI
    #error "linoper_nufft does not support MPI"
#endif

static __inline__ FNumber get_nfft_weight(LinOperNUFFT *op)
{
#if 1
    size_t rank;
    int dims[3];
    FNumber normfact;

    fft_init_params(&rank, dims, &normfact, &op->sig);
    
    return fn_sqrt(normfact);
#elif 0
    return fn_sqrt(op->n_samples);
#else
    return 1.0;
#endif
}

static void check_norm(TensorField *a, const TensorField *b, const char *nm, FNumber w)
{
    FNumber na=tensorfield_frob2(a);
    FNumber nb=tensorfield_frob2(b);
    if(na>10*w*nb)
        printf("WARNING: %s %f --> %f\n", nm, nb, na);
}

void linoper_nufft_calculate(LinOperNUFFT *op,
                             LinOperNUFFTCod *a, 
                             const LinOperNUFFTDom *b,
                             bool addto_a)
{
    FNumber w=get_nfft_weight(op);
    
    if(addto_a){
        op->plan.f_hat=(fftw_complex*)b->p;
        op->plan.f=(fftw_complex*)op->tmp_cod->p;
    
        nfft_trafo(&op->plan);
        
        tensorfield_scaleplus(a, a, op->tmp_cod, 1, 1/w, 1);
    }else{
        op->plan.f_hat=(fftw_complex*)b->p;
        op->plan.f=(fftw_complex*)a->p;
    
        nfft_trafo(&op->plan);
    
        if(w!=1.0)
            tensorfield_scale(a, 1/w);
    }
    
#if 0 //DEBUG
    check_norm(a, b, "NUFFT", 1);//a->sig.w*a->sig.h*a->sig.d);
#endif
}                                 


void linoper_nufft_calculatec(LinOperNUFFT *op,
                              LinOperNUFFTDom *a, 
                              const LinOperNUFFTCod *b,
                              bool addto_a)
{
    FNumber w=get_nfft_weight(op);

    if(addto_a){
        op->plan.f=(fftw_complex*)b->p;
        op->plan.f_hat=(fftw_complex*)op->tmp_dom->p;

        nfft_adjoint(&op->plan);
    
        tensorfield_scaleplus(a, a, op->tmp_dom, 1, 1/w, 1);
    }else{
        op->plan.f=(fftw_complex*)b->p;
        op->plan.f_hat=(fftw_complex*)a->p;
    
        nfft_adjoint(&op->plan);

        if(w!=1.0)
            tensorfield_scale(a, 1/w);
    }

#if 0 //DEBUG
    check_norm(a, b, "ANUFFT", 1);//a->sig.w*a->sig.h*a->sig.d);
#endif
}

void linoper_nufft_backproj(LinOperNUFFT *op,
                            LinOperNUFFTDom *x, 
                            const LinOperNUFFTCod *y)
{
    die("linoper_nuff_backproj unimplemented.");
    //linoper_nufft_calculatec(op, x, y, FALSE);
}


LinOperNUFFTDom *linoper_nufft_alloc_dom(LinOperNUFFT *op)
{
    return alloc_tensorfield(&op->sig);
}

static void make_fsig(LinOperNUFFT *op, Sig *sig)
{
    sig->dim=1;
    sig->order=0;
    sig->w=op->n_samples;
    sig->h=1;
    sig->d=1;
    sig->symmetric=FALSE;
    sig->is_complex=TRUE;
    sig->xres=1;
    sig->yres=1;
    sig->zres=1;
}

LinOperNUFFTCod *linoper_nufft_alloc_cod(LinOperNUFFT *op)
{
    Sig sig;

    make_fsig(op, &sig);

    return alloc_tensorfield(&sig);
}

FNumber linoper_nufft_estimate_norm(const LinOperNUFFT *op)
{
    return fn_sqrt(op->n_samples);
}

FNumber linoper_nufft_estimate_lower_bound(const LinOperNUFFT *op)
{
    return FNUMBER(0.0);
}

void linoper_nufft_deinit(LinOperNUFFT *op)
{
    if(op->is_copy){
#if 0
        // Don't want to free index if copy.
        op->plan.nfft_flags&=~NFFT_SORT_NODES;
        nfft_finalize(&op->plan);
#else
        #pragma omp critical (nfft_omp_critical_fftw_plan)
        {
            fftw_destroy_plan(op->plan.my_fftw_plan1);
            fftw_destroy_plan(op->plan.my_fftw_plan2);
            free(op->plan.g1);
        #if !USE_NFFT_INIT_HACK
            free(op->plan.g2);
        #endif
        }
#endif
    }else{
        nfft_finalize(&op->plan);
    }
    
    if(op->tmp_dom)
        destroy_obj((Obj*)op->tmp_dom);
    if(op->tmp_cod)
        destroy_obj((Obj*)op->tmp_cod);
        
    linoper_deinit(&op->linoper);
}


#if USE_NFFT_INIT_HACK

extern int nfft_next_power_of_2(const int N); // NFFT
#define WINDOW_HELP_ESTIMATE_m 6

void nfft_init_(nfft_plan *ths, int d, int *N, int M_total)
{
  unsigned nfft_flags;
  
  int *n = (int*) nfft_malloc(d*sizeof(int));
  
  for(int t = 0; t < d; t++)
    n[t] = 2*nfft_next_power_of_2(N[t]);

  int m = WINDOW_HELP_ESTIMATE_m;

  if (d > 1)
  {
#ifdef _OPENMP
    nfft_flags = PRE_PHI_HUT | PRE_FULL_PSI | MALLOC_X| // MALLOC_F_HAT | MALLOC_F |
                 FFTW_INIT | /*FFT_OUT_OF_PLACE |*/ NFFT_SORT_NODES |
		 NFFT_OMP_BLOCKWISE_ADJOINT;
#else
    nfft_flags = PRE_PHI_HUT | PRE_FULL_PSI | MALLOC_X| // MALLOC_F_HAT | MALLOC_F |
                 FFTW_INIT | /*FFT_OUT_OF_PLACE |*/ NFFT_SORT_NODES;
#endif
  }
  else
    nfft_flags = PRE_PHI_HUT | PRE_FULL_PSI | MALLOC_X| // MALLOC_F_HAT | MALLOC_F |
                 FFTW_INIT /*| FFT_OUT_OF_PLACE*/;

  unsigned fftw_flags= FFTW_ESTIMATE| FFTW_DESTROY_INPUT;
  
  nfft_init_guru(ths, d, N, M_total, n, m, nfft_flags, fftw_flags);
  
  nfft_free(n);
}

#endif

bool linoper_nufft__init(LinOperNUFFT *op, 
                          const Sig *sig, 
                          size_t n_samples,
                          const LinOperNUFFT *copyop)
{
    size_t rank;
    int dims[3];
    FNumber normfact;
    
    if(sig->order!=0)
        die("Invalid data order");

    if(sig->dim<=0)
        die("Invalid dimension");
    
    if(!sig->is_complex)
        die("NUFFT only supports complex data");

    linoper_init(&op->linoper);

    op->sig=*sig;
    op->n_samples=n_samples;
    op->tmp_cod=linoper_nufft_alloc_cod(op);
    op->tmp_dom=linoper_nufft_alloc_dom(op);

    fft_init_params(&rank, dims, &normfact, sig);

    if(copyop==NULL){
        op->is_copy=FALSE;
        
        nfft_init(&op->plan, rank, dims, n_samples);
        
        // No way to give flags with nfft_init; only nfft_init_guru
        // which is complicated to use, so just free what it unnecessarily
        // allocated.
        op->plan.nfft_flags&=~(MALLOC_F | MALLOC_F_HAT);
#if !USE_NFFT_INIT_HACK
        free(op->plan.f);
        op->plan.f=NULL;
        free(op->plan.f_hat);
        op->plan.f_hat=NULL;
#endif
    }else{
        #warning "Major abuse in order to parallelise NFFT while sharing data! \
                  Might break if there are changes to the NFFT library"
        op->is_copy=TRUE;
        op->plan=copyop->plan;
        #if 0
            // Can't avoid copying the window function!
            unsigned nfft_flags=FFTW_INIT|FFT_OUT_OF_PLACE|NFFT_OMP_BLOCKWISE_ADJOINT;
            nfft_init_guru(&op->plan, rank, dims, n_samples,
                           copyop->plan.n, copyop->plan.m, 
                           nfft_flags, copyop->plan.fftw_flags);
            op->plan.nfft_flags|=NFFT_SORT_NODES;
        #else
            #pragma omp critical (nfft_omp_critical_fftw_plan)
            {
                op->plan.g1=(fftw_complex*)nfft_malloc(op->plan.n_total*sizeof(fftw_complex));
#if !USE_NFFT_INIT_HACK
                op->plan.g2=(fftw_complex*)nfft_malloc(op->plan.n_total*sizeof(fftw_complex));
#else
                op->plan.g2=op->plan.g1;
#endif
                op->plan.my_fftw_plan1 = fftw_plan_dft(op->plan.d, op->plan.n, op->plan.g1, op->plan.g2, FFTW_FORWARD, op->plan.fftw_flags);
                op->plan.my_fftw_plan2 = fftw_plan_dft(op->plan.d, op->plan.n, op->plan.g2, op->plan.g1, FFTW_BACKWARD, op->plan.fftw_flags);
            }
        #endif
    }
    
    return TRUE;
}

bool linoper_nufft_init(LinOperNUFFT *op, 
                        const Sig *sig, 
                        size_t n_samples)
{
    return linoper_nufft__init(op, sig, n_samples, NULL);
}

LinOperNUFFT *alloc_linoper_nufft(const Sig *sig, size_t n_samples)
{
    CREATEOBJ_IMPL(LinOperNUFFT, linoper_nufft, (p, sig, n_samples));
}


LinOperNUFFT *linoper_nufft_copy(const LinOperNUFFT *op)
{
    CREATEOBJ_IMPL(LinOperNUFFT, linoper_nufft_, (p, &op->sig, op->n_samples, op));
}

typedef struct{
    double *p;
    double *freq;
    size_t rank;
    size_t size;
} NUFFTRead;

bool readfn_nufft(const double *buf, size_t idx, size_t bs, void *res_)
{
    NUFFTRead *res=(NUFFTRead*)res_;
    const size_t rank=res->rank;
    const size_t size=res->size;
    const size_t size_per_2=size/2; // For indexing complex arrays
    const size_t rank_plus_size=rank+size;
    const size_t j=idx*bs/rank_plus_size;
    double *p=res->p;
    double *freq=res->freq;
    
    assert(2*size_per_2==size);

    #pragma acc kernels loop copyin(buf[0:bs]), deviceptr(p, freq), independent
    for(size_t i=j; i<j+bs/rank_plus_size; i++){
        for(size_t k=0; k<rank; k++)
            freq[rank*i+k]=buf[rank_plus_size*i+k];
        for(size_t ell=0; ell<size_per_2; ell++)
            set_complex(p, i*size_per_2+ell, buf[rank_plus_size*i+rank+2*ell]+I*buf[rank_plus_size*i+rank+2*ell+1]);
    }
    
    return TRUE;
}

extern LinOperNUFFT *load_linoper_nufft(const char *infile, 
                                        const Sig *sig, 
                                        TensorField **f_)
{
    size_t n, k, n_samples;
    FILE *fil;
    Sig tgt_sig;
    LinOperNUFFT *op;
    TensorField *f;
    size_t rank;
    int dims[3];
    FNumber normfact;

    assert(sig->is_complex);

    fft_init_params(&rank, dims, &normfact, sig);

    // Number of doubles per tensor
    size_t size=sig_tensor_size(sig);
    // + dimensions of the FFT
    size_t rank_plus_size=rank+size;
    
    fil=fopen(infile, "r");
    
    if(fil==NULL)
        die_err_obj(infile);

    fseek(fil, 0, SEEK_END);
    n=ftell(fil);
    fseek(fil, 0, SEEK_SET);

    if(n%(rank_plus_size*sizeof(double))!=0)
        die_obj(infile, "Invalid input file size");
    n_samples=n/(rank_plus_size*sizeof(double));
    
    op=alloc_linoper_nufft(sig, n_samples);
        
    if(op==NULL)
        die_err();

    f=linoper_nufft_alloc_cod(op);
    
    if(f==NULL)
        die_err();
    
    NUFFTRead sr;
    sr.p=f->p;
    sr.freq=op->plan.x;
    sr.rank=rank;
    sr.size=size;
    k=read_doubles(1, rank_plus_size*n_samples, fil, readfn_nufft, &sr);

    if(k!=1)
        die_err_obj(infile);
    
    fclose(fil);
    
    *f_=f;
    
    if(op->plan.nfft_flags&PRE_ONE_PSI)
        nfft_precompute_one_psi(&op->plan);
    
    return op;
}

extern LinOperNUFFT *load_linoper_nufft_nodata(const char *infile, 
                                               const Sig *sig, 
                                               Sig *fsig)
{
    size_t n, k, n_samples;
    FILE *fil;
    LinOperNUFFT *op;
    TensorField *f;
    
    assert(sig->is_complex);

    size_t rank;
    int dims[3];
    FNumber normfact;

    assert(sig->is_complex);

    fft_init_params(&rank, dims, &normfact, sig);
    
    fil=fopen(infile, "r");
    
    if(fil==NULL)
        die_err_obj(infile);

    fseek(fil, 0, SEEK_END);
    n=ftell(fil);
    fseek(fil, 0, SEEK_SET);

    if(n%(rank*sizeof(double))!=0)
        die_obj(infile, "Invalid input file size");
    n_samples=n/(rank*sizeof(double));
    
    op=alloc_linoper_nufft(sig, n_samples);
        
    if(op==NULL)
        die_err();

    NUFFTRead sr;
    sr.p=NULL;
    sr.freq=op->plan.x;
    sr.rank=rank;
    sr.size=0;
    k=read_doubles(1, rank*n_samples, fil, readfn_nufft, &sr);

    if(k!=1)
        die_err_obj(infile);
    
    fclose(fil);
    
    make_fsig(op, fsig);
    
    if(op->plan.nfft_flags&PRE_ONE_PSI)
        nfft_precompute_one_psi(&op->plan);
    
    return op;
}


static DynFunTab linoper_nufft_dynfuntab[]={
    {linoper_calculate, linoper_nufft_calculate},
    {linoper_calculatec, linoper_nufft_calculatec},
    {linoper_backproj, linoper_nufft_backproj},
    {(DynFun*)linoper_alloc_dom, (DynFun*)linoper_nufft_alloc_dom},
    {(DynFun*)linoper_alloc_cod, (DynFun*)linoper_nufft_alloc_cod},
    {(DynFun*)linoper_estimate_norm, (DynFun*)linoper_nufft_estimate_norm},
    {(DynFun*)linoper_estimate_lower_bound, (DynFun*)linoper_nufft_estimate_lower_bound},
    END_DYNFUNTAB,
};

IMPLCLASS(LinOperNUFFT, LinOper, linoper_nufft_deinit, linoper_nufft_dynfuntab);

