/*
 * dticode/src/linoper_parallel.c
 *
 * Written 2013 by Tuomo Valkonen <tuomov@iki.fi>, University of Cambridge.
 *
 */

#include <assert.h>
#include <libtu/objp.h>
#include <libtu/types.h>
#include <libtu/misc.h>
#include <libtu/output.h>
#include "linoper.h"
#include "diffoper.h"
#include "linoper_parallel.h"

void linoper_parallel_calculate(LinOperParallel *op,
                                LinOperParallelCod *y, 
                                const LinOperParallelDom *x, 
                                bool addto_y)
{
    assert(x->n==op->n);
    assert(y->n==op->n);
    
    //#pragma omp parallel for
    for(size_t i=0; i<op->n; i++)
        linoper_calculate(op->lop, y->p[i], x->p[i], addto_y);
}

void linoper_parallel_backproj(LinOperParallel *op,
                               LinOperParallelDom *x, 
                               const LinOperParallelCod *y)
{
    assert(x->n==op->n);
    assert(y->n==op->n);
    
    //#pragma omp parallel for
    for(size_t i=0; i<op->n; i++)
        linoper_backproj(op->lop, x->p[i], y->p[i]);
}

void linoper_parallel_calculatec(LinOperParallel *op,
                                 LinOperParallelDom *x, 
                                 const LinOperParallelCod *y, 
                                 bool addto_x)
{
    assert(x->n==op->n);
    assert(y->n==op->n);
    
    //#pragma omp parallel for
    for(size_t i=0; i<op->n; i++)
        linoper_calculatec(op->lop, x->p[i], y->p[i], addto_x);
}


LinOperParallelDom *linoper_parallel_alloc_dom(LinOperParallel *op)
{
    SplitVar *sv=alloc_splitvar(op->n);
    
    for(size_t i=0; i<op->n; i++)
        sv->p[i]=linoper_alloc_dom(op->lop);
        
    return sv;
}

LinOperParallelCod *linoper_parallel_alloc_cod(LinOperParallel *op)
{
    SplitVar *sv=alloc_splitvar(op->n);
    
    for(size_t i=0; i<op->n; i++)
        sv->p[i]=linoper_alloc_cod(op->lop);
        
    return sv;
}

FNumber linoper_parallel_constraint_scale(LinOperParallel *op, const LinOperParallelDom *x)
{
    FNumber sc=0;

    assert(x->n==op->n);
    
    //#pragma omp parallel for
    for(size_t i=0; i<op->n; i++)
        sc=fn_max(sc, diffoper_constraint_scale((DOper*)op->lop, x->p[i]));
        
    return sc;
}

FNumber linoper_parallel_constraint_conjugate(LinOperParallel *op, LinOperParallelDom *x)
{
    FNumber res=0;
    
    assert(x->n==op->n);

    //#pragma omp parallel for
    for(size_t i=0; i<op->n; i++)
        res+=diffoper_constraint_conjugate((DOper*)op->lop, x->p[i]);
    
    return res;
}

void linoper_parallel_projdom(LinOperParallel *op, LinOperParallelDom *x)
{
    assert(x->n==op->n);

    for(size_t i=0; i<op->n; i++)
        linoper_projdom(op->lop, x->p[i]);
}

FNumber linoper_parallel_estimate_norm(LinOperParallel *op)
{
    return linoper_estimate_norm(op->lop);
}

FNumber linoper_parallel_estimate_lower_bound(LinOperParallel *op)
{
    return linoper_estimate_lower_bound(op->lop);
}

bool linoper_parallel_init(LinOperParallel *op, 
                           LinOper *lop, 
                           size_t n)
{
    op->lop=lop;
    op->n=n;
    
    return linoper_init(&op->loper);
}

void linoper_parallel_deinit(LinOperParallel *op)
{
    destroy_obj((Obj*)op->lop);
    linoper_deinit(&op->loper);
}

LinOperParallel *create_linoper_parallel(LinOper *lop, size_t n)
{
    CREATEOBJ_IMPL(LinOperParallel, linoper_parallel, (p, lop, n));
}

static DynFunTab linoper_parallel_dynfuntab[]={
    {linoper_calculate, linoper_parallel_calculate},
    {linoper_calculatec, linoper_parallel_calculatec},
    {diffoper_backproj, linoper_parallel_backproj},
    {linoper_projdom, linoper_parallel_projdom},
    {(DynFun*)linoper_alloc_dom, (DynFun*)linoper_parallel_alloc_dom},
    {(DynFun*)linoper_alloc_cod, (DynFun*)linoper_parallel_alloc_cod},
    {(DynFun*)diffoper_constraint_scale, (DynFun*)linoper_parallel_constraint_scale},
    {(DynFun*)diffoper_constraint_conjugate, (DynFun*)linoper_parallel_constraint_conjugate},
    {(DynFun*)linoper_estimate_norm, (DynFun*)linoper_parallel_estimate_norm},
    {(DynFun*)linoper_estimate_lower_bound, (DynFun*)linoper_parallel_estimate_lower_bound},
    END_DYNFUNTAB,
};

IMPLCLASS(LinOperParallel, LinOper, linoper_parallel_deinit, linoper_parallel_dynfuntab);
