/*
 * dticode/src/linoper_pointwise.c
 *
 * Written 2013 by Tuomo Valkonen <tuomov@iki.fi>, University of Cambridge.
 *
 */


#include <libtu/types.h>
#include <libtu/objp.h>
#include <libtu/misc.h>
#include <libtu/output.h>
#include <string.h>
#include "tensor.h"
#include "vector.h"
#include "number.h"
#include "linoper.h"
#include "linoper_pointwise.h"
#include "scaleplus.h"
#include "project_posdef.h"
#include "posdef.h"
#include "sumdot1.h"
#include "scaleproj.h"
#include "matrix.h"
#include "tensornorm.h"

static void amat_projector(TensorField *res,
                           TensorField *c,
                           FNumber tau, FNumber idweight, FNumber cweight,
                           bool constr,
                           const TensorField *f,
                           LinOperPointwise *op)
{
    // M = idweight I+tau A^*A
    // constrained: MX-S=tau A^* f + cweight C, X,S>=0, X o S = 0
    // unconstrained: MX=tau A^* f + cweight C
    Matrix *mmat=alloc_matrix(&op->amat->sigm, &op->amat->sigm);
    Matrix *mmat_copy=alloc_matrix(&op->amat->sigm, &op->amat->sigm);
    size_t m=op->amat->m;
    size_t n=op->amat->n;
    size_t count=c->nelem;
    FNumber * DEVPTR mp=mmat->p;
    FNumber * DEVPTR ap=op->amat->p;
    FNumber * DEVPTR acp=op->amat_conj->p;
    FNumber * DEVPTR cp=c->p;
    FNumber * DEVPTR resp=res->p;
    
    assert(m==res->elemsize && m==c->elemsize);
    assert(res->nelem==c->nelem);
    
    // Construct M
    for(size_t i=0; i<m; i++){
        for(size_t j=0; j<m; j++){
            FNumber tmp=0;
            for(size_t k=0; k<n; k++)
                tmp=tmp+acp[i+k*m]*ap[k+j*n];
            mp[i+j*m]=(i==j ? idweight+tau*tmp : tau*tmp);
        }
    }
    // c := tau A^* f + cweight c
    tensorfield_linop(c, op->amat_conj, f, cweight, tau);

    memcpy(mmat_copy->p, mmat->p, m*m*sizeof(double));
    
    #define LUDIM m
    #define NRHS count
    #define A mp
    #define C cp
    #define RES resp
    #define GETM(a, i, j) (a[(j)+(i)*LUDIM])
    #define SETM(a, i, j, x) a[(j)+(i)*LUDIM]=(x)
    #define GETV(v, i, k) (v[(i)+(k)*LUDIM])
    #define SETV(v, i, k, x) v[i+(k)*LUDIM]=(x)
    #define LUCORE_RHS_PARALLEL 1
    #include "lusolve_core.h"
    #undef LUCORE_RHS_PARALLEL
    #undef LUDIM
    #undef NRHS
    #undef A
    #undef C
    #undef RES
    #undef GETM
    #undef SETM
    #undef GETV
    #undef SETV
    
    if(constr){
        // now M*res=c, but res is not necessarily positive definite.
        size_t dim=res->sig.dim;
        FNumber gamma=0.9;
        FNumber sigma=0.5;//1-gamma/sqrt(dim);
        FNumber rho=op->iprho;//0.001;
        FNumber minm=idweight+tau*op->a_squared_bound;
        int iters=100;
        
        tensorfield_scaleproj(res, mmat_copy, minm, 
                              sigma, gamma, iters, rho);
    }
    
    free_matrix(mmat);
    free_matrix(mmat_copy);
}                           

void linoper_pointwise_calculate(LinOperPointwise *op,
                                 LinOperPointwiseCod *a, 
                                 const LinOperPointwiseDom *b,
                                 bool addto_a)
{
    tensorfield_linop(a, op->amat, b, (addto_a ? 1.0 : 0), 1.0);
}                                 

void linoper_pointwise_calculatec(LinOperPointwise *op,
                                  LinOperPointwiseDom *a, 
                                  const LinOperPointwiseCod *b,
                                  bool addto_a)
{
    tensorfield_linop(a, op->amat_conj, b, (addto_a ? 1.0 : 0), 1.0);
}

void linoper_pointwise_backproj(LinOperPointwise *op,
                                LinOperPointwiseDom *x, 
                                const LinOperPointwiseCod *y)
{
    die("linoper_pointwise_backproj unimplemented");
}                                

// Solves (zeta I + tau A^* A) u = c + tau A^* f
//
// Problem follows from:
//
// min_u { |c-u|^2/(2tau)+|f-Au|^2/2}
// OC: (u-c)/tau + A^*A u - A^*f=0.
//     (I+tau A^* A) u = c + tau A^*f ; ==> zeta = 1, tau = given
//
// sup_u { <c,u> - |f-Au|^2/2} = min_u { -<c,u> + |f-Au|^2/2}
// OC: -c + A^*A u - A^*f=0.
//     A^* A u = c + A^*f ; ==> zeta = 0, tau = 1
void linoper_pointwise_gsolve(LinOperPointwise *op,
                              LinOperPointwiseDom *u,
                              FNumber tau, FNumber zeta, FNumber cweight,
                              const LinOperPointwiseCod *f)
{
    if(op->amat->is_identity)
        tensorfield_scaleplus(u, u, f, cweight, tau, zeta+tau);
    else
        amat_projector(u, u, tau, zeta, cweight, FALSE, f, op);
}

// Same as above, positivity-constrained
void linoper_pointwise_nonneg_gsolve(LinOperPointwiseNonNeg *op,
                                     LinOperPointwiseDom *u,
                                     FNumber tau, FNumber zeta, FNumber cweight,
                                     const LinOperPointwiseCod *f)
{
    if(tau==FNUMBER(0.0) && zeta==FNUMBER(1.0)){
        if(cweight!=FNUMBER(1.0))
            tensorfield_scale(u, cweight);
        tensorfield_project_posdef(u);
    }else if(op->linoperpw.amat->is_identity){
        tensorfield_scaleplus(u, u, f, cweight, tau, zeta+tau);
        tensorfield_project_posdef(u);
    }else{
        amat_projector(u, u, tau, zeta, cweight, TRUE, f, &op->linoperpw);
    }
}

extern FNumber linoper_pointwise_g(LinOperPointwise *op,
                                   const LinOperPointwiseDom *x,
                                   const LinOperPointwiseCod *f,
                                   FNumber *m)
{
    if(m!=NULL)
        *m=0;
    
    return tensorfield_linop_frob2dist22(f, op->amat, x);
}

extern FNumber linoper_pointwise_nonneg_g(LinOperPointwiseNonNeg *op,
                                          const LinOperPointwiseDom *x,
                                          const LinOperPointwiseCod *f,
                                          FNumber *m)
{
    if(m!=NULL)
        *m=0;
    
    if(!tensorfield_posdef(x, -FNUMBER_EPS))
        return INFINITY;
    else
        return tensorfield_linop_frob2dist22(f, op->linoperpw.amat, x);
}
                              
// Calculate G^*(x), with weight of constrained part split into c.
extern FNumber linoper_pointwise_gc(LinOperPointwise *op,
                                    const LinOperPointwiseDom *minus_x,
                                    const LinOperPointwiseCod *f,
                                    LinOperPointwiseDom *xwork,
                                    FNumber *c)
{
    *c=0;
    
    tensorfield_copy(xwork, minus_x);
    amat_projector(xwork, xwork, 1, 0, -1, FALSE, f, op);
    
    return (-tensorfield_sumdot1(minus_x, xwork, FALSE, 1)
            -tensorfield_linop_frob2dist22(f, op->amat, xwork));
}

extern FNumber linoper_pointwise_nonneg_gc(LinOperPointwiseNonNeg *op,
                                           const LinOperPointwiseDom *minus_x,
                                           const LinOperPointwiseCod *f,
                                           LinOperPointwiseDom *xwork,
                                           FNumber *c)
{
    *c=0;
    
    tensorfield_copy(xwork, minus_x);
    amat_projector(xwork, xwork, 1, 0, -1, TRUE, f, &op->linoperpw);
    
    return (-tensorfield_sumdot1(minus_x, xwork, FALSE, 1)
            -tensorfield_linop_frob2dist22(f, op->linoperpw.amat, xwork));
}

void linoper_pointwise_projdom(LinOperPointwise *op, LinOperPointwiseDom *x)
{
}

void linoper_pointwise_nonneg_projdom(LinOperPointwiseNonNeg *op, LinOperPointwiseDom *x)
{
    tensorfield_project_posdef(x);
}


LinOperPointwiseDom *linoper_pointwise_alloc_dom(LinOperPointwise *op)
{
    Sig sig=op->sig;
    sig.order=op->amat->sigm.order;
    sig.dim=op->amat->sigm.dim;
    sig.symmetric=op->amat->sigm.symmetric;
    sig.is_complex=op->amat->sigm.is_complex;
    
    return alloc_tensorfield(&sig);
}

LinOperPointwiseCod *linoper_pointwise_alloc_cod(LinOperPointwise *op)
{
    Sig sig=op->sig;
    sig.order=op->amat->sign.order;
    sig.dim=op->amat->sign.dim;
    sig.symmetric=op->amat->sign.symmetric;
    sig.is_complex=op->amat->sign.is_complex;

    return alloc_tensorfield(&sig);
}

FNumber linoper_pointwise_estimate_norm(const LinOperPointwise *op)
{
    return matrix_opnorm(op->amat);
}

FNumber linoper_pointwise_estimate_lower_bound(const LinOperPointwise *op)
{
    return fn_sqrt(op->a_squared_bound);
}

void linoper_pointwise_deinit(LinOperPointwise *op)
{
    free_matrix(op->amat_conj);
    
    linoper_l2_deinit(&op->linoperl2);
}

bool linoper_pointwise_init(LinOperPointwise *op, 
                            const Sig *sig, 
                            const Matrix *amat)
{
    linoper_l2_init(&op->linoperl2);
    
    op->sig=*sig;
    op->amat=amat;
    op->amat_conj=matrix_conjugate(amat);
    op->iprho=FNUMBER(1e-8);
    op->a_squared_bound=matrix_normsquared_lowerbound(amat);
    
    if(op->a_squared_bound<=0)
        die("Singular pointwise operators not supported.");
        
    return TRUE;
}

void linoper_pointwise_nonneg_deinit(LinOperPointwiseNonNeg *op)
{
    linoper_pointwise_deinit(&op->linoperpw);
}

bool linoper_pointwise_nonneg_init(LinOperPointwiseNonNeg *op, 
                                   const Sig *sig, 
                                   const Matrix *amat)
{
    return linoper_pointwise_init(&op->linoperpw, sig, amat);
}

LinOperPointwise *alloc_linoper_pointwise(const Sig *sig, 
                                          const Matrix *amat)
{
    CREATEOBJ_IMPL(LinOperPointwise, linoper_pointwise, (p, sig, amat));
}

LinOperPointwiseNonNeg *alloc_linoper_pointwise_nonneg(const Sig *sig, 
                                                       const Matrix *amat)
{
    CREATEOBJ_IMPL(LinOperPointwiseNonNeg, linoper_pointwise_nonneg, (p, sig, amat));
}

FNumber linoper_pointwise_nonneg_constraint_scale(LinOperPointwiseNonNeg *op, const LinOperPointwiseDom *x)
{
    // Artificial constraint: norm{x}<=1 pointwise. Ignore real constarint y>1, as it is assumed to hold.
    return tensorfield_frobinf(x);
}

FNumber linoper_pointwise_nonneg_constraint_conjugate(LinOperPointwiseNonNeg *op, LinOperPointwiseDom *x)
{
    // Calculate sup_y <z,x> subject to y>=0 pointwise and norm{y}<=1 pointwise
    linoper_pointwise_nonneg_gsolve(op, x, 0.0, 1.0, 1.0, x);
    return tensorfield_frob1(x);
}

static DynFunTab linoper_pointwise_dynfuntab[]={
    {linoper_calculate, linoper_pointwise_calculate},
    {linoper_calculatec, linoper_pointwise_calculatec},
    {linoper_backproj, linoper_pointwise_backproj},
    {linoper_l2_gsolve, linoper_pointwise_gsolve},
    {linoper_projdom, linoper_pointwise_projdom},
    {(DynFun*)linoper_l2_g, (DynFun*)linoper_pointwise_g},
    {(DynFun*)linoper_l2_gc, (DynFun*)linoper_pointwise_gc},
    {(DynFun*)linoper_alloc_dom, (DynFun*)linoper_pointwise_alloc_dom},
    {(DynFun*)linoper_alloc_cod, (DynFun*)linoper_pointwise_alloc_cod},
    {(DynFun*)linoper_estimate_norm, (DynFun*)linoper_pointwise_estimate_norm},
    {(DynFun*)linoper_estimate_lower_bound, (DynFun*)linoper_pointwise_estimate_lower_bound},
    END_DYNFUNTAB,
};

IMPLCLASS(LinOperPointwise, LinOperL2, linoper_pointwise_deinit, linoper_pointwise_dynfuntab);

static DynFunTab linoper_pointwise_nonneg_dynfuntab[]={
    {linoper_l2_gsolve, linoper_pointwise_nonneg_gsolve},
    {linoper_projdom, linoper_pointwise_nonneg_projdom},
    {(DynFun*)linoper_l2_gc, (DynFun*)linoper_pointwise_nonneg_gc},
    {(DynFun*)linoper_l2_g, (DynFun*)linoper_pointwise_nonneg_g},
    {(DynFun*)diffoper_constraint_scale, (DynFun*)linoper_pointwise_nonneg_constraint_scale},
    {(DynFun*)diffoper_constraint_conjugate, (DynFun*)linoper_pointwise_nonneg_constraint_conjugate},
    END_DYNFUNTAB,
};

IMPLCLASS(LinOperPointwiseNonNeg, LinOperPointwise, linoper_pointwise_nonneg_deinit, linoper_pointwise_nonneg_dynfuntab);
