/*
 * dticode/src/linoper_sparsefft.c
 *
 * Written 2013 by Tuomo Valkonen <tuomov@iki.fi>, University of Cambridge.
 *
 */


#include <libtu/types.h>
#include <libtu/objp.h>
#include <libtu/misc.h>
#include <libtu/output.h>
#include "tensor.h"
#include "vector.h"
#include "number.h"
#include "linoper.h"
#include "linoper_sparsefft.h"
#include "scaleplus.h"
#include "complexx.h"
#include "sumdot1.h"
#include "tensornorm.h"
#include "doubleio.h"

#if _OPENACC
    #error "linoper_sparsefft GPU does not support OpenACC"
#endif

#if _MPI
    #error "linoper_sparsefft does not support MPI"
#endif


void linoper_sparsefft_calculate(LinOperSparseFFT *op,
                                 LinOperSparseFFTCod *a, 
                                 const LinOperSparseFFTDom *b,
                                 bool addto_a)
{
    size_t rank;
    int dims[3];
    FNumber normfact;
    
    fft_init_params(&rank, dims, &normfact, &b->sig);

    if(op->full){
        do_fft(rank, dims, a->p, b->p);
        tensorfield_scale(a, 1/fn_sqrt(normfact));
    }else{
        FNumber w=fn_sqrt(normfact);
        FNumber *tmp_p=op->tmp->p;
        FNumber *a_p=a->p;
        size_t n_samples=SPARSEFFT_N(op);
        size_t *idx=SPARSEFFT_IDX(op);
        
        do_fft(rank, dims, tmp_p, b->p);
        
        if(addto_a){
            #pragma omp parallel for
            #pragma acc kernels loop, independent, deviceptr(idx, tmp_p, a_p)
            for(size_t i=0; i<n_samples; i++)
                add_complex(a_p, i, get_complex(tmp_p, idx[i])/w);
        }else{
            #pragma omp parallel for
            #pragma acc kernels loop, independent, deviceptr(idx, tmp_p, a_p)
            for(size_t i=0; i<n_samples; i++)
                set_complex(a_p, i, get_complex(tmp_p, idx[i])/w);
        }
    }
}                                 

void linoper_sparsefft_calculatec_do(LinOperSparseFFT *op,
                                     LinOperSparseFFTDom *tmp, 
                                     const LinOperSparseFFTCod *b,
                                     LinOperSparseFFTDom *addto)
{
    size_t rank;
    int dims[3];
    FNumber normfact;
    
    fft_init_params(&rank, dims, &normfact, &tmp->sig);

    if(op->full){
        do_ifft(rank, dims, tmp->p, b->p);
    }else{
        FNumber *tmp_p=tmp->p;
        FNumber *b_p=b->p;
        size_t n_samples=SPARSEFFT_N(op);
        size_t *idx=SPARSEFFT_IDX(op);

        tensorfield_zero(tmp);
        
        #pragma omp parallel for
        #pragma acc kernels loop, independent, deviceptr(idx, tmp_p, b_p)
        for(size_t i=0; i<n_samples; i++)
            set_complex(tmp_p, idx[i], get_complex(b_p, i));
        
        do_ifft(rank, dims, tmp_p, tmp_p);
    }

    //printf("CC %f -> %f | %f\n", vector_norm(b), vector_norm(tmp), (addto ? vector_norm(addto) : -1));
    
    // According to the rules for combine.c, fidelity function
    // implementation has to add Kb to a
    if(addto)
        tensorfield_scaleplus(addto, addto, tmp, 1, 1/fn_sqrt(normfact), 1);
    else
        tensorfield_scale(tmp, 1/fn_sqrt(normfact));

}                                  

void linoper_sparsefft_calculatec(LinOperSparseFFT *op,
                                  LinOperSparseFFTDom *a, 
                                  const LinOperSparseFFTCod *b,
                                  bool addto_a)
{
    if(addto_a)
        linoper_sparsefft_calculatec_do(op, op->tmp, b, a);
    else
        linoper_sparsefft_calculatec_do(op, a, b, NULL);
}

void linoper_sparsefft_backproj(LinOperSparseFFT *op,
                                LinOperSparseFFTDom *x, 
                                const LinOperSparseFFTCod *y)
{
    linoper_sparsefft_calculatec(op, x, y, FALSE);
}

// Solve for y: zeta y - cweight u + tau K^*(Ky-f)  \in -N_{constraint set}
void linoper_sparsefft_gsolve(LinOperSparseFFT *op,
                              LinOperSparseFFTDom *u,
                              FNumber tau, FNumber zeta, FNumber cweight,
                              const LinOperSparseFFTCod *f)
{
    size_t rank;
    int dims[3];
    FNumber normfact;
    
    fft_init_params(&rank, dims, &normfact, &u->sig);

    // Solve for u_out:
    // (zeta I+tau SF^* SF) u_out = tau SF^* f + cweight u_in
    // Computing u' = F(tau S^* F^* f + u_in) = tau S^* f + cweight F u_in,
    // We get the condition (zeta+tau S^*S)F u_out = u', which is easy to solve.
    
    do_fft(rank, dims, u->p, u->p);
    
    if(op->full){
        // Need to normalise u by 1/sqrt(normfact)
        //tensorfield_scaleplus(u, fp->f, u, tau, 1/sqrt(normfact), 1);
        // Fu_out=u/(1+tau)
        // Need further normalisation of the whole thing by 1/sqrt(normfact)
        //tensorfield_scale(u, 1/(sqrt(normfact)*(1+tau)));
        tensorfield_scaleplus(u, u, f, cweight/fn_sqrt(normfact), tau,
                              fn_sqrt(normfact)*(zeta+tau));
    }else{
        //samples_tensorfield_scaleplus(u, u, op->samples, op->n_samples, 
        //                              1/sqrt(normfact), tau, sqrt(normfact)*(1+tau));
        FNumber *u_p=u->p;
        FNumber *f_p=f->p;
        size_t n_samples=SPARSEFFT_N(op);
        size_t *idx=SPARSEFFT_IDX(op);

#if 1
        const FNumber sigma=zeta+tau;
        const FNumber alpha=zeta/sigma;
        const FNumber beta=tau/(sigma*fn_sqrt(normfact));
        
        // Solves for components outside the sampling range,
        // while also scaling u wrongly outside the range; this
        // is taken into account in alpha.
        tensorfield_scale(u, cweight/(zeta*normfact));
        
        #pragma omp parallel for
        #pragma acc kernels loop, independent, deviceptr(idx, f_p, u_p)
        for(size_t i=0; i<n_samples; i++){
            size_t j=idx[i];
            set_complex(u_p, j, alpha*get_complex(u_p, j)+beta*get_complex(f_p, i));
        }
#else
        const FNumber sigma=zeta+tau;
        const FNumber gamma=cweight/(zeta*normfact);
        const FNumber alpha=cweight/(sigma*normfact);
        const FNumber beta=tau/(sigma*fn_sqrt(normfact));
        const size_t nelem=u->nelem;
        
        #pragma acc kernels deviceptr(idx, f_p, u_p)
        {
            #pragma omp parallel for
            #pragma acc loop, independent, 
            for(size_t i=0; i<n_samples; i++){
                // ASSUMPTION: samples sorted by idx
                size_t j=idx[i];
                size_t prev=(i==0 ? 0 : idx[i-1]+1);
                #pragma acc loop independent, reduction(+:c0)
                for(size_t k=prev; k<j; k++)
                    set_complex(u_p, k, gamma*get_complex(u_p, k));
                set_complex(u_p, j, alpha*get_complex(u_p, j)+beta*get_complex(f_p, i));
            }
            
            size_t prev=idx[n_samples-1]+1;
            
            #pragma omp parallel for
            #pragma acc loop, independent
            for(size_t k=prev; k<nelem; k++)
                set_complex(u_p, k, gamma*get_complex(u_p, k));
        }
#endif
    }

    do_ifft(rank, dims, u->p, u->p);
}

FNumber linoper_sparsefft_g(LinOperSparseFFT *op,
                            const LinOperSparseFFTDom *x,
                            const LinOperSparseFFTCod *f,
                            FNumber *m)
{
    size_t rank;
    int dims[3];
    FNumber normfact;
    FNumber g;
    TensorField *tmp=op->tmp;
    
    fft_init_params(&rank, dims, &normfact, &x->sig);

    do_fft(rank, dims, tmp->p, x->p);

    if(op->full){
        tensorfield_scale(tmp, 1/fn_sqrt(normfact));
        g=tensorfield_frob2dist22(tmp, f);
        if(m!=NULL)
            *m=0;
    }else{
        double a=fn_sqrt(normfact);
        size_t n_samples=SPARSEFFT_N(op);
        size_t *idx=SPARSEFFT_IDX(op);
        FNumber *tmp_p=tmp->p;
        FNumber *f_p=f->p;
        
        g=0;
        
        #pragma omp parallel for
        #pragma acc kernels loop, independent, reduction(+:g), deviceptr(idx, tmp_p, f_p)
        for(size_t i=0; i<n_samples; i++){
            g+=fn_cabs2(get_complex(f_p, i)-get_complex(tmp_p, idx[i])/a);
        }
        g/=2;
        
        if(m!=NULL)
            *m=tensorfield_frobinf(tmp)/a;
    }
    
    return g;
}
                              
// Calculate G^*(x), with weight of constrained part split into c.
FNumber linoper_sparsefft_gc(LinOperSparseFFT *op,
                             const LinOperSparseFFTDom *minus_x,
                             const LinOperSparseFFTCod *f,
                             LinOperSparseFFTDom *xwork,
                             FNumber *c)
{
    size_t rank;
    int dims[3];
    FNumber normfact;
    FNumber gcm;
    
    fft_init_params(&rank, dims, &normfact, &minus_x->sig);

    if(op->full){
        do_ifft(rank, dims, xwork->p, f->p);
        tensorfield_scale(xwork, 1/fn_sqrt(normfact));
        gcm=(-tensorfield_sumdot1(minus_x, xwork, FALSE, 1)
             +tensorfield_frob222(minus_x));
        if(c!=NULL)
            *c=0;
    }else{
        double a=fn_sqrt(normfact);
        size_t n_samples=SPARSEFFT_N(op);
        size_t *idx=SPARSEFFT_IDX(op);
        size_t xwork_nelem=xwork->nelem;
        FNumber *xwork_p=xwork->p;
        FNumber *f_p=f->p;
        FNumber cc=0;
        
        do_fft(rank, dims, xwork_p, minus_x->p);

        if(c!=NULL){
            // sup_u <x,u>-|f-SFu|^2/2, (x=-K^* y)
            // => SFu-f=SFx
            // => <x,u>-|f-SFu|^2/2 
            //    = <Fx,Fu> -|f-SFu|^2/2
            //    = <(I-S)Fx,(I-S)Fu> + <SFx,SFu> -|f-SFu|^2/2
            //    = <(I-S)Fx,(I-S)Fu> + <SFx,SFx> + <SFx,f> -|SFx|^2/2
            //    = <(I-S)Fx,(I-S)Fu> + <SFx,f> + |SFx|^2/2
            // First term is infinite, unless we add a bount
            // |(I-S)Fu|_infty < M, in which case
            // we get M * |(I-S)Fu|_1
            #pragma acc kernels deviceptr(idx, xwork_p, f_p)
            {
                #pragma omp parallel for
                #pragma acc loop, independent, reduction(+:gcm), reduction(+:cc)
                for(size_t i=0; i<n_samples; i++){
                    // ASSUMPTION: samples sorted by idx
                    size_t j=idx[i];
                    FNumber c0=0;
                    size_t prev=(i==0 ? 0 : idx[i-1]+1);
                    #pragma acc loop independent, reduction(+:c0)
                    for(size_t k=prev; k<j; k++)
                        c0+=fn_cabs(get_complex(xwork_p, k)/a);
                    FNumber_complex x=-get_complex(xwork_p, j)/a;
                    gcm+=ciprod(get_complex(f_p, i), x)+fn_cabs2(x)/2;
                    cc+=c0;
                }
                
                size_t prev=idx[n_samples-1]+1;
                
                #pragma omp parallel for
                #pragma acc loop, independent, reduction(+:cc)
                for(size_t k=prev; k<xwork_nelem; k++)
                    cc+=fn_cabs(get_complex(xwork_p, k)/a);
            }
            
            *c=cc;
        }
    }
    
    return gcm;
}

LinOperSparseFFTDom *linoper_sparsefft_alloc_dom(LinOperSparseFFT *op)
{
    return alloc_tensorfield(&op->sig);
}

static void make_fsig(LinOperSparseFFT *op, Sig *sig)
{
    sig->dim=1;
    sig->order=0;
    sig->w=op->n_samples;
    sig->h=1;
    sig->d=1;
    sig->symmetric=FALSE;
    sig->is_complex=TRUE;
    sig->xres=1;
    sig->yres=1;
    sig->zres=1;
}

LinOperSparseFFTCod *linoper_sparsefft_alloc_cod(LinOperSparseFFT *op)
{
    if(op->full){
        return alloc_tensorfield(&op->sig);
    }else{
        Sig sig;

        make_fsig(op, &sig);
        
        return alloc_tensorfield(&sig);
    }
}

FNumber linoper_sparsefft_estimate_norm(const LinOperSparseFFT *op)
{
    return FNUMBER(1.0);
}

FNumber linoper_sparsefft_estimate_lower_bound(const LinOperSparseFFT *op)
{
    return (op->full ? FNUMBER(1.0) : FNUMBER(0.0));
}

void linoper_sparsefft_deinit(LinOperSparseFFT *op)
{
    if(op->tmp)
        destroy_obj((Obj*)op->tmp);
        
    linoper_l2_deinit(&op->linoperl2);
}

bool linoper_sparsefft_init(LinOperSparseFFT *op, 
                            const Sig *sig, 
                            size_t n_samples)
{
    if(sig->order!=0)
        die("Invalid data order");
    
    if(!sig->is_complex)
        die("SparseFFT only supports complex data");

    linoper_l2_init(&op->linoperl2);
    
    op->sig=*sig;
    op->n_samples=n_samples;
    op->full=(n_samples==sig_tensor_count(sig));
    op->tmp=alloc_tensorfield(sig);
    op->idx=HWALLOC(n_samples*sizeof(size_t));
    
    if(op->idx==NULL)
        die_err();
        
    return TRUE;
}

LinOperSparseFFT *alloc_linoper_sparsefft(const Sig *sig, 
                                          size_t n_samples)
{
    CREATEOBJ_IMPL(LinOperSparseFFT, linoper_sparsefft, (p, sig, n_samples));
}

LinOperSparseFFT *linoper_sparsefft_copy(const LinOperSparseFFT *copyop)
{
    LinOperSparseFFT *op=alloc_linoper_sparsefft(&copyop->sig, copyop->n_samples);
    
    memcpy_dev(op->idx, copyop->idx, sizeof(size_t)*op->n_samples);
    
    return op;
}


typedef struct{
    size_t *idx;
    double *p;
} SampleRead;

bool readfn_fourier_samples(const double *buf, size_t idx, size_t bs,
                            void *res_)
{
    SampleRead *res=(SampleRead*)res_;
    size_t j=idx*bs/3;

    for(size_t i=j; i<j+bs/3; i++){
        res->idx[i]=(size_t)buf[3*i];
        set_complex(res->p, i, buf[3*i+1]+I*buf[3*i+2]);
    }
    
    return TRUE;
}

static int compare_sample_idx(const void *a_, const void *b_)
{
    size_t ai=(size_t)((const double *)a_)[0];
    size_t bi=(size_t)((const double *)b_)[0];

    return (ai<bi ? -1 : ((ai>bi) ? 1 : 0));
}

LinOperSparseFFT *load_linoper_sparsefft(const char *infile, 
                                         const Sig *sig, 
                                         bool sparse_infile,
                                         TensorField **f_)
{
    size_t n, k, n_samples;
    FILE *fil;
    LinOperSparseFFT *op;
    TensorField *f;
    
    fil=fopen(infile, "r");
    
    if(fil==NULL)
        die_err_obj(infile);

    fseek(fil, 0, SEEK_END);
    n=ftell(fil);
    fseek(fil, 0, SEEK_SET);

    if(sparse_infile){
        if(n%(3*sizeof(double))!=0)
            die_obj(infile, "Invalid input file size");
        n_samples=n/(3*sizeof(double));
    }else{
        n_samples=n/(2*sizeof(double));
    }
    
    op=alloc_linoper_sparsefft(sig, n_samples);
        
    if(op==NULL)
        die_err();
            
    f=linoper_sparsefft_alloc_cod(op);
        
    if(f==NULL)
        die_err();
            
    if(sparse_infile){
        // Read into temporary buffer for qsort
        double *buf=ALLOC_N(double, 3*n_samples);
        
        if(buf==NULL)
            die_err();

        k=read_doubles(1, 3*n_samples, fil, readfn_simple, buf);

        if(k!=1)
            die_err_obj(infile);
        
        // Sort for scanning through elements not in the sparse basis.
        qsort(buf, n_samples, 3*sizeof(double), compare_sample_idx);

        {
            SampleRead sr;
            sr.p=f->p;
            sr.idx=SPARSEFFT_IDX(op);
            
            readfn_fourier_samples(buf, 0, 3*n_samples, &sr);
        }
        
        if(op->idx[n_samples-1]>=sig_tensor_count(sig))
            die_obj(infile, "Invalid sample index");
        
        free(buf);
    }else{
        if(!op->full)
            die_obj(infile, "Invalid number of samples");
        
        k=read_doubles(1, 2*n_samples, fil, readfn_simple_acc, f->p);
        
        if(k!=1)
            die_err_obj(infile);
        
        size_t *idx=SPARSEFFT_IDX(op);
        #pragma acc loop deviceptr(idx), independent
        for(size_t i=0; i<n_samples; i++)
            idx[i]=i;
    }
    
    fclose(fil);
    
    *f_=f;
    
    return op;
}


LinOperSparseFFT *load_linoper_sparsefft_nodata(const char *infile, 
                                                const Sig *sig, 
                                                bool sparse_infile,
                                                Sig *fsig)
{
    if(!sparse_infile){
        *fsig=*fsig;
        return alloc_linoper_sparsefft(sig, sig_tensor_count(sig));
    }else{
        size_t n, k, n_samples;
        FILE *fil;
        LinOperSparseFFT *op;
        TensorField *f;
        
        fil=fopen(infile, "r");
        
        if(fil==NULL)
            die_err_obj(infile);

        fseek(fil, 0, SEEK_END);
        n=ftell(fil);
        fseek(fil, 0, SEEK_SET);

        if(n%(sizeof(double))!=0)
            die_obj(infile, "Invalid input file size");
            
        n_samples=n/sizeof(double);
        
        op=alloc_linoper_sparsefft(sig, n_samples);
            
        if(op==NULL)
            die_err();
                
        // Read into temporary buffer for qsort
        double *buf=ALLOC_N(double, n_samples);
        
        if(buf==NULL)
            die_err();

        k=read_doubles(1, n_samples, fil, readfn_simple, buf);

        if(k!=1)
            die_err_obj(infile);
        
        // Sort for scanning through elements not in the sparse basis.
        qsort(buf, n_samples, sizeof(double), compare_sample_idx);

        for(size_t i=0; i<n_samples; i++)
            op->idx[i]=(size_t)buf[i];
        
        if(op->idx[n_samples-1]>=sig_tensor_count(sig))
            die_obj(infile, "Invalid sample index");
        
        free(buf);
        
        fclose(fil);
        
        make_fsig(op, fsig);
        
        return op;
    }
}

static DynFunTab linoper_sparsefft_dynfuntab[]={
    {linoper_calculate, linoper_sparsefft_calculate},
    {linoper_calculatec, linoper_sparsefft_calculatec},
    {linoper_backproj, linoper_sparsefft_backproj},
    {linoper_l2_gsolve, linoper_sparsefft_gsolve},
    {(DynFun*)linoper_l2_g, (DynFun*)linoper_sparsefft_g},
    {(DynFun*)linoper_l2_gc, (DynFun*)linoper_sparsefft_gc},
    {(DynFun*)linoper_alloc_dom, (DynFun*)linoper_sparsefft_alloc_dom},
    {(DynFun*)linoper_alloc_cod, (DynFun*)linoper_sparsefft_alloc_cod},
    {(DynFun*)linoper_estimate_norm, (DynFun*)linoper_sparsefft_estimate_norm},
    {(DynFun*)linoper_estimate_lower_bound, (DynFun*)linoper_sparsefft_estimate_lower_bound},
    END_DYNFUNTAB,
};

IMPLCLASS(LinOperSparseFFT, LinOperL2, linoper_sparsefft_deinit, linoper_sparsefft_dynfuntab);


