/*
 * dticode/src/linoper_sparsemat.c
 *
 * Written 2014 by Tuomo Valkonen <tuomov@iki.fi>, University of Cambridge.
 *
 */


#include <libtu/types.h>
#include <libtu/objp.h>
#include <libtu/misc.h>
#include <libtu/output.h>
#include "tensor.h"
#include "vector.h"
#include "number.h"
#include "linoper.h"
#include "linoper_sparsemat.h"
#include "scaleplus.h"
#include "complexx.h"
#include "sumdot1.h"
#include "tensornorm.h"
#include "doubleio.h"
#include "sparse.h"
#include "acc_support.h"

#define SPARSE_MMULT(resdim, res, dim, vec) \
    { \
        _Pragma("acc loop seq") \
        _Pragma("omp parallel for") \
        for(size_t _i=0; _i<dim; _i++){ \
            const FNumber _val=vec[_i]; \
            const size_t _end2=pm_jc[_i+1]; \
            _Pragma("acc loop seq") \
            for(size_t _j=pm_jc[_i]; _j<_end2; _j++){ \
                const size_t _k=pm_ir[_j]; \
                res[_k]+=pm_pr[_j]*_val; \
            } \
        } \
    }

#define SPARSE_MMULT_TRANSPOSE(resdim, res, outw, dim, vec, inw) \
    { \
        _Pragma("acc loop seq") \
        _Pragma("omp parallel for") \
        for(size_t _i=0; _i<dim; _i++){ \
            const size_t _end2=pm_jc[_i+1]; \
            FNumber _val=0; \
            _Pragma("acc loop seq") \
            for(size_t _j=pm_jc[_i]; _j<_end2; _j++){ \
                const size_t _k=pm_ir[_j]; \
                _val+=pm_pr[_j]*inw(_k)*vec[_k]; \
            } \
            res[_i]+=_val/outw(_i); \
        } \
    }

void linoper_sparsemat_calculate(LinOperSparseMat *op,
                                 LinOperSparseMatCod *a, 
                                 const LinOperSparseMatDom *b,
                                 bool addto_a)
{
    const FNumber * const restrict pm_pr=op->pm.pr;
    const size_t * const restrict pm_ir=op->pm.ir;
    const size_t * const restrict pm_jc=op->pm.jc;
    const size_t irsize=op->pm.irsize;
    const size_t pmcols=op->pm.m;
    const size_t pmrows=op->pm.n;
    size_t resdim=sig_tensorfield_size(&a->sig);
    size_t dim=sig_tensorfield_size(&b->sig);
    FNumber *a_p=a->p;
    const FNumber *b_p=b->p;
    
    assert(pmcols==dim && pmrows==resdim);
    
    if(!addto_a)
        tensorfield_zero(a);

    SPARSE_MMULT(resdim, a_p, dim, b_p);
}


void linoper_sparsemat_calculatec(LinOperSparseMat *op,
                                  LinOperSparseMatDom *a, 
                                  const LinOperSparseMatCod *b,
                                  bool addto_a)
{
    const FNumber * const restrict pm_pr=op->pm.pr;
    const size_t * const restrict pm_ir=op->pm.ir;
    const size_t * const restrict pm_jc=op->pm.jc;
    const size_t irsize=op->pm.irsize;
    const size_t pmcols=op->pm.m;
    const size_t pmrows=op->pm.n;
    size_t resdim=sig_tensorfield_size(&a->sig);
    size_t dim=sig_tensorfield_size(&b->sig);
    FNumber *a_p=a->p;
    const FNumber *b_p=b->p;
    
    assert(pmrows==dim && pmcols==resdim);
    
    if(!addto_a)
        tensorfield_zero(a);
    
    if((!op->sigcod.symmetric || op->sigcod.order==0) && 
       (!op->sigdom.symmetric || op->sigdom.order==0)){
        // Optimise case with both weight matrices identities
        #define inw(i) FNUMBER(1.0)
        #define outw(i) FNUMBER(1.0)
        SPARSE_MMULT_TRANSPOSE(resdim, a_p, outw, dim, b_p, inw);
        #undef inw
        #undef outw
    }else{
        FNumber *inw_p=sig_get_weight(&op->sigcod);
        FNumber *outw_p=sig_get_weight(&op->sigdom);
        size_t out_sz=sig_tensor_size(&a->sig);
        size_t in_sz=sig_tensor_size(&b->sig);
        #define inw(i) inw_p[i%in_sz]
        #define outw(i) outw_p[i%out_sz]
        SPARSE_MMULT_TRANSPOSE(resdim, a_p, outw, dim, b_p, inw);
        #undef inw
        #undef outw
    }
}

void linoper_sparsemat_backproj(LinOperSparseMat *op,
                                LinOperSparseMatDom *x, 
                                const LinOperSparseMatCod *y)
{
    die("linoper_sparsemat_backproj unimplemented.");
}


LinOperSparseMatDom *linoper_sparsemat_alloc_dom(LinOperSparseMat *op)
{
    return alloc_tensorfield(&op->sigdom);
}

LinOperSparseMatCod *linoper_sparsemat_alloc_cod(LinOperSparseMat *op)
{
    return alloc_tensorfield(&op->sigcod);
}

FNumber linoper_sparsemat_estimate_norm(const LinOperSparseMat *op)
{
    return op->ubnd;
}

FNumber linoper_sparsemat_estimate_lower_bound(const LinOperSparseMat *op)
{
    return op->lbnd;
}

void linoper_sparsemat_deinit(LinOperSparseMat *op)
{
    free(op->pm.jc);
    free(op->pm.ir);
    free(op->pm.pr);
    
    linoper_deinit(&op->linoper);
}


bool linoper_sparsemat_init(LinOperSparseMat *op, 
                            const Sig *sigcod, 
                            const Sig *sigdom,
                            SparseMatrix *pm,
                            FNumber lbnd,
                            FNumber ubnd)
{
    op->sigcod=*sigcod;
    op->sigdom=*sigdom;
    op->pm=*pm;
    op->lbnd=lbnd;
    op->ubnd=ubnd;
    
    return TRUE;
}

LinOperSparseMat *alloc_linoper_sparsemat(const Sig *sigcod, 
                                          const Sig *sigdom,
                                          SparseMatrix *pm,
                                          FNumber lbnd,
                                          FNumber ubnd)
{
    CREATEOBJ_IMPL(LinOperSparseMat, linoper_sparsemat, 
                   (p, sigcod, sigdom, pm, lbnd, ubnd));
}

bool readfn_sparsemat_jc(const double *buf, size_t idx, size_t bs, void *res_)
{
    SparseMatrix *pm=(SparseMatrix*)res_;
    size_t * DEVPTR jc=pm->jc;
    
    assert(bs==pm->m+1);
    
    #pragma acc kernels loop copyout(buf[0:bs]), deviceptr(jc), independent
    for(size_t i=0; i<bs; i++)
        jc[i]=buf[i];

    return TRUE;
}

bool readfn_sparsemat_ir(const double *buf, size_t idx, size_t bs, void *res_)
{
    SparseMatrix *pm=(SparseMatrix*)res_;
    size_t * DEVPTR ir=pm->ir;
    
    assert(bs==pm->irsize);
    
    #pragma acc kernels loop copyout(buf[0:bs]), deviceptr(ir), independent
    for(size_t i=0; i<bs; i++)
        ir[i]=buf[i];
        
    return TRUE;
}

extern LinOperSparseMat *load_linoper_sparsemat(const char *infile, 
                                                const Sig *sigcod, 
                                                const Sig *sigdom)
{
    FNumber bnds[2];
    SparseMatrix pm;
    size_t maxidx=0;
    size_t k;
    FILE *fil;
    
    pm.m=sig_tensorfield_size(sigdom);
    pm.n=sig_tensorfield_size(sigcod);
    
    pm.jc=(size_t*)HWALLOC(sizeof(size_t)*(pm.m+1));
    if(pm.jc==NULL)
        die_err();
        
    // Load jc which has known size. This will then give the size of the rest.
    
    fil=fopen(infile, "r");
    
    if(fil==NULL)
        die_err_obj(infile);

    k=read_doubles(1, 2, fil, readfn_simple, &bnds);
    
    if(k!=1)
        die_obj(infile, "Invalid sparse matrix");
        
    if(bnds[0]<0 || bnds[0] > bnds[1])
        die_obj(infile, "Invalid matrix bounds");

    k=read_doubles(1, pm.m+1, fil, readfn_sparsemat_jc, &pm);
    
    if(k!=1)
        die_obj(infile, "Invalid sparse matrix");
        
    // Verify ordering of jc, and extract index upper bound
    {
        size_t * DEVPTR jc=pm.jc;
        size_t m=pm.m;
        bool error=FALSE;
        
        #pragma acc kernels loop deviceptr(jc)
        for(size_t i=0; i<m; i++){
            if(jc[i]>jc[i+1])
                error=TRUE;
        }
        
        if(error)
            die("Incorrectly ordered matrix");
            
        #pragma acc kernels deviceptr(jc)
        {
            maxidx=jc[m];
        }
    }
    
    pm.irsize=maxidx;
    pm.ir=(size_t*)HWALLOC(sizeof(size_t)*maxidx);
    pm.pr=(FNumber*)HWALLOC(sizeof(FNumber)*maxidx);

    if(pm.ir==NULL || pm.pr==NULL)
        die_err();
    
    // Load ir and pr now that we know their size
    
    k=read_doubles(1, maxidx, fil, readfn_sparsemat_ir, &pm);
    
    if(k!=1)
        die_obj(infile, "Invalid sparse matrix");

    k=read_doubles(1, maxidx, fil, readfn_simple_acc, pm.pr);
    
    if(k!=1)
        die_obj(infile, "Invalid sparse matrix");
        
    size_t p1=ftell(fil);
    fseek(fil, 0, SEEK_END);
    size_t p2=ftell(fil);

    if(p1!=p2)
        die_obj(infile, "Invalid sparse matrix");

    LinOperSparseMat *ret=alloc_linoper_sparsemat(sigcod, sigdom, &pm, bnds[0], bnds[1]);
    
    if(ret==NULL)
        die_err();
        
    return ret;
}

static DynFunTab linoper_sparsemat_dynfuntab[]={
    {linoper_calculate, linoper_sparsemat_calculate},
    {linoper_calculatec, linoper_sparsemat_calculatec},
    {linoper_backproj, linoper_sparsemat_backproj},
    {(DynFun*)linoper_alloc_dom, (DynFun*)linoper_sparsemat_alloc_dom},
    {(DynFun*)linoper_alloc_cod, (DynFun*)linoper_sparsemat_alloc_cod},
    {(DynFun*)linoper_estimate_norm, (DynFun*)linoper_sparsemat_estimate_norm},
    {(DynFun*)linoper_estimate_lower_bound, (DynFun*)linoper_sparsemat_estimate_lower_bound},
    END_DYNFUNTAB,
};

IMPLCLASS(LinOperSparseMat, LinOper, linoper_sparsemat_deinit, linoper_sparsemat_dynfuntab);

