/*
 * dticode/src/linoper_split.c
 *
 * Written 2013 by Tuomo Valkonen <tuomov@iki.fi>, University of Cambridge.
 *
 */

#include <assert.h>
#include <libtu/objp.h>
#include <libtu/types.h>
#include <libtu/misc.h>
#include <libtu/output.h>
#include "linoper.h"
#include "diffoper.h"
#include "linoper_split.h"

void linoper_split_calculate(LinOperSplit *op,
                             LinOperSplitCod *y, 
                             const LinOperSplitDom *x, 
                             bool addto_y)
{
    assert(x->n==op->n);
    assert(y->n==op->n);
    
    #pragma omp parallel for
    for(size_t i=0; i<op->n; i++)
        linoper_calculate(op->lops[i], y->p[i], x->p[i], addto_y);
}

void linoper_split_backproj(LinOperSplit *op,
                               LinOperSplitDom *x, 
                               const LinOperSplitCod *y)
{
    assert(x->n==op->n);
    assert(y->n==op->n);
    
    #pragma omp parallel for
    for(size_t i=0; i<op->n; i++)
        linoper_backproj(op->lops[i], x->p[i], y->p[i]);
}

void linoper_split_calculatec(LinOperSplit *op,
                                 LinOperSplitDom *x, 
                                 const LinOperSplitCod *y, 
                                 bool addto_x)
{
    assert(x->n==op->n);
    assert(y->n==op->n);
    
    #pragma omp parallel for
    for(size_t i=0; i<op->n; i++)
        linoper_calculatec(op->lops[i], x->p[i], y->p[i], addto_x);
}


LinOperSplitDom *linoper_split_alloc_dom(LinOperSplit *op)
{
    SplitVar *sv=alloc_splitvar(op->n);
    
    for(size_t i=0; i<op->n; i++)
        sv->p[i]=linoper_alloc_dom(op->lops[i]);
        
    return sv;
}

LinOperSplitCod *linoper_split_alloc_cod(LinOperSplit *op)
{
    SplitVar *sv=alloc_splitvar(op->n);
    
    for(size_t i=0; i<op->n; i++)
        sv->p[i]=linoper_alloc_cod(op->lops[i]);
        
    return sv;
}

FNumber linoper_split_constraint_scale(LinOperSplit *op, const LinOperSplitDom *x)
{
    FNumber sc=0;

    assert(x->n==op->n);
    
    #pragma omp parallel for
    for(size_t i=0; i<op->n; i++)
        sc=fn_max(sc, diffoper_constraint_scale((DOper*)op->lops[i], x->p[i]));
        
    return sc;
}

FNumber linoper_split_constraint_conjugate(LinOperSplit *op, LinOperSplitDom *x)
{
    FNumber res=0;
    
    assert(x->n==op->n);

    #pragma omp parallel for
    for(size_t i=0; i<op->n; i++)
        res+=diffoper_constraint_conjugate((DOper*)op->lops[i], x->p[i]);
    
    return res;
}

FNumber linoper_split_estimate_norm(LinOperSplit *op)
{
    FNumber res=0;
    
    #pragma omp parallel for
    for(size_t i=0; i<op->n; i++)
        res=fn_max(res, linoper_estimate_norm((LinOper*)op->lops[i]));

    return res;
}

void linoper_split_projdom(LinOperSplit *op, LinOperSplitDom *x)
{
    #pragma omp parallel for
    for(size_t i=0; i<op->n; i++)
        linoper_projdom(op->lops[i], x->p[i]);
}

FNumber linoper_split_estimate_lower_bound(LinOperSplit *op)
{
    FNumber res=INFINITY;

    #pragma omp parallel for
    for(size_t i=0; i<op->n; i++)
        res=fn_min(res, linoper_estimate_lower_bound((LinOper*)op->lops[i]));
        
    return res;
}

bool linoper_split_init(LinOperSplit *op, 
                        LinOper **lops, 
                        size_t n)
{
    op->lops=ALLOC_N(LinOper*, n);
    
    for(size_t i=0; i<n; i++)
        op->lops[i]=lops[i];

    op->n=n;
    
    return linoper_init(&op->loper);
}

void linoper_split_deinit(LinOperSplit *op)
{
    for(size_t i=0; i<op->n; i++)
        destroy_obj((Obj*)op->lops[i]);
    free(op->lops);
    linoper_deinit(&op->loper);
}

LinOperSplit *create_linoper_split(LinOper **lops, size_t n)
{
    CREATEOBJ_IMPL(LinOperSplit, linoper_split, (p, lops, n));
}

static DynFunTab linoper_split_dynfuntab[]={
    {linoper_calculate, linoper_split_calculate},
    {linoper_calculatec, linoper_split_calculatec},
    {diffoper_backproj, linoper_split_backproj},
    {linoper_projdom, linoper_split_projdom},
    {(DynFun*)linoper_alloc_dom, (DynFun*)linoper_split_alloc_dom},
    {(DynFun*)linoper_alloc_cod, (DynFun*)linoper_split_alloc_cod},
    {(DynFun*)diffoper_constraint_scale, (DynFun*)linoper_split_constraint_scale},
    {(DynFun*)diffoper_constraint_conjugate, (DynFun*)linoper_split_constraint_conjugate},
    {(DynFun*)linoper_estimate_norm, (DynFun*)linoper_split_estimate_norm},
    {(DynFun*)linoper_estimate_lower_bound, (DynFun*)linoper_split_estimate_lower_bound},
    END_DYNFUNTAB,
};

IMPLCLASS(LinOperSplit, LinOper, linoper_split_deinit, linoper_split_dynfuntab);
