/*
 * dticode/src/matrix.c
 *
 * Written 2012 by Tuomo Valkonen <tuomov@iki.fi>, University of Graz.
 *
 */


#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <string.h>
#include <stddef.h>
#include <math.h>
//#include <cblas.h>
#include <libtu/output.h>
#include <libtu/misc.h>
#include "matrix.h"
#include "tensor.h"
#include "acc_support.h"
#include "number.h"
#include "doubleio.h"
#include "scaleplus.h"
#include "misc.h"
#include "tensornorm.h"

#define IS_ID(X) X 
//#define IS_ID(X) FALSE

/* lapack */

typedef double doublereal;
typedef float real;
typedef ptrdiff_t integer;

extern int dsyev_(char *jobz, char *uplo, integer *n, doublereal *a,
                  integer *lda, doublereal *w, doublereal *work, integer *lwork, 
                  integer *info);
extern int ssyev_(char *jobz, char *uplo, integer *n, real *a, 
                  integer *lda, real *w, real *work, integer *lwork, 
                  integer *info);
           
/* !lapack */
           

Matrix *alloc_matrix(const Sig *sign, 
                     const Sig *sigm)
{                               
    size_t n=sig_tensor_size(sign);
    size_t m=sig_tensor_size(sigm);
    Matrix *mat=ALLOC(Matrix);
    if(mat==NULL)
        die_err();
    mat->p=malloc(sizeof(FNumber)*n*m);
    if(mat->p==NULL)
        die_err();
    mat->n=n;
    mat->m=m;
    mat->is_identity=0;
    mat->sign=*sign;
    mat->sigm=*sigm;
    return mat;
}


static void make_identity(Matrix *mat)
{
    size_t n=mat->n, m=mat->m;
    FNumber *p=mat->p;
    
    for(size_t i=0; i<n; i++){
        for(size_t j=0; j<m; j++)
            p[i+j*n]=0;
        p[i+i*n]=1;
    }
    mat->is_identity=1;
}

Matrix *load_matrix(const Sig *sign, 
                    const Sig *sigm,
                    const char *file)
{
    Matrix *mat=alloc_matrix(sign, sigm);
    
    if(file==NULL){
        if(mat->n!=mat->m)
            die("Mismatching tensors for identity operation");
        make_identity(mat);
    }else{
        read_doublefile(1, mat->n*mat->m, file,
                        readfn_simple, mat->p);
    }
    return mat;
}

// <y, Ax>_F = <y, EXPAND A_pres SYMPRES x>_F
//           = <W_y SYMPRES y, A_pres SYMPRES x>_Rn
//           = <W_y y_pres, A_pres x_pres>_Rn
//           = <A_pres^T W_y y_pres, x_pres>_Rn
//           = <A_pres^T W_y y_pres, W_x^(-1) W_x x_pres>
//           = <W_x^(-1) A_pres^T W_y y_pres, W_x x_pres>
//           = <A_pres^* y_pres, W_x x_pres>
//           = <EXPAND A_pres^* SYMPRES y, x>_F
//           = <A^* y, x>_F
// where _pres indicates symmetric presentation.
Matrix *matrix_conjugate(const Matrix *m1)
{
    Matrix *m2=alloc_matrix(&m1->sigm, &m1->sign);
    if(IS_ID(m1->is_identity)){
        make_identity(m2);
    }else{
        FNumber *wy=sig_get_weight_host(&m1->sign);
        FNumber *wx=sig_get_weight_host(&m1->sigm);
        FNumber *p1=m1->p;
        FNumber *p2=m2->p;
        size_t n=m1->n;
        size_t m=m1->m;
        
        for(size_t i=0; i<m; i++){
            for(size_t j=0; j<n; j++){
                FNumber wyj=(wy ? wy[j] : 1);
                FNumber wxi=(wx ? wx[i] : 1);
                p2[i+j*m]=p1[j+n*i]*wyj/wxi;
            }
        }
    }
    
    return m2;
}

void tensorfield_linop(TensorField *y,
                       const Matrix *a,
                       const TensorField *x,
                       FNumber yfact,
                       FNumber xfact)
{
#if 0
    fprintf(stderr, "(%d %d) <- (%d %d, %d %d) (%d %d)",
            y->sig.dim, y->sig.order, 
            a->sign.dim, a->sign.order,
            a->sigm.dim, a->sigm.order,
            x->sig.dim, x->sig.order);
#endif

    if(IS_ID(a->is_identity)){
        if(yfact==FNUMBER(0.0) && xfact==FNUMBER(1.0))
            tensorfield_copy(y, x);
        else
            tensorfield_scaleplus(y, y, x, yfact, xfact, 1);
    }else{
        FNumber * DEVPTR yp=y->p;
        const FNumber * DEVPTR xp=x->p;
        const FNumber *ap=a->p;
        const size_t count=x->nelem;
        const size_t n=a->n;
        const size_t m=a->m;

        assert(y->nelem==x->nelem);
        assert(y->elemsize==n);
        assert(x->elemsize==m);
        
        #pragma acc kernels loop deviceptr(yp, xp), \
                copyin(ap[0:n*m]), independent
        #pragma omp parallel for
        for(size_t k=0; k<count; k++){
            size_t ypoff=k*n;
            size_t xpoff=k*m;
            #pragma acc loop vector, independent
            for(size_t i=0; i<n; i++){
                FNumber tmp=0;
                for(size_t j=0; j<m; j++){
                    tmp+=ap[i+n*j]*xp[xpoff+j];
                }
                /*if(yfact==0)
                    yp[k*n+i]=xfact*tmp;
                else*/
                yp[ypoff+i]=yfact*yp[ypoff+i]+xfact*tmp;
            }
        }
    }
}                       

void free_matrix(Matrix *m)
{
    free(m->p);
    free(m);
}

Matrix *matrix_multiply(const Matrix *a, const Matrix *b)
{
    assert(a->m==b->n);
    Matrix *r=alloc_matrix(&a->sign, &b->sigm);

#if 1
    size_t n=r->n;
    size_t m=r->m;
    size_t p=a->m;
    FNumber *ap=a->p;
    FNumber *bp=b->p;
    FNumber *rp=r->p;
    for(size_t i=0; i<n; i++){
        for(size_t j=0; j<m; j++){
            FNumber tmp=0;
            for(size_t k=0; k<p; k++)
                tmp+=ap[i+k*n]*bp[k+j*p];
            rp[i+j*n]=tmp;
        }
    }
#else
#if USE_FLOAT
    cblas_sgemm
#else
    cblas_dgemm
#endif    
               (CblasColMajor, 
                CblasNoTrans,
                CblasNoTrans, 
                a->n, a->m, b->n, 1.0, 
                a->p, a->n,
                b->p, b->n,
                0.0, 
                r->p, r->n);
#endif

    return r;
}


static void symeigs_inplace(FNumber *a, ptrdiff_t dim, FNumber *ev)
{
    ptrdiff_t one=1;
    ptrdiff_t res=0;
    ptrdiff_t worksize;
    char job='n';
    char uplo='u';
    
    worksize=3*dim;
    FNumber *work=malloc(sizeof(FNumber)*worksize);
    assert(work!=NULL);

#if USE_FLOAT
    ssyev_(&job, &uplo, &dim, a, &dim, ev, work, &worksize, &res);
#else
    dsyev_(&job, &uplo, &dim, a, &dim, ev, work, &worksize, &res);
#endif
           
    free(work);
    
    assert(res==0);
}

static FNumber minabs_symeig_inplace(FNumber *a, size_t dim)
{
    FNumber *ev=malloc(sizeof(FNumber)*dim);
    assert(ev!=NULL);
    symeigs_inplace(a, dim, ev);
    FNumber mi=INFINITY;
    for(size_t i=0; i<dim; i++)
        mi=fn_min(mi, fn_abs(ev[i]));
    free(ev);
    return mi;
}

static FNumber maxabs_symeig_inplace(FNumber *a, size_t dim)
{
    FNumber *ev=malloc(sizeof(FNumber)*dim);
    assert(ev!=NULL);
    symeigs_inplace(a, dim, ev);
    FNumber ma=0;
    for(size_t i=0; i<dim; i++)
        ma=fn_max(ma, fn_abs(ev[i]));
    free(ev);
    return ma;
}


// Calculate W_x^(1/2) A_pres^* A_pres W_x^(-1/2)
// (see below for enlightening comments)
static Matrix *matrix_conjprod_normweight(const Matrix *a)
{
    Matrix *b=matrix_conjugate(a);
    Matrix *a2=matrix_multiply(b, a);
    free_matrix(b);
    {
        FNumber *wn=sig_get_weight_host(&a2->sign);
        FNumber *a2p=a2->p;
        size_t n=a2->n;
        
        for(size_t i=0; i<n; i++){
            for(size_t j=0; j<n; j++){
                FNumber w1=(wn ? wn[i] : 1);
                FNumber w2=(wn ? wn[j] : 1);
                a2p[i+j*n]*=sqrt(w1/w2);
            }
        }
    }
    
    return a2;
}


// Writing A= EXPAND A_pres SYMPRES, x_prex = SYMPRES x, and minding that
// A_pres^* = W_x^(-1) A_pres^T W_y, we have
//
//   sup_x |Ax|_F/|x|_F
//     = sup_x |EXPAND A_pres SYMPRES x|_F / |x|_F
//     = sup_x |W_y^(1/2) A_pres x_pres|_Rn / |W_x^(1/2) x_pres|_Rn
//     = sup_z |W_y^(1/2) A_pres W_x^(-1/2) z|_Rn / |z|_Rn
//     = sup_{|z|_Rn=1} |W_y^(1/2) A_pres W_x^(-1/2) z|_Rn
//     = sup_{|z|_Rn=1} sqrt( < W_y^(1/2) A_pres W_x^(-1/2) z, 
//                              W_y^(1/2) A_pres W_x^(-1/2) z >_Rn )
//     = sup_{|z|_Rn=1} sqrt( < W_x^(-1/2) A_pres^T W_y A_pres W_x^(-1/2) z, z>_Rn )
//     = sup_{|z|_Rn=1} sqrt( < W_x^(1/2) A_pres^* A_pres W_x^(-1/2) z, z>_Rn )
//     = sqrt(max_i lambda_i(W_x^(1/2) A_pres^* A_pres W_x^(-1/2)) )
FNumber matrix_opnorm(const Matrix *a)
{
    Matrix *a2=matrix_conjprod_normweight(a);
    FNumber maxabs2=maxabs_symeig_inplace(a2->p, a2->m);
    free_matrix(a2);
    return fn_sqrt(maxabs2);
}

// Return gamma such that |Av|^2 > gamma |v|^2
FNumber matrix_normsquared_lowerbound(const Matrix *a)
{
    Matrix *a2=matrix_conjprod_normweight(a);
    FNumber minabs2=minabs_symeig_inplace(a2->p, a2->m);
    free_matrix(a2);
    return minabs2;
}

#define UPDATE UPDATE_PLUS
#define POWER idop
#define DIFF powerdiff
#define APPLY_WEIGHT 1
#define WORKER sumdot1_linop_diff_w
#include "sumdot1_linop_worker_template.h"

#define UPDATE UPDATE_PLUS
#define POWER idop
#define DIFF powerdiff
#define APPLY_WEIGHT 0
#define WORKER sumdot1_linop_diff_nw
#include "sumdot1_linop_worker_template.h"

FNumber tensorfield_linop_frob2dist22(const TensorField *y,
                                      const Matrix *a,
                                      const TensorField *x)
{
    if(IS_ID(a->is_identity)){
        return tensorfield_frob2dist22(y, x);
    }else{
        const size_t count=x->nelem;
        const size_t n=a->n;
        const size_t m=a->m;
        const FNumber *wm=tensorfield_get_weight(y);
        
        assert(y->nelem==x->nelem);
        assert(y->elemsize==n);
        assert(x->elemsize==m);
        
        FNumber res;
        res=(wm==NULL
             ? sumdot1_linop_diff_nw(n, m, count, y->p, x->p, a->p, wm)
             : sumdot1_linop_diff_w(n, m, count, y->p, x->p, a->p, wm));
        
        return res/2;
    }
}                       
