/*
 * dticode/src/method.c
 *
 * Written 2013 by Tuomo Valkonen <tuomov@iki.fi>, University of Cambridge.
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <libtu/util.h>
#include <libtu/output.h>
#include <libtu/optparser.h>
#include <libtu/misc.h>

#include "method.h"
#include "vector.h"
#include "timing.h"
#include "parallel.h"
#include "combine.h"
#include "combine_nonlin.h"
#include "feedback.h"
#include "feedback_nonlin.h"
#include "cpock_accel.h"
#include "cpock_nonlin.h"

static void help(OptParserOpt *opts)
{
    printf("Usage: %s [options]\n\n", libtu_progname());
    optparser_printhelp(OPTP_MIDLONG, opts);
    printf("\n");
}

bool method_do_opt(MethodParam *mp, int opt, OptParserOpt *opts)
{
    switch(opt){
        case OPT_ID('t'):
            mp->cpockp.tau=atof(optparser_get_arg()); 
            break;
        case OPT_ID('s'):
            mp->cpockp.sigma=atof(optparser_get_arg());
            break;
        case OPT_ID('P'):
            {
                const char *s=optparser_get_arg();
                if(strcmp(s, "lin")==0)
                    mp->cpockp.mode=NLCPockLin;
                else if(strcmp(s, "exact")==0)
                    mp->cpockp.mode=NLCPockExact;
                else if(strcmp(s, "interp")==0)
                    mp->cpockp.mode=NLCPockInterp;
                else
                    optparser_print_error();
                break;
            }
        case OPT_ID('M'):
            {
                const char *method=optparser_get_arg();
            
                if(strcmp(method, "bregman-cpock")==0){
                    mp->gaussnewton=FALSE;
                    mp->bregman=TRUE;
                }else if(strcmp(method, "bregman-gn")==0){
                    mp->gaussnewton=TRUE;
                    mp->bregman=TRUE;
                }else if(strcmp(method, "cpock")==0){
                    mp->gaussnewton=FALSE;
                    mp->bregman=FALSE;
                }else if(strcmp(method, "gn")==0){
                    mp->gaussnewton=TRUE;
                    mp->bregman=FALSE;
                }else{
                    die("Invalid method.");
                }
                break;
            }
        case OPT_ID('r'):
            mp->cpockp.rho=atof(optparser_get_arg());
            break;
        case OPT_ID('i'):
            mp->cpockp.maxiter=atoi(optparser_get_arg());
            break;
        case OPT_ID('I'):
            mp->maxfbiter=atoi(optparser_get_arg());
            //bregman=TRUE;
            break;
        case OPT_ID('B'):
            mp->disttarget=atof(optparser_get_arg());
            //bregman=TRUE;
            break;
        case OPT_ID('e'):
            mp->fbeps=atof(optparser_get_arg());
            //bregman=TRUE;
            break;
        case OPT_ID('w'):
            mp->sig.w=atoi(optparser_get_arg());
            break;
        case OPT_ID('h'):
            mp->sig.h=atoi(optparser_get_arg());
            break;
        case OPT_ID('d'):
            mp->sig.d=atoi(optparser_get_arg());
            break;
        case OPT_ID('X'):
            mp->sig.xres=atof(optparser_get_arg());
            break;
        case OPT_ID('Y'):
            mp->sig.yres=atof(optparser_get_arg());
            break;
        case OPT_ID('Z'):
            mp->sig.zres=atof(optparser_get_arg());
            break;
        case OPT_ID('G'):
            mp->cpockp.verbose_granularity=atoi(optparser_get_arg());
            break;
        case OPT_ID('4'):
            {
                const char *s=optparser_get_arg();
                if(strcmp(s, "zero")==0)
                    mp->initstyle=FidInitPrimalZero;
                else if(strcmp(s, "backproj")==0)
                    mp->initstyle=FidInitPrimalBackproj;
                else if(strcmp(s, "file")==0)
                    mp->initstyle=FidInitPrimalFile;
                else
                    optparser_print_error();
                break;
            }
        case OPT_ID('3'):
            mp->initfile=optparser_get_arg();
            mp->initstyle=FidInitPrimalFile;
            break;
        case OPT_ID('o'):
            mp->outfile=optparser_get_arg();
            break;
        case OPT_ID('p'):
            mp->fullsave=TRUE;
            break;
        case OPT_ID('g'):
            mp->diffalg=optparser_get_arg();
            break;
        case OPT_ID('T'):
            mp->timing=1;
            break;
        case OPT_ID('V'):
            mp->cpockp.verbose=1;
            break;
        case 'h':
            help(opts);
            exit(0);
        default:
            return FALSE;
    }
    
    return TRUE;
}

void method_params_sanity_check(MethodParam *mp)
{
    if(mp->outfile==NULL)
        die("outfile missing\n");

    if(mp->initfile==NULL && mp->initstyle==FidInitPrimalFile)
        die("initfile missing\n");
}


void method_do(MethodParam *mp, Fid *fid, FidParams *fidp, Reg *reg, RegParams *regp)
{
    Vector *u;

    // Run things
    if(mp->timing)
        timing_start();

    if(mp->gaussnewton){
        if(mp->bregman){
            u=combined_do_feedback_cpock(mp, fid, fidp, reg, regp,
                                         FidBaseshiftBregman, "Bregman-Gauss-Newton");
        }else{
            u=combined_do_feedback_cpock(mp, fid, fidp, reg, regp,
                                         FidBaseshiftSet, "Gauss-Newton");
        }
    }else{
        if(mp->bregman){
            u=combined_do_feedback_cpock(mp, fid, fidp, reg, regp,
                                         FidBaseshiftBregman, "Bregman");
        }else{
            u=combined_do_feedback_cpock(mp, fid, fidp, reg, regp,
                                         FidBaseshiftNone, "PDHGM");
        }
    }

    if(mp->timing)
        timing_end();
    
    if(mp->fullsave)
        warn("Fullsave unimplemented");
    vector_save(u, mp->outfile);
}

void method_do_nl(MethodParam *mp, FidNL *fidnl, FidParams *fidp, Reg *reg, RegParams *regp)
{
    Vector *u;
    
    if(mp->timing)
        timing_start();

    if(mp->bregman){
        u=combined_do_feedback_cpock_nonlin(mp, fidnl, fidp, reg, regp,
                                            FidBaseshiftBregman, "Bregman");
    }else{
        //u=combine_do_cpock_nonlin(&cpockp, fidnl, &fidp, reg, &regp, 
        //                          mp->initstyle, mp->initfile);
        u=combined_do_feedback_cpock_nonlin(mp, fidnl, fidp, reg, regp,
                                            FidBaseshiftNone, "NL-PDHGM");
    }

    if(mp->timing)
        timing_end();
    
    if(mp->fullsave)
        warn("Fullsave unimplemented");
    vector_save(u, mp->outfile);
}

