/*
 * dticode/src/multiply.c
 *
 * Written 2013 by Tuomo Valkonen <tuomov@iki.fi>, University of Cambridge.
 *
 */

#include <assert.h>
#include <libtu/objp.h>
#include <libtu/types.h>
#include <libtu/misc.h>
#include <libtu/output.h>
#include "complexx.h"
#include "splitvar.h"


#define MULTIPLY(A, B) ((A)*(B))
#define DIVIDE(A, B) ((A)/(B))

#define DO_MULTIPLY_A(y, x1, x2, sety, getx1, getx2, op) { \
        const FNumber *x1_p=x1->p; \
        const FNumber *x2_p=x2->p; \
        FNumber *y_p=y->p; \
        const size_t nelem=y->nelem; \
        const size_t yelsz=(y->elemsize/(y->sig.is_complex ? 2 : 1)); \
        const size_t elemsize=(x1->sig.order==0 ? yelsz : 1); \
        const size_t npts=nelem*(x1->sig.order==0 ? 1 : yelsz); \
        if(elemsize==1){ \
            _Pragma("omp parallel for") \
            _Pragma("acc kernels loop deviceptr(y_p, x1_p, x2_p), independent") \
            for(size_t j=0; j<npts; j++) \
                sety(y_p, j, op(getx2(x2_p, j), getx1(x1_p, j))); \
        }else{ \
            _Pragma("omp parallel for") \
            _Pragma("acc kernels loop deviceptr(y_p, x1_p, x2_p), independent") \
            for(size_t j=0; j<npts; j++){ \
                const __typeof__(getx1(x1_p, j)) b=getx1(x1_p, j); \
                for(size_t k=0; k<elemsize; k++){ \
                    const __typeof__(getx2(x2_p, elemsize*j+k)) a=getx2(x2_p, elemsize*j+k); \
                    sety(y_p, elemsize*j+k, op(a, b)); \
                } \
            } \
        } \
    }

#define MULTIPLY_A(y, x1, x2, addto_y, yset_type, getx1, getx2, op) { \
        if(addto_y){ \
            DO_MULTIPLY_A(y, x1, x2, add_##yset_type, getx1, getx2, op); \
        }else{ \
            DO_MULTIPLY_A(y, x1, x2, set_##yset_type, getx1, getx2, op); \
        } \
    }

void tensorfield_multiply_pointwise(TensorField *y, 
                                    const TensorField *x1, 
                                    const TensorField *x2, 
                                    bool addto_y)
{
    assert(y->nelem==x1->nelem && y->nelem==x2->nelem);
    
    assert(!y->sig.is_complex || (x1->sig.is_complex || x2->sig.is_complex));
    assert(y->sig.is_complex || (!x1->sig.is_complex && !x2->sig.is_complex));
    
    if(x1->sig.order==0){
        assert(x2->sig.order==y->sig.order && x2->sig.dim==y->sig.dim);
    }else if(x2->sig.order==0){
        assert(x1->sig.order==y->sig.order && x1->sig.dim==y->sig.dim);
        const TensorField *xt=x1;
        x1=x2;
        x2=xt;
    }else{
        assert(x1->sig.order==y->sig.order && x1->sig.dim==y->sig.dim);
        assert(x2->sig.order==y->sig.order && x2->sig.dim==y->sig.dim);
    }
    
    if(!x1->sig.is_complex && !x2->sig.is_complex){
        // y is asserted not complex
        MULTIPLY_A(y, x1, x2, addto_y, real, get_real, get_real, MULTIPLY);
    }else if(!x1->sig.is_complex && x2->sig.is_complex){
        // y is asserted complex
        MULTIPLY_A(y, x1, x2, addto_y, complex, get_real, get_complex, MULTIPLY);
    }else if(!x2->sig.is_complex && x1->sig.is_complex){
        // y is asserted complex
        MULTIPLY_A(y, x1, x2, addto_y, complex, get_complex, get_real, MULTIPLY);
    }else{
        // y is asserted complex
        MULTIPLY_A(y, x1, x2, addto_y, complex, get_complex, get_complex, MULTIPLY);
    }
}

// <a+ib, (c+id)(c'+id')> (real inner product)
// = a(cc'-dd')+b(cd'+c'd)
// = (ac+bd)c' + (bc-ad)d'
// = <(a+ib)(c-id), c'+id'> (real inner product)
void tensorfield_multiply_pointwise_conj(TensorField *y, 
                                         const TensorField *x1, 
                                         const TensorField *x2, 
                                         bool addto_y)
{
    assert(y->nelem==x1->nelem && y->nelem==x2->nelem);
    
    assert(!y->sig.is_complex || x2->sig.is_complex);
    
    if(x1->sig.order==0){
        assert(x2->sig.order==y->sig.order && x2->sig.dim==y->sig.dim);
    }else{
        assert(x1->sig.order==y->sig.order && x1->sig.dim==y->sig.dim);
        assert(x2->sig.order==y->sig.order && x2->sig.dim==y->sig.dim);
    }
    
    if(!x1->sig.is_complex && !x2->sig.is_complex){
        // y is asserted not complex
        MULTIPLY_A(y, x1, x2, addto_y, real, get_real, get_real, MULTIPLY);
    }else if(!y->sig.is_complex && !x1->sig.is_complex){
        // x2 is asserted complex
        MULTIPLY_A(y, x1, x2, addto_y, real_part, get_real, get_complex, MULTIPLY);
    }else if(!y->sig.is_complex && x1->sig.is_complex){
        // x2 is asserted complex
        MULTIPLY_A(y, x1, x2, addto_y, real_part, get_complex_conj, get_complex, MULTIPLY);
    }else if(y->sig.is_complex && !x1->sig.is_complex){
        // x2 is asserted complex
        MULTIPLY_A(y, x1, x2, addto_y, complex, get_real, get_complex, MULTIPLY);
    }else{
        // everything complex
        MULTIPLY_A(y, x1, x2, addto_y, complex, get_complex_conj, get_complex, MULTIPLY);
    }
}


void tensorfield_divide_pointwise(TensorField *y, 
                                  const TensorField *x1, 
                                  const TensorField *x2, 
                                  bool addto_y)
{
    assert(y->nelem==x1->nelem && y->nelem==x2->nelem);
    
    assert(!y->sig.is_complex || (x1->sig.is_complex || x2->sig.is_complex));
    assert(y->sig.is_complex || (!x1->sig.is_complex && !x2->sig.is_complex));
    
    if(x2->sig.order==0){
        assert(x1->sig.order==y->sig.order && x1->sig.dim==y->sig.dim);
    }else{
        assert(x1->sig.order==y->sig.order && x1->sig.dim==y->sig.dim);
        assert(x2->sig.order==y->sig.order && x2->sig.dim==y->sig.dim);
    }
    
    if(!x1->sig.is_complex && !x2->sig.is_complex){
        // y is asserted not complex
        MULTIPLY_A(y, x1, x2, addto_y, real, get_real, get_real, DIVIDE);
    }else if(!x1->sig.is_complex && x2->sig.is_complex){
        // y is asserted complex
        MULTIPLY_A(y, x1, x2, addto_y, complex, get_real, get_complex, DIVIDE);
    }else if(!x2->sig.is_complex && x1->sig.is_complex){
        // y is asserted complex
        MULTIPLY_A(y, x1, x2, addto_y, complex, get_complex, get_real, DIVIDE);
    }else{
        // y is asserted complex
        MULTIPLY_A(y, x1, x2, addto_y, complex, get_complex, get_complex, DIVIDE);
    }
}
