/*
 * dticode/tensorreg/fid_fourier_l2.c
 *
 * Written 2012 by Tuomo Valkonen <tuomov@iki.fi>, University of Cambridge.
 *
 */

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <libtu/types.h>
#include <libtu/output.h>
#include <libtu/misc.h>
#include <string.h>

#include "tensor.h"
#include "combine.h"
#include "fid_fourier_l2.h"
#include "fourier_sample.h"
#include "scaleplus.h"
#include "project_posdef.h"
#include "tensornorm.h"
#include "posdef.h"
#include "sumdot1.h"
#include "normalise.h"
#include "doubleio.h"
#include "fft.h"
#include "complexx.h"

//
// |f-Fu|^2/2 + |Mu|^2/2
// = |f-Fu|^2/2 + <lambda, Mu> - |lambda|^2/2
// = G(u) + <lambda, Ku> - F(lambda)
//
// NOTE: We use symmetric scaling of the DFT.
// This is in particular for better convergence properties
// of the Chambolle-Pock algorithm.
//
//  Fx=(1/sqrt(N)) sum_{j=0}^{N-1} x_j exp(-i 2 pi j/N)
//  F^* x=(1/sqrt(N)) sum_{j=0}^{N-1} x_j exp(i 2 pi j/N)
//


static void fourier_l2_k(FidFourierL2Dual *a, const FidFourierL2Primal *b, 
                         bool addto_a, FidFourierL2Params *fp)
{
    if(fp->mask_term_weight!=0){
        FNumber *ap=a->p;
        const FNumber *bp=b->p;
        const FNumber *mp=fp->mask->p;
        const FNumber w=fp->mask_term_weight;
        const size_t n=a->nelem;
        const size_t m=a->elemsize;
        
        if(addto_a){
            #pragma omp parallel for
            #pragma acc kernels loop deviceptr(ap, bp, mp), independent
            for(size_t i=0; i<n; i++){
                #pragma acc loop independent, vector
                for(size_t j=0; j<m; j++)
                    ap[i*m+j]+=bp[i*m+j]*mp[i]*w;
            }
        }else{
            #pragma omp parallel for
            #pragma acc kernels loop deviceptr(ap, bp, mp), independent
            for(size_t i=0; i<n; i++){
                #pragma acc loop independent, vector
                for(size_t j=0; j<m; j++)
                    ap[i*m+j]=bp[i*m+j]*mp[i]*w;
            }
        }
    }
}

static void fourier_l2_kc(FidFourierL2Primal *a, const FidFourierL2Dual *b, 
                          bool addto_a, FidFourierL2Params *fp)
{
    if(fp->mask_term_weight!=0){
        FNumber *ap=a->p;
        const FNumber *bp=b->p;
        const FNumber *mp=fp->mask->p;
        const FNumber w=fp->mask_term_weight;
        const size_t n=a->nelem;
        const size_t m=a->elemsize;
        
        if(addto_a){
            #pragma omp parallel for
            #pragma acc kernels loop deviceptr(ap, bp, mp), independent
            for(size_t i=0; i<n; i++){
                #pragma acc loop independent, vector
                for(size_t j=0; j<m; j++)
                    ap[i*m+j]+=bp[i*m+j]*mp[i]*w;
            }
        }else{
            #pragma omp parallel for
            #pragma acc kernels loop deviceptr(ap, bp, mp), independent
            for(size_t i=0; i<n; i++){
                #pragma acc loop independent, vector
                for(size_t j=0; j<m; j++)
                    ap[i*m+j]=bp[i*m+j]*mp[i]*w;
            }
        }
    }
}

static void fourier_l2_gres(FidFourierL2Primal *u, FNumber tau, 
                            FidFourierL2Params *fp)
{
    size_t rank;
    int dims[3];
    FNumber normfact;
    
    fft_init_params(&rank, dims, &normfact, &u->sig);

    // Solve for u_out:
    // (I+tau SF^* SF) u_out = tau SF^* f + u_in
    // Now u = F(tau S^* F^* f + u_in) = tau S^* f + F u_in
    // ==> (1+tau S^*S)F u_out = u
    
    do_fft(rank, dims, u->p, u->p);
    
    if(fp->f){
        // Need to normalise u by 1/sqrt(normfact)
        //tensorfield_scaleplus(u, fp->f, u, tau, 1/sqrt(normfact), 1);
        // Fu_out=u/(1+tau)
        // Need further normalisation of the whole thing by 1/sqrt(normfact)
        //tensorfield_scale(u, 1/(sqrt(normfact)*(1+tau)));
        tensorfield_scaleplus(u, u, fp->f, 1/sqrt(normfact), tau,
                              sqrt(normfact)*(1+tau));
    }else{
        samples_tensorfield_scaleplus(u, u, fp->samples, fp->n_samples, 
                                      1/sqrt(normfact), tau, sqrt(normfact)*(1+tau));
    }

    do_ifft(rank, dims, u->p, u->p);
}

static void fourier_l2_fcres(FidFourierL2Dual *lambda, FNumber tau,
                             FidFourierL2Params *fp)
{
    // min_x{ |lambda-x|^2/2tau+|x|^2/2 } 
    // => (x-lambda)/tau + x = 0
    //  => x =  lambda/(1+tau)
    if(fp->mask_term_weight!=0)
        tensorfield_scale(lambda, 1/(1+tau));
}

static void fourier_l2_pdgap(FNumber *pdgap,
                             const TensorField *u, 
                             const TensorField *y,
                             const TensorField *kcy,
                             const TensorField *ku,
                             TensorField *uwork,
                             unsigned int iter,
                             FidFourierL2Params *fp)
{
    FNumber g, gcm, fpart;
    size_t rank;
    int dims[3];
    FNumber normfact;

    fft_init_params(&rank, dims, &normfact, &u->sig);

    if(fp->mask_term_weight!=0){
        // F(Ku)+F^*(y)
        fpart=tensorfield_frob222(ku)+tensorfield_frob222(y);
    }else{
        fpart=0;
    }

    if(fp->f){
        // calculate G(u)
        do_fft(rank, dims, uwork->p, u->p);
        tensorfield_scale(uwork, 1/sqrt(normfact));
        g=tensorfield_frob2dist22(fp->f, uwork);
        
        // calculate G^*(-kcy)
        do_ifft(rank, dims, uwork->p, fp->f->p);
        tensorfield_scale(uwork, 1/sqrt(normfact));
        gcm=(-tensorfield_sumdot1(kcy, uwork, FALSE, 1)
             +tensorfield_frob222(kcy));
    }else{
        double a=sqrt(normfact);
        size_t n_samples=fp->n_samples;
        const FourierSample *samples=fp->samples;
        size_t uwork_nelem=uwork->nelem;
        FNumber *uwork_p=uwork->p;
        
        do_fft(rank, dims, uwork->p, u->p);
        g=0;
        
        #pragma omp parallel for
        #pragma acc kernels loop, independent, reduction(+:g), deviceptr(samples, uwork_p)
        for(size_t i=0; i<n_samples; i++){
            double complex x=get_complex(uwork_p, samples[i].idx)/a;
            double complex s=samples[i].val;
            g+=fn_cabs2(s-x);
        }
        g/=2;
        
        
        //FNumber mx=tensorfield_frob2(u);
        do_fft(rank, dims, uwork->p, u->p);
        FNumber mx=tensorfield_frobinf(uwork)/a;
        if(mx>fp->fu_infty_bound2)
            fp->fu_infty_bound2=mx;
        else
            mx=fp->fu_infty_bound2;
            
        do_fft(rank, dims, uwork->p, kcy->p);
        gcm=0;
        
        // sup_u <x,u>-|f-SFu|^2/2, (x=-K^* y)
        // => SFu-f=SFx
        // => <x,u>-|f-SFu|^2/2 
        //    = <Fx,Fu> -|f-SFu|^2/2
        //    = <(I-S)Fx,(I-S)Fu> + <SFx,SFu> -|f-SFu|^2/2
        //    = <(I-S)Fx,(I-S)Fu> + <SFx,SFx> + <SFx,f> -|SFx|^2/2
        //    = <(I-S)Fx,(I-S)Fu> + <SFx,f> + |SFx|^2/2
        // First term is infinite, unless we add a bount
        // |(I-S)Fu|_infty < M, in which case
        // we get M * #{i in I-S : (Fx)_i!=0}.
        #pragma acc kernels deviceptr(samples, uwork_p)
        {
            #pragma omp parallel for
            #pragma acc loop, independent, reduction(+:gcm)
            for(size_t i=0; i<n_samples; i++){
                // ASSUMPTION: samples sorted by idx
                size_t j=samples[i].idx;
                FNumber gcm0=gcm;
                size_t prev=(i==0 ? 0 : samples[i-1].idx+1);
                #pragma acc loop independent, reduction(+:gcm0)
                for(size_t k=prev; k<j; k++)
                    gcm0+=fn_cabs(get_complex(uwork_p, k)*mx/a);
                FNumber_complex f=samples[i].val;
                FNumber_complex x=-get_complex(uwork_p, j)/a;
                
                FNumber x_abs2=fn_cabs2(x);
                if(x_abs2>0)
                    mx=fn_max(mx, x_abs2);
                // Fx_i f_i + Fx_i^2/2
                gcm+=gcm0+ciprod(f, x)+x_abs2/2;
            }
            
            size_t prev=samples[n_samples-1].idx+1;
            
            #pragma omp parallel for
            #pragma acc loop, independent, reduction(+:gcm)
            for(size_t k=prev; k<uwork_nelem; k++)
                gcm+=fn_cabs(get_complex(uwork_p, k)*mx/a);
        }
    }
    
    //fprintf(stderr, "%f %f\n", g, gcm);
    
    pdgap[0]=g+gcm+fpart;
    
    if(iter==0)
        fp->pdgap0_ab=pdgap[0];
    pdgap[1]=fp->pdgap0_ab;
}

static TensorField *fourier_l2_alloc_primal(FidFourierL2Params *fp)
{
    return alloc_tensorfield(&fp->sig);
}

static Vector *fourier_l2_alloc_dual(FidFourierL2Params *fp)
{
    if(fp->mask_term_weight!=0)
        return (Vector*)alloc_tensorfield(&fp->sig);
    else
        return (Vector*)alloc_dummyvector();
}

FNumber fid_fourier_l2_baseshift(FidFourierL2Primal *u, FidBaseshiftMode mode, 
                                 FidFourierL2Params *fp)
{
    size_t rank;
    int dims[3];
    FNumber normfact;
    FNumber dist=0;
    TensorField *f_tmp=fp->f_tmp;
    FourierSample *samples_tmp=fp->samples_tmp;
    const TensorField *f_orig=fp->f_orig;
    const FourierSample *samples_orig=fp->samples_orig;
    size_t n_samples=fp->n_samples;

    // Allocate copy
    if(fp->full_f && f_tmp==NULL){
        f_tmp=alloc_tensorfield(&(f_orig->sig));
        tensorfield_copy(f_tmp, f_orig);
        fp->f=f_tmp;
        fp->f_tmp=f_tmp;
    }else if(!fp->full_f && samples_tmp==NULL){
        samples_tmp=ALLOC_N(FourierSample, n_samples);
        assert(samples_tmp!=NULL);
        memcpy(samples_tmp, samples_orig, sizeof(FourierSample)*n_samples);
        fp->samples=samples_tmp;
        fp->samples_tmp=samples_tmp;
    }

    // Do FFT. Done in-place on u - has to be restored afterwards!
    fft_init_params(&rank, dims, &normfact, &u->sig);
    do_fft(rank, dims, u->p, u->p);
    tensorfield_scale(u, 1/sqrt(normfact));

    // Calculate current distance and shift
    if(fp->full_f){
        dist=tensorfield_frob2dist22(f_orig, u);
        
        if(mode==FidBaseshiftAdd){
            // f_tmp := f_tmp - Fu
            tensorfield_scaleplus(f_tmp, f_tmp, u, 1, -1, 1);
        }else if(mode==FidBaseshiftSet){
            // f_tmp := f_orig - Fu
            tensorfield_scaleplus(f_tmp, f_orig, u, 1, -1, 1);
        }else if(mode==FidBaseshiftBregman){
            // f_tmp := f_orig + f_tmp - Fu
            tensorfield_scaleplus(f_tmp, f_tmp, f_orig, 1, 1, 1);
            tensorfield_scaleplus(f_tmp, f_tmp, u, 1, -1, 1);
        }else if(mode==FidBaseshiftNone){
            tensorfield_copy(f_tmp, f_orig);
        }else{
            die("Unknown FidBaseshiftMode");
        }
    }else{
        FNumber *u_p=u->p;
        
        #warning "TODO: Distance calculation doesn't make sense generally"

        #pragma omp parallel for
        #pragma acc kernels loop, independent, reduction(+:dist), deviceptr(samples_orig, samples_tmp, u_p)
        for(size_t i=0; i<n_samples; i++){
            double complex x=get_complex(u_p, samples_orig[i].idx);
            dist+=fn_cabs2(samples_orig[i].val-x);
            
            size_t j=samples_orig[i].idx;
            
            //assert(samples_tmp[i].idx==j);
            
            if(mode==FidBaseshiftAdd){
                samples_tmp[i].val-=get_complex(u_p, j);
            }else if(mode==FidBaseshiftSet){
                samples_tmp[i].val=samples_orig[i].val-get_complex(u_p, j);
            }else if(mode==FidBaseshiftBregman){
                samples_tmp[i].val=samples_orig[i].val+samples_tmp[i].val-get_complex(u_p, j);
            }else /*if(mode==FidBaseshiftNone)*/{
                samples_tmp[i].val=samples_orig[i].val;
            }
        }
        
        dist/=2;
    }

    // Restore u
    do_ifft(rank, dims, u->p, u->p);
    tensorfield_scale(u, 1/sqrt(normfact));

    if(fp->samples!=NULL){
        // Calculate initial estimate of L^infty bound for Fu on Range(I-S)
        FNumber new_estim=2*pow(fourier_sample_norm_inf(fp->samples, fp->n_samples), 2);
        fp->fu_infty_bound2=fn_max(fp->fu_infty_bound2, new_estim);
    }
    
    return dist;
}

void fid_fourier_l2_init_primal(FidFourierL2Primal *u, FidInitPrimalStyle initstyle, 
                                FidFourierL2Params *fp)
{
    if(initstyle==FidInitPrimalZero){
        tensorfield_zero(u);
    }else if(initstyle==FidInitPrimalBackproj){
        fourier_backproj(u, (fp->full_f ? fp->f : NULL), fp->samples, fp->n_samples);
    }else{
        die("Unknown FidInitPrimalStyle");
    }
}

static FNumber fid_fourier_l2_estimate(Estimate estim, FidFourierL2Params *fp)
{
    if(estim==EstimateKNorm){
        return (fp->mask==NULL ? 0 : fp->mask_term_weight); 
    }else if(estim==EstimateGUniformConvexity){
        // gamma = lambda_min[(SF)^*(SF)]
        //       = lambda_min[F^*S^*SF]
        //       = lambda_min[F^{-1}S^*SF]
        //       = 1 if S full and zero otherwise
        return (fp->full_f ? 1 : 0);
    }else if(estim==EstimateFCUniformConvexity){
        return INFINITY;
    }else{
        return 0;
    }
}

bool fid_fourier_l2_grad(TensorField *grad, const TensorField *u, 
                         FidFourierL2Params *fp)
{
    size_t rank;
    int dims[3];
    FNumber normfact;
    FNumber sqrtnormfact;
    
    fft_init_params(&rank, dims, &normfact, &u->sig);
    
    sqrtnormfact=sqrt(normfact);
    
    do_fft(rank, dims, grad->p, u->p);
    // Now grad=Fu*sqrt(normfact)

    if(fp->full_f){
        // Calculate S^*(SFu-f)/sqrt(normfact), also performing the 
        // normalisation by sqrt(normfact) for Fu, not performed by do_fft
        tensorfield_scaleplus(grad, grad, fp->f, 1/sqrtnormfact, -1, sqrtnormfact);
    }else{
        samples_tensorfield_scaleplus(grad, grad, fp->samples, fp->n_samples,
                                      1/sqrtnormfact, -1, sqrtnormfact);
    }

    do_ifft(rank, dims, grad->p, grad->p);
    
    return TRUE;
}

static Fid fid_fourier_l2={
    (FidK*)fourier_l2_k,
    (FidKC*)fourier_l2_kc,
    (FidFCres*)fourier_l2_fcres,
    (FidGres*)fourier_l2_gres,
    (FidPDgap*)fourier_l2_pdgap,
    (FidAllocPrimal*)fourier_l2_alloc_primal,
    (FidAllocDual*)fourier_l2_alloc_dual,
    (FidInitPrimal*)fid_fourier_l2_init_primal,
#warning "TODO: proper FidInitDual"
    (FidInitDual*)noop,
    (FidBaseshift*)fid_fourier_l2_baseshift,
    (FidEstimate*)fid_fourier_l2_estimate,
    (FidGrad*)fid_fourier_l2_grad,
};


void fid_fourier_l2_init(const TensorField *f,
                         const TensorField *mask,
                         FNumber mask_term_weight,
                         const Sig *sig,
                         size_t n_samples, 
                         const FourierSample *samples,
                         Fid **fid, 
                         FidFourierL2Params *fp)
{
    if(sig->order!=0)
        die("Invalid data order");
    
    if((f && samples) || (!f && !samples))
        die("Input data mismatch");
    
    
    fp->full_f=(f!=NULL);
    fp->sig=*sig;
    fp->mask=mask;
    fp->pdgap0_ab=0;
    fp->pdgap0_c=0;
    fp->ubound=INFINITY;
    fp->mask_term_weight=mask_term_weight;
    fp->f=f;
    fp->f_orig=f;
    fp->f_tmp=NULL;
    fp->samples=samples;
    fp->samples_orig=samples;
    fp->samples_tmp=NULL;
    fp->n_samples=n_samples;
    
    if(samples!=NULL){
        // Calculate initial estimate of L^infty bound for Fu on Range(I-S)
        fp->fu_infty_bound2=2*pow(fourier_sample_norm_inf(samples, n_samples), 2);
    }else{
        // Not used
        fp->fu_infty_bound2=INFINITY;
    }

    
    *fid=&fid_fourier_l2;
}

