/*
 * dticode/tensorreg/fid_fourier_l2alt.c
 *
 * Written 2013 by Tuomo Valkonen <tuomov@iki.fi>, University of Cambridge.
 *
 */

#include <math.h>
#include <stdio.h>
#include <string.h>
#include <libtu/types.h>
#include <libtu/output.h>
#include <libtu/misc.h>

#include "tensor.h"
#include "combine.h"
#include "combine_nonlin.h"
#include "fid_fourier_l2alt.h"
#include "scaleplus.h"
#include "project_posdef.h"
#include "tensornorm.h"
#include "posdef.h"
#include "sumdot1.h"
#include "normalise.h"
#include "complexx.h"
#include "fft.h"
#include "diffoper.h"
#include "fourier_sample.h"

//
// The approach:
//
// |Au-f|^2_F + delta_{u>=0}
//  = max_lambda  delta_{u>=0} + <lambda, Au-f> - |lambda|^2/2
//  = max_lambda  delta_{u>=0} + <lambda,A u>    - <lambda,f> - |lambda|^2/2
//  = max_lambda  G(u)         + <lambda,K u>  - F^*(lambda)
//
// F(y) = sup_x <y,x> - <x,f>-|lambda|^2/2
//      = |y-f|^2/2
//
// G(u) = delta_{u>=0}(u) 
//
// G^*(y) = sup_x <y,x>-delta_{x>=0}(x) = 
//      { INFINITY, y has pos eigenvalues
//      { 0,        y neg.semi.def
//     
// Because G^* can be infinite, we need an extra bound |u(x)|<M, 
// where M > sup_x |f(x)| to make the duality gap practical. 
//
// Therefore, we take actually:
//
// G(u) = delta_{u>=0 & |u|<=M}(u)
//
// G^*(y) = sup_x <y,x>-delta_{x>=0 & |x|<=M}(x)
//        = M<normalise_pointwise(P_{>=0}(y)), y>
//        = M|P_{>=0}(y)|_{L2Alt-frob}
// G^*(-K^*y) = M*|P_{>=0}(-K^*y)|
//

static void fourier_l2alt_k(FidFourierL2AltDual *a, const FidFourierL2AltPrimal *b, 
                            bool addto_a, FidFourierL2AltParams *fp)
{
    size_t rank;
    int dims[3];
    FNumber normfact;
    
    fft_init_params(&rank, dims, &normfact, &b->sig);

    if(fp->full_f){
        do_fft(rank, dims, a->p, b->p);
        tensorfield_scale(a, 1/sqrt(normfact));
    }else{
        FNumber w=sqrt(normfact);
        FNumber *tmp_p=fp->tmp->p;
        FNumber *a_p=a->p;
        size_t n_samples=fp->n_samples;
        const FourierSample *samples=fp->samples;
        
        do_fft(rank, dims, tmp_p, b->p);
        
        if(addto_a){
            #pragma omp parallel for
            #pragma acc kernels loop, independent, deviceptr(samples, tmp_p, a_p)
            for(size_t i=0; i<n_samples; i++){
                FNumber_complex c=get_complex(tmp_p, samples[i].idx);
                FNumber_complex ai=get_complex(a_p, i);
                set_complex(a_p, i, ai+c/w);
            }
        }else{
            #pragma omp parallel for
            #pragma acc kernels loop, independent, deviceptr(samples, tmp_p, a_p)
            for(size_t i=0; i<n_samples; i++){
                FNumber_complex c=get_complex(tmp_p, samples[i].idx);
                set_complex(a_p, i, c/w);
            }
        }
    }
}


static void fourier_l2alt_kc_do(FidFourierL2AltPrimal *tmp, const FidFourierL2AltDual *b,
                                FidFourierL2AltParams *fp, FidFourierL2AltPrimal *addto)
{
    size_t rank;
    int dims[3];
    FNumber normfact;
    
    fft_init_params(&rank, dims, &normfact, &tmp->sig);

    if(fp->full_f){
        do_ifft(rank, dims, tmp->p, b->p);
    }else{
        FNumber *tmp_p=tmp->p;
        FNumber *b_p=b->p;
        size_t n_samples=fp->n_samples;
        const FourierSample *samples=fp->samples;
        
        tensorfield_zero(tmp);
        #pragma omp parallel for
        #pragma acc kernels loop, independent, deviceptr(samples, tmp_p, b_p)
        for(size_t i=0; i<n_samples; i++){
            FNumber_complex c=get_complex(b_p, i);
            set_complex(tmp_p, samples[i].idx, c);
        }
        do_ifft(rank, dims, tmp_p, tmp_p);
    }

    // According to the rules for combine.c, fidelity function
    // implementation has to add Kb to a
    if(addto)
        tensorfield_scaleplus(addto, addto, tmp, 1, 1/sqrt(normfact), 1);
    else
        tensorfield_scale(tmp, 1/sqrt(normfact));
}

static void fourier_l2alt_kc(FidPrimal *a, const FidFourierL2AltDual *b, 
                             bool addto_a, FidFourierL2AltParams *fp)
{
    if(addto_a)
        fourier_l2alt_kc_do(fp->tmp, b, fp, (FidFourierL2AltPrimal*)a);
    else
        fourier_l2alt_kc_do((FidFourierL2AltPrimal*)a, b, fp, NULL);
}


static void fourier_l2alt_gres(FidPrimal *u, FNumber tau, 
                               FidFourierL2AltParams *fp)
{
    // Do nothing, artificial bound on u assumed large enough,
    // so only affect duality gap.
    return;
}

static void fourier_l2alt_fcres(FidFourierL2AltDual *lambda, FNumber tau,
                                FidFourierL2AltParams *fp)
{
    // min_x{ |lambda-x|^2/2tau+<f,x>+|x|^2/2 } 
    // => (x-lambda)/tau + f + x = 0
    //  => x =  (lambda-tau*f)/(1+tau)
    tensorfield_scaleplus(lambda, lambda, fp->f, 1, -tau, 1+tau);
}

static void fourier_l2alt_pdgap(FNumber *pdgap,
                                const FidPrimal *u, 
                                const FidFourierL2AltDual *y,
                                const FidPrimal *kcy,
                                const FidFourierL2AltDual *ku,
                                FidPrimal *uwork,
                                unsigned int iter,
                                FidFourierL2AltParams *fp)
{
    FNumber ab, c, mx;
    
    /* F(Ku) + G(u) +G^*(-K*y)+F^*(y) */

    if(u!=NULL){
        vector_scaleplus((Vector*)uwork, (Vector*)kcy, (Vector*)kcy, -1, 0, 1);
    
        mx=tensorfield_frobinf((FidFourierL2AltPrimal*)u);
        mx=fn_max(mx, fp->ubound);
        fp->ubound=mx;
        c=tensorfield_frob1((FidFourierL2AltPrimal*)uwork);
    }else{
        // Called through, e.g., fid_oper that implements the same as
        // above on the operator parameter domain.
        c=0;
        mx=0;
    }
    
    ab=tensorfield_frob2dist22(ku, fp->f)      // F(Ku)
       +0.0                                    // G(u)
       +(tensorfield_iprod(y, fp->f)           // F^*(y)
         +tensorfield_frob222(y));
    // mx*cconc                               // G^*(-K^* y)
    if(iter==0){
        fp->pdgap0_ab=ab;
        fp->pdgap0_c=c;
    }
    
    pdgap[0]=ab+mx*c;
    pdgap[1]=fp->pdgap0_ab+mx*fp->pdgap0_c;
}

static FidPrimal *fourier_l2alt_alloc_primal(FidFourierL2AltParams *fp)
{
    return (FidPrimal*)alloc_tensorfield(&(fp->sig));
}

static FidPrimal *fourier_l2alt_alloc_dual(FidFourierL2AltParams *fp)
{
    return (FidPrimal*)alloc_tensorfield(&(fp->f->sig));
}

void fourier_l2alt_init_dual(FidFourierL2AltDual *b, const FidFourierL2AltPrimal *a, 
                             FidFourierL2AltParams *fp)
{
    fourier_l2alt_k(b, a, FALSE, fp);
    tensorfield_scaleplus(b, b, fp->f, 1, -1, 1);
}


static FNumber fid_fourier_l2alt_baseshift(FidFourierL2AltPrimal *u, FidBaseshiftMode mode, 
                                           FidFourierL2AltParams *fp)
{
    FNumber dist;
    
    dist=fid_fourier_l2_baseshift(u, mode, fp);

    // Update TensorField shadow copy of samples..
    if(!fp->full_f){
        size_t n_samples=fp->n_samples;
        FourierSample *samples_tmp=fp->samples_tmp;
        TensorField *f_tmp=fp->f_tmp;
        
        for(size_t i=0; i<n_samples; i++){
            set_complex(f_tmp->p, i, samples_tmp[i].val);
        }
    }
    
    return dist;
}

static FNumber fid_fourier_l2alt_estimate(Estimate estim, FidFourierL2AltParams *fp)
{
    if(estim==EstimateKNorm){
        //cpock->knorm+=(mask==NULL 
        //                ? 0 
        //                : mask_term_weight); 
        return 1;
    }else if(estim==EstimateGUniformConvexity){
        return 0;
    }else if(estim==EstimateFCUniformConvexity){
        return INFINITY;
    }else{
        return 0;
    }
}

static Fid fid_fourier_l2alt={
    (FidK*)fourier_l2alt_k,
    (FidKC*)fourier_l2alt_kc,
    (FidFCres*)fourier_l2alt_fcres,
    (FidGres*)fourier_l2alt_gres,
    (FidPDgap*)fourier_l2alt_pdgap,
    (FidAllocPrimal*)fourier_l2alt_alloc_primal,
    (FidAllocDual*)fourier_l2alt_alloc_dual,
    (FidInitPrimal*)fid_fourier_l2_init_primal,
    (FidInitDual*)fourier_l2alt_init_dual,
    (FidBaseshift*)fid_fourier_l2alt_baseshift,
    (FidEstimate*)fid_fourier_l2alt_estimate,
    (FidGrad*)fid_fourier_l2_grad,
};


static TensorField *shadow_copy_sparse_f(size_t n_samples, const FourierSample *samples)
{
    TensorField *f;
    Sig sig;
    
    sig.dim=1;
    sig.order=0;
    sig.w=n_samples;
    sig.h=1;
    sig.d=1;
    sig.symmetric=FALSE;
    sig.is_complex=TRUE;
    sig.xres=1;
    sig.yres=1;
    sig.zres=1;

    f=alloc_tensorfield(&sig);
    
    FNumber *f_p=f->p;
    
    #pragma omp parallel for
    #pragma acc kernels loop, independent, deviceptr(samples, f_p)
    for(size_t i=0; i<n_samples; i++)
        set_complex(f_p, i, samples[i].val);
    
    return f;
}


void fid_fourier_l2alt_init(const TensorField *f,
                            //const TensorField *mask,
                            //FNumber mask_term_weight,
                            const Sig *sig,
                            size_t n_samples, 
                            const FourierSample *samples,
                            Fid **fid, 
                            FidFourierL2AltParams *fp)
{
    if(sig->order!=0)
        die("Invalid data order");
    
    if((f && samples) || (!f && !samples))
        die("Input data mismatch");
    
    
    if(f){
        fp->full_f=TRUE;
        fp->f=f;
        fp->f_orig=f;
        fp->f_tmp=NULL;
    }else{
        fp->full_f=FALSE;
        fp->f_tmp=shadow_copy_sparse_f(n_samples, samples);
        fp->f_orig=NULL;
        fp->f=fp->f_tmp;
    }
    fp->sig=*sig;
    fp->mask=NULL;
    fp->mask_term_weight=0;
    fp->n_samples=n_samples;
    fp->samples=samples;
    fp->samples_orig=samples;
    fp->samples_tmp=NULL;
    fp->tmp=alloc_tensorfield(sig);
    fp->pdgap0_ab=0;
    fp->pdgap0_c=0;
    
    // Calculate initial estimate of L^infty bound for u
    fp->fu_infty_bound2=tensorfield_frobinf(fp->f);
    fp->ubound=fn_sqrt(fp->fu_infty_bound2)*sig_tensorfield_size(sig);
    
    *fid=&fid_fourier_l2alt;

}
