/*
 * dticode/tensorreg/fid_frob_l1.c
 *
 * Written 2012 by Tuomo Valkonen <tuomov@iki.fi>, University of Cambridge.
 *
 */

#include <math.h>
#include <stdio.h>
#include <libtu/types.h>

#include "tensor.h"
#include "combine.h"
#include "fid_frob_l1.h"
#include "scaleplus.h"
#include "project_posdef.h"
#include "tensornorm.h"
#include "posdef.h"
#include "sumdot1.h"
#include "normalise.h"

//
// The approach:
//
// |A u-f|_F + delta_{u>=0}
//  = max_lambda  delta_{u>=0} + <lambda, A u-f> - delta_{|lambda|<=1}
//  = max_lambda  delta_{u>=0} + <lambda,A u>    - <lambda,f> - delta_{|lambda|<=1}
//  = max_lambda  G(u)         + <lambda,K u>  - F^*(lambda)
//
// F(y) = sup_x <y,x> - <x,f>-delta_{|x|<=1}
//      = |y-f|_1
//
// G(u) = delta_{u>=0}(u) 
//
// G^*(y) = sup_x <y,x>-delta_{x>=0}(x) = 
//      { INFINITY, y has pos eigenvalues
//      { 0,        y neg.semi.def
//     
// Because G^* can be infinite, we need an extra bound |u(x)|<M, 
// where M > sup_x |f(x)| to make the duality gap practical. 
//
// Therefore, we take actually:
//
// G(u) = delta_{u>=0 & |u|<=M}(u)
//
// G^*(y) = sup_x <y,x>-delta_{x>=0 & |x|<=M}(x)
//        = M<normalise_pointwise(P_{>=0}(y)), y>
//        = M|P_{>=0}(y)|_{L1-frob}
// G^*(-K^*y) = M*|P_{>=0}(-K^*y)|
//

static void frob_l1_k(FidTFL1Dual *a, const FidTFL1Primal *b, 
                      bool addto_a, FidTFL1Params *gtfp)
{
    // <lambda, K u>=<lambda, u>
    tensorfield_linop(a, gtfp->amat, b, (addto_a ? 1 : 0), 1);
}

static void frob_l1_kc(FidTFL1Primal *a, const FidTFL1Dual *b, 
                       bool addto_a, FidTFL1Params *gtfp)
{
    // <K^* lambda, u>=<lambda, u>
    //tensorfield_copy(a, b);
    // According to the rules for combine.c, fidelity function
    // implementation has to add Kb to a
    tensorfield_linop(a, gtfp->amat_conj, b, (addto_a ? 1 : 0), 1);
}

static void frob_l1_gres(FidTFL1Primal *u, FNumber tau, 
                         FidTFL1Params *gtfp)
{
    // min_u { |u-lambda|^2/2tau + delta_{u>=0}(u) + delta_{|u|<M}(u) }
    //  = P_{|u|<=M} P_{u>=0} u
    if(!gtfp->unconstr)
        tensorfield_project_posdef(u);
    tensorfield_project_frob2(u, gtfp->ubound);
}

static void frob_l1_fcres(FidTFL1Dual *lambda, FNumber tau,
                          FidTFL1Params *gtfp)
{
    // min_x{ |lambda-x|^2/2tau+<f,x>+delta_{||<1}(x) }
    // (x-lambda)/tau+f + N_{||<1}(x) \ni 0
    // x + N_{||<1}(x) \ni (lambda-f*tau) 
    // P_{||<=1} (lambda-f*tau)
    tensorfield_scaleplus(lambda, lambda, gtfp->f, 1, -tau, 1);
    tensorfield_project_frob2(lambda, 1);
}

static void frob_l1_pdgap(FNumber *pdgap,
                          const TensorField *u, 
                          const TensorField *y,
                          const TensorField *kcy,
                          const TensorField *ku,
                          TensorField *uwork,
                          unsigned int iter,
                          FidTFL1Params *gtfp)
{
    /* F(Ku) + G(u) +G^*(-K*y)+F^*(y) */
    
    tensorfield_scaleplus(uwork, kcy, kcy, -1, 0, 1);
    if(!gtfp->unconstr)
        tensorfield_project_posdef(uwork);
    
    pdgap[0]=tensorfield_sumdot1(ku, gtfp->f, TRUE, 1) // F(Ku)
             +0.0                                      // G(u)
             +gtfp->ubound*tensorfield_frob1(uwork)    // G^*(-K^* y)
             +tensorfield_iprod(y, gtfp->f);           // F^*(y)
    
    if(iter==0)
        gtfp->pdgap0=pdgap[0];
    pdgap[1]=gtfp->pdgap0;
}

static TensorField *frob_l1_alloc_primal(FidTFL1Params *gtfp)
{
    return alloc_tensorfield(&gtfp->amat->sigm);
}

static TensorField *frob_l1_alloc_dual(FidTFL1Params *gtfp)
{
    return alloc_tensorfield(&TENSORFIELD_FULLSIG(gtfp->f));
}

static void frob_l1_init_dual(FidTFL1Dual *phi, const FidTFL1Primal *u, FidTFL1Params *gtfp)
{
    tensorfield_scaleplus(phi, u, gtfp->f, 1, -1, 1);
    tensorfield_normalise_frob2(phi, 1);
}

static FNumber frob_l1_estimate(Estimate estim, FidTFL1Params *gtfp)
{
    if(estim==EstimateKNorm)
        return matrix_opnorm(gtfp->amat);
    else if(estim==EstimateGUniformConvexity)
        return 0;
    else if(estim==EstimateFCUniformConvexity)
        return INFINITY;
    else
        return 0;
}

static bool frob_l1_grad(TensorField *grad, const TensorField *u, 
                         FidTFL1Params *gtfp)
{
    tensorfield_scaleplus(grad, u, gtfp->f, 1, -1, 1);
    tensorfield_normalise_frob2(grad, 1);
    return FALSE; // Only a subgradient
}

static Fid fid_frob_l1={
    (FidK*)frob_l1_k,
    (FidKC*)frob_l1_kc,
    (FidFCres*)frob_l1_fcres,
    (FidGres*)frob_l1_gres,
    (FidPDgap*)frob_l1_pdgap,
    (FidAllocPrimal*)frob_l1_alloc_primal,
    (FidAllocDual*)frob_l1_alloc_dual,
    (FidInitPrimal*)NULL,
    (FidInitDual*)frob_l1_init_dual,
    (FidBaseshift*)NULL,
    (FidEstimate*)frob_l1_estimate,
    (FidGrad*)frob_l1_grad,
};

void fid_frob_l1_init(const TensorField *f,
                      const Matrix *amat,
                      int unconstr,
                      Fid **fid, 
                      FidTFL1Params *param)
{
    param->amat=amat;
    param->amat_conj=matrix_conjugate(amat);
    param->unconstr=unconstr;
    param->f=f;
    // See fid_frob_l2alt.c for derivation of this bound
    param->ubound=1+10*tensorfield_frobinf(f)/sqrt(matrix_normsquared_lowerbound(amat));
    param->pdgap0=0;
    
    *fid=&fid_frob_l1;
}
