/*
 * dticode/tensorreg/fid_frob_l2.c
 *
 * Written 2012 by Tuomo Valkonen <tuomov@iki.fi>, University of Graz.
 *
 */

#include <math.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <libtu/types.h>
#include <libtu/output.h>

#include "tensor.h"
#include "combine.h"
#include "fid_frob_l2.h"
#include "scaleplus.h"
#include "project_posdef.h"
#include "tensornorm.h"
#include "posdef.h"
#include "sumdot1.h"
#include "scaleproj.h"

#define IS_ID(X) X 
//#define IS_ID(X) FALSE

static void frob_l2_k(FidTFL2Dual *a, const FidTFL2Primal *b, 
                      bool addto_a, FidTFL2Params *gtfp)
{
    // Do nothing FidTFL2Dual is a dummy vector
}

static void frob_l2_kc(FidTFL2Primal *a, const FidTFL2Dual *b, 
                       bool addto_a, FidTFL2Params *gtfp)
{
    if(!addto_a)
        tensorfield_zero(a);
}

static void amat_projector(TensorField *res,
                           TensorField *c,
                           FNumber tau, FNumber idweight, FNumber cweight,
                           FidTFL2Params *gtfp)
{
    // M = idweight I+tau A^*A
    // constrained: MX-S=tau A^* f + cweight C, X,S>=0, X o S = 0
    // unconstrained: MX=tau A^* f + cweight C
    Matrix *mmat=alloc_matrix(&gtfp->amat->sigm, &gtfp->amat->sigm);
    Matrix *mmat_copy=alloc_matrix(&gtfp->amat->sigm, &gtfp->amat->sigm);
    size_t m=gtfp->amat->m;
    size_t n=gtfp->amat->n;
    size_t count=c->nelem;
    FNumber * DEVPTR mp=mmat->p;
    FNumber * DEVPTR ap=gtfp->amat->p;
    FNumber * DEVPTR acp=gtfp->amat_conj->p;
    FNumber * DEVPTR cp=c->p;
    FNumber * DEVPTR resp=res->p;
    
    assert(m==res->elemsize && m==c->elemsize);
    assert(res->nelem==c->nelem);
    
    // Construct M
    for(size_t i=0; i<m; i++){
        for(size_t j=0; j<m; j++){
            FNumber tmp=0;
            for(size_t k=0; k<n; k++)
                tmp=tmp+acp[i+k*m]*ap[k+j*n];
            mp[i+j*m]=(i==j ? idweight+tau*tmp : tau*tmp);
        }
    }
    // c := tau A^* f + cweight c
    tensorfield_linop(c, gtfp->amat_conj, gtfp->f, cweight, tau);

    memcpy(mmat_copy->p, mmat->p, m*m*sizeof(double));
    
    #define LUDIM m
    #define NRHS count
    #define A mp
    #define C cp
    #define RES resp
    #define GETM(a, i, j) (a[(j)+(i)*LUDIM])
    #define SETM(a, i, j, x) a[(j)+(i)*LUDIM]=(x)
    #define GETV(v, i, k) (v[(i)+(k)*LUDIM])
    #define SETV(v, i, k, x) v[i+(k)*LUDIM]=(x)
    #define LUCORE_RHS_PARALLEL 1
    #include "lusolve_core.h"
    #undef LUCORE_RHS_PARALLEL
    #undef LUDIM
    #undef NRHS
    #undef A
    #undef C
    #undef RES
    #undef GETM
    #undef SETM
    #undef GETV
    #undef SETV
    
    if(!gtfp->unconstr){
        // now M*res=c, but res is not necessarily positive definite.
        size_t dim=res->sig.dim;
        FNumber gamma=0.9;
        FNumber sigma=0.5;//1-gamma/sqrt(dim);
        FNumber rho=gtfp->iprho;//0.001;
        FNumber minm=idweight+tau*gtfp->a_squared_bound;
        int iters=100;
        
        tensorfield_scaleproj(res, mmat_copy, minm, 
                              sigma, gamma, iters, rho);
    }
    
    free_matrix(mmat);
    free_matrix(mmat_copy);
}                           
                        

static void frob_l2_gres(TensorField *u, FNumber tau, FidTFL2Params *gtfp)
{
    if(IS_ID(gtfp->amat->is_identity)){
        tensorfield_scaleplus(u, u, gtfp->f, 1, tau, 1+tau);
        if(!gtfp->unconstr)
            tensorfield_project_posdef(u);
    }else{
        // min_u { |c-u|^2/(2tau)+|f-Au|^2/2 + delta_{>=0}(u)}
        // (u-c)/tau + A^*A u - A^*f + N_{>=0)(u)=0.
        // (I+tau A^* A) u + N_{>=0}(u) = c + tau A^*f 
        amat_projector(u, u, tau, 1, 1, gtfp);
    }
}


static void frob_l2_pdgap(FNumber *pdgap,
                          const TensorField *u, 
                          const void *y, /* = NULL */
                          const TensorField *minus_divp,
                          const void *kx, /* = NULL */
                          TensorField *uwork,
                          unsigned int iter,
                          FidTFL2Params *gtfp)
{
    FNumber g, gcm;
    
    /* calculate G(u) */
    if(gtfp->unconstr || tensorfield_posdef(u, -FNUMBER_EPS)){
        if(IS_ID(gtfp->amat->is_identity)){
            g=tensorfield_frob2dist22(gtfp->f, u);
        }else{
            g=tensorfield_linop_frob2dist22(gtfp->f, gtfp->amat, u);
        }
    }else{
        g=INFINITY;
    }

    // calculate G^*(-minus_divp) 
    if(IS_ID(gtfp->amat->is_identity)){
        tensorfield_scaleplus(uwork, gtfp->f, minus_divp, 1, -1, 1);
        if(!gtfp->unconstr)
            tensorfield_project_posdef(uwork);
        gcm=(-tensorfield_sumdot1(minus_divp, uwork, FALSE, 1)
             -tensorfield_frob2dist22(gtfp->f, uwork));
    }else{
        // sup( <u,-minus_divp> - |f-Au|^2/2 - \delta_{u>=0} )
        //  ==> -minus_divp - A^*A u + A^* f - N_{>=0}(u) = 0
        //  ==> A^*A u + N_{>=0}(u) = -minus_divp + A^*f
        tensorfield_copy(uwork, minus_divp);
        amat_projector(uwork, uwork, 1, 0, -1, gtfp);
        gcm=(-tensorfield_sumdot1(minus_divp, uwork, FALSE, 1)
             -tensorfield_linop_frob2dist22(gtfp->f, gtfp->amat, uwork));
        //FNumber a=-tensorfield_sumdot1(minus_divp, uwork, FALSE, 1);
        //FNumber b=-tensorfield_linop_frob2dist22(gtfp->f, gtfp->amat, uwork);
        //fprintf(stderr, "%f %f %f [%f]\n", g, a, b,
        //    tensorfield_linop_frob2dist22(gtfp->f, gtfp->amat, minus_divp));
        //gcm=a+b;
    }
    
    pdgap[0]=g+gcm;
    
    if(iter==0)
        gtfp->pdgap0=pdgap[0];
    pdgap[1]=gtfp->pdgap0;
}

static TensorField *frob_l2_alloc_primal(FidTFL2Params *gtfp)
{
    return alloc_tensorfield(&gtfp->amat->sigm);
}

static Vector *frob_l2_alloc_dual(FidTFL2Params *gtfp)
{
    return alloc_dummyvector();
}

static FNumber frob_l2_estimate(Estimate estim, FidTFL2Params *gtfp)
{
    if(estim==EstimateKNorm)
        return 0;
    else if(estim==EstimateGUniformConvexity)
        return gtfp->a_squared_bound;
    else if(estim==EstimateFCUniformConvexity)
        return INFINITY;
    else
        return 0;
}

static bool frob_l2_grad(TensorField *grad, const TensorField *u, 
                         FidTFL2Params *gtfp)
{
    tensorfield_scaleplus(grad, u, gtfp->f, 1, -1, 1);
    return TRUE; // A proper gradient
}

static Fid fid_frob={
    (FidK*)frob_l2_k,
    (FidKC*)frob_l2_kc,
    (FidFCres*)noop,
    (FidGres*)frob_l2_gres,
    (FidPDgap*)frob_l2_pdgap,
    (FidAllocPrimal*)frob_l2_alloc_primal,
    (FidAllocDual*)frob_l2_alloc_dual,
    (FidInitPrimal*)NULL,
    (FidInitDual*)noop,
    (FidBaseshift*)NULL,
    (FidEstimate*)frob_l2_estimate,
    (FidGrad*)frob_l2_grad,
};

void fid_frob_l2_init(const TensorField *f,
                      const Matrix *amat,
                      int unconstr,
                      Fid **fid, 
                      FidTFL2Params *param)
{
    // |f-Au|^2/2
    // grad = A^*Au-A^*f
    // |f-Av|^2/2 - |f-Au|^2/2
    // = |Av|^2/2 - |Au|^2/2 - <A^* f, v-u>
    // = |Av|^2/2 - |Au|^2/2 - <A^*Au,v-u> + <grad, v-u>
    // = |Av|^2/2 + |Au|^2/2 - <A^*Au,v> + <grad, v-u>
    // = |A(v-u)|^2/2 + <grad, v-u>
    // >= gamma|v-u|^2/2 + <grad, v-u>
    // where gamma=min_i lambda_i( A^*A)
    
    param->a_squared_bound=matrix_normsquared_lowerbound(amat);
    param->unconstr=unconstr;
    param->f=f;
    param->pdgap0=0;
    param->amat=amat;
    param->amat_conj=matrix_conjugate(amat);
    
    *fid=&fid_frob;
}
