/*
 * dticode/tensorreg/fid_frob_l2alt.c
 *
 * Written 2012 by Tuomo Valkonen <tuomov@iki.fi>, University of Graz.
 *
 */

#include <math.h>
#include <stdio.h>
#include <libtu/types.h>

#include "tensor.h"
#include "combine.h"
#include "fid_frob_l2alt.h"
#include "scaleplus.h"
#include "project_posdef.h"
#include "tensornorm.h"
#include "posdef.h"
#include "sumdot1.h"
#include "normalise.h"

//
// The approach:
//
// |Au-f|^2_F + delta_{u>=0}
//  = max_lambda  delta_{u>=0} + <lambda, Au-f> - |lambda|^2/2
//  = max_lambda  delta_{u>=0} + <lambda,A u>    - <lambda,f> - |lambda|^2/2
//  = max_lambda  G(u)         + <lambda,K u>  - F^*(lambda)
//
// F(y) = sup_x <y,x> - <x,f>-|lambda|^2/2
//      = |y-f|^2/2
//
// G(u) = delta_{u>=0}(u) 
//
// G^*(y) = sup_x <y,x>-delta_{x>=0}(x) = 
//      { INFINITY, y has pos eigenvalues
//      { 0,        y neg.semi.def
//     
// Because G^* can be infinite, we need an extra bound |u(x)|<M, 
// where M > sup_x |f(x)| to make the duality gap practical. 
//
// Therefore, we take actually:
//
// G(u) = delta_{u>=0 & |u|<=M}(u)
//
// G^*(y) = sup_x <y,x>-delta_{x>=0 & |x|<=M}(x)
//        = M<normalise_pointwise(P_{>=0}(y)), y>
//        = M|P_{>=0}(y)|_{L2Alt-frob}
// G^*(-K^*y) = M*|P_{>=0}(-K^*y)|
//
//
// *****
//
// THE DIFFERENCE to fid_frob_l1 is only in F^* 
//   ==> frob_l2alt_fcres and frob_l2alt_pdgap
//

static void frob_l2alt_k(FidTFL2AltDual *a, const FidTFL2AltPrimal *b, 
                         bool addto_a, FidTFL2AltParams *gtfp)
{
    tensorfield_linop(a, gtfp->amat, b, (addto_a ? 1.0 : 0), 1.0);
}

static void frob_l2alt_kc(FidTFL2AltPrimal *a, const FidTFL2AltDual *b, 
                          bool addto_a, FidTFL2AltParams *gtfp)
{
    // According to the rules for combine.c, fidelity function
    // implementation has to add Kb to a
    tensorfield_linop(a, gtfp->amat_conj, b, (addto_a ? 1.0 : 0), 1.0);
}

static void frob_l2alt_gres(FidTFL2AltPrimal *u, FNumber tau, 
                            FidTFL2AltParams *gtfp)
{
    // min_u { |u-lambda|^2/2tau + delta_{u>=0}(u) + delta_{|u|<M}(u) }
    //  = P_{|u|<=M} P_{u>=0} u
    if(!gtfp->unconstr)
        tensorfield_project_posdef(u);
    tensorfield_project_frob2(u, gtfp->ubound);
}

static void frob_l2alt_fcres(FidTFL2AltDual *lambda, FNumber tau,
                             FidTFL2AltParams *gtfp)
{
    // min_x{ |lambda-x|^2/2tau+<f,x>+|x|^2/2 } 
    // => (x-lambda)/tau + f + x = 0
    //  => x =  (lambda-tau*f)/(1+tau)
    tensorfield_scaleplus(lambda, lambda, gtfp->f, 1, -tau, 1+tau);
}

static void frob_l2alt_pdgap(FNumber *pdgap,
                             const TensorField *u, 
                             const TensorField *y,
                             const TensorField *kcy,
                             const TensorField *ku,
                             TensorField *uwork,
                             unsigned int iter,
                             FidTFL2AltParams *gtfp)
{
    /* F(Ku) + G(u) +G^*(-K*y)+F^*(y) */
    
    tensorfield_scaleplus(uwork, kcy, kcy, -1, 0, 1);
    if(!gtfp->unconstr)
        tensorfield_project_posdef(uwork);
    
    pdgap[0]=tensorfield_frob2dist22(ku, gtfp->f)      // F(Ku)
             +0.0                                      // G(u)
             +gtfp->ubound*tensorfield_frob1(uwork)    // G^*(-K^* y)
             +(tensorfield_iprod(y, gtfp->f)           // F^*(y)
               +tensorfield_frob222(y));
               
    if(iter==0)
        gtfp->pdgap0=pdgap[0];
    pdgap[1]=gtfp->pdgap0;
}

static TensorField *frob_l2alt_alloc_primal(FidTFL2AltParams *gtfp)
{
    return alloc_tensorfield(&gtfp->amat->sigm);
}

static TensorField *frob_l2alt_alloc_dual(FidTFL2AltParams *gtfp)
{
    return alloc_tensorfield(&TENSORFIELD_FULLSIG(gtfp->f));
}

static void frob_l2alt_init_dual(FidTFL2AltDual *phi, const FidTFL2AltPrimal *u, 
                                 FidTFL2AltParams *gtfp)
{
    frob_l2alt_k(phi, u, FALSE, gtfp);
    tensorfield_scaleplus(phi, phi, gtfp->f, 1, -1, 1);
}

static FNumber frob_l2alt_estimate(Estimate estim, FidTFL2AltParams *gtfp)
{
    if(estim==EstimateKNorm)
        return matrix_opnorm(gtfp->amat);
    else if(estim==EstimateGUniformConvexity)
        return 0;
    else if(estim==EstimateFCUniformConvexity)
        return INFINITY;
    else
        return 0;
}

static bool frob_l2alt_grad(TensorField *grad, const TensorField *u, 
                            FidTFL2AltParams *gtfp)
{
    tensorfield_scaleplus(grad, u, gtfp->f, 1, -1, 1);
    return TRUE; // A proper gradient
}

static Fid fid_frob_l2alt={
    (FidK*)frob_l2alt_k,
    (FidKC*)frob_l2alt_kc,
    (FidFCres*)frob_l2alt_fcres,
    (FidGres*)frob_l2alt_gres,
    (FidPDgap*)frob_l2alt_pdgap,
    (FidAllocPrimal*)frob_l2alt_alloc_primal,
    (FidAllocDual*)frob_l2alt_alloc_dual,
    (FidInitPrimal*)NULL,
    (FidInitDual*)frob_l2alt_init_dual,
    (FidBaseshift*)NULL,
    (FidEstimate*)frob_l2alt_estimate,
    (FidGrad*)frob_l2alt_grad,
};


void fid_frob_l2alt_init(const TensorField *f,
                         const Matrix *amat,
                         int unconstr,
                         Fid **fid, 
                         FidTFL2AltParams *param)
{
    param->amat=amat;
    param->amat_conj=matrix_conjugate(amat);
    param->unconstr=unconstr;
    param->f=f;

    // |f-Au|^2 <= |f|^2
    // => |Au|^2 <= 2<f, A u> <= 2|f||Au|
    // => |Au| <= 2|f|
    // => |A||u| <= 2|f|

    // Downscale profitable at x if
    //
    // |f-kAu|^2 <= |f-Au|^2 (k<1) at x
    //
    // <= (k^2-1)|Au|^2 <= 2(k-1) <f, Au>
    // <= (1-k^2)|Au|^2 >= 2(1-k) <f, Au>
    // <= (1-k^2)|Au|^2 >= 2(1-k) |f||Au|
    // <= (1-k^2)|Au| >= 2(1-k)|f|
    // <= (1-k^2)c|u| >= 2(1-k)|f|
    // <= |u| >= 2(1-k)/(1-k^2) |f|/c
    // <= |u| >= 2(1-k)/(1-k^2) |f|_infty/|A| 
    // 
    // where c=sqrt(min_i lambda_i(A^*A))
    //
    // k--> 1: 
    //    |u| >= |f|_infty/c
    //
    // We multiply this bound a little bit, for numerical inaccuracies,
    // and +1 in case |f|_infty=0
    param->ubound=1+10*tensorfield_frobinf(f)/sqrt(matrix_normsquared_lowerbound(amat));
    param->pdgap0=0;
    
    *fid=&fid_frob_l2alt;
}
