/*
 * dticode/src/project_posdef.c
 *
 * Written 2012 by Tuomo Valkonen <tuomov@iki.fi>, University of Graz.
 *
 */

#include <string.h>
#include <stdio.h>
#include <stddef.h>
#include <math.h>
#include <libtu/output.h>

#include "sym2multi3.h"
#include "sym2multi2.h"
#include "sylvester.h"
#include "tensor.h"
#include "misc.h"
#include "varmatrix.h"
#include "acc_support.h"
#include "number.h"

/*#include "eig_qr.h"*/
#include "eig_qr.c"

#define RECONSTRUCT(a, off, l, ex, ey, ez) \
    if(l>0){ \
        a[off+I_XX]+=l*ex*ex; \
        a[off+I_XY]+=l*ex*ey; \
        a[off+I_YY]+=l*ey*ey; \
        a[off+I_XZ]+=l*ex*ez; \
        a[off+I_YZ]+=l*ey*ez; \
        a[off+I_ZZ]+=l*ez*ez; \
    }

static const char *project_worker3(const size_t end, 
                                   FNumber * const restrict DEVPTR ga)
{
    const size_t dimend=I_N*end;
    const char *error=NULL;

    #pragma omp parallel for
    #pragma acc kernels loop deviceptr(ga), independent
    for(size_t i=0; i<dimend; i+=I_N){
        if(!sylvester_3(ga+i, 0)){
            int converged=0;
#if 1
            /* - Manual inlining for PGCC to parallelise 
             * - Matrices split into variables for efficiency
             */
            DEFINE_INIT_VARMATRIX_SYM3(amat, (ga+i));
            DEFINE_VARMATRIX_3(evmat);
            #define A(I) VAR_VARMATRIX(amat, I)
            #define EV(I) VAR_VARMATRIX(evmat, I)
            #define calc_evs 1
            #include "eig_qr_core.h"
#else
            FNumber ev[INS_N];
            converged=eig_qr(ev, ga+i, 1);
            #define A(I) ga[i+I]
            #define EV(I) ev[I]
#endif

#if _OPENACC
#warning "Error-handling disabled with OpenACC"
#else
            if(!converged){
                #pragma omp critical
                error="eig error";
            }
#endif

            /* reconstruct with projected eigenvalues */
            const FNumber l1=A(I_XX), l2=A(I_YY), l3=A(I_ZZ);
            
            //veczero(ga+i, I_N);
            ga[i+I_XX]=0; ga[i+I_XY]=0; ga[i+I_XZ]=0; 
            ga[i+I_YY]=0; ga[i+I_YZ]=0; ga[i+I_ZZ]=0;
            
            RECONSTRUCT(ga, i, l1, EV(INS_XX), EV(INS_YX), EV(INS_ZX));
            RECONSTRUCT(ga, i, l2, EV(INS_XY), EV(INS_YY), EV(INS_ZY));
            RECONSTRUCT(ga, i, l3, EV(INS_XZ), EV(INS_YZ), EV(INS_ZZ));
            
            #undef EV
            #undef A
        }
    }
    return error;
}

static const char *project_worker2(const size_t end, 
                                   FNumber * const restrict DEVPTR ga)
{
    const size_t dimend=I2_N*end;
    const char *error=NULL;

    #pragma omp parallel for
    #pragma acc kernels loop deviceptr(ga), independent
    for(size_t i=0; i<dimend; i+=I2_N){
        if(!sylvester_2(ga+i, 0)){
    #if 1
            DEFINE_VARMATRIX_2(qmat);
            FNumber l1, l2;
            #define A(I) ga[i+I]
            #define Q(I) VAR_VARMATRIX(qmat,I)
            #define L1 l1
            #define L2 l2
            /* Manual inlining needed for PGCC to parallelise */
            DO_EIG2(L1, L2, Q, A, 1);
    #else
            FNumber q[INS2_N];
            FNumber l[2];
            #define A(I) ga[i+I]
            #define Q(I) q[I]
            #define L1 l[0]
            #define L2 l[1]
            eig2(q, l, a+i, 1);
    #endif

            if(L1<0 || L2<0){
                L1=(L1<0 ? 0 : L1);
                L2=(L2<0 ? 0 : L2);
                ga[i+I2_XX]=L1*Q(I2_XX)*Q(I2_XX)+L2*Q(I2_XY)*Q(I2_XY);
                ga[i+I2_XY]=L1*Q(I2_XX)*Q(I2_YX)+L2*Q(I2_XY)*Q(I2_YY);
                ga[i+I2_YY]=L1*Q(I2_YX)*Q(I2_YX)+L2*Q(I2_YY)*Q(I2_YY);
            }
            #undef A
            #undef Q
            #undef L1
            #undef L2
        }
    }
    return error;
}

static const char *project_worker1(size_t end,
                                   FNumber * const restrict DEVPTR ga)
{
    size_t i;
    
    #pragma omp parallel for
    #pragma acc kernels loop deviceptr(ga), independent
    for(i=0; i<end; i++){
        if(ga[i]<0)
            ga[i]=0;
    }
    return NULL;
}

static const char *do_project(size_t dim, size_t n, FNumber * DEVPTR a)
{
    if(dim==1)
        return project_worker1(n, a);
    else if(dim==I2_N)
        return project_worker2(n, a);
    else if(dim==I_N)
        return project_worker3(n, a);
    else
        return "Unsupported or erroneous tensor type";
}

#ifndef MEX
void tensorfield_project_posdef(TensorField *t)
{
    const char *err;
    
    if(!t->sig.symmetric)
        die("Pos.def projection not implemented for non-symmetric tensors");
    if(t->sig.is_complex)
        die("Pos.def projection not implemented for complex tensors");
    if(t->sig.order!=2)
        die("Pos.def projection requires order 2 tensor");
        
    err=do_project(t->elemsize, t->nelem, t->p);
    
    if(err)
        die(err);
}
#endif
