/*
 * dticode/src/reg_operl1.c
 *
 * Written 2014 by Tuomo Valkonen <tuomov@iki.fi>, EPN Quito / Modemat.
 *
 */

#include <stdio.h>
#include <math.h>
#include <assert.h>
#include <libtu/output.h>
#include "tensor.h"
#include "tensornorm.h"
#include "cpock_accel.h"
#include "diff.h"
#include "project_posdef.h"
#include "normalise.h"
#include "combine.h"
#include "reg_operl1.h"
//#include "stability.h"
#include "scaleplus.h"

#define WEIGHT_PART2(x) 1
#define WEIGHT_PART1(x) x

#define HUBERW(l1p) (l1p->huber*WEIGHT_PART2(l1p->alpha)/WEIGHT_PART1(l1p->alpha))

typedef TensorField OperL1Primal;
typedef TensorField OperL1Dual;
typedef void OperL1ExtraPrimal;

static void do_operl1_k(OperL1Dual *a, const OperL1Primal *b, bool addto_a, RegOperL1Params *l1p)
{
    linoper_calculate(l1p->oper, (Vector*)a, (const Vector *)b, addto_a);
}

static void operl1_k(OperL1Dual *a, const CombinedPrimal *b, bool addto_a, RegOperL1Params *l1p)
{
    do_operl1_k(a, (OperL1Primal*)COMBINED_U(b), addto_a, l1p);
}

static void operl1_kc(CombinedPrimal *a, const OperL1Dual *b, bool addto_au, bool addto_aw, RegOperL1Params *l1p)
{
    linoper_calculatec(l1p->oper, (Vector*)COMBINED_U(a), (const Vector*)b, addto_au);
}

static void operl1_fcres(OperL1Dual *a, FNumber tau, RegOperL1Params *l1p)
{
    // min |x-a|^2/tau + delta_C + (gamma*alpha_part2/(2*alpha_part1)) |x|^2
    // ==> (1+tau*gamma*alpha_part2/alpha_part1)x + N_C = a
    if(l1p->huber!=0)
        tensorfield_scale(a, 1.0/(1.0+tau*HUBERW(l1p)));
    tensorfield_project_frob2(a, WEIGHT_PART1(l1p->alpha));
}


static void operl1_pdgap(FNumber *pdgap, 
                         const CombinedPrimal *x, 
                         const OperL1Dual *phi, /* y */
                         const CombinedPrimal *kcy, 
                         const OperL1Dual *eu,  /* kx */
                         CombinedPrimal *xwork,
                         unsigned int iter,
                         RegOperL1Params *l1p)
{
    TensorField *u=(TensorField*)COMBINED_U(x);
    TensorField *minus_divp=(TensorField*)COMBINED_U(kcy);
    TensorField *uwork=(TensorField*)COMBINED_U(xwork);
    
    if(l1p->huber!=0)
        die("duality gap not implemented for huberised OperL1.");
        
    pdgap[0]=WEIGHT_PART1(l1p->alpha)*tensorfield_frob1(eu);
    if(iter==0)
        l1p->pdgap0=pdgap[0];
    pdgap[1]=l1p->pdgap0;
}


static OperL1Dual *operl1_alloc_dual(RegOperL1Params *l1p)
{
    TensorField* ret=(TensorField*)linoper_alloc_cod(l1p->oper);
    assert(OBJ_IS(ret, TensorField));
    return ret;
}

static void operl1_init_dual_extra(OperL1ExtraPrimal *w, OperL1Dual *psi, const OperL1Primal *u, 
                                   RegOperL1Params *l1p)
{
    do_operl1_k(psi, u, FALSE, l1p);
    tensorfield_normalise_frob2(psi, WEIGHT_PART1(l1p->alpha));
}

static FNumber operl1_estimate(Estimate estim, RegOperL1Params *l1p)
{
    if(estim==EstimateKNorm){
        return WEIGHT_PART2(l1p->alpha)*linoper_estimate_norm(l1p->oper);
    }else if(estim==EstimateGUniformConvexity){
        return INFINITY;
    }else if(estim==EstimateFCUniformConvexity){
        return 0;
    }else{
        return 0;
    }
}

static Reg reg_operl1={
    (RegK*)operl1_k,
    (RegKC*)operl1_kc,
    (RegFCres*)operl1_fcres,
    (RegGres*)noop,
    (RegPDgap*)operl1_pdgap,
    (RegAllocPrimal*)alloc_dummyvector,
    (RegAllocDual*)operl1_alloc_dual,
    (RegInitDualExtra*)operl1_init_dual_extra,
    (RegEstimate*)operl1_estimate,
};

void reg_operl1_init(LinOper *oper, FNumber alpha, FNumber huber,
                     Reg **reg, RegOperL1Params *l1p)
{
    l1p->alpha=alpha;
    l1p->huber=huber;
    l1p->oper=oper;
    l1p->pdgap0=0;
    
    *reg=&reg_operl1;
}
