/*
 * dticode/src/reg_operl2.c
 *
 * Written 2014 by Tuomo Valkonen <tuomov@iki.fi>, EPN Quito / Modemat.
 *
 */

#include <stdio.h>
#include <math.h>
#include <assert.h>
#include <libtu/output.h>
#include "tensor.h"
#include "tensornorm.h"
#include "cpock_accel.h"
#include "diff.h"
#include "project_posdef.h"
#include "normalise.h"
#include "combine.h"
#include "reg_operl2.h"
#include "linoper_grad.h"
#include "scaleplus.h"

typedef Vector OperL2Primal;
typedef Vector OperL2Dual;
typedef void OperL2ExtraPrimal;

static void do_operl2_k(OperL2Dual *a, const OperL2Primal *b, bool addto_a, RegOperL2Params *l2p)
{
    linoper_calculate(l2p->oper, (Vector*)a, (const Vector *)b, addto_a);
}

static void operl2_k(OperL2Dual *a, const CombinedPrimal *b, bool addto_a, RegOperL2Params *l2p)
{
    do_operl2_k(a, (OperL2Primal*)COMBINED_U(b), addto_a, l2p);
}

static void operl2_kc(CombinedPrimal *a, const OperL2Dual *b, bool addto_au, bool addto_aw, RegOperL2Params *l2p)
{
    linoper_calculatec(l2p->oper, (Vector*)COMBINED_U(a), (const Vector*)b, addto_au);
}

static void operl2_fcres(OperL2Dual *a, FNumber tau, RegOperL2Params *l2p)
{
    // min |x-a|^2/2tau + |x|^2/2alpha
    // ==> (alpha+tau)x  = alpha*a
    vector_scale(a, l2p->alpha/(l2p->alpha+tau));
}


static void operl2_pdgap(FNumber *pdgap, 
                         const CombinedPrimal *x, 
                         const OperL2Dual *phi, /* y */
                         const CombinedPrimal *kcy, 
                         const OperL2Dual *eu,  /* kx */
                         CombinedPrimal *xwork,
                         unsigned int iter,
                         RegOperL2Params *l2p)
{
    Vector *u=(Vector*)COMBINED_U(x);
    Vector *minus_divp=(Vector*)COMBINED_U(kcy);
    Vector *uwork=(Vector*)COMBINED_U(xwork);
    
    pdgap[0]=l2p->alpha*vector_norm22(eu)+vector_norm22(phi)/l2p->alpha;
    if(iter==0)
        l2p->pdgap0=pdgap[0];
    pdgap[1]=l2p->pdgap0;
}


static OperL2Dual *operl2_alloc_dual(RegOperL2Params *l2p)
{
    Vector* ret=(Vector*)linoper_alloc_cod(l2p->oper);
    assert(OBJ_IS(ret, Vector));
    return ret;
}

static void operl2_init_dual_extra(OperL2ExtraPrimal *w, OperL2Dual *psi, const OperL2Primal *u, 
                                   RegOperL2Params *l2p)
{
    do_operl2_k(psi, u, FALSE, l2p);
}

static FNumber operl2_estimate(Estimate estim, RegOperL2Params *l2p)
{
    if(estim==EstimateKNorm){
        return linoper_estimate_norm(l2p->oper);
    }else if(estim==EstimateGUniformConvexity){
        return INFINITY;
    }else if(estim==EstimateFCUniformConvexity){
        return 0;
    }else{
        return 0;
    }
}

static Reg reg_operl2={
    (RegK*)operl2_k,
    (RegKC*)operl2_kc,
    (RegFCres*)operl2_fcres,
    (RegGres*)noop,
    (RegPDgap*)operl2_pdgap,
    (RegAllocPrimal*)alloc_dummyvector,
    (RegAllocDual*)operl2_alloc_dual,
    (RegInitDualExtra*)operl2_init_dual_extra,
    (RegEstimate*)operl2_estimate,
};

void reg_operl2_init(LinOper *oper, FNumber alpha, 
                     Reg **reg, RegOperL2Params *l2p)
{
    l2p->alpha=alpha;
    l2p->oper=oper;
    l2p->pdgap0=0;
    
    *reg=&reg_operl2;
}

void reg_gradl2_init(const Sig *sig, bool symmetric, 
                     FNumber alpha, const char *diffalg,
                     Reg **reg, RegOperL2Params *l2p)
{
    LinOper *op=(LinOper*)alloc_linoper_grad(sig, symmetric, diffalg);

    assert(op!=NULL);
    
    reg_operl2_init(op, alpha, reg, l2p);
}
                     
