/*
 * dticode/src/reg_split.c
 *
 * Written 2013 by Tuomo Valkonen <tuomov@iki.fi>, University of Cambridge.
 *
 */

#include <stdio.h>
#include <math.h>
#include <assert.h>
#include <libtu/misc.h>
#include <libtu/output.h>
#include "combine.h"
#include "reg_split.h"
#include "splitvar.h"

////////

#include "splitp.h"

////////

static void reg_split_k(SplitDual *a, const CombinedPrimal *b, bool addto_a, RegSplitParams *spp)
{
    #pragma omp parallel for
    for(size_t i=0; i<spp->n; i++){
        RegDual *a_i;
        CombinedPrimal *b_i;
        
        get_dual_i(&a_i, a, i);
        get_combined_primal_i_const(&b_i, b, i);
        
        spp->reg[i]->k(a_i, b_i, addto_a, spp->params[i]);
        
        release_combined_primal(b_i);
    }
}

static void reg_split_kc(CombinedPrimal *a, const SplitDual *b, bool addto_au, bool addto_aw, RegSplitParams *spp)
{
    #pragma omp parallel for
    for(size_t i=0; i<spp->n; i++){
        CombinedPrimal *a_i;
        const RegDual *b_i;
        
        get_combined_primal_i(&a_i, a, i);
        get_dual_i_const(&b_i, b, i);
        
        spp->reg[i]->kc(a_i, b_i, addto_au, addto_aw, spp->params[i]);
        
        release_combined_primal(a_i);

        addto_au=TRUE;
    }
}

////////

static void reg_split_fcres(SplitDual *a, FNumber tau, RegSplitParams *spp)
{
    #pragma omp parallel for
    for(size_t i=0; i<spp->n; i++){
        RegDual *a_i;
        get_dual_i(&a_i, a, i);
        
        spp->reg[i]->fcres(a_i, tau, spp->params[i]);
    }
}

static void reg_split_gres(SplitExtraPrimal *a, FNumber tau, RegSplitParams *spp)
{
    #pragma omp parallel for
    for(size_t i=0; i<spp->n; i++){
        RegExtraPrimal *a_i;
        get_extraprimal_i(&a_i, a, i);
        
        spp->reg[i]->gres(a_i, tau, spp->params[i]);
    }
}

////////

static void reg_split_pdgap(FNumber *pdgap, 
                            const CombinedPrimal *x, 
                            const SplitDual *y,
                            const CombinedPrimal *kcy, 
                            const SplitDual *kx,
                            CombinedPrimal *xwork,
                            unsigned int iter,
                            RegSplitParams *spp)
{
    pdgap[0]=0;
    pdgap[1]=0;
    
    #pragma omp parallel for
    for(size_t i=0; i<spp->n; i++){
        CombinedPrimal *x_i;
        const RegDual *y_i;
        CombinedPrimal *kcy_i;
        const RegDual *kx_i;
        CombinedPrimal *xwork_i;
        FNumber pdgap_i[2];
        
        get_combined_primal_i_const(&x_i, x, i);
        get_combined_primal_i_const(&kcy_i, kcy, i);
        get_combined_primal_i(&xwork_i, xwork, i);
        get_dual_i_const(&y_i, y, i);
        get_dual_i_const(&kx_i, kx, i);
        
        spp->reg[i]->pdgap(pdgap_i, x_i, y_i, kcy_i, kx_i, xwork_i, iter, spp->params[i]);
        
        release_combined_primal(x_i);
        release_combined_primal(kcy_i);
        release_combined_primal(xwork_i);
        
        pdgap[0]+=pdgap_i[0];
        pdgap[1]+=pdgap_i[1];
    }
}

////////


static SplitExtraPrimal *reg_split_alloc_primal(RegSplitParams *spp)
{
    SplitDual *d=alloc_splitvar(spp->n);
    
    for(size_t i=0; i<spp->n; i++)
        d->p[i]=spp->reg[i]->alloc_primal(spp->params[i]);
        
    return d;
}

static SplitDual *reg_split_alloc_dual(RegSplitParams *spp)
{
    SplitDual *d=alloc_splitvar(spp->n);
    
    for(size_t i=0; i<spp->n; i++)
        d->p[i]=spp->reg[i]->alloc_dual(spp->params[i]);
        
    return d;
}

static void reg_split_init_dual_extra(SplitExtraPrimal *w, SplitDual *psi, const SplitPrimal *u, 
                                      RegSplitParams *spp)
{
    for(size_t i=0; i<spp->n; i++){
        const FidPrimal *u_i;
        RegExtraPrimal *w_i;
        RegDual *psi_i;
        
        get_primal_i_const(&u_i, u, i);
        get_extraprimal_i(&w_i, w, i);
        get_dual_i(&psi_i, psi, i);
        
        spp->reg[i]->init_dual_extra(w_i, psi_i, u_i, spp->params[i]);
    }
    
}

static FNumber reg_split_estimate(Estimate estim, RegSplitParams *spp)
{
    if(estim==EstimateKNorm){
        FNumber knorm=0;
        #pragma omp parallel for
        for(size_t i=0; i<spp->n; i++)
            knorm+=spp->reg[i]->estimate(EstimateKNorm, spp->params[i]);
        return knorm;
    }else if(estim==EstimateGUniformConvexity || estim==EstimateFCUniformConvexity){
        FNumber uc=0;
        #pragma omp parallel for
        for(size_t i=0; i<spp->n; i++)
            uc=fn_min(uc, spp->reg[i]->estimate(estim, spp->params[i]));
        return uc;
    }else{
        return 0;
    }
}

////////

static Reg reg_split={
    (RegK*)reg_split_k,
    (RegKC*)reg_split_kc,
    (RegFCres*)reg_split_fcres,
    (RegGres*)reg_split_gres,
    (RegPDgap*)reg_split_pdgap,
    (RegAllocPrimal*)reg_split_alloc_primal,
    (RegAllocDual*)reg_split_alloc_dual,
    (RegInitDualExtra*)reg_split_init_dual_extra,
    (RegEstimate*)reg_split_estimate
};

////////

void reg_split_init(Reg **reg, RegSplitParams *spp)
{
    *reg=&reg_split;
}
