/*
 * dticode/src/reg_splitsame.c
 *
 * Written 2014 by Tuomo Valkonen <tuomov@iki.fi>, EPN Quito / Modemat.
 *
 */

#include <stdio.h>
#include <math.h>
#include <assert.h>
#include <libtu/misc.h>
#include <libtu/output.h>
#include "combine.h"
#include "reg_splitsame.h"
#include "splitvar.h"

////////

#include "splitp.h"

////////

static void reg_splitsame_k(SplitSameDual *a, const CombinedPrimal *b, bool addto_a, RegSplitSameParams *spp)
{
    #pragma omp parallel for
    for(size_t i=0; i<spp->n; i++){
        RegDual *a_i;
        CombinedPrimal *b_i;
        
        get_dual_i(&a_i, a, i);
        get_combined_primal_i_xonly_const(&b_i, b, i);
        
        spp->reg[i]->k(a_i, b_i, addto_a, spp->params[i]);
        
        release_combined_primal_xonly(b_i);
    }
}

static void reg_splitsame_kc(CombinedPrimal *a, const SplitSameDual *b, bool addto_au, bool addto_aw, RegSplitSameParams *spp)
{
    #pragma omp parallel for
    for(size_t i=0; i<spp->n; i++){
        CombinedPrimal *a_i;
        const RegDual *b_i;
        
        get_combined_primal_i_xonly(&a_i, a, i);
        get_dual_i_const(&b_i, b, i);
        
        spp->reg[i]->kc(a_i, b_i, addto_au, addto_aw, spp->params[i]);
        addto_au=TRUE;
        
        release_combined_primal_xonly(a_i);

        addto_au=TRUE;
    }
}

////////

static void reg_splitsame_fcres(SplitSameDual *a, FNumber tau, RegSplitSameParams *spp)
{
    #pragma omp parallel for
    for(size_t i=0; i<spp->n; i++){
        RegDual *a_i;
        get_dual_i(&a_i, a, i);
        
        spp->reg[i]->fcres(a_i, tau, spp->params[i]);
    }
}

static void reg_splitsame_gres(SplitSameExtraPrimal *a, FNumber tau, RegSplitSameParams *spp)
{
    #pragma omp parallel for
    for(size_t i=0; i<spp->n; i++){
        RegExtraPrimal *a_i;
        get_extraprimal_i(&a_i, a, i);
        
        spp->reg[i]->gres(a_i, tau, spp->params[i]);
    }
}

////////

static void reg_splitsame_pdgap(FNumber *pdgap, 
                            const CombinedPrimal *x, 
                            const SplitSameDual *y,
                            const CombinedPrimal *kcy, 
                            const SplitSameDual *kx,
                            CombinedPrimal *xwork,
                            unsigned int iter,
                            RegSplitSameParams *spp)
{
    pdgap[0]=0;
    pdgap[1]=0;
    
    #pragma omp parallel for
    for(size_t i=0; i<spp->n; i++){
        CombinedPrimal *x_i;
        const RegDual *y_i;
        CombinedPrimal *kcy_i;
        const RegDual *kx_i;
        CombinedPrimal *xwork_i;
        FNumber pdgap_i[2];
        
        get_combined_primal_i_xonly_const(&x_i, x, i);
        get_combined_primal_i_xonly_const(&kcy_i, kcy, i);
        get_combined_primal_i_xonly(&xwork_i, xwork, i);
        get_dual_i_const(&y_i, y, i);
        get_dual_i_const(&kx_i, kx, i);
        
        spp->reg[i]->pdgap(pdgap_i, x_i, y_i, kcy_i, kx_i, xwork_i, iter, spp->params[i]);
        
        release_combined_primal_xonly(x_i);
        release_combined_primal_xonly(kcy_i);
        release_combined_primal_xonly(xwork_i);
        
        pdgap[0]+=pdgap_i[0];
        pdgap[1]+=pdgap_i[1];
    }
}

////////


static SplitSameExtraPrimal *reg_splitsame_alloc_primal(RegSplitSameParams *spp)
{
    SplitSameDual *d=alloc_splitvar(spp->n);
    
    for(size_t i=0; i<spp->n; i++)
        d->p[i]=spp->reg[i]->alloc_primal(spp->params[i]);
        
    return d;

}

static SplitSameDual *reg_splitsame_alloc_dual(RegSplitSameParams *spp)
{
    SplitSameDual *d=alloc_splitvar(spp->n);
    
    for(size_t i=0; i<spp->n; i++)
        d->p[i]=spp->reg[i]->alloc_dual(spp->params[i]);
        
    return d;
}

static void reg_splitsame_init_dual_extra(SplitSameExtraPrimal *w, SplitSameDual *psi, const SplitSamePrimal *u, 
                                          RegSplitSameParams *spp)
{
    for(size_t i=0; i<spp->n; i++){
        RegExtraPrimal *w_i;
        RegDual *psi_i;
        
        get_extraprimal_i(&w_i, w, i);
        get_dual_i(&psi_i, psi, i);
        
        spp->reg[i]->init_dual_extra(w_i, psi_i, u, spp->params[i]);
    }
    
}

static FNumber reg_splitsame_estimate(Estimate estim, RegSplitSameParams *spp)
{
    if(estim==EstimateKNorm){
        FNumber knorm2=0;
        #pragma omp parallel for
        for(size_t i=0; i<spp->n; i++)
            knorm2+=fn_pow2(spp->reg[i]->estimate(EstimateKNorm, spp->params[i]));
        return fn_sqrt(knorm2);
    }else if(estim==EstimateGUniformConvexity || estim==EstimateFCUniformConvexity){
        FNumber uc=0;
        #pragma omp parallel for
        for(size_t i=0; i<spp->n; i++)
            uc=fn_min(uc, spp->reg[i]->estimate(estim, spp->params[i]));
        return uc;
    }else{
        return 0;
    }
}

////////

static Reg reg_splitsame={
    (RegK*)reg_splitsame_k,
    (RegKC*)reg_splitsame_kc,
    (RegFCres*)reg_splitsame_fcres,
    (RegGres*)reg_splitsame_gres,
    (RegPDgap*)reg_splitsame_pdgap,
    (RegAllocPrimal*)reg_splitsame_alloc_primal,
    (RegAllocDual*)reg_splitsame_alloc_dual,
    (RegInitDualExtra*)reg_splitsame_init_dual_extra,
    (RegEstimate*)reg_splitsame_estimate
};

////////

void reg_splitsame_init(Reg **reg, RegSplitSameParams *spp)
{
    *reg=&reg_splitsame;
}
