/*
 * dticode/src/reg_tdiv.c
 *
 * Written 2012 by Tuomo Valkonen <tuomov@iki.fi>, University of Graz.
 *
 */

#include <stdio.h>
#include <math.h>
#include <assert.h>
#include <libtu/output.h>
#include "tensor.h"
#include "tensornorm.h"
#include "cpock_accel.h"
#include "diff.h"
#include "project_posdef.h"
#include "normalise.h"
#include "combine.h"
#include "reg_tdiv.h"
#include "stability.h"
#include "scaleplus.h"

#define HUBERW(tvp) (tvp->huber*WEIGHT_PART2(tvp->alpha)/WEIGHT_PART1(tvp->alpha))
#define TDIV_BDRY 1 // Neumann

typedef TensorField TDivPrimal;
typedef TensorField TDivDual;
typedef void TDivExtraPrimal;

static void do_tdiv_k(TDivDual *a, const TDivPrimal *b, bool addto_a, RegTDivParams *tvp)
{
    if(tvp->td)
        tensorfield_negsymdiv(a, b, (addto_a ? 1.0 : 0.0), WEIGHT_PART2(tvp->alpha), TDIV_BDRY, tvp->diffalg);
    else
        tensorfield_negdiv(a, b, (addto_a ? 1.0 : 0.0), WEIGHT_PART2(tvp->alpha), TDIV_BDRY, tvp->diffalg);
}

static void tdiv_k(TDivDual *a, const CombinedPrimal *b, bool addto_a, RegTDivParams *tvp)
{
    do_tdiv_k(a, (TDivPrimal*)COMBINED_U(b), addto_a, tvp);
}

static void tdiv_kc(CombinedPrimal *a, const TDivDual *b, bool addto_au, bool addto_aw, RegTDivParams *tvp)
{
    // We may symmetrise here for TDiv, because the primal 
    // variable is (assumed) symmetric
    if(tvp->td)
        tensorfield_symgrad((TensorField*)COMBINED_U(a), b, (addto_au ? 1.0 : 0.0), WEIGHT_PART2(tvp->alpha), TDIV_BDRY, tvp->diffalg);
    else
        tensorfield_grad((TensorField*)COMBINED_U(a), b, (addto_au ? 1.0 : 0.0), WEIGHT_PART2(tvp->alpha), TDIV_BDRY, tvp->diffalg);
}

static void tdiv_fcres(TDivDual *a, FNumber tau, RegTDivParams *tvp)
{
    // min |x-a|^2/tau + delta_C + (gamma*alpha_part2/(2*alpha_part1)) |x|^2
    // ==> (1+tau*gamma*alpha_part2/alpha_part1)x + N_C = a
    if(tvp->huber!=0)
        tensorfield_scale(a, 1.0/(1.0+tau*HUBERW(tvp)));
    tensorfield_project_frob2(a, WEIGHT_PART1(tvp->alpha));
}


static void tdiv_pdgap(FNumber *pdgap, 
                     const CombinedPrimal *x, 
                     const TDivDual *phi, /* y */
                     const CombinedPrimal *kcy, 
                     const TDivDual *eu,  /* kx */
                     CombinedPrimal *xwork,
                     unsigned int iter,
                     RegTDivParams *tvp)
{
    TensorField *u=(TensorField*)COMBINED_U(x);
    TensorField *minus_divp=(TensorField*)COMBINED_U(kcy);
    TensorField *uwork=(TensorField*)COMBINED_U(xwork);

    if(tvp->huber!=0)
        die("duality gap not implemented for huberised TV.");
    
    pdgap[0]=WEIGHT_PART1(tvp->alpha)*tensorfield_frob1(eu);
    if(iter==0)
        tvp->pdgap0=pdgap[0];
    pdgap[1]=tvp->pdgap0;
}


static TDivDual *tdiv_alloc_dual(RegTDivParams *tvp)
{
    Sig dsig=tvp->sig; 
    dsig.order--; 
    dsig.symmetric=tvp->td;
    
    return alloc_tensorfield(&dsig);
}

static void tdiv_init_dual_extra(TDivExtraPrimal *w, TDivDual *psi, const TDivPrimal *u, 
                               RegTDivParams *tvp)
{
    do_tdiv_k(psi, u, FALSE, tvp);
    tensorfield_normalise_frob2(psi, WEIGHT_PART1(tvp->alpha));
}

static FNumber tdiv_estimate(Estimate estim, RegTDivParams *tvp)
{
    if(estim==EstimateKNorm){
        FNumber l2=calc_l2fact(&(tvp->sig), tvp->diffalg);
        return WEIGHT_PART2(tvp->alpha)*fn_sqrt(l2);
    }else if(estim==EstimateGUniformConvexity){
        return INFINITY;
    }else if(estim==EstimateFCUniformConvexity){
        return 0;
    }else{
        return 0;
    }
}

static Reg reg_tdiv={
    (RegK*)tdiv_k,
    (RegKC*)tdiv_kc,
    (RegFCres*)tdiv_fcres,
    (RegGres*)noop,
    (RegPDgap*)tdiv_pdgap,
    (RegAllocPrimal*)alloc_dummyvector,
    (RegAllocDual*)tdiv_alloc_dual,
    (RegInitDualExtra*)tdiv_init_dual_extra,
    (RegEstimate*)tdiv_estimate,
};

void reg_tdiv_init(const Sig *sig, bool td, FNumber alpha, FNumber huber,
                   const char *diffalg, Reg **reg, RegTDivParams *tvp)
{
    assert(sig->order>=1);
    
    tvp->alpha=alpha;
    tvp->huber=huber;
    tvp->diffalg=(diffalg[0] ? diffalg[0] : 'f');
    tvp->td=td;
    tvp->sig=*sig;
    tvp->pdgap0=0;
    
    *reg=&reg_tdiv;
}
