/*
 * dticode/src/reg_tgv2.c
 *
 * Written 2012 by Tuomo Valkonen <tuomov@iki.fi>, University of Graz.
 *
 */


#include <stdio.h>
#include <math.h>
#include <libtu/misc.h>
#include <libtu/output.h>
#include "tensor.h"
#include "tensornorm.h"
#include "cpock_accel.h"
#include "diff.h"
#include "project_posdef.h"
#include "normalise.h"
#include "scaleplus.h"
#include "combine.h"
#include "number.h"
#include "reg_tgv2.h"
#include "splitvar.h"

//#define TGV2_FACTORS_IN_K 1

typedef TensorField TGV2ExtraPrimal;

typedef SplitVar TGV2Dual;

#define DUAL_PHI(spv) *((TensorField**)&((spv)->p[0]))
#define DUAL_PSI(spv) *((TensorField**)&((spv)->p[1]))

static void do_tgv2_k(TGV2Dual *a, const TensorField *u, const TensorField *w,
                      bool addto_a, RegTGV2Params *tgv2p)
{
    TensorField *phi=DUAL_PHI(a);
    TensorField *psi=DUAL_PSI(a);
    
    tensorfield_symgrad(phi, u, (addto_a ? 1.0 : 0.0), 1.0, 0, tgv2p->diffalg1);
    tensorfield_symgrad(psi, w, (addto_a ? 1.0 : 0.0), 1.0, 0, tgv2p->diffalg2);
#if TGV2_FACTORS_IN_K
    // phi=alpha(eu-w)
    tensorfield_scaleplus(phi, phi, w, tgv2p->alpha, -tgv2p->alpha, 1);
    // psi=beta(ew)
    tensorfield_scaleplus(psi, psi, psi, tgv2p->beta, 0, 1);
#else
    // phi=eu-w 
    tensorfield_scaleplus(phi, phi, w, 1, -1, 1);
#endif
}

static void tgv2_k(TGV2Dual *a, const CombinedPrimal *b, bool addto_a, RegTGV2Params *tgv2p)
{
    TensorField *u=(TensorField*)COMBINED_U(b);
    TensorField *w=(TensorField*)COMBINED_W(b);
    
    do_tgv2_k(a, u, w, addto_a, tgv2p);
}

static void tgv2_kc(CombinedPrimal *a, const TGV2Dual *b, 
                    bool addto_au, bool addto_aw, 
                    RegTGV2Params *tgv2p)
{
    TensorField *u=(TensorField*)COMBINED_U(a);
    TensorField *w=(TensorField*)COMBINED_W(a);
    TensorField *phi=DUAL_PHI(b);
    TensorField *psi=DUAL_PSI(b);
        
    tensorfield_negsymdiv(u, phi, (addto_au ? 1.0 : 0.0), 1.0, 0, tgv2p->diffalg1);
    tensorfield_negsymdiv(w, psi, (addto_aw ? 1.0 : 0.0), 1.0, 0, tgv2p->diffalg2);
#if TGV2_FACTORS_IN_K
    // w=beta(-div psi)-alpha(phi)
    tensorfield_scaleplus(w, w, phi, tgv2p->beta, -tgv2p->alpha, 1);
    // u=alpha(-div phi)
    tensorfield_scaleplus(u, u, u, tgv2p->alpha, 0, 1);
#else
    // w=(-div psi)-phi
    tensorfield_scaleplus(w, w, phi, 1, -1, 1);
#endif
}

static void tgv2_fcres(TGV2Dual *a, FNumber tau, RegTGV2Params *tgv2p)
{
    TensorField *phi=DUAL_PHI(a);
    TensorField *psi=DUAL_PSI(a);

#if TGV2_FACTORS_IN_K
    tensorfield_project_frob2(phi, 1);
    tensorfield_project_frob2(psi, 1);
#else
    tensorfield_project_frob2(phi, tgv2p->alpha);
    tensorfield_project_frob2(psi, tgv2p->beta);
#endif
}


static void tgv2_pdgap(FNumber *pdgap, 
                       const CombinedPrimal *uw,
                       const TGV2Dual *phipsi,
                       const CombinedPrimal *kc,
                       const TGV2Dual *k, 
                       CombinedPrimal *uwork,
                       unsigned int iter,
                       RegTGV2Params *tgv2p)
{
    // k->phi = eu-w; k->psi = ew
    // kc->phi = - div phi; kc->psi = -(phi + div psi)
    TensorField *u=(TensorField*)COMBINED_U(uw); 
    TensorField *w=(TensorField*)COMBINED_W(uw);
    TensorField *euMw=DUAL_PHI(k);
    TensorField *ew=DUAL_PSI(k);
    TensorField *minus_divphi=(TensorField*)COMBINED_U(kc);
    TensorField *minus_phiPdivq=(TensorField*)COMBINED_W(kc);
    
    FNumber m2=tgv2p->m2;
    FNumber wnorm=tensorfield_frob1(w);
    
    if(wnorm>m2){
        m2=wnorm;
        tgv2p->m2=wnorm;
    }
    
    FNumber a=tensorfield_frob1(euMw);
    FNumber b=tensorfield_frob1(ew);
    FNumber c=tensorfield_frobinf(minus_phiPdivq);
#if TGV2_FACTORS_IN_K
    FNumber ab = a + b;
#else
    FNumber ab = tgv2p->alpha*a + tgv2p->beta*b;
#endif    
    if(iter==0){
        tgv2p->pdgap0_ab=ab;
        tgv2p->pdgap0_c=c;
    }
    pdgap[0]=ab+m2*c;
    pdgap[1]=tgv2p->pdgap0_ab+m2*tgv2p->pdgap0_c;
}

static TGV2ExtraPrimal *tgv2_alloc_primal(RegTGV2Params *tgv2p)
{
    Sig dsig=tgv2p->sig; dsig.order++; dsig.symmetric=1;
    
    return alloc_tensorfield(&dsig);
}

static TGV2Dual *tgv2_alloc_dual(RegTGV2Params *tgv2p)
{
    Sig dsig=tgv2p->sig; dsig.order++; dsig.symmetric=1;
    Sig d2sig=tgv2p->sig; d2sig.order+=2; d2sig.symmetric=1;
    
    TGV2Dual *res=alloc_splitvar(2);
    
    if(res==NULL)
        die_err();
    
    res->p[0]=(Vector*)alloc_tensorfield(&dsig);
    res->p[1]=(Vector*)alloc_tensorfield(&d2sig);

    return res;
}

static void tgv2_init_dual_extra(TGV2ExtraPrimal *w, TGV2Dual *dual, const TensorField *u, 
                                 RegTGV2Params *tgv2p)
{
    // Just pick something...
    tensorfield_zero(w);
    
    do_tgv2_k(dual, u, w, FALSE, tgv2p);
    
#if TGV2_FACTORS_IN_K
    tensorfield_normalise_frob2(DUAL_PHI(dual), 1);
    tensorfield_normalise_frob2(DUAL_PSI(dual), 1);
#else
    tensorfield_normalise_frob2(DUAL_PHI(dual), tgv2p->alpha);
    tensorfield_normalise_frob2(DUAL_PSI(dual), tgv2p->beta);
#endif
}


static FNumber tgv2_estimate(Estimate estim, RegTGV2Params *tgv2p)
{
    if(estim==EstimateKNorm){
        Sig dsig=tgv2p->sig; dsig.order++;
        // |Eu - w|^2 <= (1+theta)|Eu|^2 + (1+1/theta)|w|^2
        //            <= (1+theta)e2|u|^2 + (1+1/theta)|w|^2
        // |Ew|^2 <= e3|w|^2
        // Hence |Eu-w|^2 + |Ew|^2 <= (1+theta)e2|u|^2 + (1+1/theta+e3)|w|^2
        //                         <=  l2 (|u|^2+|w|^2)
        // for suitable l2.
        // Solving 1+theta=1+1/theta+e3 we get
        // theta = (-d+sqrt(d^2+4*e2))/2 with  d=e2-e3-1 (=-1!)
        // and so l2=(1+theta)e2
        FNumber e2=calc_l2fact(&(tgv2p->sig), tgv2p->diffalg1);
        FNumber e3=calc_l2fact(&dsig, tgv2p->diffalg2);
    #if TGV2_FACTORS_IN_K
        e3=e3*(beta*beta)/(alpha*alpha);
    #endif
        FNumber d=(e2-e3-1);
        FNumber l2=e2+(-d+sqrt(d*d+4*e2))/2;
    #if TGV2_FACTORS_IN_K
        l2=l2*(alpha*alpha);
    #endif
        return fn_sqrt(l2);
    }else if(estim==EstimateGUniformConvexity){
        return 0; // G(w) = 0
    }else if(estim==EstimateFCUniformConvexity){
        return 0;
    }else{
        return 0;
    }
}


static Reg reg_tgv2={
    (RegK*)tgv2_k,
    (RegKC*)tgv2_kc,
    (RegFCres*)tgv2_fcres,
    (RegGres*)noop,
    (RegPDgap*)tgv2_pdgap,
    (RegAllocPrimal*)tgv2_alloc_primal,
    (RegAllocDual*)tgv2_alloc_dual,
    (RegInitDualExtra*)tgv2_init_dual_extra,
    (RegEstimate*)tgv2_estimate,
};


void reg_tgv2_init(const Sig *sig, FNumber alpha, FNumber beta, const char *diffalg, 
                   Reg **reg, RegTGV2Params *tgv2p)
{
    tgv2p->sig=*sig;
    tgv2p->alpha=alpha;
    tgv2p->beta=beta;
    tgv2p->diffalg1=(diffalg[0] ? diffalg[0] : 'f');
    tgv2p->diffalg2=(diffalg[0] && diffalg[1] ? diffalg[1] : tgv2p->diffalg1);
    tgv2p->m2=-1;
    tgv2p->pdgap0_ab=0;
    tgv2p->pdgap0_c=0;

    *reg=&reg_tgv2;
}

