/*
 * dticode/src/reg_tv.c
 *
 * Written 2012 by Tuomo Valkonen <tuomov@iki.fi>, University of Graz.
 *
 */

#include <stdio.h>
#include <math.h>
#include <assert.h>
#include <libtu/output.h>
#include "tensor.h"
#include "tensornorm.h"
#include "cpock_accel.h"
#include "diff.h"
#include "project_posdef.h"
#include "normalise.h"
#include "combine.h"
#include "reg_tv.h"

typedef TensorField TVPrimal;
typedef TensorField TVDual;
typedef void TVExtraPrimal;

static void do_tv_k(TVDual *a, const TVPrimal *b, bool addto_a, RegTVParams *tvp)
{
    if(tvp->td)
        tensorfield_symgrad(a, b, (addto_a ? 1.0 : 0.0), 1.0, 0, tvp->diffalg);
    else
        tensorfield_grad(a, b, (addto_a ? 1.0 : 0.0), 1.0, 0, tvp->diffalg);
}

static void tv_k(TVDual *a, const CombinedPrimal *b, bool addto_a, RegTVParams *tvp)
{
    do_tv_k(a, (TVPrimal*)COMBINED_U(b), addto_a, tvp);
}

static void tv_kc(CombinedPrimal *a, const TVDual *b, bool addto_au, bool addto_aw, RegTVParams *tvp)
{
    // We may symmetrise here for TV, because the primal 
    // variable is (assumed) symmetric
    tensorfield_negsymdiv((TensorField*)COMBINED_U(a), b, (addto_au ? 1.0 : 0.0), 1.0, 0, tvp->diffalg);
}

static void tv_fcres(TVDual *a, FNumber tau, RegTVParams *tvp)
{
    tensorfield_project_frob2(a, tvp->alpha);
}


static void tv_pdgap(FNumber *pdgap, 
                     const CombinedPrimal *x, 
                     const TVDual *phi, /* y */
                     const CombinedPrimal *kcy, 
                     const TVDual *eu,  /* kx */
                     CombinedPrimal *xwork,
                     unsigned int iter,
                     RegTVParams *tvp)
{
    TensorField *u=(TensorField*)COMBINED_U(x);
    TensorField *minus_divp=(TensorField*)COMBINED_U(kcy);
    TensorField *uwork=(TensorField*)COMBINED_U(xwork);
    
    pdgap[0]=(tvp->alpha)*tensorfield_frob1(eu);
    if(iter==0)
        tvp->pdgap0=pdgap[0];
    pdgap[1]=tvp->pdgap0;
}


static TVDual *tv_alloc_dual(RegTVParams *tvp)
{
    Sig dsig=tvp->sig; 
    dsig.order++; 
    dsig.symmetric=tvp->td;
    
    return alloc_tensorfield(&dsig);
}

static void tv_init_dual_extra(TVExtraPrimal *w, TVDual *psi, const TVPrimal *u, 
                               RegTVParams *tvp)
{
    do_tv_k(psi, u, FALSE, tvp);
    tensorfield_normalise_frob2(psi, tvp->alpha);
}

static FNumber tv_estimate(Estimate estim, RegTVParams *tvp)
{
    if(estim==EstimateKNorm){
        FNumber l2=calc_l2fact(&(tvp->sig), tvp->diffalg);
        return fn_sqrt(l2);
    }else if(estim==EstimateGUniformConvexity){
        return INFINITY;
    }else if(estim==EstimateFCUniformConvexity){
        return 0;
    }else{
        return 0;
    }
}

static Reg reg_tv={
    (RegK*)tv_k,
    (RegKC*)tv_kc,
    (RegFCres*)tv_fcres,
    (RegGres*)noop,
    (RegPDgap*)tv_pdgap,
    (RegAllocPrimal*)alloc_dummyvector,
    (RegAllocDual*)tv_alloc_dual,
    (RegInitDualExtra*)tv_init_dual_extra,
    (RegEstimate*)tv_estimate,
};

void reg_tv_init(const Sig *sig, bool td, FNumber alpha, const char *diffalg, 
                 Reg **reg, RegTVParams *tvp)
{
    tvp->alpha=alpha;
    tvp->diffalg=(diffalg[0] ? diffalg[0] : 'f');
    tvp->td=td;
    tvp->sig=*sig;
    tvp->pdgap0=0;
    
    *reg=&reg_tv;
}
