/*
 * dticode/src/scaleproj_core.c
 *
 * Written 2012 by Tuomo Valkonen <tuomov@iki.fi>, University of Graz.
 *
 */

#ifndef DIM
#error "This file needs to be compiled with DIM set."
#endif

#include <math.h>
#include <libtu/output.h>
#include <assert.h>
#include <stdio.h>

#include "number.h"
#include "sylvester.h"
#include "misc.h"
#include "diff_rules.h" /* CAT */ 

#define DEBUG 0

#if DIM == 1

void scaleproj_1(FNumber *v, size_t count, const FNumber *m_mat, 
                 FNumber minm, FNumber sigma, FNumber gamma,
                 int iters, FNumber rho)
{
    #pragma omp parallel for
    #pragma acc kernels loop deviceptr(v), independent
    for(size_t elem=0; elem<count; elem++){
        if(v[elem]<0)
            v[elem]=0;
    }
}

#else

#if DIM==2

# include "scaleproj_symmat.2.h"
# include "build/make_lmat.2.h"

#elif DIM==3

# include "scaleproj_symmat.3.h"
# include "build/make_lmat.3.h"

#else // DIM

# include "scaleproj_symmat.h"
# error "Not presently implemented for DIM. Please provide make_lmat.DIM.h."

#endif // DIM

#define MAKE_LD_LPM CAT(MAKE_LD_LPM, DIM)
//#define MAKE_LMAT CAT(make_lmat, DIM)
#define SYLVESTER CAT(sylvester, DIM)

// Solve M p + N_{>=0}(p) = M p_param, p >=0,
// where p_param is the initial (parameter) value of p 
// (given in v, which also contains the result)
void CAT(scaleproj, DIM)(FNumber * restrict v, 
                         size_t count, const FNumber *m_mat, 
                         FNumber minm, FNumber sigma, FNumber gamma,
                         int iters, FNumber rho)
{
    #define SZ symsize(DIM)

#if 0 //_OPENACC
    FNumber ld_lpm_mat_[count*SZ*SZ], p_[count*SZ], d_[count*SZ], dp_[count*SZ], dd_[count*SZ];
#endif

    #pragma omp parallel
    #pragma acc kernels copyin(m_mat[0:SZ*SZ]), deviceptr(v)
    {
        FNumber ld_lpm_mat[SZ*SZ], p[SZ], d[SZ], dp[SZ], dd[SZ];

        #pragma omp for
        #pragma acc loop, independent, private(p[0:SZ], d[0:SZ], dp[0:SZ], dd[0:SZ], ld_lpm_mat[0:SZ*SZ])
        for(size_t off=0; off<count*SZ; off+=SZ){
#if !_OPENACC
            FNumber * restrict p_param=v+off;
#else
            #define p_param (v+off)
#endif

            FNumber mu_init, mu;
            int iter;
            
            // If v is already positive definite, nothing needs to be done.
            if(SYLVESTER(p_param, -FNUMBER_EPS))
                continue;
            
            // Initialisation. Calculate p, d>=0, |p o d - mu| <= gamma mu,
            // such that M p - d = mmat * p_param. This is done by setting
            // p = p_param + beta*I, d = beta * M I for suitably large beta
            // and gamma.
            //
            // In this situation:
            //
            //   mu = <p,d>/m = beta*tr(M p_param)/m + beta^2 tr(MI)/m
            //                =beta*tr_Mpp/m + beta^2*tr_MI/m
            //   p o d = beta (M p_param) + beta^2 (MI)
            //   Pep (p o d) = beta ( M p_param - tr(M p_param)/m)
            //                 + beta^2 (MI - tr(MI)/m)
            //               = beta Pep (M p_param) + beta^2 Pep (MI)
            //               = beta Pe_Mpp + beta^2 Pe_MI
            //
            // Need to calculate tr, Pep for: p_param o MI, I o MI
            //
            // Condition 1:
            //
            //   |Pep (p o d)| <= gamma mu
            //     <=> |Pe_Mpp + beta Pe_MI| < w tr_Mpp + w beta tr_MI (w=gamma/m)
            //     <=> 1) tr_Mpp + beta tr_MI >=0
            //            <=> beta >= beta1 := -tr_Mpp/tr_MI  (tr_MI>=0!)
            //         2) n2_Pe_Mpp + beta^2 n2_Pe_MI + beta iprod_2Pe_MI_Mpp
            //             <= w^2 tr2_Mpp + w^2beta^2 tr2_MI + 2w^2 beta tr_Mpp_MI
            //            <=> 
            //            beta^2(n2_Pe_MI-w^2tr2_MI) + beta(iprod_2Pe_MI_Mpp-2 w^2 tr_Mpp_MI)
            //              + (n2_Pe_Mpp - w^2 tr2_Mpp) <= 0
            //            <=> beta^2 a2 + beta a1 + a0 <=0
            //
            // Condition 2:
            //
            //   p >= 0
            //     <=> p_param + beta*I >= 0
            //     <= lambda_min p_param + beta >= 0
            //     <= beta >= beta0 := |p_param|_F.
            
            // We will use beta0 for a stopping criterion as well.
            FNumber beta0=symfrob(p_param, DIM);
            {
                make_eye(dd, 1, DIM); // dd = I

                MATVECMULT(d, m_mat, dd, SZ, SZ); // d = MI
                MATVECMULT(p, m_mat, p_param, SZ, SZ); // p = M p_param

                FNumber tr_MI=symtrace(d, DIM); // tr (MI)
                FNumber tr_Mpp=symtrace(p, DIM); // tr (M p_param)

                symproj_pe(dd, d, DIM); // dd = Pep MI
                symproj_pe(dp, p, DIM); // dp = Pep (M p_param)
                FNumber n2_Pe_MI=symiprod(dd, dd, DIM); // |Pep MI|^2
                FNumber n2_Pe_Mpp=symiprod(dp, dp, DIM); // |Pep Mp_param|^2
                FNumber iprod_2Pe_MI_Mpp=2*symiprod(dd, dp, DIM); // 2<Pep MI, Pep Mp_param>

                // Gamma has to be big enough that increasing beta
                // eventually causes |Pep (p o d)| <= gamma mu to hold.
                FNumber gamma_limit=sqrt(n2_Pe_MI)/(tr_MI/DIM);
#if DEBUG
                assert(gamma_limit<1);
#endif
                gamma=fn_max(gamma, (1+gamma_limit)/2);

                //fprintf(stderr, "gamma: %f\n", gamma);

                FNumber w=gamma/dimrank(DIM);
                FNumber w2=w*w;
                FNumber tr2_Mpp_w=tr_Mpp*tr_Mpp*w2;
                FNumber tr2_MI_w=tr_MI*tr_MI*w2;
                FNumber p_2tr_Mpp_MI_w=2*tr_MI*tr_Mpp*w2;
                
                FNumber beta1=-tr_Mpp/tr_MI;
                
                // Solve
                //   beta^2(n2_Pe_MI-tr2_MI_w) + beta(iprod_2Pe_MI_Mpp-p_2tr_Mpp_MI_w)
                //   + (n2_Pe_Mpp - tr2_Mpp_w) <= 0
                // <=> beta^2 a2 + beta a1 + a0 <=0
                FNumber a2=n2_Pe_MI-tr2_MI_w;
                FNumber a1=iprod_2Pe_MI_Mpp-p_2tr_Mpp_MI_w;
                FNumber a0=n2_Pe_Mpp-tr2_Mpp_w;
                FNumber dis=a1*a1-4*a2*a0;

#if DEBUG
                // Big enough gamma should ensure this.
                assert(a2<=0);
#endif
                // Larger solution to a2*beta+a1+a0=0:
                //
                //FNumber beta2=(-a1+fn_sqrt(dis))/(2*a2);
                //
                // Approximate: sqrt(x^2+y) <= sqrt(x^2)+ y/(2sqrt(x^2))
                //                           = |x| + y/(2|x|)
                //
                // => (-a1+sqrt(a1^2-4 a2 a0))/(2 a2)
                //    <= |a1|/|a2| + 2|a0|
                FNumber beta2=fn_abs(a2/a2)+2*fn_abs(a0);
                
                FNumber beta=fn_max(fn_max(beta0, beta1), beta2);
                
                vecscale(d, beta, SZ); // d = beta MI
                make_eye(p, beta, DIM);
                vecadd(p, p_param, 1, SZ); // p = beta I + p_param

                mu_init=symiprod(p, d, DIM)/dimrank(DIM);
            }

            
#if DEBUG
            assert(SYLVESTER(p, -FNUMBER_EPS));
            assert(SYLVESTER(d, -FNUMBER_EPS));
            
            symprod(dp, p, d, DIM);
            symproj_pe(dd, dp, DIM);
            if(gamma*mu_init < symfrob(dd, DIM)){
                die("%s:%d:%s: |P|=%g < mu*gamma=%g\n", 
                    __FILE__, __LINE__, __func__,
                    symfrob(dd, DIM), gamma*mu_init);
            }
#endif
            mu=mu_init;

            //fprintf(stderr, "INIT %f -> %f\n", symfrob(p_param, DIM), symfrob(p, DIM));
            
            #pragma acc loop seq
            for(iter=0; iter<iters; iter++){
                // Calculate L(d) + L(p)M
                MAKE_LD_LPM(ld_lpm_mat, p, d, m_mat);
                //MAKE_LMAT(tmp_mat, p);
                //MAKE_LMAT(ld_lpm_mat, d);
                //matmatmult(ld_lpm_mat, tmp_mat, m_mat, 1, SZ, SZ, SZ);
                
                // Calculate right-hand side
                make_dirvec(dp, p, d, sigma*mu, DIM);
                
                // Solve dp in-place
                #define LUDIM SZ
                #define NRHS 1
                #define A ld_lpm_mat
                #define C dp
                #define RES dp
                #define GETM(a, i, j) (a[(j)+(i)*LUDIM])
                #define SETM(a, i, j, x) a[(j)+(i)*LUDIM]=(x)
                #define GETV(v, i, k) (v[(i)+(k)*LUDIM])
                #define SETV(v, i, k, x) v[i+(k)*LUDIM]=(x)
                #include "lusolve_core.h"
                #undef LUDIM
                #undef NRHS
                #undef A
                #undef C
                #undef RES
                #undef GETM
                #undef SETM
                #undef GETV
                #undef SETV

                // Calculate dd = M dp
                MATVECMULT(dd, m_mat, dp, SZ, SZ);
                
                // calculate alpha
                FNumber ndp=symfrob(dp, DIM);
                FNumber ndd=symfrob(dd, DIM);
                FNumber hat_alpha=sigma*gamma*mu/(2*ndp*ndd);
                FNumber bar_alpha=fn_min(hat_alpha, 1);
                FNumber ip=symiprod(dp, dd, DIM);
                FNumber check_alpha=(1-sigma)*DIM*mu/(2*ip);
                // A bit shorter stop due to numerical inaccuracies
                FNumber alpha=FNUMBER(0.9)*fn_min(bar_alpha, check_alpha);
                //assert(alpha==bar_alpha);
                
                // update p, d, and mu
                vecadd(p, dp, alpha, SZ);
                vecadd(d, dd, alpha, SZ);
                mu=symiprod(p, d, DIM)/dimrank(DIM);

                /*
                MATVECMULT(dd, m_mat, p, SZ, SZ);
                vecadd(dd, d, -1, SZ); // dd = Mp-d
                MATVECMULT(dp, m_mat, p_param, SZ, SZ); // dp = Mp_param
                vecadd(dd, dp, -1, SZ); // dd = Mp-d-Mp_param
                fprintf(stderr, "DIST: %f\n", symfrob(dd, DIM));
                */
#if DEBUG
                assert(SYLVESTER(p, -FNUMBER_EPS));
                assert(SYLVESTER(d, -FNUMBER_EPS));
#elseif 0
                // Stop if we go over the boundary
                // (due to numerical inaccuracies)
                {
                    if(!SYLVESTER(p, -FNUMBER_EPS)){
                        // Restore positive definite p
                        vecadd(p, dp, -alpha, SZ);
                        break;
                    }

                    if(!SYLVESTER(d, -FNUMBER_EPS))
                        break;
                }
#endif

                //fprintf(stderr, "%d: rho: %f, mu: %f\n", iter+1, mu/mu_init, mu);
                
                // check stopping criterion
                
                // In addition to standard duality gap stopping criterion,
                // in order to not require large rho, require also that
                // at least |p| < (1+rho)|c|, because for a projection 
                // we should have |p| < |c|. Unfortunately calculating
                // the smallest negative eigenvalue of c for a better
                // criterion is too expensive.
                if(mu/mu_init<rho && symfrob(p, DIM)<(1+rho)*beta0)
                    break;
            }

            veccopy(p_param, p, SZ);

            //fprintf(stderr, "RES -> %f\n", symfrob(p, DIM));
        }
    }
}

//#undef SZ

#endif /* DIM != 1 */
