/*
 * dticode/src/scaleproj_symmat.h
 *
 * Written 2012 by Tuomo Valkonen <tuomov@iki.fi>, University of Graz.
 *
 */
 
#ifndef _SCALEPROJ_SYMMAT_H
#define _SCALEPROJ_SYMMAT_H

#include <stdio.h>
#include <string.h>
#include <stddef.h>
#include <math.h>
#include <assert.h>

#include "number.h"

/* */


#define symsize(dim) (dim*(dim+1)/2)
#define dimrank(dim) (dim)


/* Iterate over diagonal entries 
 */
#define SYMMAT_ITERATE_DIAG(dim, i, k, DIAG) \
    { \
        size_t i, k; \
        for(size_t i=0, k=0; i<dim; i++, k+=i+1) DIAG; \
    }

/* Iterate over all entries (symmetric entries not duplicated), 
 * column by column, with different handler for diagonal entries.
 */
#define SYMMAT_ITERATE2(dim, i, k, start, end, OFFDIAG, DIAG) \
    { \
        size_t i, k, start, end; \
        for(i=0, start=0, end=0; i<dim; i++, start=end+1, end=start+i){ \
            for(k=start; k<end; k++){ \
                OFFDIAG; \
            }; \
            DIAG; \
        } \
    }

/* The same as above, but same operations for diagonal and off-diagonal 
 * elements.
 */
#define SYMMAT_ITERATE(dim, i, k, start, end, OP) \
    { \
        size_t i, start, end, k; \
        for(i=0, start=0, end=0; i<dim; i++, start=end+1, end=start+i){ \
            for(k=start; k<=end; k++) OP; \
        } \
    }

#define SYMMAT_ITERATE_DOUBLE(dim, k, r1, r2, INITOP, OP, FINALOP) \
    { \
        size_t k, r1, r2; \
        size_t _j, _i, _l, _cstart, _cend, _rstart, _rend, _e1, _e2; \
        /* iterate LHS columns */ \
        /*fprintf(stderr, "--\n");*/ \
        for(_i=0, _cstart=0, _cend=0; _i<dim; _i++, _cstart=_cend+1, _cend=_cstart+_i){ \
            /* iterate LHS rows for column */ \
            /* k = LHS index; k-_cstart = ROW, */ \
            /* _rstart = beginning of corresponding row for RHS */ \
            /* _rend = end of consequtive indices for the row */ \
            /*fprintf(stderr, "Y %d %d\n", _cstart, _cend);*/ \
            for(_l=0, k=_cstart, _rstart=0, _rend=0; k<=_cend; k++, _l++, _rstart=_rend+1, _rend=_rstart+_l){ \
                /*fprintf(stderr, "X %d %d\n", _rstart, _rend);*/ \
                if(_cend-_cstart < _rend-_rstart){ \
                    /* more consequtive row indices than column */ \
                    r1=_cstart; r2=_rstart; \
                    _e1=_cend; _e2=_rend; \
                }else{ \
                    /* more consequtive column indices than row */ \
                    r2=_cstart; r1=_rstart; \
                    _e2=_cend; _e1=_rend; \
                } \
                INITOP; \
                /* both RHS1 and RHS2 have consequtive indices */ \
                for(_j=0; r1<_e1; _j++, r1++, r2++) \
                    OP; \
                /* RHS1 has non-consequtive indices, RHS2 consequtive */ \
                for( ; r2<_e2; _j++, r1+=_j, r2++) \
                    OP; \
                /* Both RHS1 and RHS2 have non-consequtive indices */ \
                for( ; _j<dim; _j++, r1+=_j, r2+=_j) \
                    OP; \
                FINALOP; \
            } \
        } \
    }

#define SYMMAT_ITERATE_SYMPROD(dim, k, val, FINALOP) \
    SYMMAT_ITERATE_DOUBLE(dim, k, _r1, _r2, \
        FNumber val=0, \
        {val+=(p[_r1]*d[_r2]+p[_r2]*d[_r1]);}, \
        FINALOP \
    );
            

static __inline__ void make_eye(FNumber *p, FNumber scale, size_t dim)
{
    SYMMAT_ITERATE2(dim, i, k, start, end, {
        p[k]=0; // off-diagonal
    }, {
        p[k]=scale; // diagonal
    });
}

static __inline__ FNumber symtrace(const FNumber *p, size_t dim)
{
    FNumber res=0;
    
    SYMMAT_ITERATE_DIAG(dim, i, k, {
        res+=p[k];
    });
    
    return res;
}

static __inline__ void symproj_pe(FNumber *proj, const FNumber *p, size_t dim)
{
    FNumber mu=symtrace(p, dim)/dimrank(dim);

    SYMMAT_ITERATE2(dim, i, k, start, end, {
        proj[k]=p[k]; // off-diagonal
    }, {
        proj[k]=p[k]-mu; // diagonal
    });
}


static __inline__ FNumber symiprod(const FNumber *p, const FNumber *d, size_t dim)
{
    FNumber res=0;
    
    SYMMAT_ITERATE2(dim, i, k, start, end, {
        res+=2*(p[k]*d[k]);
    }, {
        res+=p[k]*d[k];
    });
    
    return res;
}


static __inline__ FNumber symfrob(const FNumber *p, size_t dim)
{
    return sqrt(symiprod(p, p, dim));
}

 
static __inline__ void symprod(FNumber *res, const FNumber *p, const FNumber *d,
                               size_t dim)
{
    SYMMAT_ITERATE_SYMPROD(dim, k, val, {
        res[k]=val*FNUMBER(0.5);
    })
}

static __inline__ void make_dirvec(FNumber *v, const FNumber *p, const FNumber *d, 
                                   FNumber emult, size_t dim)
{
    symprod(v, p, d, dim);

    SYMMAT_ITERATE2(dim, i, k, start, end, {
        v[k]=-v[k]; // off-diagonal
    }, {
        v[k]=emult-v[k]; // diagonal
    });
}

#endif /* _SCALEPROJ_SYMMAT_H */
