/*
 * dticode/src/splitvar.c
 *
 * Written 2013 by Tuomo Valkonen <tuomov@iki.fi>, University of Cambridge.
 *
 */

#include <libtu/misc.h>
#include <libtu/output.h>
#include <libtu/objp.h>
#include "splitvar.h"
#include "tensor.h"

static bool splitvar_init(SplitVar *p, size_t n)
{
    if(p==NULL)
        die_err();
    
    p->n=n;
    p->p=ALLOC_N(Vector*, n);
    if(p->p==NULL)
        die_err();
        
    for(size_t i=0; i<n; i++)
        p->p[i]=NULL;    
    
    return TRUE;
}

SplitVar *alloc_splitvar(size_t n)
{
    CREATEOBJ_IMPL(SplitVar, splitvar, (p, n));
}

SplitVar *alloc_splitvar_tensorfields(size_t n, const Sig *sig)
{
    SplitVar *v=alloc_splitvar(n);
    
    for(size_t i=0; i<n; i++)
        v->p[i]=(Vector*)alloc_tensorfield(sig);
    
    return v;
}

void splitvar_deinit(SplitVar *p)
{
    for(size_t i=0; i<p->n; i++){
        if(p->p[i])
            destroy_obj((Obj*)p->p[i]);
    }
    vector_deinit(&p->vector);
}

static void splitvar_scale(SplitVar *a, FNumber gamma)
{
    for(size_t i=0; i<a->n; i++){
        if(a->p[i]!=NULL)
            vector_scale(a->p[i], gamma);
    }
}

static void splitvar_copy(SplitVar *a, const SplitVar *b)
{
    assert(a->n==b->n);
    
    for(size_t i=0; i<a->n; i++){
        if(a->p[i]!=NULL)
            vector_copy(a->p[i], b->p[i]);
    }
}

static void splitvar_scaleplus(SplitVar *a, const SplitVar *b, const SplitVar *c,
                               FNumber gamma, FNumber tau, FNumber sigma)
{
    assert(a->n==b->n);
    
    for(size_t i=0; i<a->n; i++){
        if(a->p[i]!=NULL)
            vector_scaleplus(a->p[i], b->p[i], c->p[i], gamma, tau, sigma);
    }
}

static void splitvar_zero(SplitVar *a)
{
    for(size_t i=0; i<a->n; i++){
        if(a->p[i]!=NULL)
            vector_zero(a->p[i]);
    }
}

static FNumber splitvar_iprod(const SplitVar *a, const SplitVar *b)
{
    FNumber ret=0;
    
    assert(a->n==b->n);
    
    for(size_t i=0; i<a->n; i++)
        ret+=vector_iprod(a->p[i], b->p[i]);
        
    return ret;
}

static FNumber splitvar_dist(const SplitVar *a, const SplitVar *b)
{
    FNumber ret=0;
    
    assert(a->n==b->n);
    
    for(size_t i=0; i<a->n; i++)
        ret+=vector_dist(a->p[i], b->p[i]);
        
    return ret;
}

static FNumber splitvar_norm(const SplitVar *a)
{
    FNumber ret=0;
    
    for(size_t i=0; i<a->n; i++)
        ret+=vector_norm(a->p[i]);
        
    return ret;
}

static void splitvar_save(const SplitVar *a, const char *outfile)
{
    for(size_t i=0; i<a->n; i++){
        char *tmp=NULL;
        libtu_asprintf(&tmp, "%s.component%d", outfile, i);
        assert(tmp!=NULL);
        if(a->p[i]!=NULL)
            vector_save(a->p[i], tmp);
        free(tmp);
    }
}

static bool splitvar_load(SplitVar *a, const char *outfile)
{
    for(size_t i=0; i<a->n; i++){
        char *tmp=NULL;
        libtu_asprintf(&tmp, "%s.component%d", outfile, i);
        assert(tmp!=NULL);
        if(a->p[i]!=NULL){
            if(!vector_load(a->p[i], tmp))
                return FALSE;
        }
        free(tmp);
    }
    return TRUE;
}

static DynFunTab splitvar_dynfuntab[]={
    {vector_scale, splitvar_scale},
    {vector_scaleplus, splitvar_scaleplus},
    {vector_copy, splitvar_copy},
    {vector_zero, splitvar_zero},
    {vector_save, splitvar_save},
    {(DynFun*)vector_load, (DynFun*)splitvar_load},
    {(DynFun*)vector_iprod, (DynFun*)splitvar_iprod},
    {(DynFun*)vector_dist, (DynFun*)splitvar_dist},
    {(DynFun*)vector_norm, (DynFun*)splitvar_norm},
    END_DYNFUNTAB
};

IMPLCLASS(SplitVar, Vector, splitvar_deinit, splitvar_dynfuntab);
