/*
 * dticode/src/tensor.c
 *
 * Written 2012 by Tuomo Valkonen <tuomov@iki.fi>, University of Graz.
 *
 */


#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <string.h>
#include <libtu/output.h>
#include <libtu/misc.h>
#include <libtu/objp.h>
#if _OPENACC
#include <openacc.h>
#endif
#if _MPI
#include <mpi.h>
#endif
#include "acc_support.h"

#include "tensor.h"
#include "tensornorm.h"
#include "lut.h"
#include "misc.h"
#include "number.h"
#include "doubleio.h"
#include "scaleplus.h"


void sig_check_max(const Sig *sig)
{
    if(sig->dim > MAX_DIM || sig->dim < 1)
        die("signature dimension out of LUT range");
    if(sig->order > MAX_ORDER || sig->order < 0)
        die("signature order out of LUT range");
}

void sig_check_match(const Sig *sig1, const Sig *sig2)
{
    if(sig1->dim!=sig2->dim || sig2->order!=sig2->order || 
       sig1->symmetric!=sig2->symmetric || sig1->is_complex!=sig2->is_complex){
        die("Tensor type mismatch");
    }
    if(sig1->w!=sig2->w || sig1->h!=sig2->h || sig1->d!=sig2->d){
        die("Tensor field size mismatch");
    }
}

size_t sig_tensor_size_c(const Sig *sig)
{
    size_t sz;
    
    if(sig->order==0){
        sz=1;
    }else if(sig->order==1){
        sz=sig->dim;
    }else{
        sig_check_max(sig);

        sz=(sig->symmetric ? tsize_sym[sig->dim-1][sig->order]
                           : tsize_nonsym[sig->dim-1][sig->order]);
    }
    
    return sz;
}

size_t sig_tensor_size(const Sig *sig)
{
    size_t sz=sig_tensor_size_c(sig);
    return (sig->is_complex ? 2*sz : sz);
}

size_t sig_slice_count(const Sig *sig)
{
    return sig->d;
}
    
size_t sig_slice_size(const Sig *sig)
{
    return sig_tensor_size(sig)*(sig->w)*(sig->h);
}

size_t sig_tensor_count(const Sig *sig)
{
    return sig_slice_count(sig)*(sig->w)*(sig->h);
}

size_t sig_tensorfield_size(const Sig *sig)
{
    return sig_tensor_size(sig)*sig_tensor_count(sig);
}

#if _MPI

size_t rank_slices(size_t totalslices, unsigned int numtasks, 
                   unsigned int rank)
{
    if(numtasks>totalslices){
        return (rank<totalslices ? 1 : 0);
    }else{
        size_t baseslices=totalslices/numtasks;
        size_t plus1count=totalslices%numtasks;
        
        return baseslices+(rank<plus1count ? 1 : 0);
    }
}

size_t rank_startslice(size_t totalslices, unsigned int numtasks, 
                       unsigned int rank)
{
    size_t baseslices=totalslices/numtasks;
    size_t plus1count=totalslices%numtasks;

    if(rank<plus1count)
        return (baseslices+1)*rank;
    else
        return (baseslices+1)*plus1count+baseslices*(rank-plus1count);
}

size_t slice_rank(size_t totalslices, unsigned int numtasks, size_t i)
{
    size_t baseslices=totalslices/numtasks;
    size_t plus1count=totalslices%numtasks;
    size_t plus1end=plus1count*(baseslices+1);
    
    if(i<plus1end)
        return i/(baseslices+1);
    else
        return plus1count+(i-plus1end)/baseslices;
}

#endif


bool tensorfield_init(TensorField *t, const Sig *sig)
{
    size_t bytes;
    
    if(!vector_init(&t->vector))
        return FALSE;

#if _MPI
    int numtasks, rank;
    size_t myslices;
    size_t slicesize;
        
    MPI_Comm_rank(MPI_COMM_WORLD, &rank);
    MPI_Comm_size(MPI_COMM_WORLD, &numtasks);
    
    slicesize=sig_slice_size(sig);
    myslices=rank_slices(sig_slice_count(sig), numtasks, rank);
    
    assert(myslices>0);
    
    t->sig=*sig;
    t->fullsig=*sig;
    t->sig.d=myslices;
    
    MPI_Type_contiguous(slicesize, FNUMBER_MPI_TYPE, &(t->slicetype));
    MPI_Type_commit(&(t->slicetype));
    
    bytes=sizeof(FNumber)*slicesize*(myslices+2);
    
    t->fullp=HWALLOC(bytes);

    if(t->fullp==NULL)
        die_err();

#if _OPENACC
    size_t tbytes=sizeof(FNumber)*slicesize*4;
    t->transferbuf=malloc(tbytes);
    if(t->transferbuf==NULL)
        die_err();
#endif
    
    t->p=t->fullp+slicesize;
#else /* !_MPI */
    t->sig=*sig;

    bytes=sizeof(FNumber)*sig_tensorfield_size(sig);
    
    t->p=HWALLOC(bytes);
    
    if(t->p==NULL)
        die_err();
#endif /* _MPI */
    t->nelem=sig_tensor_count(&t->sig);
    t->elemsize=sig_tensor_size(&t->sig);
    
    return TRUE;
}

TensorField *alloc_tensorfield(const Sig *sig)
{
    CREATEOBJ_IMPL(TensorField, tensorfield, (p, sig));
}

void tensorfield_deinit(TensorField *t)
{
#if _MPI
    HWFREE(t->fullp);
    MPI_Type_free(&t->slicetype);
#if _OPENACC
    free(t->transferbuf);
#endif
#else
    HWFREE(t->p);
#endif
}

void free_tensorfield(TensorField *t)
{
    destroy_obj((Obj*)t);
}

static bool readfn_tensorfield(const double *buf, size_t idx, size_t bs, 
                               void *res_)
{
    FNumber * DEVPTR res=&((FNumber*)res_)[idx*bs];
    
    #pragma acc kernels loop deviceptr(res), copyin(buf[0:bs]), independent
    for(size_t i=0; i<bs; i++)
        res[i]=buf[i];
        
    return TRUE;
}

static bool writefn_tensorfield(double *buf, size_t idx, size_t bs, 
                                void *data_)
{
    FNumber * DEVPTR data=&((FNumber*)data_)[idx*bs];
    
    #pragma acc kernels loop deviceptr(data), copyout(buf[0:bs]), independent
    for(size_t i=0; i<bs; i++)
        buf[i]=data[i];

    return TRUE;
}

#if _MPI


typedef struct{
    FNumber *ptr;
    size_t totalslices;
    int numtasks;
} LoadParam;

static bool readfn_mpi(const double *buf, size_t idx, size_t bs, 
                        void *lp_)
{
    LoadParam *lp=(LoadParam*)lp_;
    int target=slice_rank(lp->totalslices, lp->numtasks, idx);
    
    if(target==0){
        return readfn_tensorfield(buf, idx, bs, lp->ptr);
    }else{
        int tag=idx-rank_startslice(lp->totalslices, lp->numtasks, target);
        int rs=rank_slices(lp->totalslices, lp->numtasks, target);
        MPI_Send((void*)buf, bs, MPI_DOUBLE, target, tag, 
                 MPI_COMM_WORLD);
        return 1;
    }
}

static bool writefn_mpi(double *buf, size_t idx, size_t bs, 
                        void *lp_)
{
    LoadParam *lp=(LoadParam*)lp_;
    int target=slice_rank(lp->totalslices, lp->numtasks, idx);
    
    if(target==0){
        return writefn_tensorfield(buf, idx, bs, lp->ptr);
    }else{
        int tag=idx-rank_startslice(lp->totalslices, lp->numtasks, target);
        MPI_Recv(buf, bs, MPI_DOUBLE, target, tag, 
                 MPI_COMM_WORLD, MPI_STATUS_IGNORE);
        return 1;
    }
}

static bool *do_load_tensorfield(TensorField *t, const char *file)
{
    
size_t myslices, slicesize;
    int numtasks, rank;
    LoadParam lp;
    
    MPI_Comm_rank(MPI_COMM_WORLD, &rank);
    MPI_Comm_size(MPI_COMM_WORLD, &numtasks);
    
    myslices=sig_slice_count(&t->sig);
    slicesize=sig_slice_size(&t->sig);
    
    lp.ptr=t->p;
    lp.totalslices=sig_slice_count(&t->fullsig);
    lp.numtasks=numtasks;

    if(rank==0){
        read_doublefile(lp.totalslices, slicesize, file, readfn_mpi, &lp);
    }else{
        double *buf=malloc(sizeof(double)*slicesize);
        if(buf==NULL)
            die_err();
        for(size_t idx=0; idx<myslices; idx++){
            MPI_Recv(buf, slicesize, MPI_DOUBLE, 0, idx, 
                     MPI_COMM_WORLD, MPI_STATUS_IGNORE);
            readfn_tensorfield(buf, idx, slicesize, lp.ptr);
        }
        free(buf);
    }
    
    MPI_Barrier(MPI_COMM_WORLD);
    
    return TRUE;
}


void save_tensorfield(const TensorField *t, const char *file)
{
    size_t myslices, slicesize;
    int numtasks, rank;
    LoadParam lp;

    MPI_Comm_rank(MPI_COMM_WORLD, &rank);
    MPI_Comm_size(MPI_COMM_WORLD, &numtasks);
    
    myslices=sig_slice_count(&t->sig);
    slicesize=sig_slice_size(&t->sig);
    
    lp.ptr=t->p;
    lp.totalslices=sig_slice_count(&t->fullsig);
    lp.numtasks=numtasks;
    
    if(rank==0){
        write_doublefile(lp.totalslices, slicesize, file, writefn_mpi, &lp);
    }else{
        double *buf=malloc(sizeof(double)*slicesize);
        if(buf==NULL)
            die_err();
        for(size_t idx=0; idx<myslices; idx++){
            writefn_tensorfield(buf, idx, slicesize, lp.ptr);
            MPI_Send(buf, slicesize, MPI_DOUBLE, 0, idx, MPI_COMM_WORLD);
        }
        free(buf);
    }
    
    MPI_Barrier(MPI_COMM_WORLD);
}

#else /* !_MPI */


static bool do_load_tensorfield(TensorField *t, const char *file)
{
    size_t nslices=sig_slice_count(&t->sig);
    size_t slicesize=sig_slice_size(&t->sig);

    read_doublefile(nslices, slicesize, file, readfn_tensorfield, t->p);

    return TRUE;
}

void save_tensorfield(const TensorField *t, const char *file)
{
    size_t nslices=sig_slice_count(&t->sig);
    size_t slicesize=sig_slice_size(&t->sig);

    write_doublefile(nslices, slicesize, file, writefn_tensorfield, t->p);
}

#endif /* _MPI */

TensorField *load_tensorfield(const Sig *sig, const char *file)
{
    TensorField *t=alloc_tensorfield(sig);

    if(!do_load_tensorfield(t, file))
        die_err_obj(file);

    return t;
}

/* */

static FNumber* ipweight_devptr[1][MAX_DIM][MAX_ORDER+1]={NULL,};

FNumber *sig_get_weight_host(const Sig *sig)
{
    if(sig->order<=1)
        return NULL;
        
    sig_check_max(sig);
    
    if(sig->symmetric){
        return (sig->is_complex
                ? ipweightcomplex[sig->dim-1][sig->order]
                : ipweight[sig->dim-1][sig->order]);
    }else{
        return NULL;
    }
}

FNumber * DEVPTR sig_get_weight(const Sig *sig)
{
    FNumber *wm=sig_get_weight_host(sig);
#ifdef _OPENACC
    if(wm!=NULL){
        FNumber *devwm=ipweight_devptr[sig->is_complex][sig->dim-1][sig->order];
        if(devwm==NULL){
            size_t n=sig_tensor_size(sig);
            devwm=vector_to_dev(wm, n, FALSE);
            ipweight_devptr[sig->is_complex][sig->dim-1][sig->order]=devwm;
        }
        return devwm;
    }
    return NULL;
#else
    return wm;
#endif
}


FNumber *tensorfield_get_weight(const TensorField *t)
{
    return sig_get_weight(&t->sig);
}


/* */

void tensorfield_zero(TensorField *t)
{
    size_t s=t->nelem*t->elemsize;
    FNumber *tp=t->p;
    #pragma acc kernels loop deviceptr(tp), independent
    for(size_t i=0; i<s; i++)
        tp[i]=0;
}

void tensorfield_copy(TensorField *t, const TensorField *u)
{
    size_t s=t->nelem*t->elemsize;
    FNumber * DEVPTR tp=t->p;
    FNumber * DEVPTR up=u->p;
    
    sig_check_match(&t->sig, &u->sig);
    
    #pragma acc kernels loop deviceptr(tp, up), independent
    for(size_t i=0; i<s; i++)
        tp[i]=up[i];
}

static DynFunTab tensorfield_dynfuntab[]={
    {vector_scale, tensorfield_scale},
    {vector_scaleplus, tensorfield_scaleplus},
    {vector_copy, tensorfield_copy},
    {vector_zero, tensorfield_zero},
    {vector_save, save_tensorfield},
    {(DynFun*)vector_load, (DynFun*)do_load_tensorfield},
    {(DynFun*)vector_iprod, (DynFun*)tensorfield_iprod},
    {(DynFun*)vector_norm22, (DynFun*)tensorfield_frob222},
    {(DynFun*)vector_dist22, (DynFun*)tensorfield_frob2dist22},
    END_DYNFUNTAB
};

IMPLCLASS(TensorField, Vector, tensorfield_deinit, tensorfield_dynfuntab);
