/*
 * dticode/src/tensorreg.c
 *
 * Written 2012 by Tuomo Valkonen <tuomov@iki.fi>, University of Graz.
 *
 */


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <libtu/util.h>
#include <libtu/output.h>
#include <libtu/optparser.h>
#include <libtu/misc.h>

#include "parallel.h"
#include "cpock_accel.h"
#include "tensor.h"
#include "reg.h"
#include "reg_split.h"
#include "fid.h"
#include "fid_oper.h"
#include "fid_split.h"
#include "fftop.h"
#include "linoper_id.h"
#include "linoper_pointwise.h"
#include "linoper_compose.h"
#include "linoper_parallel.h"
#include "linoper_multiply_by.h"
#include "linoper_sparsefft.h"
#include "linoper_nufft.h"
#include "diffoper.h"
#include "diffoper_dti.h"
#include "diffoper_dtis.h"
#include "diffoper_multiply.h"
#include "diffoper_split.h"
#include "diffoper_splitsame.h"
#include "diffoper_compose.h"
#include "timing.h"
#include "combine.h"
#include "matrix.h"
#include "vector.h"
#include "method.h"
#include "multiply.h"
#include "complexx.h"

#define PARALLEL_NFFT_FOR_CHANNELS 1

static OptParserOpt opts[]={
    QUICK_OPT('R', "reg",     "Regulariser of u: tv|tgv2|none,alpha,beta"),
    QUICK_OPT('1', "s0reg",   "Regulariser of s0: tv|tgv2|none,alpha,beta"),
    QUICK_OPT('F', "fid",     "Fidelity: l2|l1|l2alt|l2alt-nl|po"),
    QUICK_OPT('q', "iprho",   "Target fractional duality gap for the IP method"),
    QUICK_OPT('f', "infile",  "Input file (raw doubles)"),
    QUICK_OPT('l', "lbound",  "Input file (raw doubles)"),
    QUICK_OPT('r', "rbound",  "Input file (raw doubles)"),
    QUICK_OPT('A', "amat",    "Weight matrix A (raw doubles)"),
    QUICK_OPT('E', "wvec",    "Weight vector w (raw doubles)"),
    QUICK_OPT('0', "s0file",  "Signal for zero diffusion sensitising gradient or coil sensitivities, or both (raw doubles; size same as infile)"),
    QUICK_OPT('2', "fqfile",  "Frequency file"),
    QUICK_OPT('O', "order",   "Tensor order"),
    QUICK_OPT('D', "dim",     "Tensor dimension"),
    QUICK_OPT('Q', "forder",  "Data tensor order (default: order)"),
    QUICK_OPT('W', "fdim",    "Data tensor dimension (default: dim)"),
    QUICK_OPT('k', "fdomain", "Data domain: image|image-complex|kspace|kspace-sparse|kspace-nonuniform"),
    QUICK_OPT('S', "s0est",   "Estimate s0 or coil sensitivities: no|s0|coil|"),
    QUICK_OPT('8', "ndiff",   "Number of diffusion directions (when separate amat for different directions)"),
    QUICK_OPT('9', "nchan","Number of channels"),
    {OPT_ID('U'), "unconstr", 0, NULL, "Solve unconstrained problem"},
    METHOD_OPTS,
    END_OPTPARSEROPTS
};

int main(int argc, char *argv[])
{

bool bounds=FALSE;
MethodParam mp=DEFAULT_METHOD_PARAM;
    
    mp.sig.dim=3;
    mp.sig.order=2;
    mp.sig.symmetric=1;
    
    Sig dwisig={
        .dim = 0,
        .order = 0,
        .symmetric = TRUE,
        .is_complex = FALSE,
        .w = 0,
        .h = 0,
        .d = 1,
        .xres = 1,
        .yres = 1,
        .zres = 1
    };
    
    Sig s0sig;
    
    const char *datadomain="image";
    const char *s0est="no";
    const char *fidname="l2";
    const char *regspec="tv,1";
    const char *s0regspec=NULL;
    FNumber iprho=10e-10;
    const char *infile=NULL;
	
    const char *lbfile=NULL;
	const char *rbfile=NULL;
	
    const char *s0file=NULL;
    const char *amatfile=NULL;
    const char *wvecfile=NULL;
    const char *fqfile=NULL;
    Matrix *amat=NULL;
    int opt;
    bool constr=TRUE;
    FidNL *fidnl=NULL;
    FidP fidp;
    FidP fidp_base;
    RegP regp;
    Fid *fid;
    Fid *fid_base;
    Reg *reg;
    bool nonlinear=FALSE, in_kspace=FALSE;
    size_t nchan=0, ndiff=0;
    Reg *reg_split[2];
    RegP regp_split[2];
    void *regp_split_p[2]={&regp_split[0], &regp_split[1]};

    parallel_init1(&argc, &argv);

#if 0 /*_OPENMP*/
    omp_set_nested(1);
    omp_set_max_active_levels(2);
#endif
  
    libtu_init(argv[0]);
    optparser_init(argc, argv, OPTP_MIDLONG, opts);
    
    while((opt=optparser_get_opt())){
        switch(opt){
            case OPT_ID('R'):
                regspec=optparser_get_arg();
                break;
            case OPT_ID('1'):
                s0regspec=optparser_get_arg();
                break;
            case OPT_ID('F'):
                fidname=optparser_get_arg();
                break;
            case OPT_ID('q'):
                iprho=atof(optparser_get_arg());
                break;
            case OPT_ID('f'):
                infile=optparser_get_arg();
                break;
            case OPT_ID('l'):
                lbfile=optparser_get_arg();
		bounds=TRUE;
                break;
            case OPT_ID('r'):
                rbfile=optparser_get_arg();
                break;
            case OPT_ID('0'):
                s0file=optparser_get_arg();
                break;
            case OPT_ID('2'):
                fqfile=optparser_get_arg();
                break;
            case OPT_ID('A'):
                amatfile=optparser_get_arg();
                break;
            case OPT_ID('E'):
                wvecfile=optparser_get_arg();
                break;
            case OPT_ID('Q'):
                dwisig.order=atoi(optparser_get_arg());
                break;
            case OPT_ID('W'):
                dwisig.dim=atoi(optparser_get_arg());
                break;
            case OPT_ID('O'):
                mp.sig.order=atoi(optparser_get_arg());
                break;
            case OPT_ID('D'):
                mp.sig.dim=atoi(optparser_get_arg());
                break;
            case OPT_ID('U'):
                constr=FALSE;
                break;
            case OPT_ID('S'):
                s0est=optparser_get_arg();
                break;
            case OPT_ID('k'):
                datadomain=optparser_get_arg();
                break;
            case OPT_ID('8'):
                ndiff=atoi(optparser_get_arg());
                break;
            case OPT_ID('9'):
                nchan=atoi(optparser_get_arg());
                break;
            default:
                if(!method_do_opt(&mp, opt, opts)){
                    optparser_print_error();
                    return 1;
                }
                break;
        }
    }
    
    nonlinear=(strcmp(fidname, "l2alt-nl")==0);
    in_kspace=(strstr(datadomain, "kspace")!=NULL);
    
    if(infile==NULL)
        die("infile missing\n");
        
    if(nchan>=1 && !nonlinear)
        die("Multiple channels only supported by the non-linear models.");

    method_params_sanity_check(&mp);

    parallel_init2();

    fft_init();
 
    // Signature for DWI signals or general weights in image space
    dwisig.w=mp.sig.w;
    dwisig.h=mp.sig.h;
    dwisig.d=mp.sig.d;
    if(dwisig.order==0)
        dwisig.order=(in_kspace ? 0 : mp.sig.order);
    if(dwisig.dim==0)
        dwisig.dim=mp.sig.dim;
    if(strcmp(datadomain, "image-complex")==0 || in_kspace)
        dwisig.is_complex=TRUE;
    
    // Signature for s0 or coil sensitivites to be estimated in image space
    s0sig=dwisig;
    if(strcmp(s0est, "coil")==0){
        s0sig.is_complex=TRUE;
        die("Coil sensitivity estimation unimplemented.");
    }else if(strcmp(s0est, "s0")==0){
        s0sig.is_complex=FALSE;
    }else if(strcmp(s0est, "no")!=0){
        die("Invalid s0est parameter");
    }
    
    char *amatfile_diff=(char*)amatfile;
    char *infile_diff=(char*)infile;
	char *lbfile_diff=(char*)lbfile;
	char *rbfile_diff=(char*)rbfile;
	
    char *s0file_diff=(char*)s0file;
    char *fqfile_diff=(char*)fqfile;
    char *wvecfile_diff=(char*)wvecfile;
    
    size_t ndiff_real=(ndiff < 1 ? 1 : ndiff);
    size_t nchan_real=(nchan < 1 ? 1 : nchan);
    
  
    
    DOper **ops_diff=ALLOC_N(DOper*, ndiff_real);
    Fid **fid_diff=ALLOC_N(Fid*, ndiff_real);
    FidP *fidp_diff=ALLOC_N(FidP, ndiff_real);
    void **fidp_diff_p=ALLOC_N(void*, ndiff_real);
    
    if(ops_diff==NULL || fid_diff==NULL || fidp_diff==NULL || fidp_diff_p==NULL)
        die_err();
        
    // Construct operator for every signal
    for(size_t diff=0; diff<ndiff_real; diff++){
        LinOper *fftop=NULL;
        LinOper *oppw=NULL;
        DOper *ddti=NULL;
        Sig fsig;
        
        fidp_diff_p[diff]=&fidp_diff[diff];
        
        if(ndiff>=1 || nchan>=1){
            if(amatfile!=NULL)
                libtu_asprintf(&amatfile_diff, "%s.diff%d", amatfile, diff);
            if(fqfile!=NULL)
                libtu_asprintf(&fqfile_diff, "%s.diff%d", fqfile, diff);
        }
        
        // Load the DWI gradient matrix
        {
            Sig sig=dwisig;
            sig.is_complex=FALSE;
            amat=load_matrix(&sig, &mp.sig, amatfile_diff);
        }
        
        //
        // Load image/k-space data and create linear operator
        // into data domain.
        //
        if(strcmp(datadomain, "image")==0 || strcmp(datadomain, "image-complex")==0){
            fsig=dwisig;
            fftop=(LinOper*)alloc_linoper_id(&fsig);
        }else{
            fftop=load_fftop_nodata(fqfile_diff, &dwisig, datadomain, &fsig);
        }
      
        //
        // Create DTI operator
        //
        if(nonlinear){
            TensorField *s00=NULL;
            
            assert(s0file!=NULL);
            
            if(nchan==0){
                if(ndiff>=1)
                    libtu_asprintf(&s0file_diff, "%s.diff%d", s0file, diff);
                    
                s00=load_tensorfield(&dwisig, s0file_diff);

                if(ndiff>=1)
                    free(s0file_diff);
            }
            
            // Init nonlinear operator
            if(strcmp(s0est, "no")==0){
                if(!constr)
                    ddti=(DOper*)create_diffoper_dti(&mp.sig, amat, s00);
                else
                    ddti=(DOper*)create_diffoper_dti_nonneg(&mp.sig, amat, s00);
            }else{
                if(!constr)
                    ddti=(DOper*)create_diffoper_dtis(&mp.sig, amat, s00);
                else
                    ddti=(DOper*)create_diffoper_dtis_nonneg(&mp.sig, amat, s00);
            }
        }else{
            if(strcmp(datadomain, "image")!=0)
                die("Invalid data domain %s for linear problem.", datadomain);
            if(constr)
                oppw=(LinOper*)alloc_linoper_pointwise_nonneg(&mp.sig, amat);
            else
                oppw=(LinOper*)alloc_linoper_pointwise(&mp.sig, amat);
        }

        //
        // Load channels for current diffusion direction
        //

        DOper **ops_chan=ALLOC_N(DOper*, nchan_real);
        Fid **fid_chan=ALLOC_N(Fid*, nchan_real);
        FidP *fidp_chan=ALLOC_N(FidP, nchan_real);
        void **fidp_chan_p=ALLOC_N(void*, nchan_real);

        if(ops_chan==NULL || fid_chan==NULL || fidp_chan==NULL || fidp_chan_p==NULL)
            die_err();
        
        for(size_t chan=0; chan<nchan_real; chan++){
            TensorField *s0=NULL;
            LinOper *wop=NULL;
            
            fidp_chan_p[chan]=&fidp_chan[chan];
            
            if(ndiff>=1 || nchan>=1){
                if(wvecfile!=NULL)
                    libtu_asprintf(&wvecfile_diff, "%s.diff%d.component%d", wvecfile, diff, chan);
                if(s0file!=NULL)
                    libtu_asprintf(&s0file_diff, "%s.diff%d.component%d", s0file, diff, chan);
                libtu_asprintf(&infile_diff, "%s.diff%d.component%d", infile, diff, chan);
            }
            
	  
            TensorField *f=load_tensorfield(&fsig, infile_diff);
	    TensorField *lb=load_tensorfield(&fsig, infile_diff);
	    TensorField *rb=load_tensorfield(&fsig, infile_diff);

	    if (bounds==TRUE) {
	    	
lb=load_tensorfield(&fsig, lbfile_diff);

	    	rb=load_tensorfield(&fsig, rbfile_diff);}
	    	
	     	
            if(nonlinear && nchan>=1){
                if(s0file_diff==NULL)
                    die("%s missing\n", s0file_diff);
                s0=load_tensorfield(&dwisig, s0file_diff);
                ops_chan[chan]=(DOper*)create_linoper_multiply_by(&s0sig, s0);
            }
            
            if(wvecfile_diff!=NULL){
                Sig wsig=f->sig; wsig.is_complex=FALSE;
                TensorField *wvec=load_tensorfield(&wsig, wvecfile_diff);
                tensorfield_multiply_pointwise(f, wvec, f, FALSE);
                wop=(LinOper*)create_linoper_multiply_by(&f->sig, wvec);
            }else{
                wop=(LinOper*)alloc_linoper_id(&f->sig);
            }
            
           
 if(!nonlinear){
                // Only single channel supported. Problem the composition below;
                // should create multiple copies of oppw, or support an operator
                // outside the collection of fidelities for different channels,
                // like fid_oper does for the nonlinear problem.
                assert(nchan<1);
                if(OBJ_IS(wop, LinOperId)){
                    destroy_obj((Obj*)wop);
                    wop=oppw;
                }else{
                    wop=(LinOper*)create_linoper_compose(wop, oppw);
                }
            }else{
#if PARALLEL_NFFT_FOR_CHANNELS
                LinOper *this_fftop=NULL;
                
#warning "TODO: general linoper_copy operation"
                if(chan==0)
                    this_fftop=fftop;
                else if(OBJ_IS(fftop, LinOperSparseFFT))
                    this_fftop=(LinOper*)linoper_sparsefft_copy((LinOperSparseFFT*)fftop);
                else if(OBJ_IS(fftop, LinOperNUFFT))
                    this_fftop=(LinOper*)linoper_nufft_copy((LinOperNUFFT*)fftop);
                else if(OBJ_IS(fftop, LinOperId))
                    this_fftop=(LinOper*)alloc_linoper_id(&fsig);
                else
                    die("Internal error: Invalid FFT operator.");
                
                if(OBJ_IS(wop, LinOperId)){
                    destroy_obj((Obj*)wop);
                    wop=this_fftop;
                }else if(OBJ_IS(this_fftop, LinOperId)){
                    destroy_obj((Obj*)this_fftop);
                }else{
                    wop=(LinOper*)create_linoper_compose(wop, this_fftop);
                }
#endif
            }
         
            init_fidelity(&fid_chan[chan], &fidp_chan[chan], fidname, (LinOper*)wop, f, lb, rb);
            

            if(ndiff>=1 || nchan>=1){
                if(wvecfile_diff!=NULL)
                    free(wvecfile_diff);
                if(s0file_diff!=NULL)
                    free(s0file_diff);
                free(infile_diff);
            }
        }
        
        if(nchan<1){
            fid_diff[diff]=fid_chan[0];
            fidp_diff[diff]=fidp_chan[0];
            if(nonlinear){
#if PARALLEL_NFFT_FOR_CHANNELS
                ops_diff[diff]=ddti;
#else
                ops_diff[diff]=(DOper*)create_diffoper_compose((DOper*)fftop, ddti);
#endif
            }
        }else{
            fidp_diff[diff].split.n=nchan_real;
            fidp_diff[diff].split.fid=fid_chan;
            fidp_diff[diff].split.params=fidp_chan_p;
            
            fid_split_init(&fid_diff[diff], &fidp_diff[diff].split);
            
            if(nonlinear){
                DOper *ds0=(DOper*)create_diffoper_splitsame(ops_chan, nchan_real);
                DOper *ddtis0=(DOper*)create_diffoper_compose(ds0, ddti);
#if PARALLEL_NFFT_FOR_CHANNELS
                ops_diff[diff]=ddtis0;
#else
                DOper *dparfft=(DOper*)create_linoper_parallel(fftop, nchan_real);
                ops_diff[diff]=(DOper*)create_diffoper_compose((DOper*)dparfft, ddtis0);
#endif
            }
        }
        
        if(ndiff>=1 || nchan>=1){
            if(amatfile_diff!=NULL)
                free(amatfile_diff);
            if(fqfile_diff!=NULL)
                free(fqfile_diff);
        }
    }
      
    if(ndiff_real>1){
        fidp_base.split.n=ndiff_real;
        fidp_base.split.fid=fid_diff;
        fidp_base.split.params=fidp_diff_p;
        fid_split_init(&fid_base, &fidp_base.split);
    }else{
        fid_base=fid_diff[0];
        fidp_base=fidp_diff[0];
    }
   
    if(!nonlinear){
        fid=fid_base;
        fidp=fidp_base;
    }else{
        DOper *doper;
        
        if(ndiff_real>1)
            doper=(DOper*)create_diffoper_splitsame(ops_diff, ndiff_real);
        else
            doper=ops_diff[0];
        
        if(mp.gaussnewton){
            fid_oper_init(fid_base, (FidParams*)&fidp_base, 
                          doper, &fid, &fidp.oper);
        }else{
            fid_oper_init_nonlin(fid_base, (FidParams*)&fidp_base, 
                                 doper, &fidnl, &fidp.oper);
        }
    }

    init_regulariser(&reg_split[0], &regp_split[0], regspec, &mp.sig, mp.diffalg);
    
    if(nonlinear && strcmp(s0est, "no")!=0){
        s0regspec=(s0regspec ? s0regspec : regspec);
        init_regulariser(&reg_split[1], &regp_split[1], s0regspec, &s0sig, mp.diffalg);
        regp.split.reg=reg_split;
        regp.split.params=regp_split_p;
        regp.split.n=2;
        reg_split_init(&reg, &regp.split);
    }else{
        regp=regp_split[0];
        reg=reg_split[0];
    }
    
    if(nonlinear && !mp.gaussnewton)
        method_do_nl(&mp, fidnl, &fidp, reg, &regp);
    else
        method_do(&mp, fid, &fidp, reg, &regp);
    
    /*
    if(fullsave){
        char *nm;
        libtu_asprintf(&nm, "%s.primal", outfile);
        vector_save((Vector*)pres, nm);
        free(nm);
        libtu_asprintf(&nm, "%s.dual", outfile);
        vector_save((Vector*)dres, nm);
        free(nm);
    }*/
    
    fft_finalise();
    
    parallel_finalise();
    

    return 0;
}
