/*
 * dticode/src/timing.c
 *
 * Written 2012 by Tuomo Valkonen <tuomov@iki.fi>, University of Graz.
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <time.h>
#ifdef _OPENACC
#include <openacc.h>
#endif
#include <libtu/output.h>

static int started=0;
static struct timespec real1={0,0}, real2={0,0};
static struct timespec cpu1={0,0}, cpu2={0,0};
static long gpu1=0, gpu2=0;
static long gput1=0, gput2=0;
static clock_t clock1, clock2;

#define NANOSEC 1e-9
#define MICROSEC 1e-6
#define D(X) ((double)(X))
#define TDOUBLESECS(T) (D(T.tv_sec)+NANOSEC*D(T.tv_nsec))
#define TDIFF(T1, T2) timespec_diff(&T1, &T2)
#define GPUTDIFF(T1, T2) ((D(T2)-D(T1))*MICROSEC)

static double timespec_diff(const struct timespec *t1, 
                            const struct timespec *t2)
{
    struct timespec diff;
    if((t2->tv_nsec-t1->tv_nsec)<0){
        diff.tv_sec=t2->tv_sec-t1->tv_sec-1;
        diff.tv_nsec=1000000000+t2->tv_nsec-t1->tv_nsec;
    }else{
        diff.tv_sec=t2->tv_sec-t1->tv_sec;
        diff.tv_nsec=t2->tv_nsec-t1->tv_nsec;
    }
    return TDOUBLESECS(diff);
}

void timing_start()
{
    clock1=clock();
#if defined(_POSIX_MONOTONIC_CLOCK) && defined(CLOCK_MONOTONIC)
    if(clock_gettime(CLOCK_MONOTONIC, &real1)!=0)
        die_err();
#endif
#if defined(_POSIX_CPUTIME) && defined(_CLOCK_PROCESS_CPUTIME_ID)
    if(clock_gettime(CLOCK_PROCESS_CPUTIME_ID, &cpu1)!=0)
        die_err();
#endif
#if defined(_OPENACC) && defined(__PGI)
    /* PGCC-specific, not OpenACC! */
    acc_enable_time(acc_device_nvidia);
    gpu1=acc_exec_time(acc_device_nvidia);
    gput1=acc_total_time(acc_device_nvidia);
#endif
    started=1;
}

void timing_snapshot()
{
    clock2=clock();
#if defined(_POSIX_MONOTONIC_CLOCK) && defined(CLOCK_MONOTONIC)
    if(clock_gettime(CLOCK_MONOTONIC, &real2)!=0)
        die_err();
#endif
#if defined(_POSIX_CPUTIME) && defined(_CLOCK_PROCESS_CPUTIME_ID)
    if(clock_gettime(CLOCK_PROCESS_CPUTIME_ID, &cpu2)!=0)
        die_err();
#endif
    started=0;
#if defined(_OPENACC) && defined(__PGI)
    /* PGCC-specific, not OpenACC! */
    gpu2=acc_exec_time(acc_device_nvidia);
    gput2=acc_total_time(acc_device_nvidia);
#endif
}

void timing_end()
{
    timing_snapshot();
#if defined(_OPENACC) && defined(__PGI)
    acc_disable_time(acc_device_nvidia);
#endif
}

void timing_print(const char *prefix)
{
    if(prefix==NULL)
        prefix="";
    //printf("%d(%d) %f %f %d(%d) %f %f\n", real1.tv_sec, FIX_NSECS(real1.tv_nsec), TDOUBLESECS(real1), real1.tv_nsec*NANOSEC, real2.tv_sec, FIX_NSECS(real2.tv_nsec), TDOUBLESECS(real2), real2.tv_nsec*NANOSEC);
#if defined(_POSIX_MONOTONIC_CLOCK) && defined(CLOCK_MONOTONIC)
    printf("%stime real %fs\n", prefix, TDIFF(real1, real2));
#endif
#if defined(_POSIX_CPUTIME) && defined(_CLOCK_PROCESS_CPUTIME_ID)
    printf("%stime cpu %fs\n", prefix, TDIFF(cpu1, cpu2));
#endif
    printf("%stime cpu(clock) %fs\n", prefix, ((double)(clock2-clock1)/(double)CLOCKS_PER_SEC));
#if defined(_OPENACC) && defined(__PGI)
    printf("%stime gpu exec %fs\n", prefix, GPUTDIFF(gpu1, gpu2));
    printf("%stime gpu total %fs\n", prefix, GPUTDIFF(gput1, gput2));
#endif
}
